
#include <stdio.h>
#include <malloc.h>
#include <string.h>
#include <assert.h>
#include <unistd.h>
#include <math.h>

#include <rsx/commands.h>
#include <rsx/nv40.h>
#include <rsx/reality.h>

#include <io/pad.h>

#include <sysmodule/sysmodule.h>

#include "dice.bin.h"
#include "texture.h"
#include "rsxutil.h"
#include "nv_shaders.h"
#include "vshader.vcg.h"

typedef struct
{
       float data[4][4];
} realityMatrix;


realityMatrix realityMatrixMul(realityMatrix *left,realityMatrix *right)
{
       int row,col;
       realityMatrix result;

       for(row=0;row<4;++row)
       {
               for(col=0;col<4;++col)
               {
                       int i;
                       result.data[row][col] = 0;
                       for(i=0;i<4;++i)
                       {
                               result.data[row][col]+=left->data[row][i]*right->data[i][col];
                       }
               }
       }

       return result;
}

realityMatrix realityMatrixIdentity()
{
       int row,col;
       realityMatrix result;

       for(row=0;row<4;++row)
       {
               for(col=0;col<4;++col)
               {
                       if(row==col)
                               result.data[row][col]=1;
                       else
                               result.data[row][col]=0;
               }
       }

       return result;
}

realityMatrix realityMatrixTranslation(float x,float y,float z)
{
       realityMatrix result=realityMatrixIdentity();
       result.data[3][0]=x;
       result.data[3][1]=y;
       result.data[3][2]=z;

       return result;
}

realityMatrix realityMatrixScale(float x,float y,float z)
{
       realityMatrix result=realityMatrixIdentity();
       result.data[0][0]=x;
       result.data[1][1]=y;
       result.data[2][2]=z;

       return result;
}

realityMatrix realityMatrixTranspose(realityMatrix *src)
{
       int row,col;

       realityMatrix result;

       for(row=0;row<4;++row)
       {
               for(col=0;col<4;++col)
               {
                       result.data[row][col] = src->data[col][row];
               }
       }
       return result;
}

realityMatrix realityMatrixRotationX(float angle)
{
       realityMatrix result=realityMatrixIdentity();

       result.data[1][1]=cos(angle);
       result.data[2][1]=sin(angle);

       result.data[1][2]=-sin(angle);
       result.data[2][2]=cos(angle);

       return result;
}

realityMatrix realityMatrixRotationY(float angle)
{
       realityMatrix result=realityMatrixIdentity();

       result.data[0][0]=cos(angle);
       result.data[2][0]=-sin(angle);

       result.data[0][2]=sin(angle);
       result.data[2][2]=cos(angle);

       return result;
}

realityMatrix realityMatrixRotationZ(float angle)
{
       realityMatrix result=realityMatrixIdentity();

       result.data[0][0]=cos(angle);
       result.data[1][0]=sin(angle);

       result.data[0][1]=-sin(angle);
       result.data[1][1]=cos(angle);

       return result;
}

int currentBuffer = 0;
//realityTexture *ball; // Texture.

u32 *tx_mem;
u32 tx_offset;
Image dice;

const struct Vector
{
float x,y,z,w;
} VertexBufferMain[]=
{
  // f 453/3/1 452/2/1 1/1/1
  {-0.42742792, 0.14125545, -0.01973337,1.0},
  {-0.49998266, 0.14125545, -0.01973337,1.0},
  {-0.49998266, 0.14125545, 0,1.0},
  // f 454/5/3 453/3/2 2/4/2
  {-0.41928023, 0.14089021, -0.01973337,1.0},
  {-0.42742792, 0.14125545, -0.01973337,1.0},
  {-0.42742792, 0.14125545, 0,1.0},
  // f 455/7/5 454/5/3 3/6/4
  {-0.4126853, 0.13995226, -0.01973337,1.0},
  {-0.41928023, 0.14089021, -0.01973337,1.0},
  {-0.41928023, 0.14089021, 0,1.0},
  // f 456/9/7 455/7/5 4/8/6
  {-0.40741441, 0.13865802, -0.01973337,1.0},
  {-0.4126853, 0.13995226, -0.01973337,1.0},
  {-0.4126853, 0.13995226, 0,1.0},
  // f 457/11/9 456/9/7 5/10/8
  {-0.40326801, 0.13719702, -0.01973337,1.0},
  {-0.40741441, 0.13865802, -0.01973337,1.0},
  {-0.40741441, 0.13865802, 0,1.0},
  // f 458/13/11 457/11/9 6/12/10
  {-0.39946407, 0.13541128, -0.01973337,1.0},
  {-0.40326801, 0.13719702, -0.01973337,1.0},
  {-0.40326801, 0.13719702, 0,1.0},
  // f 459/15/13 458/13/11 7/14/12
  {-0.39655605, 0.13367519, -0.01973337,1.0},
  {-0.39946407, 0.13541128, -0.01973337,1.0},
  {-0.39946407, 0.13541128, 0,1.0},
  // f 460/17/15 459/15/13 8/16/14
  {-0.39440098, 0.13212398, -0.01973337,1.0},
  {-0.39655605, 0.13367519, -0.01973337,1.0},
  {-0.39655605, 0.13367519, 0,1.0},
  // f 461/19/17 460/17/15 9/18/16
  {-0.39192128, 0.12998199, -0.01973337,1.0},
  {-0.39440098, 0.13212398, -0.01973337,1.0},
  {-0.39440098, 0.13212398, 0,1.0},
  // f 462/22/19 461/21/17 10/20/18
  {-0.38967857, 0.1276319, -0.01973337,1.0},
  {-0.39192128, 0.12998199, -0.01973337,1.0},
  {-0.39192128, 0.12998199, 0,1.0},
  // f 463/24/21 462/22/19 11/23/20
  {-0.38767204, 0.12509105, -0.01973337,1.0},
  {-0.38967857, 0.1276319, -0.01973337,1.0},
  {-0.38967857, 0.1276319, 0,1.0},
  // f 464/26/23 463/24/21 12/25/22
  {-0.3859016, 0.1223594, -0.01973337,1.0},
  {-0.38767204, 0.12509105, -0.01973337,1.0},
  {-0.38767204, 0.12509105, 0,1.0},
  // f 465/28/25 464/26/23 13/27/24
  {-0.38436708, 0.11943695, -0.01973337,1.0},
  {-0.3859016, 0.1223594, -0.01973337,1.0},
  {-0.3859016, 0.1223594, 0,1.0},
  // f 466/30/27 465/28/25 14/29/26
  {-0.38283733, 0.11567822, -0.01973337,1.0},
  {-0.38436708, 0.11943695, -0.01973337,1.0},
  {-0.38436708, 0.11943695, 0,1.0},
  // f 467/32/29 466/30/27 15/31/28
  {-0.38164759, 0.11164479, -0.01973337,1.0},
  {-0.38283733, 0.11567822, -0.01973337,1.0},
  {-0.38283733, 0.11567822, 0,1.0},
  // f 468/34/31 467/32/29 16/33/30
  {-0.38079774, 0.10733657, -0.01973337,1.0},
  {-0.38164759, 0.11164479, -0.01973337,1.0},
  {-0.38164759, 0.11164479, 0,1.0},
  // f 469/36/33 468/34/31 17/35/32
  {-0.3802878, 0.10275365, -0.01973337,1.0},
  {-0.38079774, 0.10733657, -0.01973337,1.0},
  {-0.38079774, 0.10733657, 0,1.0},
  // f 470/38/35 469/36/33 18/37/34
  {-0.38011783, 0.097896, -0.01973337,1.0},
  {-0.3802878, 0.10275365, -0.01973337,1.0},
  {-0.3802878, 0.10275365, 0,1.0},
  // f 471/40/37 470/38/35 19/39/36
  {-0.38036993, 0.09208981, -0.01973337,1.0},
  {-0.38011783, 0.097896, -0.01973337,1.0},
  {-0.38011783, 0.097896, 0,1.0},
  // f 472/42/39 471/40/37 20/41/38
  {-0.38112643, 0.08665, -0.01973337,1.0},
  {-0.38036993, 0.09208981, -0.01973337,1.0},
  {-0.38036993, 0.09208981, 0,1.0},
  // f 473/44/41 472/42/39 21/43/40
  {-0.38217598, 0.08227888, -0.01973337,1.0},
  {-0.38112643, 0.08665, -0.01973337,1.0},
  {-0.38112643, 0.08665, 0,1.0},
  // f 474/46/43 473/44/41 22/45/42
  {-0.38359609, 0.07817698, -0.01973337,1.0},
  {-0.38217598, 0.08227888, -0.01973337,1.0},
  {-0.38217598, 0.08227888, 0,1.0},
  // f 475/48/45 474/46/43 23/47/44
  {-0.38506255, 0.07496432, -0.01973337,1.0},
  {-0.38359609, 0.07817698, -0.01973337,1.0},
  {-0.38359609, 0.07817698, 0,1.0},
  // f 476/50/47 475/48/45 24/49/46
  {-0.38678622, 0.07193859, -0.01973337,1.0},
  {-0.38506255, 0.07496432, -0.01973337,1.0},
  {-0.38506255, 0.07496432, 0,1.0},
  // f 477/52/49 476/50/47 25/51/48
  {-0.38876721, 0.06909983, -0.01973337,1.0},
  {-0.38678622, 0.07193859, -0.01973337,1.0},
  {-0.38678622, 0.07193859, 0,1.0},
  // f 478/54/51 477/52/49 26/53/50
  {-0.39053714, 0.06696337, -0.01973337,1.0},
  {-0.38876721, 0.06909983, -0.01973337,1.0},
  {-0.38876721, 0.06909983, 0,1.0},
  // f 479/56/53 478/54/51 27/55/52
  {-0.39298108, 0.06446107, -0.01973337,1.0},
  {-0.39053714, 0.06696337, -0.01973337,1.0},
  {-0.39053714, 0.06696337, 0,1.0},
  // f 480/59/55 479/58/53 28/57/54
  {-0.39568928, 0.06217263, -0.01973337,1.0},
  {-0.39298108, 0.06446107, -0.01973337,1.0},
  {-0.39298108, 0.06446107, 0,1.0},
  // f 481/61/57 480/59/55 29/60/56
  {-0.39866793, 0.06012511, -0.01973337,1.0},
  {-0.39568928, 0.06217263, -0.01973337,1.0},
  {-0.39568928, 0.06217263, 0,1.0},
  // f 482/63/59 481/61/57 30/62/58
  {-0.40191758, 0.05831847, -0.01973337,1.0},
  {-0.39866793, 0.06012511, -0.01973337,1.0},
  {-0.39866793, 0.06012511, 0,1.0},
  // f 483/65/61 482/63/59 31/64/60
  {-0.40617442, 0.05646847, -0.01973337,1.0},
  {-0.40191758, 0.05831847, -0.01973337,1.0},
  {-0.40191758, 0.05831847, 0,1.0},
  // f 484/67/63 483/65/61 32/66/62
  {-0.4108212, 0.05496534, -0.01973337,1.0},
  {-0.40617442, 0.05646847, -0.01973337,1.0},
  {-0.40617442, 0.05646847, 0,1.0},
  // f 485/69/65 484/67/63 33/68/64
  {-0.41673514, 0.0536501, -0.01973337,1.0},
  {-0.4108212, 0.05496534, -0.01973337,1.0},
  {-0.4108212, 0.05496534, 0,1.0},
  // f 486/71/67 485/69/65 34/70/66
  {-0.42414385, 0.05272509, -0.01973337,1.0},
  {-0.41673514, 0.0536501, -0.01973337,1.0},
  {-0.41673514, 0.0536501, 0,1.0},
  // f 487/73/69 486/71/67 35/72/68
  {-0.43224552, 0.05241676, -0.01973337,1.0},
  {-0.42414385, 0.05272509, -0.01973337,1.0},
  {-0.42414385, 0.05272509, 0,1.0},
  // f 488/75/70 487/73/70 36/74/70
  {-0.45614135, 0.05241676, -0.01973337,1.0},
  {-0.43224552, 0.05241676, -0.01973337,1.0},
  {-0.43224552, 0.05241676, 0,1.0},
  // f 489/78/71 488/77/71 37/76/71
  {-0.45614135, 0, -0.01973337,1.0},
  {-0.45614135, 0.05241676, -0.01973337,1.0},
  {-0.45614135, 0.05241676, 0,1.0},
  // f 490/81/70 489/80/70 38/79/70
  {-0.49998266, 0, -0.01973337,1.0},
  {-0.45614135, 0, -0.01973337,1.0},
  {-0.45614135, 0, 0,1.0},
  // f 452/84/72 490/83/72 39/82/72
  {-0.49998266, 0.14125545, -0.01973337,1.0},
  {-0.49998266, 0, -0.01973337,1.0},
  {-0.49998266, 0, 0,1.0},
  // f 492/87/1 491/86/1 40/85/1
  {-0.4454461, 0.081034, -0.01973337,1.0},
  {-0.45614135, 0.081034, -0.01973337,1.0},
  {-0.45614135, 0.081034, 0,1.0},
  // f 493/89/74 492/87/73 41/88/73
  {-0.440256, 0.08124619, -0.01973337,1.0},
  {-0.4454461, 0.081034, -0.01973337,1.0},
  {-0.4454461, 0.081034, 0,1.0},
  // f 494/91/76 493/89/74 42/90/75
  {-0.43655068, 0.08173544, -0.01973337,1.0},
  {-0.440256, 0.08124619, -0.01973337,1.0},
  {-0.440256, 0.08124619, 0,1.0},
  // f 495/93/78 494/91/76 43/92/77
  {-0.43400544, 0.0823124, -0.01973337,1.0},
  {-0.43655068, 0.08173544, -0.01973337,1.0},
  {-0.43655068, 0.08173544, 0,1.0},
  // f 496/95/80 495/93/78 44/94/79
  {-0.43205836, 0.08294371, -0.01973337,1.0},
  {-0.43400544, 0.0823124, -0.01973337,1.0},
  {-0.43400544, 0.0823124, 0,1.0},
  // f 497/97/82 496/95/80 45/96/81
  {-0.43060106, 0.08356626, -0.01973337,1.0},
  {-0.43205836, 0.08294371, -0.01973337,1.0},
  {-0.43205836, 0.08294371, 0,1.0},
  // f 498/99/84 497/97/82 46/98/83
  {-0.42929417, 0.08427649, -0.01973337,1.0},
  {-0.43060106, 0.08356626, -0.01973337,1.0},
  {-0.43060106, 0.08356626, 0,1.0},
  // f 499/101/86 498/99/84 47/100/85
  {-0.42858201, 0.0847447, -0.01973337,1.0},
  {-0.42929417, 0.08427649, -0.01973337,1.0},
  {-0.42929417, 0.08427649, 0,1.0},
  // f 500/103/88 499/101/86 48/102/87
  {-0.42771694, 0.08541812, -0.01973337,1.0},
  {-0.42858201, 0.0847447, -0.01973337,1.0},
  {-0.42858201, 0.0847447, 0,1.0},
  // f 501/105/90 500/103/88 49/104/89
  {-0.42656487, 0.0865057, -0.01973337,1.0},
  {-0.42771694, 0.08541812, -0.01973337,1.0},
  {-0.42771694, 0.08541812, 0,1.0},
  // f 502/108/92 501/107/90 50/106/91
  {-0.42555979, 0.08766402, -0.01973337,1.0},
  {-0.42656487, 0.0865057, -0.01973337,1.0},
  {-0.42656487, 0.0865057, 0,1.0},
  // f 503/110/94 502/108/92 51/109/93
  {-0.42483458, 0.08868335, -0.01973337,1.0},
  {-0.42555979, 0.08766402, -0.01973337,1.0},
  {-0.42555979, 0.08766402, 0,1.0},
  // f 504/112/96 503/110/94 52/111/95
  {-0.42421165, 0.08975184, -0.01973337,1.0},
  {-0.42483458, 0.08868335, -0.01973337,1.0},
  {-0.42483458, 0.08868335, 0,1.0},
  // f 505/114/98 504/112/96 53/113/97
  {-0.42369071, 0.09086943, -0.01973337,1.0},
  {-0.42421165, 0.08975184, -0.01973337,1.0},
  {-0.42421165, 0.08975184, 0,1.0},
  // f 506/116/100 505/114/98 54/115/99
  {-0.42327198, 0.09203622, -0.01973337,1.0},
  {-0.42369071, 0.09086943, -0.01973337,1.0},
  {-0.42369071, 0.09086943, 0,1.0},
  // f 507/118/102 506/116/100 55/117/101
  {-0.42295536, 0.09325211, -0.01973337,1.0},
  {-0.42327198, 0.09203622, -0.01973337,1.0},
  {-0.42327198, 0.09203622, 0,1.0},
  // f 508/120/104 507/118/102 56/119/103
  {-0.42271024, 0.09477603, -0.01973337,1.0},
  {-0.42295536, 0.09325211, -0.01973337,1.0},
  {-0.42295536, 0.09325211, 0,1.0},
  // f 509/122/106 508/120/104 57/121/105
  {-0.42264283, 0.09556455, -0.01973337,1.0},
  {-0.42271024, 0.09477603, -0.01973337,1.0},
  {-0.42271024, 0.09477603, 0,1.0},
  // f 510/124/108 509/122/106 58/123/107
  {-0.42261013, 0.09664342, -0.01973337,1.0},
  {-0.42264283, 0.09556455, -0.01973337,1.0},
  {-0.42264283, 0.09556455, 0,1.0},
  // f 511/126/110 510/124/108 59/125/109
  {-0.42269701, 0.0984653, -0.01973337,1.0},
  {-0.42261013, 0.09664342, -0.01973337,1.0},
  {-0.42261013, 0.09664342, 0,1.0},
  // f 512/128/112 511/126/110 60/127/111
  {-0.42290977, 0.09996384, -0.01973337,1.0},
  {-0.42269701, 0.0984653, -0.01973337,1.0},
  {-0.42269701, 0.0984653, 0,1.0},
  // f 513/130/114 512/128/112 61/129/113
  {-0.42325002, 0.10140404, -0.01973337,1.0},
  {-0.42290977, 0.09996384, -0.01973337,1.0},
  {-0.42290977, 0.09996384, 0,1.0},
  // f 514/132/116 513/130/114 62/131/115
  {-0.42371821, 0.10278598, -0.01973337,1.0},
  {-0.42325002, 0.10140404, -0.01973337,1.0},
  {-0.42325002, 0.10140404, 0,1.0},
  // f 515/134/118 514/132/116 63/133/117
  {-0.4243139, 0.10410966, -0.01973337,1.0},
  {-0.42371821, 0.10278598, -0.01973337,1.0},
  {-0.42371821, 0.10278598, 0,1.0},
  // f 516/136/120 515/134/118 64/135/119
  {-0.42503724, 0.10537505, -0.01973337,1.0},
  {-0.4243139, 0.10410966, -0.01973337,1.0},
  {-0.4243139, 0.10410966, 0,1.0},
  // f 517/138/122 516/136/120 65/137/121
  {-0.42588824, 0.10658219, -0.01973337,1.0},
  {-0.42503724, 0.10537505, -0.01973337,1.0},
  {-0.42503724, 0.10537505, 0,1.0},
  // f 518/140/124 517/138/122 66/139/123
  {-0.42636168, 0.10716394, -0.01973337,1.0},
  {-0.42588824, 0.10658219, -0.01973337,1.0},
  {-0.42588824, 0.10658219, 0,1.0},
  // f 519/142/126 518/140/124 67/141/125
  {-0.42704237, 0.10791686, -0.01973337,1.0},
  {-0.42636168, 0.10716394, -0.01973337,1.0},
  {-0.42636168, 0.10716394, 0,1.0},
  // f 520/145/128 519/144/126 68/143/127
  {-0.42780152, 0.10862726, -0.01973337,1.0},
  {-0.42704237, 0.10791686, -0.01973337,1.0},
  {-0.42704237, 0.10791686, 0,1.0},
  // f 521/147/130 520/145/128 69/146/129
  {-0.4286606, 0.10927848, -0.01973337,1.0},
  {-0.42780152, 0.10862726, -0.01973337,1.0},
  {-0.42780152, 0.10862726, 0,1.0},
  // f 522/149/132 521/147/130 70/148/131
  {-0.42961943, 0.10987046, -0.01973337,1.0},
  {-0.4286606, 0.10927848, -0.01973337,1.0},
  {-0.4286606, 0.10927848, 0,1.0},
  // f 523/151/134 522/149/132 71/150/133
  {-0.43095866, 0.11052722, -0.01973337,1.0},
  {-0.42961943, 0.10987046, -0.01973337,1.0},
  {-0.42961943, 0.10987046, 0,1.0},
  // f 524/153/136 523/151/134 72/152/135
  {-0.43245378, 0.11109147, -0.01973337,1.0},
  {-0.43095866, 0.11052722, -0.01973337,1.0},
  {-0.43095866, 0.11052722, 0,1.0},
  // f 525/155/138 524/153/136 73/154/137
  {-0.43445414, 0.11164644, -0.01973337,1.0},
  {-0.43245378, 0.11109147, -0.01973337,1.0},
  {-0.43245378, 0.11109147, 0,1.0},
  // f 526/157/140 525/155/138 74/156/139
  {-0.43707186, 0.11212561, -0.01973337,1.0},
  {-0.43445414, 0.11164644, -0.01973337,1.0},
  {-0.43445414, 0.11164644, 0,1.0},
  // f 527/159/142 526/157/140 75/158/141
  {-0.43999577, 0.11242347, -0.01973337,1.0},
  {-0.43707186, 0.11212561, -0.01973337,1.0},
  {-0.43707186, 0.11212561, 0,1.0},
  // f 528/161/144 527/159/142 76/160/143
  {-0.44371176, 0.11254187, -0.01973337,1.0},
  {-0.43999577, 0.11242347, -0.01973337,1.0},
  {-0.43999577, 0.11242347, 0,1.0},
  // f 529/163/70 528/161/70 77/162/70
  {-0.45614135, 0.11254187, -0.01973337,1.0},
  {-0.44371176, 0.11254187, -0.01973337,1.0},
  {-0.44371176, 0.11254187, 0,1.0},
  // f 491/166/71 529/165/71 78/164/71
  {-0.45614135, 0.081034, -0.01973337,1.0},
  {-0.45614135, 0.11254187, -0.01973337,1.0},
  {-0.45614135, 0.11254187, 0,1.0},
  // f 531/169/145 530/168/145 79/167/145
  {-0.32336509, 0.04933343, -0.01973337,1.0},
  {-0.36489382, 0.04673185, -0.01973337,1.0},
  {-0.36489382, 0.04673185, 0,1.0},
  // f 532/172/148 531/171/147 80/170/146
  {-0.32255679, 0.04475466, -0.01973337,1.0},
  {-0.32336509, 0.04933343, -0.01973337,1.0},
  {-0.32336509, 0.04933343, 0,1.0},
  // f 533/174/150 532/172/148 81/173/149
  {-0.32163715, 0.04136299, -0.01973337,1.0},
  {-0.32255679, 0.04475466, -0.01973337,1.0},
  {-0.32255679, 0.04475466, 0,1.0},
  // f 534/176/152 533/174/150 82/175/151
  {-0.3206183, 0.03864001, -0.01973337,1.0},
  {-0.32163715, 0.04136299, -0.01973337,1.0},
  {-0.32163715, 0.04136299, 0,1.0},
  // f 535/178/154 534/176/152 83/177/153
  {-0.31956071, 0.03648167, -0.01973337,1.0},
  {-0.3206183, 0.03864001, -0.01973337,1.0},
  {-0.3206183, 0.03864001, 0,1.0},
  // f 536/180/156 535/178/154 84/179/155
  {-0.31882712, 0.03525797, -0.01973337,1.0},
  {-0.31956071, 0.03648167, -0.01973337,1.0},
  {-0.31956071, 0.03648167, 0,1.0},
  // f 537/182/158 536/180/156 85/181/157
  {-0.31787291, 0.03391673, -0.01973337,1.0},
  {-0.31882712, 0.03525797, -0.01973337,1.0},
  {-0.31882712, 0.03525797, 0,1.0},
  // f 538/184/160 537/182/158 86/183/159
  {-0.31646615, 0.03228737, -0.01973337,1.0},
  {-0.31787291, 0.03391673, -0.01973337,1.0},
  {-0.31787291, 0.03391673, 0,1.0},
  // f 539/187/162 538/186/160 87/185/161
  {-0.31494385, 0.03082954, -0.01973337,1.0},
  {-0.31646615, 0.03228737, -0.01973337,1.0},
  {-0.31646615, 0.03228737, 0,1.0},
  // f 540/189/164 539/187/162 88/188/163
  {-0.31330577, 0.02954321, -0.01973337,1.0},
  {-0.31494385, 0.03082954, -0.01973337,1.0},
  {-0.31494385, 0.03082954, 0,1.0},
  // f 541/191/166 540/189/164 89/190/165
  {-0.31155208, 0.02842839, -0.01973337,1.0},
  {-0.31330577, 0.02954321, -0.01973337,1.0},
  {-0.31330577, 0.02954321, 0,1.0},
  // f 542/193/168 541/191/166 90/192/167
  {-0.30968276, 0.02748508, -0.01973337,1.0},
  {-0.31155208, 0.02842839, -0.01973337,1.0},
  {-0.31155208, 0.02842839, 0,1.0},
  // f 543/195/170 542/193/168 91/194/169
  {-0.30769795, 0.02671328, -0.01973337,1.0},
  {-0.30968276, 0.02748508, -0.01973337,1.0},
  {-0.30968276, 0.02748508, 0,1.0},
  // f 544/197/172 543/195/170 92/196/171
  {-0.30559742, 0.02611299, -0.01973337,1.0},
  {-0.30769795, 0.02671328, -0.01973337,1.0},
  {-0.30769795, 0.02671328, 0,1.0},
  // f 545/199/174 544/197/172 93/198/173
  {-0.30292422, 0.02561904, -0.01973337,1.0},
  {-0.30559742, 0.02611299, -0.01973337,1.0},
  {-0.30559742, 0.02611299, 0,1.0},
  // f 546/201/176 545/199/174 94/200/175
  {-0.3010495, 0.02542694, -0.01973337,1.0},
  {-0.30292422, 0.02561904, -0.01973337,1.0},
  {-0.30292422, 0.02561904, 0,1.0},
  // f 547/203/178 546/201/176 95/202/177
  {-0.29860207, 0.02534119, -0.01973337,1.0},
  {-0.3010495, 0.02542694, -0.01973337,1.0},
  {-0.3010495, 0.02542694, 0,1.0},
  // f 548/205/180 547/203/178 96/204/179
  {-0.29537636, 0.02548324, -0.01973337,1.0},
  {-0.29860207, 0.02534119, -0.01973337,1.0},
  {-0.29860207, 0.02534119, 0,1.0},
  // f 549/207/182 548/205/180 97/206/181
  {-0.29274219, 0.025848, -0.01973337,1.0},
  {-0.29537636, 0.02548324, -0.01973337,1.0},
  {-0.29537636, 0.02548324, 0,1.0},
  // f 550/209/184 549/207/182 98/208/183
  {-0.29061744, 0.02635129, -0.01973337,1.0},
  {-0.29274219, 0.025848, -0.01973337,1.0},
  {-0.29274219, 0.025848, 0,1.0},
  // f 551/211/186 550/209/184 99/210/185
  {-0.28893006, 0.02691947, -0.01973337,1.0},
  {-0.29061744, 0.02635129, -0.01973337,1.0},
  {-0.29061744, 0.02635129, 0,1.0},
  // f 552/213/188 551/211/186 100/212/187
  {-0.28736609, 0.02761392, -0.01973337,1.0},
  {-0.28893006, 0.02691947, -0.01973337,1.0},
  {-0.28893006, 0.02691947, 0,1.0},
  // f 553/215/190 552/213/188 101/214/189
  {-0.2859256, 0.02843463, -0.01973337,1.0},
  {-0.28736609, 0.02761392, -0.01973337,1.0},
  {-0.28736609, 0.02761392, 0,1.0},
  // f 554/217/192 553/215/190 102/216/191
  {-0.28503397, 0.02905191, -0.01973337,1.0},
  {-0.2859256, 0.02843463, -0.01973337,1.0},
  {-0.2859256, 0.02843463, 0,1.0},
  // f 555/219/194 554/217/192 103/218/193
  {-0.28419709, 0.02972531, -0.01973337,1.0},
  {-0.28503397, 0.02905191, -0.01973337,1.0},
  {-0.28503397, 0.02905191, 0,1.0},
  // f 556/221/196 555/219/194 104/220/195
  {-0.28287992, 0.03098026, -0.01973337,1.0},
  {-0.28419709, 0.02972531, -0.01973337,1.0},
  {-0.28419709, 0.02972531, 0,1.0},
  // f 557/224/198 556/223/196 105/222/197
  {-0.28190863, 0.0320995, -0.01973337,1.0},
  {-0.28287992, 0.03098026, -0.01973337,1.0},
  {-0.28287992, 0.03098026, 0,1.0},
  // f 558/226/200 557/224/198 106/225/199
  {-0.28108308, 0.033259, -0.01973337,1.0},
  {-0.28190863, 0.0320995, -0.01973337,1.0},
  {-0.28190863, 0.0320995, 0,1.0},
  // f 559/228/202 558/226/200 107/227/201
  {-0.2805064, 0.03425597, -0.01973337,1.0},
  {-0.28108308, 0.033259, -0.01973337,1.0},
  {-0.28108308, 0.033259, 0,1.0},
  // f 560/230/204 559/228/202 108/229/203
  {-0.28003085, 0.03528089, -0.01973337,1.0},
  {-0.2805064, 0.03425597, -0.01973337,1.0},
  {-0.2805064, 0.03425597, 0,1.0},
  // f 561/232/206 560/230/204 109/231/205
  {-0.27965656, 0.03633375, -0.01973337,1.0},
  {-0.28003085, 0.03528089, -0.01973337,1.0},
  {-0.28003085, 0.03528089, 0,1.0},
  // f 562/234/208 561/232/206 110/233/207
  {-0.27938333, 0.03741456, -0.01973337,1.0},
  {-0.27965656, 0.03633375, -0.01973337,1.0},
  {-0.27965656, 0.03633375, 0,1.0},
  // f 563/236/210 562/234/208 111/235/209
  {-0.27921134, 0.03852331, -0.01973337,1.0},
  {-0.27938333, 0.03741456, -0.01973337,1.0},
  {-0.27938333, 0.03741456, 0,1.0},
  // f 564/238/212 563/236/210 112/237/211
  {-0.27915674, 0.03920197, -0.01973337,1.0},
  {-0.27921134, 0.03852331, -0.01973337,1.0},
  {-0.27921134, 0.03852331, 0,1.0},
  // f 565/240/214 564/238/212 113/239/213
  {-0.27913854, 0.0398907, -0.01973337,1.0},
  {-0.27915674, 0.03920197, -0.01973337,1.0},
  {-0.27915674, 0.03920197, 0,1.0},
  // f 566/242/216 565/240/214 114/241/215
  {-0.27918667, 0.04097757, -0.01973337,1.0},
  {-0.27913854, 0.0398907, -0.01973337,1.0},
  {-0.27913854, 0.0398907, 0,1.0},
  // f 567/244/218 566/242/216 115/243/217
  {-0.27933127, 0.04204133, -0.01973337,1.0},
  {-0.27918667, 0.04097757, -0.01973337,1.0},
  {-0.27918667, 0.04097757, 0,1.0},
  // f 568/246/220 567/244/218 116/245/219
  {-0.2795721, 0.04308196, -0.01973337,1.0},
  {-0.27933127, 0.04204133, -0.01973337,1.0},
  {-0.27933127, 0.04204133, 0,1.0},
  // f 569/248/222 568/246/220 117/247/221
  {-0.27990934, 0.04409945, -0.01973337,1.0},
  {-0.2795721, 0.04308196, -0.01973337,1.0},
  {-0.2795721, 0.04308196, 0,1.0},
  // f 570/250/224 569/248/222 118/249/223
  {-0.280343, 0.04509383, -0.01973337,1.0},
  {-0.27990934, 0.04409945, -0.01973337,1.0},
  {-0.27990934, 0.04409945, 0,1.0},
  // f 571/252/226 570/250/224 119/251/225
  {-0.28099045, 0.04625655, -0.01973337,1.0},
  {-0.280343, 0.04509383, -0.01973337,1.0},
  {-0.280343, 0.04509383, 0,1.0},
  // f 572/254/228 571/252/226 120/253/227
  {-0.2817767, 0.04738599, -0.01973337,1.0},
  {-0.28099045, 0.04625655, -0.01973337,1.0},
  {-0.28099045, 0.04625655, 0,1.0},
  // f 573/256/230 572/254/228 121/255/229
  {-0.2827017, 0.04848211, -0.01973337,1.0},
  {-0.2817767, 0.04738599, -0.01973337,1.0},
  {-0.2817767, 0.04738599, 0,1.0},
  // f 574/259/232 573/258/230 122/257/231
  {-0.28395623, 0.04971883, -0.01973337,1.0},
  {-0.2827017, 0.04848211, -0.01973337,1.0},
  {-0.2827017, 0.04848211, 0,1.0},
  // f 575/261/234 574/259/232 123/260/233
  {-0.28480846, 0.0504095, -0.01973337,1.0},
  {-0.28395623, 0.04971883, -0.01973337,1.0},
  {-0.28395623, 0.04971883, 0,1.0},
  // f 576/263/236 575/261/234 124/262/235
  {-0.28610268, 0.05126418, -0.01973337,1.0},
  {-0.28480846, 0.0504095, -0.01973337,1.0},
  {-0.28480846, 0.0504095, 0,1.0},
  // f 577/265/238 576/263/236 125/264/237
  {-0.28834203, 0.0524445, -0.01973337,1.0},
  {-0.28610268, 0.05126418, -0.01973337,1.0},
  {-0.28610268, 0.05126418, 0,1.0},
  // f 578/267/240 577/265/238 126/266/239
  {-0.29195365, 0.05393434, -0.01973337,1.0},
  {-0.28834203, 0.0524445, -0.01973337,1.0},
  {-0.28834203, 0.0524445, 0,1.0},
  // f 579/269/242 578/267/240 127/268/241
  {-0.2980513, 0.05587222, -0.01973337,1.0},
  {-0.29195365, 0.05393434, -0.01973337,1.0},
  {-0.29195365, 0.05393434, 0,1.0},
  // f 580/271/244 579/269/242 128/270/243
  {-0.30631042, 0.05790896, -0.01973337,1.0},
  {-0.2980513, 0.05587222, -0.01973337,1.0},
  {-0.2980513, 0.05587222, 0,1.0},
  // f 581/273/246 580/271/244 129/272/245
  {-0.3220557, 0.06216396, -0.01973337,1.0},
  {-0.30631042, 0.05790896, -0.01973337,1.0},
  {-0.30631042, 0.05790896, 0,1.0},
  // f 582/275/248 581/273/246 130/274/247
  {-0.33171731, 0.06576931, -0.01973337,1.0},
  {-0.3220557, 0.06216396, -0.01973337,1.0},
  {-0.3220557, 0.06216396, 0,1.0},
  // f 583/277/250 582/275/248 131/276/249
  {-0.33843592, 0.06902439, -0.01973337,1.0},
  {-0.33171731, 0.06576931, -0.01973337,1.0},
  {-0.33171731, 0.06576931, 0,1.0},
  // f 584/279/252 583/277/250 132/278/251
  {-0.34292337, 0.07174607, -0.01973337,1.0},
  {-0.33843592, 0.06902439, -0.01973337,1.0},
  {-0.33843592, 0.06902439, 0,1.0},
  // f 585/281/254 584/279/252 133/280/253
  {-0.34519771, 0.07337592, -0.01973337,1.0},
  {-0.34292337, 0.07174607, -0.01973337,1.0},
  {-0.34292337, 0.07174607, 0,1.0},
  // f 586/283/256 585/281/254 134/282/255
  {-0.34726104, 0.07506003, -0.01973337,1.0},
  {-0.34519771, 0.07337592, -0.01973337,1.0},
  {-0.34519771, 0.07337592, 0,1.0},
  // f 587/285/258 586/283/256 135/284/257
  {-0.35004342, 0.07771154, -0.01973337,1.0},
  {-0.34726104, 0.07506003, -0.01973337,1.0},
  {-0.34726104, 0.07506003, 0,1.0},
  // f 588/288/260 587/287/258 136/286/259
  {-0.35247061, 0.08053235, -0.01973337,1.0},
  {-0.35004342, 0.07771154, -0.01973337,1.0},
  {-0.35004342, 0.07771154, 0,1.0},
  // f 589/290/262 588/288/260 137/289/261
  {-0.35422194, 0.08301231, -0.01973337,1.0},
  {-0.35247061, 0.08053235, -0.01973337,1.0},
  {-0.35247061, 0.08053235, 0,1.0},
  // f 590/292/264 589/290/262 138/291/263
  {-0.35572663, 0.08560983, -0.01973337,1.0},
  {-0.35422194, 0.08301231, -0.01973337,1.0},
  {-0.35422194, 0.08301231, 0,1.0},
  // f 591/294/266 590/292/264 139/293/265
  {-0.35698465, 0.0883249, -0.01973337,1.0},
  {-0.35572663, 0.08560983, -0.01973337,1.0},
  {-0.35572663, 0.08560983, 0,1.0},
  // f 592/296/268 591/294/266 140/295/267
  {-0.35799599, 0.09115754, -0.01973337,1.0},
  {-0.35698465, 0.0883249, -0.01973337,1.0},
  {-0.35698465, 0.0883249, 0,1.0},
  // f 593/298/270 592/296/268 141/297/269
  {-0.35876068, 0.09410771, -0.01973337,1.0},
  {-0.35799599, 0.09115754, -0.01973337,1.0},
  {-0.35799599, 0.09115754, 0,1.0},
  // f 594/300/272 593/298/270 142/299/271
  {-0.35935268, 0.0978031, -0.01973337,1.0},
  {-0.35876068, 0.09410771, -0.01973337,1.0},
  {-0.35876068, 0.09410771, 0,1.0},
  // f 595/302/274 594/300/272 143/301/273
  {-0.35951549, 0.09971424, -0.01973337,1.0},
  {-0.35935268, 0.0978031, -0.01973337,1.0},
  {-0.35935268, 0.0978031, 0,1.0},
  // f 596/304/276 595/302/274 144/303/275
  {-0.35959435, 0.10232831, -0.01973337,1.0},
  {-0.35951549, 0.09971424, -0.01973337,1.0},
  {-0.35951549, 0.09971424, 0,1.0},
  // f 597/306/278 596/304/276 145/305/277
  {-0.35943285, 0.10578167, -0.01973337,1.0},
  {-0.35959435, 0.10232831, -0.01973337,1.0},
  {-0.35959435, 0.10232831, 0,1.0},
  // f 598/308/280 597/306/278 146/307/279
  {-0.35894817, 0.10917331, -0.01973337,1.0},
  {-0.35943285, 0.10578167, -0.01973337,1.0},
  {-0.35943285, 0.10578167, 0,1.0},
  // f 599/310/282 598/308/280 147/309/281
  {-0.35814032, 0.11250332, -0.01973337,1.0},
  {-0.35894817, 0.10917331, -0.01973337,1.0},
  {-0.35894817, 0.10917331, 0,1.0},
  // f 600/312/284 599/310/282 148/311/283
  {-0.35700944, 0.11577167, -0.01973337,1.0},
  {-0.35814032, 0.11250332, -0.01973337,1.0},
  {-0.35814032, 0.11250332, 0,1.0},
  // f 601/314/286 600/312/284 149/313/285
  {-0.35555524, 0.11897835, -0.01973337,1.0},
  {-0.35700944, 0.11577167, -0.01973337,1.0},
  {-0.35700944, 0.11577167, 0,1.0},
  // f 602/316/288 601/314/286 150/315/287
  {-0.35448223, 0.12095121, -0.01973337,1.0},
  {-0.35555524, 0.11897835, -0.01973337,1.0},
  {-0.35555524, 0.11897835, 0,1.0},
  // f 603/318/290 602/316/288 151/317/289
  {-0.35328317, 0.12289994, -0.01973337,1.0},
  {-0.35448223, 0.12095121, -0.01973337,1.0},
  {-0.35448223, 0.12095121, 0,1.0},
  // f 604/320/292 603/318/290 152/319/291
  {-0.3513914, 0.12553017, -0.01973337,1.0},
  {-0.35328317, 0.12289994, -0.01973337,1.0},
  {-0.35328317, 0.12289994, 0,1.0},
  // f 605/322/294 604/320/292 153/321/293
  {-0.34925041, 0.12799795, -0.01973337,1.0},
  {-0.3513914, 0.12553017, -0.01973337,1.0},
  {-0.3513914, 0.12553017, 0,1.0},
  // f 606/325/296 605/324/294 154/323/295
  {-0.3464981, 0.13061942, -0.01973337,1.0},
  {-0.34925041, 0.12799795, -0.01973337,1.0},
  {-0.34925041, 0.12799795, 0,1.0},
  // f 607/327/298 606/325/296 155/326/297
  {-0.34342036, 0.13302872, -0.01973337,1.0},
  {-0.3464981, 0.13061942, -0.01973337,1.0},
  {-0.3464981, 0.13061942, 0,1.0},
  // f 608/329/300 607/327/298 156/328/299
  {-0.34001693, 0.13522591, -0.01973337,1.0},
  {-0.34342036, 0.13302872, -0.01973337,1.0},
  {-0.34342036, 0.13302872, 0,1.0},
  // f 609/331/302 608/329/300 157/330/301
  {-0.3372508, 0.13673456, -0.01973337,1.0},
  {-0.34001693, 0.13522591, -0.01973337,1.0},
  {-0.34001693, 0.13522591, 0,1.0},
  // f 610/333/304 609/331/302 158/332/303
  {-0.33430135, 0.13812394, -0.01973337,1.0},
  {-0.3372508, 0.13673456, -0.01973337,1.0},
  {-0.3372508, 0.13673456, 0,1.0},
  // f 611/335/306 610/333/304 159/334/305
  {-0.33000627, 0.13975506, -0.01973337,1.0},
  {-0.33430135, 0.13812394, -0.01973337,1.0},
  {-0.33430135, 0.13812394, 0,1.0},
  // f 612/337/308 611/335/306 160/336/307
  {-0.32459944, 0.14125092, -0.01973337,1.0},
  {-0.33000627, 0.13975506, -0.01973337,1.0},
  {-0.33000627, 0.13975506, 0,1.0},
  // f 613/339/310 612/337/308 161/338/309
  {-0.31716415, 0.1425917, -0.01973337,1.0},
  {-0.32459944, 0.14125092, -0.01973337,1.0},
  {-0.32459944, 0.14125092, 0,1.0},
  // f 614/341/312 613/339/310 162/340/311
  {-0.30881959, 0.14339617, -0.01973337,1.0},
  {-0.31716415, 0.1425917, -0.01973337,1.0},
  {-0.31716415, 0.1425917, 0,1.0},
  // f 615/343/314 614/341/312 163/342/313
  {-0.29956564, 0.1436643, -0.01973337,1.0},
  {-0.30881959, 0.14339617, -0.01973337,1.0},
  {-0.30881959, 0.14339617, 0,1.0},
  // f 616/345/316 615/343/314 164/344/315
  {-0.28827497, 0.14317694, -0.01973337,1.0},
  {-0.29956564, 0.1436643, -0.01973337,1.0},
  {-0.29956564, 0.1436643, 0,1.0},
  // f 617/347/318 616/345/316 165/346/317
  {-0.27996331, 0.14205326, -0.01973337,1.0},
  {-0.28827497, 0.14317694, -0.01973337,1.0},
  {-0.28827497, 0.14317694, 0,1.0},
  // f 618/349/320 617/347/318 166/348/319
  {-0.27327496, 0.14050667, -0.01973337,1.0},
  {-0.27996331, 0.14205326, -0.01973337,1.0},
  {-0.27996331, 0.14205326, 0,1.0},
  // f 619/351/322 618/349/320 167/350/321
  {-0.26796278, 0.13873047, -0.01973337,1.0},
  {-0.27327496, 0.14050667, -0.01973337,1.0},
  {-0.27327496, 0.14050667, 0,1.0},
  // f 620/353/324 619/351/322 168/352/323
  {-0.26381099, 0.13689388, -0.01973337,1.0},
  {-0.26796278, 0.13873047, -0.01973337,1.0},
  {-0.26796278, 0.13873047, 0,1.0},
  // f 621/355/326 620/353/324 169/354/325
  {-0.26124883, 0.1355084, -0.01973337,1.0},
  {-0.26381099, 0.13689388, -0.01973337,1.0},
  {-0.26381099, 0.13689388, 0,1.0},
  // f 622/357/328 621/355/326 170/356/327
  {-0.2582778, 0.13359529, -0.01973337,1.0},
  {-0.26124883, 0.1355084, -0.01973337,1.0},
  {-0.26124883, 0.1355084, 0,1.0},
  // f 623/359/330 622/357/328 171/358/329
  {-0.2555505, 0.1314625, -0.01973337,1.0},
  {-0.2582778, 0.13359529, -0.01973337,1.0},
  {-0.2582778, 0.13359529, 0,1.0},
  // f 624/361/332 623/359/330 172/360/331
  {-0.25305346, 0.12909168, -0.01973337,1.0},
  {-0.2555505, 0.1314625, -0.01973337,1.0},
  {-0.2555505, 0.1314625, 0,1.0},
  // f 625/364/334 624/363/332 173/362/333
  {-0.25078672, 0.12648289, -0.01973337,1.0},
  {-0.25305346, 0.12909168, -0.01973337,1.0},
  {-0.25305346, 0.12909168, 0,1.0},
  // f 626/366/336 625/364/334 174/365/335
  {-0.24875031, 0.1236361, -0.01973337,1.0},
  {-0.25078672, 0.12648289, -0.01973337,1.0},
  {-0.25078672, 0.12648289, 0,1.0},
  // f 627/368/338 626/366/336 175/367/337
  {-0.24694414, 0.12055133, -0.01973337,1.0},
  {-0.24875031, 0.1236361, -0.01973337,1.0},
  {-0.24875031, 0.1236361, 0,1.0},
  // f 628/370/340 627/368/338 176/369/339
  {-0.24508065, 0.11653544, -0.01973337,1.0},
  {-0.24694414, 0.12055133, -0.01973337,1.0},
  {-0.24694414, 0.12055133, 0,1.0},
  // f 629/372/342 628/370/340 177/371/341
  {-0.2435489, 0.11217681, -0.01973337,1.0},
  {-0.24508065, 0.11653544, -0.01973337,1.0},
  {-0.24508065, 0.11653544, 0,1.0},
  // f 630/374/344 629/372/342 178/373/343
  {-0.24234867, 0.1074755, -0.01973337,1.0},
  {-0.2435489, 0.11217681, -0.01973337,1.0},
  {-0.2435489, 0.11217681, 0,1.0},
  // f 631/376/346 630/374/344 179/375/345
  {-0.24136762, 0.10155746, -0.01973337,1.0},
  {-0.24234867, 0.1074755, -0.01973337,1.0},
  {-0.24234867, 0.1074755, 0,1.0},
  // f 632/379/347 631/378/347 180/377/347
  {-0.28251091, 0.09914861, -0.01973337,1.0},
  {-0.24136762, 0.10155746, -0.01973337,1.0},
  {-0.24136762, 0.10155746, 0,1.0},
  // f 633/382/350 632/381/349 181/380/348
  {-0.28321767, 0.10241412, -0.01973337,1.0},
  {-0.28251091, 0.09914861, -0.01973337,1.0},
  {-0.28251091, 0.09914861, 0,1.0},
  // f 634/384/352 633/382/350 182/383/351
  {-0.28404248, 0.1050337, -0.01973337,1.0},
  {-0.28321767, 0.10241412, -0.01973337,1.0},
  {-0.28321767, 0.10241412, 0,1.0},
  // f 635/386/354 634/384/352 183/385/353
  {-0.28491592, 0.10710733, -0.01973337,1.0},
  {-0.28404248, 0.1050337, -0.01973337,1.0},
  {-0.28404248, 0.1050337, 0,1.0},
  // f 636/388/356 635/386/354 184/387/355
  {-0.28577727, 0.10872237, -0.01973337,1.0},
  {-0.28491592, 0.10710733, -0.01973337,1.0},
  {-0.28491592, 0.10710733, 0,1.0},
  // f 637/390/358 636/388/356 185/389/357
  {-0.28674284, 0.11018759, -0.01973337,1.0},
  {-0.28577727, 0.10872237, -0.01973337,1.0},
  {-0.28577727, 0.10872237, 0,1.0},
  // f 638/392/360 637/390/358 186/391/359
  {-0.28781238, 0.11150294, -0.01973337,1.0},
  {-0.28674284, 0.11018759, -0.01973337,1.0},
  {-0.28674284, 0.11018759, 0,1.0},
  // f 639/394/362 638/392/360 187/393/361
  {-0.2885831, 0.11229659, -0.01973337,1.0},
  {-0.28781238, 0.11150294, -0.01973337,1.0},
  {-0.28781238, 0.11150294, 0,1.0},
  // f 640/397/364 639/396/362 188/395/363
  {-0.28940025, 0.11302363, -0.01973337,1.0},
  {-0.2885831, 0.11229659, -0.01973337,1.0},
  {-0.2885831, 0.11229659, 0,1.0},
  // f 641/399/366 640/397/364 189/398/365
  {-0.29071814, 0.11400183, -0.01973337,1.0},
  {-0.28940025, 0.11302363, -0.01973337,1.0},
  {-0.28940025, 0.11302363, 0,1.0},
  // f 642/401/368 641/399/366 190/400/367
  {-0.29215118, 0.11485515, -0.01973337,1.0},
  {-0.29071814, 0.11400183, -0.01973337,1.0},
  {-0.29071814, 0.11400183, 0,1.0},
  // f 643/403/370 642/401/368 191/402/369
  {-0.29369941, 0.11558357, -0.01973337,1.0},
  {-0.29215118, 0.11485515, -0.01973337,1.0},
  {-0.29215118, 0.11485515, 0,1.0},
  // f 644/405/372 643/403/370 192/404/371
  {-0.2953628, 0.11618715, -0.01973337,1.0},
  {-0.29369941, 0.11558357, -0.01973337,1.0},
  {-0.29369941, 0.11558357, 0,1.0},
  // f 645/407/374 644/405/372 193/406/373
  {-0.29744899, 0.11673349, -0.01973337,1.0},
  {-0.2953628, 0.11618715, -0.01973337,1.0},
  {-0.2953628, 0.11618715, 0,1.0},
  // f 646/409/376 645/407/374 194/408/375
  {-0.30002517, 0.11714971, -0.01973337,1.0},
  {-0.29744899, 0.11673349, -0.01973337,1.0},
  {-0.29744899, 0.11673349, 0,1.0},
  // f 647/411/378 646/409/376 195/410/377
  {-0.30173922, 0.11729716, -0.01973337,1.0},
  {-0.30002517, 0.11714971, -0.01973337,1.0},
  {-0.30002517, 0.11714971, 0,1.0},
  // f 648/413/380 647/411/378 196/412/379
  {-0.30390158, 0.11735957, -0.01973337,1.0},
  {-0.30173922, 0.11729716, -0.01973337,1.0},
  {-0.30173922, 0.11729716, 0,1.0},
  // f 649/415/382 648/413/380 197/414/381
  {-0.30679607, 0.11723047, -0.01973337,1.0},
  {-0.30390158, 0.11735957, -0.01973337,1.0},
  {-0.30390158, 0.11735957, 0,1.0},
  // f 650/417/384 649/415/382 198/416/383
  {-0.30889514, 0.11694124, -0.01973337,1.0},
  {-0.30679607, 0.11723047, -0.01973337,1.0},
  {-0.30679607, 0.11723047, 0,1.0},
  // f 651/419/386 650/417/384 199/418/385
  {-0.31057414, 0.11655263, -0.01973337,1.0},
  {-0.30889514, 0.11694124, -0.01973337,1.0},
  {-0.30889514, 0.11694124, 0,1.0},
  // f 652/421/388 651/419/386 200/420/387
  {-0.31189597, 0.11611877, -0.01973337,1.0},
  {-0.31057414, 0.11655263, -0.01973337,1.0},
  {-0.31057414, 0.11655263, 0,1.0},
  // f 653/423/390 652/421/388 201/422/389
  {-0.31310949, 0.11559201, -0.01973337,1.0},
  {-0.31189597, 0.11611877, -0.01973337,1.0},
  {-0.31189597, 0.11611877, 0,1.0},
  // f 654/425/392 653/423/390 202/424/391
  {-0.31403819, 0.115082, -0.01973337,1.0},
  {-0.31310949, 0.11559201, -0.01973337,1.0},
  {-0.31310949, 0.11559201, 0,1.0},
  // f 655/427/394 654/425/392 203/426/393
  {-0.31472713, 0.1146275, -0.01973337,1.0},
  {-0.31403819, 0.115082, -0.01973337,1.0},
  {-0.31403819, 0.115082, 0,1.0},
  // f 656/429/396 655/427/394 204/428/395
  {-0.31536773, 0.1141317, -0.01973337,1.0},
  {-0.31472713, 0.1146275, -0.01973337,1.0},
  {-0.31472713, 0.1146275, 0,1.0},
  // f 657/431/398 656/429/396 205/430/397
  {-0.31623718, 0.1133369, -0.01973337,1.0},
  {-0.31536773, 0.1141317, -0.01973337,1.0},
  {-0.31536773, 0.1141317, 0,1.0},
  // f 658/434/400 657/433/398 206/432/399
  {-0.31699577, 0.11250187, -0.01973337,1.0},
  {-0.31623718, 0.1133369, -0.01973337,1.0},
  {-0.31623718, 0.1133369, 0,1.0},
  // f 659/436/402 658/434/400 207/435/401
  {-0.31764328, 0.1116266, -0.01973337,1.0},
  {-0.31699577, 0.11250187, -0.01973337,1.0},
  {-0.31699577, 0.11250187, 0,1.0},
  // f 660/438/404 659/436/402 208/437/403
  {-0.31809804, 0.11086644, -0.01973337,1.0},
  {-0.31764328, 0.1116266, -0.01973337,1.0},
  {-0.31764328, 0.1116266, 0,1.0},
  // f 661/440/406 660/438/404 209/439/405
  {-0.31847572, 0.11007833, -0.01973337,1.0},
  {-0.31809804, 0.11086644, -0.01973337,1.0},
  {-0.31809804, 0.11086644, 0,1.0},
  // f 662/442/408 661/440/406 210/441/407
  {-0.31877634, 0.10926232, -0.01973337,1.0},
  {-0.31847572, 0.11007833, -0.01973337,1.0},
  {-0.31847572, 0.11007833, 0,1.0},
  // f 663/444/410 662/442/408 211/443/409
  {-0.31899998, 0.10841836, -0.01973337,1.0},
  {-0.31877634, 0.10926232, -0.01973337,1.0},
  {-0.31877634, 0.10926232, 0,1.0},
  // f 664/446/412 663/444/410 212/445/411
  {-0.31914634, 0.10754646, -0.01973337,1.0},
  {-0.31899998, 0.10841836, -0.01973337,1.0},
  {-0.31899998, 0.10841836, 0,1.0},
  // f 665/448/414 664/446/412 213/447/413
  {-0.31919718, 0.1070099, -0.01973337,1.0},
  {-0.31914634, 0.10754646, -0.01973337,1.0},
  {-0.31914634, 0.10754646, 0,1.0},
  // f 666/450/416 665/448/414 214/449/415
  {-0.31922197, 0.10627884, -0.01973337,1.0},
  {-0.31919718, 0.1070099, -0.01973337,1.0},
  {-0.31919718, 0.1070099, 0,1.0},
  // f 667/452/418 666/450/416 215/451/417
  {-0.31919011, 0.10561109, -0.01973337,1.0},
  {-0.31922197, 0.10627884, -0.01973337,1.0},
  {-0.31922197, 0.10627884, 0,1.0},
  // f 668/454/420 667/452/418 216/453/419
  {-0.31909472, 0.10495687, -0.01973337,1.0},
  {-0.31919011, 0.10561109, -0.01973337,1.0},
  {-0.31919011, 0.10561109, 0,1.0},
  // f 669/456/422 668/454/420 217/455/421
  {-0.31893578, 0.1043161, -0.01973337,1.0},
  {-0.31909472, 0.10495687, -0.01973337,1.0},
  {-0.31909472, 0.10495687, 0,1.0},
  // f 670/458/424 669/456/422 218/457/423
  {-0.31871322, 0.10368883, -0.01973337,1.0},
  {-0.31893578, 0.1043161, -0.01973337,1.0},
  {-0.31893578, 0.1043161, 0,1.0},
  // f 671/460/426 670/458/424 219/459/425
  {-0.31842694, 0.10307506, -0.01973337,1.0},
  {-0.31871322, 0.10368883, -0.01973337,1.0},
  {-0.31871322, 0.10368883, 0,1.0},
  // f 672/462/428 671/460/426 220/461/427
  {-0.31807724, 0.10247475, -0.01973337,1.0},
  {-0.31842694, 0.10307506, -0.01973337,1.0},
  {-0.31842694, 0.10307506, 0,1.0},
  // f 673/464/429 672/462/428 221/463/336
  {-0.31757355, 0.10177226, -0.01973337,1.0},
  {-0.31807724, 0.10247475, -0.01973337,1.0},
  {-0.31807724, 0.10247475, 0,1.0},
  // f 674/466/431 673/464/429 222/465/430
  {-0.31697828, 0.1010891, -0.01973337,1.0},
  {-0.31757355, 0.10177226, -0.01973337,1.0},
  {-0.31757355, 0.10177226, 0,1.0},
  // f 675/469/433 674/468/431 223/467/432
  {-0.31653062, 0.10064447, -0.01973337,1.0},
  {-0.31697828, 0.1010891, -0.01973337,1.0},
  {-0.31697828, 0.1010891, 0,1.0},
  // f 676/471/435 675/469/433 224/470/434
  {-0.31604224, 0.10020851, -0.01973337,1.0},
  {-0.31653062, 0.10064447, -0.01973337,1.0},
  {-0.31653062, 0.10064447, 0,1.0},
  // f 677/473/437 676/471/435 225/472/436
  {-0.31549466, 0.09976391, -0.01973337,1.0},
  {-0.31604224, 0.10020851, -0.01973337,1.0},
  {-0.31604224, 0.10020851, 0,1.0},
  // f 678/475/439 677/473/437 226/474/438
  {-0.31465751, 0.09921506, -0.01973337,1.0},
  {-0.31549466, 0.09976391, -0.01973337,1.0},
  {-0.31549466, 0.09976391, 0,1.0},
  // f 679/477/441 678/475/439 227/476/440
  {-0.31342909, 0.09856663, -0.01973337,1.0},
  {-0.31465751, 0.09921506, -0.01973337,1.0},
  {-0.31465751, 0.09921506, 0,1.0},
  // f 680/479/443 679/477/441 228/478/442
  {-0.31112865, 0.09761478, -0.01973337,1.0},
  {-0.31342909, 0.09856663, -0.01973337,1.0},
  {-0.31342909, 0.09856663, 0,1.0},
  // f 681/481/445 680/479/443 229/480/444
  {-0.30683568, 0.09628404, -0.01973337,1.0},
  {-0.31112865, 0.09761478, -0.01973337,1.0},
  {-0.31112865, 0.09761478, 0,1.0},
  // f 682/483/447 681/481/445 230/482/446
  {-0.30139637, 0.0950054, -0.01973337,1.0},
  {-0.30683568, 0.09628404, -0.01973337,1.0},
  {-0.30683568, 0.09628404, 0,1.0},
  // f 683/485/449 682/483/447 231/484/448
  {-0.27602455, 0.08855294, -0.01973337,1.0},
  {-0.30139637, 0.0950054, -0.01973337,1.0},
  {-0.30139637, 0.0950054, 0,1.0},
  // f 684/487/451 683/485/449 232/486/450
  {-0.26727593, 0.08554854, -0.01973337,1.0},
  {-0.27602455, 0.08855294, -0.01973337,1.0},
  {-0.27602455, 0.08855294, 0,1.0},
  // f 685/489/453 684/487/451 233/488/452
  {-0.26039758, 0.08252749, -0.01973337,1.0},
  {-0.26727593, 0.08554854, -0.01973337,1.0},
  {-0.26727593, 0.08554854, 0,1.0},
  // f 686/491/455 685/489/453 234/490/454
  {-0.25571474, 0.07988161, -0.01973337,1.0},
  {-0.26039758, 0.08252749, -0.01973337,1.0},
  {-0.26039758, 0.08252749, 0,1.0},
  // f 687/493/457 686/491/455 235/492/456
  {-0.25196159, 0.07729269, -0.01973337,1.0},
  {-0.25571474, 0.07988161, -0.01973337,1.0},
  {-0.25571474, 0.07988161, 0,1.0},
  // f 688/495/459 687/493/457 236/494/458
  {-0.24899231, 0.07482632, -0.01973337,1.0},
  {-0.25196159, 0.07729269, -0.01973337,1.0},
  {-0.25196159, 0.07729269, 0,1.0},
  // f 689/497/461 688/495/459 237/496/460
  {-0.24636841, 0.07220456, -0.01973337,1.0},
  {-0.24899231, 0.07482632, -0.01973337,1.0},
  {-0.24899231, 0.07482632, 0,1.0},
  // f 690/500/463 689/499/461 238/498/462
  {-0.24435572, 0.06978304, -0.01973337,1.0},
  {-0.24636841, 0.07220456, -0.01973337,1.0},
  {-0.24636841, 0.07220456, 0,1.0},
  // f 691/502/465 690/500/463 239/501/464
  {-0.24332435, 0.0683459, -0.01973337,1.0},
  {-0.24435572, 0.06978304, -0.01973337,1.0},
  {-0.24435572, 0.06978304, 0,1.0},
  // f 692/504/467 691/502/465 240/503/466
  {-0.24237932, 0.06686991, -0.01973337,1.0},
  {-0.24332435, 0.0683459, -0.01973337,1.0},
  {-0.24332435, 0.0683459, 0,1.0},
  // f 693/506/469 692/504/467 241/505/468
  {-0.24071984, 0.06382233, -0.01973337,1.0},
  {-0.24237932, 0.06686991, -0.01973337,1.0},
  {-0.24237932, 0.06686991, 0,1.0},
  // f 694/508/471 693/506/469 242/507/470
  {-0.23934902, 0.0606613, -0.01973337,1.0},
  {-0.24071984, 0.06382233, -0.01973337,1.0},
  {-0.24071984, 0.06382233, 0,1.0},
  // f 695/510/473 694/508/471 243/509/472
  {-0.23826675, 0.05738679, -0.01973337,1.0},
  {-0.23934902, 0.0606613, -0.01973337,1.0},
  {-0.23934902, 0.0606613, 0,1.0},
  // f 696/512/475 695/510/473 244/511/474
  {-0.23747315, 0.05399883, -0.01973337,1.0},
  {-0.23826675, 0.05738679, -0.01973337,1.0},
  {-0.23826675, 0.05738679, 0,1.0},
  // f 697/514/477 696/512/475 245/513/476
  {-0.23692529, 0.05005173, -0.01973337,1.0},
  {-0.23747315, 0.05399883, -0.01973337,1.0},
  {-0.23747315, 0.05399883, 0,1.0},
  // f 698/516/479 697/514/477 246/515/478
  {-0.23674259, 0.04596103, -0.01973337,1.0},
  {-0.23692529, 0.05005173, -0.01973337,1.0},
  {-0.23692529, 0.05005173, 0,1.0},
  // f 699/518/481 698/516/479 247/517/480
  {-0.23693499, 0.04166777, -0.01973337,1.0},
  {-0.23674259, 0.04596103, -0.01973337,1.0},
  {-0.23674259, 0.04596103, 0,1.0},
  // f 700/520/483 699/518/481 248/519/482
  {-0.2375122, 0.03748308, -0.01973337,1.0},
  {-0.23693499, 0.04166777, -0.01973337,1.0},
  {-0.23693499, 0.04166777, 0,1.0},
  // f 701/522/485 700/520/483 249/521/484
  {-0.23847421, 0.0334069, -0.01973337,1.0},
  {-0.2375122, 0.03748308, -0.01973337,1.0},
  {-0.2375122, 0.03748308, 0,1.0},
  // f 702/524/487 701/522/485 250/523/486
  {-0.23982103, 0.02943926, -0.01973337,1.0},
  {-0.23847421, 0.0334069, -0.01973337,1.0},
  {-0.23847421, 0.0334069, 0,1.0},
  // f 703/526/489 702/524/487 251/525/488
  {-0.24155261, 0.02558015, -0.01973337,1.0},
  {-0.23982103, 0.02943926, -0.01973337,1.0},
  {-0.23982103, 0.02943926, 0,1.0},
  // f 704/528/491 703/526/489 252/527/490
  {-0.24283025, 0.02322332, -0.01973337,1.0},
  {-0.24155261, 0.02558015, -0.01973337,1.0},
  {-0.24155261, 0.02558015, 0,1.0},
  // f 705/530/493 704/528/491 253/529/492
  {-0.24425827, 0.02090889, -0.01973337,1.0},
  {-0.24283025, 0.02322332, -0.01973337,1.0},
  {-0.24283025, 0.02322332, 0,1.0},
  // f 706/532/495 705/530/493 254/531/494
  {-0.24681623, 0.01738155, -0.01973337,1.0},
  {-0.24425827, 0.02090889, -0.01973337,1.0},
  {-0.24425827, 0.02090889, 0,1.0},
  // f 707/534/497 706/532/495 255/533/496
  {-0.24967997, 0.01413788, -0.01973337,1.0},
  {-0.24681623, 0.01738155, -0.01973337,1.0},
  {-0.24681623, 0.01738155, 0,1.0},
  // f 708/537/499 707/536/497 256/535/498
  {-0.25284964, 0.01117787, -0.01973337,1.0},
  {-0.24967997, 0.01413788, -0.01973337,1.0},
  {-0.24967997, 0.01413788, 0,1.0},
  // f 709/539/501 708/537/499 257/538/500
  {-0.25632519, 0.00850154, -0.01973337,1.0},
  {-0.25284964, 0.01117787, -0.01973337,1.0},
  {-0.25284964, 0.01117787, 0,1.0},
  // f 710/541/503 709/539/501 258/540/502
  {-0.26010665, 0.00610887, -0.01973337,1.0},
  {-0.25632519, 0.00850154, -0.01973337,1.0},
  {-0.25632519, 0.00850154, 0,1.0},
  // f 711/543/505 710/541/503 259/542/504
  {-0.26262525, 0.0047575, -0.01973337,1.0},
  {-0.26010665, 0.00610887, -0.01973337,1.0},
  {-0.26010665, 0.00610887, 0,1.0},
  // f 712/545/507 711/543/505 260/544/506
  {-0.2652635, 0.00351693, -0.01973337,1.0},
  {-0.26262525, 0.0047575, -0.01973337,1.0},
  {-0.26262525, 0.0047575, 0,1.0},
  // f 713/547/509 712/545/507 261/546/508
  {-0.27034768, 0.00157564, -0.01973337,1.0},
  {-0.2652635, 0.00351693, -0.01973337,1.0},
  {-0.2652635, 0.00351693, 0,1.0},
  // f 714/549/511 713/547/509 262/548/510
  {-0.27653128, -0.00013098, -0.01973337,1.0},
  {-0.27034768, 0.00157564, -0.01973337,1.0},
  {-0.27034768, 0.00157564, 0,1.0},
  // f 715/551/513 714/549/511 263/550/512
  {-0.2839832, -0.00146073, -0.01973337,1.0},
  {-0.27653128, -0.00013098, -0.01973337,1.0},
  {-0.27653128, -0.00013098, 0,1.0},
  // f 716/553/515 715/551/513 264/552/514
  {-0.29134852, -0.00217183, -0.01973337,1.0},
  {-0.2839832, -0.00146073, -0.01973337,1.0},
  {-0.2839832, -0.00146073, 0,1.0},
  // f 717/555/517 716/553/515 265/554/516
  {-0.29927659, -0.00240886, -0.01973337,1.0},
  {-0.29134852, -0.00217183, -0.01973337,1.0},
  {-0.29134852, -0.00217183, 0,1.0},
  // f 718/557/519 717/555/517 266/556/518
  {-0.31280473, -0.00185386, -0.01973337,1.0},
  {-0.29927659, -0.00240886, -0.01973337,1.0},
  {-0.29927659, -0.00240886, 0,1.0},
  // f 719/559/521 718/557/519 267/558/520
  {-0.32235077, -0.00061066, -0.01973337,1.0},
  {-0.31280473, -0.00185386, -0.01973337,1.0},
  {-0.31280473, -0.00185386, 0,1.0},
  // f 720/561/523 719/559/521 268/560/522
  {-0.32977268, 0.0010599, -0.01973337,1.0},
  {-0.32235077, -0.00061066, -0.01973337,1.0},
  {-0.32235077, -0.00061066, 0,1.0},
  // f 721/563/525 720/561/523 269/562/524
  {-0.33544296, 0.0029247, -0.01973337,1.0},
  {-0.32977268, 0.0010599, -0.01973337,1.0},
  {-0.32977268, 0.0010599, 0,1.0},
  // f 722/565/527 721/563/525 270/564/526
  {-0.33968067, 0.00478395, -0.01973337,1.0},
  {-0.33544296, 0.0029247, -0.01973337,1.0},
  {-0.33544296, 0.0029247, 0,1.0},
  // f 723/567/529 722/565/527 271/566/528
  {-0.34347504, 0.00692071, -0.01973337,1.0},
  {-0.33968067, 0.00478395, -0.01973337,1.0},
  {-0.33968067, 0.00478395, 0,1.0},
  // f 724/569/531 723/567/529 272/568/530
  {-0.34619153, 0.00882991, -0.01973337,1.0},
  {-0.34347504, 0.00692071, -0.01973337,1.0},
  {-0.34347504, 0.00692071, 0,1.0},
  // f 725/571/533 724/569/531 273/570/532
  {-0.34744334, 0.00985111, -0.01973337,1.0},
  {-0.34619153, 0.00882991, -0.01973337,1.0},
  {-0.34619153, 0.00882991, 0,1.0},
  // f 726/573/535 725/571/533 274/572/534
  {-0.34918806, 0.01146617, -0.01973337,1.0},
  {-0.34744334, 0.00985111, -0.01973337,1.0},
  {-0.34744334, 0.00985111, 0,1.0},
  // f 727/576/537 726/575/535 275/574/536
  {-0.352837, 0.01549848, -0.01973337,1.0},
  {-0.34918806, 0.01146617, -0.01973337,1.0},
  {-0.34918806, 0.01146617, 0,1.0},
  // f 728/578/539 727/576/537 276/577/538
  {-0.35558888, 0.01918923, -0.01973337,1.0},
  {-0.352837, 0.01549848, -0.01973337,1.0},
  {-0.352837, 0.01549848, 0,1.0},
  // f 729/580/541 728/578/539 277/579/540
  {-0.35799384, 0.02309645, -0.01973337,1.0},
  {-0.35558888, 0.01918923, -0.01973337,1.0},
  {-0.35558888, 0.01918923, 0,1.0},
  // f 730/582/543 729/580/541 278/581/542
  {-0.36005208, 0.0272201, -0.01973337,1.0},
  {-0.35799384, 0.02309645, -0.01973337,1.0},
  {-0.35799384, 0.02309645, 0,1.0},
  // f 731/584/545 730/582/543 279/583/544
  {-0.3620148, 0.0323046, -0.01973337,1.0},
  {-0.36005208, 0.0272201, -0.01973337,1.0},
  {-0.36005208, 0.0272201, 0,1.0},
  // f 732/586/547 731/584/545 280/585/546
  {-0.36350545, 0.03768371, -0.01973337,1.0},
  {-0.3620148, 0.0323046, -0.01973337,1.0},
  {-0.3620148, 0.0323046, 0,1.0},
  // f 733/588/549 732/586/547 281/587/548
  {-0.36428103, 0.04170631, -0.01973337,1.0},
  {-0.36350545, 0.03768371, -0.01973337,1.0},
  {-0.36350545, 0.03768371, 0,1.0},
  // f 530/590/551 733/588/549 282/589/550
  {-0.36489382, 0.04673185, -0.01973337,1.0},
  {-0.36428103, 0.04170631, -0.01973337,1.0},
  {-0.36428103, 0.04170631, 0,1.0},
  // f 735/593/1 734/592/1 283/591/1
  {-0.17112531, 0.14125545, -0.01973337,1.0},
  {-0.21477383, 0.14125545, -0.01973337,1.0},
  {-0.21477383, 0.14125545, 0,1.0},
  // f 736/596/71 735/595/71 284/594/71
  {-0.17112531, 0.03478392, -0.01973337,1.0},
  {-0.17112531, 0.14125545, -0.01973337,1.0},
  {-0.17112531, 0.14125545, 0,1.0},
  // f 737/599/1 736/598/1 285/597/1
  {-0.10300279, 0.03478392, -0.01973337,1.0},
  {-0.17112531, 0.03478392, -0.01973337,1.0},
  {-0.17112531, 0.03478392, 0,1.0},
  // f 738/602/71 737/601/71 286/600/71
  {-0.10300279, 0, -0.01973337,1.0},
  {-0.10300279, 0.03478392, -0.01973337,1.0},
  {-0.10300279, 0.03478392, 0,1.0},
  // f 739/605/70 738/604/70 287/603/70
  {-0.21477383, 0, -0.01973337,1.0},
  {-0.10300279, 0, -0.01973337,1.0},
  {-0.10300279, 0, 0,1.0},
  // f 734/608/72 739/607/72 288/606/72
  {-0.21477383, 0.14125545, -0.01973337,1.0},
  {-0.21477383, 0, -0.01973337,1.0},
  {-0.21477383, 0, 0,1.0},
  // f 741/611/71 740/610/71 289/609/71
  {-0.00057813, 0, -0.01973337,1.0},
  {-0.00057813, 0.1436643, -0.01973337,1.0},
  {-0.00057813, 0.1436643, 0,1.0},
  // f 742/614/70 741/613/70 290/612/70
  {-0.04027612, 0, -0.01973337,1.0},
  {-0.00057813, 0, -0.01973337,1.0},
  {-0.00057813, 0, 0,1.0},
  // f 743/617/72 742/616/72 291/615/72
  {-0.04027612, 0.0941382, -0.01973337,1.0},
  {-0.04027612, 0, -0.01973337,1.0},
  {-0.04027612, 0, 0,1.0},
  // f 744/620/554 743/619/553 292/618/552
  {-0.05012761, 0.08727793, -0.01973337,1.0},
  {-0.04027612, 0.0941382, -0.01973337,1.0},
  {-0.04027612, 0.0941382, 0,1.0},
  // f 745/622/556 744/620/554 293/621/555
  {-0.05456021, 0.08462132, -0.01973337,1.0},
  {-0.05012761, 0.08727793, -0.01973337,1.0},
  {-0.05012761, 0.08727793, 0,1.0},
  // f 746/624/558 745/622/556 294/623/557
  {-0.05892068, 0.0822866, -0.01973337,1.0},
  {-0.05456021, 0.08462132, -0.01973337,1.0},
  {-0.05456021, 0.08462132, 0,1.0},
  // f 747/626/560 746/624/558 295/625/559
  {-0.06907402, 0.0778543, -0.01973337,1.0},
  {-0.05892068, 0.0822866, -0.01973337,1.0},
  {-0.05892068, 0.0822866, 0,1.0},
  // f 748/628/562 747/626/560 296/627/561
  {-0.08151577, 0.07361472, -0.01973337,1.0},
  {-0.06907402, 0.0778543, -0.01973337,1.0},
  {-0.06907402, 0.0778543, 0,1.0},
  // f 749/631/72 748/630/72 297/629/72
  {-0.08151577, 0.10579707, -0.01973337,1.0},
  {-0.08151577, 0.07361472, -0.01973337,1.0},
  {-0.08151577, 0.07361472, 0,1.0},
  // f 750/634/564 749/633/563 298/632/563
  {-0.06960668, 0.11019515, -0.01973337,1.0},
  {-0.08151577, 0.10579707, -0.01973337,1.0},
  {-0.08151577, 0.10579707, 0,1.0},
  // f 751/636/566 750/634/564 299/635/565
  {-0.06128185, 0.11416088, -0.01973337,1.0},
  {-0.06960668, 0.11019515, -0.01973337,1.0},
  {-0.06960668, 0.11019515, 0,1.0},
  // f 752/638/568 751/636/566 300/637/567
  {-0.05518406, 0.11779124, -0.01973337,1.0},
  {-0.06128185, 0.11416088, -0.01973337,1.0},
  {-0.06128185, 0.11416088, 0,1.0},
  // f 753/640/570 752/638/568 301/639/569
  {-0.05269908, 0.11952468, -0.01973337,1.0},
  {-0.05518406, 0.11779124, -0.01973337,1.0},
  {-0.05518406, 0.11779124, 0,1.0},
  // f 754/642/572 753/640/570 302/641/571
  {-0.05039332, 0.12131011, -0.01973337,1.0},
  {-0.05269908, 0.11952468, -0.01973337,1.0},
  {-0.05269908, 0.11952468, 0,1.0},
  // f 755/644/574 754/642/572 303/643/573
  {-0.04615372, 0.12510261, -0.01973337,1.0},
  {-0.05039332, 0.12131011, -0.01973337,1.0},
  {-0.05039332, 0.12131011, 0,1.0},
  // f 756/647/576 755/646/574 304/645/575
  {-0.04229955, 0.12923431, -0.01973337,1.0},
  {-0.04615372, 0.12510261, -0.01973337,1.0},
  {-0.04615372, 0.12510261, 0,1.0},
  // f 757/649/578 756/647/576 305/648/577
  {-0.03883079, 0.13370511, -0.01973337,1.0},
  {-0.04229955, 0.12923431, -0.01973337,1.0},
  {-0.04229955, 0.12923431, 0,1.0},
  // f 758/651/580 757/649/578 306/650/579
  {-0.03574746, 0.13851516, -0.01973337,1.0},
  {-0.03883079, 0.13370511, -0.01973337,1.0},
  {-0.03883079, 0.13370511, 0,1.0},
  // f 759/653/582 758/651/580 307/652/581
  {-0.03304954, 0.1436643, -0.01973337,1.0},
  {-0.03574746, 0.13851516, -0.01973337,1.0},
  {-0.03574746, 0.13851516, 0,1.0},
  // f 740/656/1 759/655/1 308/654/1
  {-0.00057813, 0.1436643, -0.01973337,1.0},
  {-0.03304954, 0.1436643, -0.01973337,1.0},
  {-0.03304954, 0.1436643, 0,1.0},
  // f 761/659/72 760/658/72 309/657/72
  {0.11947938, 0.08055224, -0.01973337,1.0},
  {0.11947938, 0.05116415, -0.01973337,1.0},
  {0.11947938, 0.05116415, 0,1.0},
  // f 762/662/1 761/661/1 310/660/1
  {0.18692741, 0.08055224, -0.01973337,1.0},
  {0.11947938, 0.08055224, -0.01973337,1.0},
  {0.11947938, 0.08055224, 0,1.0},
  // f 763/665/71 762/664/71 311/663/71
  {0.18692741, 0.02033077, -0.01973337,1.0},
  {0.18692741, 0.08055224, -0.01973337,1.0},
  {0.18692741, 0.08055224, 0,1.0},
  // f 764/668/585 763/667/584 312/666/583
  {0.17215054, 0.01111929, -0.01973337,1.0},
  {0.18692741, 0.02033077, -0.01973337,1.0},
  {0.18692741, 0.02033077, 0,1.0},
  // f 765/670/587 764/668/585 313/669/586
  {0.16200872, 0.0059812, -0.01973337,1.0},
  {0.17215054, 0.01111929, -0.01973337,1.0},
  {0.17215054, 0.01111929, 0,1.0},
  // f 766/672/589 765/670/587 314/671/588
  {0.15755193, 0.00410277, -0.01973337,1.0},
  {0.16200872, 0.0059812, -0.01973337,1.0},
  {0.16200872, 0.0059812, 0,1.0},
  // f 767/674/591 766/672/589 315/673/590
  {0.15267344, 0.00236068, -0.01973337,1.0},
  {0.15755193, 0.00410277, -0.01973337,1.0},
  {0.15755193, 0.00410277, 0,1.0},
  // f 768/676/593 767/674/591 316/675/592
  {0.1452087, 0.00034603, -0.01973337,1.0},
  {0.15267344, 0.00236068, -0.01973337,1.0},
  {0.15267344, 0.00236068, 0,1.0},
  // f 769/678/595 768/676/593 317/677/594
  {0.13569312, -0.00130996, -0.01973337,1.0},
  {0.1452087, 0.00034603, -0.01973337,1.0},
  {0.1452087, 0.00034603, 0,1.0},
  // f 770/680/597 769/678/595 318/679/596
  {0.12684548, -0.00213413, -0.01973337,1.0},
  {0.13569312, -0.00130996, -0.01973337,1.0},
  {0.13569312, -0.00130996, 0,1.0},
  // f 771/682/599 770/680/597 319/681/598
  {0.11735957, -0.00240886, -0.01973337,1.0},
  {0.12684548, -0.00213413, -0.01973337,1.0},
  {0.12684548, -0.00213413, 0,1.0},
  // f 772/684/601 771/682/599 320/683/600
  {0.1058239, -0.00191491, -0.01973337,1.0},
  {0.11735957, -0.00240886, -0.01973337,1.0},
  {0.11735957, -0.00240886, 0,1.0},
  // f 773/686/603 772/684/601 321/685/602
  {0.09702906, -0.00074863, -0.01973337,1.0},
  {0.1058239, -0.00191491, -0.01973337,1.0},
  {0.1058239, -0.00191491, 0,1.0},
  // f 774/688/605 773/686/603 322/687/604
  {0.08974966, 0.00088758, -0.01973337,1.0},
  {0.09702906, -0.00074863, -0.01973337,1.0},
  {0.09702906, -0.00074863, 0,1.0},
  // f 775/690/607 774/688/605 323/689/606
  {0.08378498, 0.0028085, -0.01973337,1.0},
  {0.08974966, 0.00088758, -0.01973337,1.0},
  {0.08974966, 0.00088758, 0,1.0},
  // f 776/692/609 775/690/607 324/691/608
  {0.08029877, 0.00423204, -0.01973337,1.0},
  {0.08378498, 0.0028085, -0.01973337,1.0},
  {0.08378498, 0.0028085, 0,1.0},
  // f 777/694/611 776/692/609 325/693/610
  {0.07636081, 0.00616668, -0.01973337,1.0},
  {0.08029877, 0.00423204, -0.01973337,1.0},
  {0.08029877, 0.00423204, 0,1.0},
  // f 778/696/613 777/694/611 326/695/612
  {0.07147124, 0.00912545, -0.01973337,1.0},
  {0.07636081, 0.00616668, -0.01973337,1.0},
  {0.07636081, 0.00616668, 0,1.0},
  // f 779/698/615 778/696/613 327/697/614
  {0.06694677, 0.01251342, -0.01973337,1.0},
  {0.07147124, 0.00912545, -0.01973337,1.0},
  {0.07147124, 0.00912545, 0,1.0},
  // f 780/700/617 779/698/615 328/699/616
  {0.06278734, 0.0163306, -0.01973337,1.0},
  {0.06694677, 0.01251342, -0.01973337,1.0},
  {0.06694677, 0.01251342, 0,1.0},
  // f 781/703/619 780/702/617 329/701/618
  {0.05899298, 0.02057697, -0.01973337,1.0},
  {0.06278734, 0.0163306, -0.01973337,1.0},
  {0.06278734, 0.0163306, 0,1.0},
  // f 782/705/621 781/703/619 330/704/620
  {0.0555637, 0.02525254, -0.01973337,1.0},
  {0.05899298, 0.02057697, -0.01973337,1.0},
  {0.05899298, 0.02057697, 0,1.0},
  // f 783/707/623 782/705/621 331/706/622
  {0.05360577, 0.02839273, -0.01973337,1.0},
  {0.0555637, 0.02525254, -0.01973337,1.0},
  {0.0555637, 0.02525254, 0,1.0},
  // f 784/709/625 783/707/623 332/708/624
  {0.05179045, 0.03170058, -0.01973337,1.0},
  {0.05360577, 0.02839273, -0.01973337,1.0},
  {0.05360577, 0.02839273, 0,1.0},
  // f 785/711/627 784/709/625 333/710/626
  {0.04893376, 0.03796792, -0.01973337,1.0},
  {0.05179045, 0.03170058, -0.01973337,1.0},
  {0.05179045, 0.03170058, 0,1.0},
  // f 786/713/629 785/711/627 334/712/628
  {0.04664213, 0.04455995, -0.01973337,1.0},
  {0.04893376, 0.03796792, -0.01973337,1.0},
  {0.04893376, 0.03796792, 0,1.0},
  // f 787/715/631 786/713/629 335/714/630
  {0.04491555, 0.05147664, -0.01973337,1.0},
  {0.04664213, 0.04455995, -0.01973337,1.0},
  {0.04664213, 0.04455995, 0,1.0},
  // f 788/717/633 787/715/631 336/716/632
  {0.04365986, 0.05954267, -0.01973337,1.0},
  {0.04491555, 0.05147664, -0.01973337,1.0},
  {0.04491555, 0.05147664, 0,1.0},
  // f 789/719/635 788/717/633 337/718/634
  {0.0432413, 0.06457467, -0.01973337,1.0},
  {0.04365986, 0.05954267, -0.01973337,1.0},
  {0.04365986, 0.05954267, 0,1.0},
  // f 790/721/637 789/719/635 338/720/636
  {0.04307039, 0.07062773, -0.01973337,1.0},
  {0.0432413, 0.06457467, -0.01973337,1.0},
  {0.0432413, 0.06457467, 0,1.0},
  // f 791/723/639 790/721/637 339/722/638
  {0.04353208, 0.08051129, -0.01973337,1.0},
  {0.04307039, 0.07062773, -0.01973337,1.0},
  {0.04307039, 0.07062773, 0,1.0},
  // f 792/725/641 791/723/639 340/724/640
  {0.04475309, 0.08899045, -0.01973337,1.0},
  {0.04353208, 0.08051129, -0.01973337,1.0},
  {0.04353208, 0.08051129, 0,1.0},
  // f 793/727/643 792/725/641 341/726/642
  {0.04650446, 0.09620979, -0.01973337,1.0},
  {0.04475309, 0.08899045, -0.01973337,1.0},
  {0.04475309, 0.08899045, 0,1.0},
  // f 794/729/645 793/727/643 342/728/644
  {0.04858016, 0.10229938, -0.01973337,1.0},
  {0.04650446, 0.09620979, -0.01973337,1.0},
  {0.04650446, 0.09620979, 0,1.0},
  // f 795/731/647 794/729/645 343/730/646
  {0.05045747, 0.10666426, -0.01973337,1.0},
  {0.04858016, 0.10229938, -0.01973337,1.0},
  {0.04858016, 0.10229938, 0,1.0},
  // f 796/733/649 795/731/647 344/732/648
  {0.05260947, 0.11085567, -0.01973337,1.0},
  {0.05045747, 0.10666426, -0.01973337,1.0},
  {0.05045747, 0.10666426, 0,1.0},
  // f 797/735/651 796/733/649 345/734/650
  {0.05588891, 0.11611768, -0.01973337,1.0},
  {0.05260947, 0.11085567, -0.01973337,1.0},
  {0.05260947, 0.11085567, 0,1.0},
  // f 798/737/653 797/735/651 346/736/652
  {0.05913073, 0.12037592, -0.01973337,1.0},
  {0.05588891, 0.11611768, -0.01973337,1.0},
  {0.05588891, 0.11611768, 0,1.0},
  // f 799/739/655 798/737/653 347/738/654
  {0.06272005, 0.12431127, -0.01973337,1.0},
  {0.05913073, 0.12037592, -0.01973337,1.0},
  {0.05913073, 0.12037592, 0,1.0},
  // f 800/742/657 799/741/655 348/740/656
  {0.06665686, 0.12792365, -0.01973337,1.0},
  {0.06272005, 0.12431127, -0.01973337,1.0},
  {0.06272005, 0.12431127, 0,1.0},
  // f 801/744/659 800/742/657 349/743/658
  {0.07158156, 0.13165666, -0.01973337,1.0},
  {0.06665686, 0.12792365, -0.01973337,1.0},
  {0.06665686, 0.12792365, 0,1.0},
  // f 802/746/661 801/744/659 350/745/660
  {0.07557292, 0.13417959, -0.01973337,1.0},
  {0.07158156, 0.13165666, -0.01973337,1.0},
  {0.07158156, 0.13165666, 0,1.0},
  // f 803/748/663 802/746/661 351/747/662
  {0.08055224, 0.13682316, -0.01973337,1.0},
  {0.07557292, 0.13417959, -0.01973337,1.0},
  {0.07557292, 0.13417959, 0,1.0},
  // f 804/750/665 803/748/663 352/749/664
  {0.08540048, 0.1388372, -0.01973337,1.0},
  {0.08055224, 0.13682316, -0.01973337,1.0},
  {0.08055224, 0.13682316, 0,1.0},
  // f 805/752/667 804/750/665 353/751/666
  {0.09146211, 0.14068431, -0.01973337,1.0},
  {0.08540048, 0.1388372, -0.01973337,1.0},
  {0.08540048, 0.1388372, 0,1.0},
  // f 806/754/669 805/752/667 354/753/668
  {0.09895155, 0.14221673, -0.01973337,1.0},
  {0.09146211, 0.14068431, -0.01973337,1.0},
  {0.09146211, 0.14068431, 0,1.0},
  // f 807/756/671 806/754/669 355/755/670
  {0.10810695, 0.14327027, -0.01973337,1.0},
  {0.09895155, 0.14221673, -0.01973337,1.0},
  {0.09895155, 0.14221673, 0,1.0},
  // f 808/758/673 807/756/671 356/757/672
  {0.11919031, 0.1436643, -0.01973337,1.0},
  {0.10810695, 0.14327027, -0.01973337,1.0},
  {0.10810695, 0.14327027, 0,1.0},
  // f 809/760/675 808/758/673 357/759/674
  {0.13603304, 0.14298595, -0.01973337,1.0},
  {0.11919031, 0.1436643, -0.01973337,1.0},
  {0.11919031, 0.1436643, 0,1.0},
  // f 810/762/677 809/760/675 358/761/676
  {0.1455756, 0.14170393, -0.01973337,1.0},
  {0.13603304, 0.14298595, -0.01973337,1.0},
  {0.13603304, 0.14298595, 0,1.0},
  // f 811/764/679 810/762/677 359/763/678
  {0.15019752, 0.14067285, -0.01973337,1.0},
  {0.1455756, 0.14170393, -0.01973337,1.0},
  {0.1455756, 0.14170393, 0,1.0},
  // f 812/766/681 811/764/679 360/765/680
  {0.15421513, 0.13942471, -0.01973337,1.0},
  {0.15019752, 0.14067285, -0.01973337,1.0},
  {0.15019752, 0.14067285, 0,1.0},
  // f 813/768/683 812/766/681 361/767/682
  {0.15870138, 0.13754196, -0.01973337,1.0},
  {0.15421513, 0.13942471, -0.01973337,1.0},
  {0.15421513, 0.13942471, 0,1.0},
  // f 814/770/685 813/768/683 362/769/684
  {0.16246878, 0.13552782, -0.01973337,1.0},
  {0.15870138, 0.13754196, -0.01973337,1.0},
  {0.15870138, 0.13754196, 0,1.0},
  // f 815/772/687 814/770/685 363/771/686
  {0.16598019, 0.13321085, -0.01973337,1.0},
  {0.16246878, 0.13552782, -0.01973337,1.0},
  {0.16246878, 0.13552782, 0,1.0},
  // f 816/774/689 815/772/687 364/773/688
  {0.16923565, 0.13059105, -0.01973337,1.0},
  {0.16598019, 0.13321085, -0.01973337,1.0},
  {0.16598019, 0.13321085, 0,1.0},
  // f 817/776/691 816/774/689 365/775/690
  {0.17126371, 0.12867633, -0.01973337,1.0},
  {0.16923565, 0.13059105, -0.01973337,1.0},
  {0.16923565, 0.13059105, 0,1.0},
  // f 818/779/693 817/778/691 366/777/692
  {0.17348598, 0.12627235, -0.01973337,1.0},
  {0.17126371, 0.12867633, -0.01973337,1.0},
  {0.17126371, 0.12867633, 0,1.0},
  // f 819/781/695 818/779/693 367/780/694
  {0.17612036, 0.12290923, -0.01973337,1.0},
  {0.17348598, 0.12627235, -0.01973337,1.0},
  {0.17348598, 0.12627235, 0,1.0},
  // f 820/783/697 819/781/695 368/782/696
  {0.17875814, 0.11880992, -0.01973337,1.0},
  {0.17612036, 0.12290923, -0.01973337,1.0},
  {0.17612036, 0.12290923, 0,1.0},
  // f 821/785/699 820/783/697 369/784/698
  {0.18109146, 0.11432907, -0.01973337,1.0},
  {0.17875814, 0.11880992, -0.01973337,1.0},
  {0.17875814, 0.11880992, 0,1.0},
  // f 822/787/701 821/785/699 370/786/700
  {0.1831203, 0.10946664, -0.01973337,1.0},
  {0.18109146, 0.11432907, -0.01973337,1.0},
  {0.18109146, 0.11432907, 0,1.0},
  // f 823/789/703 822/787/701 371/788/702
  {0.18500036, 0.10367728, -0.01973337,1.0},
  {0.1831203, 0.10946664, -0.01973337,1.0},
  {0.1831203, 0.10946664, 0,1.0},
  // f 824/792/704 823/791/704 372/790/704
  {0.14289349, 0.09616163, -0.01973337,1.0},
  {0.18500036, 0.10367728, -0.01973337,1.0},
  {0.18500036, 0.10367728, 0,1.0},
  // f 825/795/707 824/794/706 373/793/705
  {0.14196849, 0.09862459, -0.01973337,1.0},
  {0.14289349, 0.09616163, -0.01973337,1.0},
  {0.14289349, 0.09616163, 0,1.0},
  // f 826/797/709 825/795/707 374/796/708
  {0.14085847, 0.10089514, -0.01973337,1.0},
  {0.14196849, 0.09862459, -0.01973337,1.0},
  {0.14196849, 0.09862459, 0,1.0},
  // f 827/799/711 826/797/709 375/798/710
  {0.13973548, 0.10272409, -0.01973337,1.0},
  {0.14085847, 0.10089514, -0.01973337,1.0},
  {0.14085847, 0.10089514, 0,1.0},
  // f 828/801/713 827/799/711 376/800/712
  {0.1384708, 0.10440568, -0.01973337,1.0},
  {0.13973548, 0.10272409, -0.01973337,1.0},
  {0.13973548, 0.10272409, 0,1.0},
  // f 829/803/715 828/801/713 377/802/714
  {0.13706453, 0.10594001, -0.01973337,1.0},
  {0.1384708, 0.10440568, -0.01973337,1.0},
  {0.1384708, 0.10440568, 0,1.0},
  // f 830/806/717 829/805/715 378/804/716
  {0.13574637, 0.1071379, -0.01973337,1.0},
  {0.13706453, 0.10594001, -0.01973337,1.0},
  {0.13706453, 0.10594001, 0,1.0},
  // f 831/808/719 830/806/717 379/807/718
  {0.13407704, 0.10839865, -0.01973337,1.0},
  {0.13574637, 0.1071379, -0.01973337,1.0},
  {0.13574637, 0.1071379, 0,1.0},
  // f 832/810/721 831/808/719 380/809/720
  {0.13226987, 0.10950262, -0.01973337,1.0},
  {0.13407704, 0.10839865, -0.01973337,1.0},
  {0.13407704, 0.10839865, 0,1.0},
  // f 833/812/723 832/810/721 381/811/722
  {0.13032858, 0.1104404, -0.01973337,1.0},
  {0.13226987, 0.10950262, -0.01973337,1.0},
  {0.13226987, 0.10950262, 0,1.0},
  // f 834/814/725 833/812/723 382/813/724
  {0.12825319, 0.11121202, -0.01973337,1.0},
  {0.13032858, 0.1104404, -0.01973337,1.0},
  {0.13032858, 0.1104404, 0,1.0},
  // f 835/816/727 834/814/725 383/815/726
  {0.12571715, 0.11189037, -0.01973337,1.0},
  {0.12825319, 0.11121202, -0.01973337,1.0},
  {0.12825319, 0.11121202, 0,1.0},
  // f 836/818/729 835/816/727 384/817/728
  {0.12300596, 0.1123516, -0.01973337,1.0},
  {0.12571715, 0.11189037, -0.01973337,1.0},
  {0.12571715, 0.11189037, 0,1.0},
  // f 837/820/731 836/818/729 385/819/730
  {0.12085765, 0.11255514, -0.01973337,1.0},
  {0.12300596, 0.1123516, -0.01973337,1.0},
  {0.12300596, 0.1123516, 0,1.0},
  // f 838/822/733 837/820/731 386/821/732
  {0.11822677, 0.11263821, -0.01973337,1.0},
  {0.12085765, 0.11255514, -0.01973337,1.0},
  {0.12085765, 0.11255514, 0,1.0},
  // f 839/824/735 838/822/733 387/823/734
  {0.11486509, 0.11249463, -0.01973337,1.0},
  {0.11822677, 0.11263821, -0.01973337,1.0},
  {0.11822677, 0.11263821, 0,1.0},
  // f 840/826/737 839/824/735 388/825/736
  {0.11167126, 0.11206378, -0.01973337,1.0},
  {0.11486509, 0.11249463, -0.01973337,1.0},
  {0.11486509, 0.11249463, 0,1.0},
  // f 841/828/739 840/826/737 389/827/738
  {0.10864536, 0.11134578, -0.01973337,1.0},
  {0.11167126, 0.11206378, -0.01973337,1.0},
  {0.11167126, 0.11206378, 0,1.0},
  // f 842/830/741 841/828/739 390/829/740
  {0.10625204, 0.11052801, -0.01973337,1.0},
  {0.10864536, 0.11134578, -0.01973337,1.0},
  {0.10864536, 0.11134578, 0,1.0},
  // f 843/832/743 842/830/741 391/831/742
  {0.10397528, 0.10951076, -0.01973337,1.0},
  {0.10625204, 0.11052801, -0.01973337,1.0},
  {0.10625204, 0.11052801, 0,1.0},
  // f 844/834/745 843/832/743 392/833/744
  {0.1018151, 0.10829411, -0.01973337,1.0},
  {0.10397528, 0.10951076, -0.01973337,1.0},
  {0.10397528, 0.10951076, 0,1.0},
  // f 845/836/747 844/834/745 393/835/746
  {0.09977153, 0.106878, -0.01973337,1.0},
  {0.1018151, 0.10829411, -0.01973337,1.0},
  {0.1018151, 0.10829411, 0,1.0},
  // f 846/838/749 845/836/747 394/837/748
  {0.09784453, 0.10526244, -0.01973337,1.0},
  {0.09977153, 0.106878, -0.01973337,1.0},
  {0.09977153, 0.106878, 0,1.0},
  // f 847/840/751 846/838/749 395/839/750
  {0.09674428, 0.10419735, -0.01973337,1.0},
  {0.09784453, 0.10526244, -0.01973337,1.0},
  {0.09784453, 0.10526244, 0,1.0},
  // f 848/843/753 847/842/751 396/841/752
  {0.09534263, 0.10266554, -0.01973337,1.0},
  {0.09674428, 0.10419735, -0.01973337,1.0},
  {0.09674428, 0.10419735, 0,1.0},
  // f 849/845/755 848/843/753 397/844/754
  {0.09372242, 0.1005549, -0.01973337,1.0},
  {0.09534263, 0.10266554, -0.01973337,1.0},
  {0.09534263, 0.10266554, 0,1.0},
  // f 850/847/757 849/845/755 398/846/756
  {0.09227277, 0.09821206, -0.01973337,1.0},
  {0.09372242, 0.1005549, -0.01973337,1.0},
  {0.09372242, 0.1005549, 0,1.0},
  // f 851/849/759 850/847/757 399/848/758
  {0.09099369, 0.09563697, -0.01973337,1.0},
  {0.09227277, 0.09821206, -0.01973337,1.0},
  {0.09227277, 0.09821206, 0,1.0},
  // f 852/851/761 851/849/759 400/850/760
  {0.08968385, 0.09224036, -0.01973337,1.0},
  {0.09099369, 0.09563697, -0.01973337,1.0},
  {0.09099369, 0.09563697, 0,1.0},
  // f 853/853/763 852/851/761 401/852/762
  {0.08846615, 0.08785504, -0.01973337,1.0},
  {0.08968385, 0.09224036, -0.01973337,1.0},
  {0.08968385, 0.09224036, 0,1.0},
  // f 854/855/765 853/853/763 402/854/764
  {0.0874838, 0.08228591, -0.01973337,1.0},
  {0.08846615, 0.08785504, -0.01973337,1.0},
  {0.08846615, 0.08785504, 0,1.0},
  // f 855/857/767 854/855/765 403/856/766
  {0.08698238, 0.07692527, -0.01973337,1.0},
  {0.0874838, 0.08228591, -0.01973337,1.0},
  {0.0874838, 0.08228591, 0,1.0},
  // f 856/859/769 855/857/767 404/858/768
  {0.08681526, 0.0711095, -0.01973337,1.0},
  {0.08698238, 0.07692527, -0.01973337,1.0},
  {0.08698238, 0.07692527, 0,1.0},
  // f 857/861/771 856/859/769 405/860/770
  {0.08723264, 0.06165352, -0.01973337,1.0},
  {0.08681526, 0.0711095, -0.01973337,1.0},
  {0.08681526, 0.0711095, 0,1.0},
  // f 858/863/773 857/861/771 406/862/772
  {0.08819505, 0.0548603, -0.01973337,1.0},
  {0.08723264, 0.06165352, -0.01973337,1.0},
  {0.08723264, 0.06165352, 0,1.0},
  // f 859/865/775 858/863/773 407/864/774
  {0.08932994, 0.05016376, -0.01973337,1.0},
  {0.08819505, 0.0548603, -0.01973337,1.0},
  {0.08819505, 0.0548603, 0,1.0},
  // f 860/867/777 859/865/775 408/866/776
  {0.09057174, 0.046547, -0.01973337,1.0},
  {0.08932994, 0.05016376, -0.01973337,1.0},
  {0.08932994, 0.05016376, 0,1.0},
  // f 861/869/779 860/867/777 409/868/778
  {0.09179632, 0.04382134, -0.01973337,1.0},
  {0.09057174, 0.046547, -0.01973337,1.0},
  {0.09057174, 0.046547, 0,1.0},
  // f 862/871/781 861/869/779 410/870/780
  {0.09319337, 0.04135774, -0.01973337,1.0},
  {0.09179632, 0.04382134, -0.01973337,1.0},
  {0.09179632, 0.04382134, 0,1.0},
  // f 863/873/783 862/871/781 411/872/782
  {0.09411437, 0.04000539, -0.01973337,1.0},
  {0.09319337, 0.04135774, -0.01973337,1.0},
  {0.09319337, 0.04135774, 0,1.0},
  // f 864/875/785 863/873/783 412/874/784
  {0.09543897, 0.03834902, -0.01973337,1.0},
  {0.09411437, 0.04000539, -0.01973337,1.0},
  {0.09411437, 0.04000539, 0,1.0},
  // f 865/877/787 864/875/785 413/876/786
  {0.09723166, 0.03648167, -0.01973337,1.0},
  {0.09543897, 0.03834902, -0.01973337,1.0},
  {0.09543897, 0.03834902, 0,1.0},
  // f 866/880/789 865/879/787 414/878/788
  {0.09916019, 0.0348109, -0.01973337,1.0},
  {0.09723166, 0.03648167, -0.01973337,1.0},
  {0.09723166, 0.03648167, 0,1.0},
  // f 867/882/791 866/880/789 415/881/790
  {0.10122457, 0.03333668, -0.01973337,1.0},
  {0.09916019, 0.0348109, -0.01973337,1.0},
  {0.09916019, 0.0348109, 0,1.0},
  // f 868/884/793 867/882/791 416/883/792
  {0.10342481, 0.03205901, -0.01973337,1.0},
  {0.10122457, 0.03333668, -0.01973337,1.0},
  {0.10122457, 0.03333668, 0,1.0},
  // f 869/886/795 868/884/793 417/885/794
  {0.10576094, 0.03097792, -0.01973337,1.0},
  {0.10342481, 0.03205901, -0.01973337,1.0},
  {0.10342481, 0.03205901, 0,1.0},
  // f 870/888/797 869/886/795 418/887/796
  {0.10823289, 0.03009339, -0.01973337,1.0},
  {0.10576094, 0.03097792, -0.01973337,1.0},
  {0.10576094, 0.03097792, 0,1.0},
  // f 871/890/799 870/888/797 419/889/798
  {0.1113786, 0.0292914, -0.01973337,1.0},
  {0.10823289, 0.03009339, -0.01973337,1.0},
  {0.10823289, 0.03009339, 0,1.0},
  // f 872/892/801 871/890/799 420/891/800
  {0.11471993, 0.02877249, -0.01973337,1.0},
  {0.1113786, 0.0292914, -0.01973337,1.0},
  {0.1113786, 0.0292914, 0,1.0},
  // f 873/894/803 872/892/801 421/893/802
  {0.11705618, 0.02858379, -0.01973337,1.0},
  {0.11471993, 0.02877249, -0.01973337,1.0},
  {0.11471993, 0.02877249, 0,1.0},
  // f 874/896/805 873/894/803 422/895/804
  {0.11947938, 0.02852089, -0.01973337,1.0},
  {0.11705618, 0.02858379, -0.01973337,1.0},
  {0.11705618, 0.02858379, 0,1.0},
  // f 875/898/807 874/896/805 423/897/806
  {0.12381244, 0.02871166, -0.01973337,1.0},
  {0.11947938, 0.02852089, -0.01973337,1.0},
  {0.11947938, 0.02852089, 0,1.0},
  // f 876/900/809 875/898/807 424/899/808
  {0.12802409, 0.02928401, -0.01973337,1.0},
  {0.12381244, 0.02871166, -0.01973337,1.0},
  {0.12381244, 0.02871166, 0,1.0},
  // f 877/902/811 876/900/809 425/901/810
  {0.13076441, 0.02987755, -0.01973337,1.0},
  {0.12802409, 0.02928401, -0.01973337,1.0},
  {0.12802409, 0.02928401, 0,1.0},
  // f 878/904/813 877/902/811 426/903/812
  {0.13345076, 0.03064069, -0.01973337,1.0},
  {0.13076441, 0.02987755, -0.01973337,1.0},
  {0.13076441, 0.02987755, 0,1.0},
  // f 879/906/815 878/904/813 427/905/814
  {0.13790309, 0.03230322, -0.01973337,1.0},
  {0.13345076, 0.03064069, -0.01973337,1.0},
  {0.13345076, 0.03064069, 0,1.0},
  // f 880/908/817 879/906/815 428/907/816
  {0.14306615, 0.03473994, -0.01973337,1.0},
  {0.13790309, 0.03230322, -0.01973337,1.0},
  {0.13790309, 0.03230322, 0,1.0},
  // f 881/910/819 880/908/817 429/909/818
  {0.14867473, 0.03786726, -0.01973337,1.0},
  {0.14306615, 0.03473994, -0.01973337,1.0},
  {0.14306615, 0.03473994, 0,1.0},
  // f 882/913/72 881/912/72 430/911/72
  {0.14867473, 0.05116415, -0.01973337,1.0},
  {0.14867473, 0.03786726, -0.01973337,1.0},
  {0.14867473, 0.03786726, 0,1.0},
  // f 760/916/70 882/915/70 431/914/70
  {0.11947938, 0.05116415, -0.01973337,1.0},
  {0.14867473, 0.05116415, -0.01973337,1.0},
  {0.14867473, 0.05116415, 0,1.0},
  // f 884/919/1 883/918/1 432/917/1
  {0.25668794, 0.14125545, -0.01973337,1.0},
  {0.21303943, 0.14125545, -0.01973337,1.0},
  {0.21303943, 0.14125545, 0,1.0},
  // f 885/922/71 884/921/71 433/920/71
  {0.25668794, 0.09182569, -0.01973337,1.0},
  {0.25668794, 0.14125545, -0.01973337,1.0},
  {0.25668794, 0.14125545, 0,1.0},
  // f 886/925/1 885/924/1 434/923/1
  {0.30438331, 0.09182569, -0.01973337,1.0},
  {0.25668794, 0.09182569, -0.01973337,1.0},
  {0.25668794, 0.09182569, 0,1.0},
  // f 887/928/72 886/927/72 435/926/72
  {0.30438331, 0.14125545, -0.01973337,1.0},
  {0.30438331, 0.09182569, -0.01973337,1.0},
  {0.30438331, 0.09182569, 0,1.0},
  // f 888/931/1 887/930/1 436/929/1
  {0.34822458, 0.14125545, -0.01973337,1.0},
  {0.30438331, 0.14125545, -0.01973337,1.0},
  {0.30438331, 0.14125545, 0,1.0},
  // f 889/934/71 888/933/71 437/932/71
  {0.34822458, 0, -0.01973337,1.0},
  {0.34822458, 0.14125545, -0.01973337,1.0},
  {0.34822458, 0.14125545, 0,1.0},
  // f 890/937/70 889/936/70 438/935/70
  {0.30438331, 0, -0.01973337,1.0},
  {0.34822458, 0, -0.01973337,1.0},
  {0.34822458, 0, 0,1.0},
  // f 891/940/72 890/939/72 439/938/72
  {0.30438331, 0.05713812, -0.01973337,1.0},
  {0.30438331, 0, -0.01973337,1.0},
  {0.30438331, 0, 0,1.0},
  // f 892/943/70 891/942/70 440/941/70
  {0.25668794, 0.05713812, -0.01973337,1.0},
  {0.30438331, 0.05713812, -0.01973337,1.0},
  {0.30438331, 0.05713812, 0,1.0},
  // f 893/946/71 892/945/71 441/944/71
  {0.25668794, 0, -0.01973337,1.0},
  {0.25668794, 0.05713812, -0.01973337,1.0},
  {0.25668794, 0.05713812, 0,1.0},
  // f 894/949/70 893/948/70 442/947/70
  {0.21303943, 0, -0.01973337,1.0},
  {0.25668794, 0, -0.01973337,1.0},
  {0.25668794, 0, 0,1.0},
  // f 883/952/72 894/951/72 443/950/72
  {0.21303943, 0.14125545, -0.01973337,1.0},
  {0.21303943, 0, -0.01973337,1.0},
  {0.21303943, 0, 0,1.0},
  // f 896/955/1 895/954/1 444/953/1
  {0.49998266, 0.14125545, -0.01973337,1.0},
  {0.36730275, 0.14125545, -0.01973337,1.0},
  {0.36730275, 0.14125545, 0,1.0},
  // f 897/958/71 896/957/820 445/956/71
  {0.49998266, 0.10637518, -0.01973337,1.0},
  {0.49998266, 0.14125545, -0.01973337,1.0},
  {0.49998266, 0.14125545, 0,1.0},
  // f 898/961/70 897/960/70 446/959/70
  {0.45546687, 0.10637518, -0.01973337,1.0},
  {0.49998266, 0.10637518, -0.01973337,1.0},
  {0.49998266, 0.10637518, 0,1.0},
  // f 899/964/71 898/963/71 447/962/71
  {0.45546687, 0, -0.01973337,1.0},
  {0.45546687, 0.10637518, -0.01973337,1.0},
  {0.45546687, 0.10637518, 0,1.0},
  // f 900/967/70 899/966/70 448/965/70
  {0.41181844, 0, -0.01973337,1.0},
  {0.45546687, 0, -0.01973337,1.0},
  {0.45546687, 0, 0,1.0},
  // f 901/970/72 900/969/72 449/968/72
  {0.41181844, 0.10637518, -0.01973337,1.0},
  {0.41181844, 0, -0.01973337,1.0},
  {0.41181844, 0, 0,1.0},
  // f 902/973/70 901/972/70 450/971/70
  {0.36730275, 0.10637518, -0.01973337,1.0},
  {0.41181844, 0.10637518, -0.01973337,1.0},
  {0.41181844, 0.10637518, 0,1.0},
  // f 895/976/72 902/975/821 451/974/72
  {0.36730275, 0.14125545, -0.01973337,1.0},
  {0.36730275, 0.10637518, -0.01973337,1.0},
  {0.36730275, 0.10637518, 0,1.0},
  // f 939/979/822 940/978/822 941/977/822
  {-0.45614135, 0.05241676, -0.01973337,1.0},
  {-0.45614135, 0, -0.01973337,1.0},
  {-0.49998266, 0, -0.01973337,1.0},
  // f 942/980/822 939/979/822 941/977/822
  {-0.45614135, 0.081034, -0.01973337,1.0},
  {-0.45614135, 0.05241676, -0.01973337,1.0},
  {-0.49998266, 0, -0.01973337,1.0},
  // f 903/981/822 942/980/822 941/977/822
  {-0.49998266, 0.14125545, -0.01973337,1.0},
  {-0.45614135, 0.081034, -0.01973337,1.0},
  {-0.49998266, 0, -0.01973337,1.0},
  // f 943/983/822 938/982/822 939/979/822
  {-0.4454461, 0.081034, -0.01973337,1.0},
  {-0.43224552, 0.05241676, -0.01973337,1.0},
  {-0.45614135, 0.05241676, -0.01973337,1.0},
  // f 942/980/822 943/983/822 939/979/822
  {-0.45614135, 0.081034, -0.01973337,1.0},
  {-0.4454461, 0.081034, -0.01973337,1.0},
  {-0.45614135, 0.05241676, -0.01973337,1.0},
  // f 946/985/822 937/984/822 938/982/822
  {-0.43400544, 0.0823124, -0.01973337,1.0},
  {-0.42414385, 0.05272509, -0.01973337,1.0},
  {-0.43224552, 0.05241676, -0.01973337,1.0},
  // f 945/986/822 946/985/822 938/982/822
  {-0.43655068, 0.08173544, -0.01973337,1.0},
  {-0.43400544, 0.0823124, -0.01973337,1.0},
  {-0.43224552, 0.05241676, -0.01973337,1.0},
  // f 944/987/822 945/986/822 938/982/822
  {-0.440256, 0.08124619, -0.01973337,1.0},
  {-0.43655068, 0.08173544, -0.01973337,1.0},
  {-0.43224552, 0.05241676, -0.01973337,1.0},
  // f 943/983/822 944/987/822 938/982/822
  {-0.4454461, 0.081034, -0.01973337,1.0},
  {-0.440256, 0.08124619, -0.01973337,1.0},
  {-0.43224552, 0.05241676, -0.01973337,1.0},
  // f 948/989/822 936/988/822 937/984/822
  {-0.43060106, 0.08356626, -0.01973337,1.0},
  {-0.41673514, 0.0536501, -0.01973337,1.0},
  {-0.42414385, 0.05272509, -0.01973337,1.0},
  // f 947/990/822 948/989/822 937/984/822
  {-0.43205836, 0.08294371, -0.01973337,1.0},
  {-0.43060106, 0.08356626, -0.01973337,1.0},
  {-0.42414385, 0.05272509, -0.01973337,1.0},
  // f 946/985/822 947/990/822 937/984/822
  {-0.43400544, 0.0823124, -0.01973337,1.0},
  {-0.43205836, 0.08294371, -0.01973337,1.0},
  {-0.42414385, 0.05272509, -0.01973337,1.0},
  // f 950/992/822 935/991/822 936/988/822
  {-0.42858201, 0.0847447, -0.01973337,1.0},
  {-0.4108212, 0.05496534, -0.01973337,1.0},
  {-0.41673514, 0.0536501, -0.01973337,1.0},
  // f 949/993/822 950/992/822 936/988/822
  {-0.42929417, 0.08427649, -0.01973337,1.0},
  {-0.42858201, 0.0847447, -0.01973337,1.0},
  {-0.41673514, 0.0536501, -0.01973337,1.0},
  // f 948/989/822 949/993/822 936/988/822
  {-0.43060106, 0.08356626, -0.01973337,1.0},
  {-0.42929417, 0.08427649, -0.01973337,1.0},
  {-0.41673514, 0.0536501, -0.01973337,1.0},
  // f 952/995/822 934/994/822 935/991/822
  {-0.42656487, 0.0865057, -0.01973337,1.0},
  {-0.40617442, 0.05646847, -0.01973337,1.0},
  {-0.4108212, 0.05496534, -0.01973337,1.0},
  // f 951/996/822 952/995/822 935/991/822
  {-0.42771694, 0.08541812, -0.01973337,1.0},
  {-0.42656487, 0.0865057, -0.01973337,1.0},
  {-0.4108212, 0.05496534, -0.01973337,1.0},
  // f 950/992/822 951/996/822 935/991/822
  {-0.42858201, 0.0847447, -0.01973337,1.0},
  {-0.42771694, 0.08541812, -0.01973337,1.0},
  {-0.4108212, 0.05496534, -0.01973337,1.0},
  // f 953/998/822 933/997/822 934/994/822
  {-0.42555979, 0.08766402, -0.01973337,1.0},
  {-0.40191758, 0.05831847, -0.01973337,1.0},
  {-0.40617442, 0.05646847, -0.01973337,1.0},
  // f 952/995/822 953/998/822 934/994/822
  {-0.42656487, 0.0865057, -0.01973337,1.0},
  {-0.42555979, 0.08766402, -0.01973337,1.0},
  {-0.40617442, 0.05646847, -0.01973337,1.0},
  // f 953/998/822 932/999/822 933/997/822
  {-0.42555979, 0.08766402, -0.01973337,1.0},
  {-0.39866793, 0.06012511, -0.01973337,1.0},
  {-0.40191758, 0.05831847, -0.01973337,1.0},
  // f 954/1001/822 931/1000/822 932/999/822
  {-0.42483458, 0.08868335, -0.01973337,1.0},
  {-0.39568928, 0.06217263, -0.01973337,1.0},
  {-0.39866793, 0.06012511, -0.01973337,1.0},
  // f 953/998/822 954/1001/822 932/999/822
  {-0.42555979, 0.08766402, -0.01973337,1.0},
  {-0.42483458, 0.08868335, -0.01973337,1.0},
  {-0.39866793, 0.06012511, -0.01973337,1.0},
  // f 954/1001/822 930/1002/822 931/1000/822
  {-0.42483458, 0.08868335, -0.01973337,1.0},
  {-0.39298108, 0.06446107, -0.01973337,1.0},
  {-0.39568928, 0.06217263, -0.01973337,1.0},
  // f 955/1004/822 929/1003/822 930/1002/822
  {-0.42421165, 0.08975184, -0.01973337,1.0},
  {-0.39053714, 0.06696337, -0.01973337,1.0},
  {-0.39298108, 0.06446107, -0.01973337,1.0},
  // f 954/1001/822 955/1004/822 930/1002/822
  {-0.42483458, 0.08868335, -0.01973337,1.0},
  {-0.42421165, 0.08975184, -0.01973337,1.0},
  {-0.39298108, 0.06446107, -0.01973337,1.0},
  // f 955/1004/822 928/1005/822 929/1003/822
  {-0.42421165, 0.08975184, -0.01973337,1.0},
  {-0.38876721, 0.06909983, -0.01973337,1.0},
  {-0.39053714, 0.06696337, -0.01973337,1.0},
  // f 956/1007/822 927/1006/822 928/1005/822
  {-0.42369071, 0.09086943, -0.01973337,1.0},
  {-0.38678622, 0.07193859, -0.01973337,1.0},
  {-0.38876721, 0.06909983, -0.01973337,1.0},
  // f 955/1004/822 956/1007/822 928/1005/822
  {-0.42421165, 0.08975184, -0.01973337,1.0},
  {-0.42369071, 0.09086943, -0.01973337,1.0},
  {-0.38876721, 0.06909983, -0.01973337,1.0},
  // f 956/1007/822 926/1008/822 927/1006/822
  {-0.42369071, 0.09086943, -0.01973337,1.0},
  {-0.38506255, 0.07496432, -0.01973337,1.0},
  {-0.38678622, 0.07193859, -0.01973337,1.0},
  // f 957/1010/822 925/1009/822 926/1008/822
  {-0.42327198, 0.09203622, -0.01973337,1.0},
  {-0.38359609, 0.07817698, -0.01973337,1.0},
  {-0.38506255, 0.07496432, -0.01973337,1.0},
  // f 956/1007/822 957/1010/822 926/1008/822
  {-0.42369071, 0.09086943, -0.01973337,1.0},
  {-0.42327198, 0.09203622, -0.01973337,1.0},
  {-0.38506255, 0.07496432, -0.01973337,1.0},
  // f 957/1010/822 924/1011/822 925/1009/822
  {-0.42327198, 0.09203622, -0.01973337,1.0},
  {-0.38217598, 0.08227888, -0.01973337,1.0},
  {-0.38359609, 0.07817698, -0.01973337,1.0},
  // f 958/1013/822 923/1012/822 924/1011/822
  {-0.42295536, 0.09325211, -0.01973337,1.0},
  {-0.38112643, 0.08665, -0.01973337,1.0},
  {-0.38217598, 0.08227888, -0.01973337,1.0},
  // f 957/1010/822 958/1013/822 924/1011/822
  {-0.42327198, 0.09203622, -0.01973337,1.0},
  {-0.42295536, 0.09325211, -0.01973337,1.0},
  {-0.38217598, 0.08227888, -0.01973337,1.0},
  // f 959/1015/822 922/1014/822 923/1012/822
  {-0.42271024, 0.09477603, -0.01973337,1.0},
  {-0.38036993, 0.09208981, -0.01973337,1.0},
  {-0.38112643, 0.08665, -0.01973337,1.0},
  // f 958/1013/822 959/1015/822 923/1012/822
  {-0.42295536, 0.09325211, -0.01973337,1.0},
  {-0.42271024, 0.09477603, -0.01973337,1.0},
  {-0.38112643, 0.08665, -0.01973337,1.0},
  // f 961/1017/822 921/1016/822 922/1014/822
  {-0.42261013, 0.09664342, -0.01973337,1.0},
  {-0.38011783, 0.097896, -0.01973337,1.0},
  {-0.38036993, 0.09208981, -0.01973337,1.0},
  // f 960/1018/822 961/1017/822 922/1014/822
  {-0.42264283, 0.09556455, -0.01973337,1.0},
  {-0.42261013, 0.09664342, -0.01973337,1.0},
  {-0.38036993, 0.09208981, -0.01973337,1.0},
  // f 959/1015/822 960/1018/822 922/1014/822
  {-0.42271024, 0.09477603, -0.01973337,1.0},
  {-0.42264283, 0.09556455, -0.01973337,1.0},
  {-0.38036993, 0.09208981, -0.01973337,1.0},
  // f 962/1020/822 920/1019/822 921/1016/822
  {-0.42269701, 0.0984653, -0.01973337,1.0},
  {-0.3802878, 0.10275365, -0.01973337,1.0},
  {-0.38011783, 0.097896, -0.01973337,1.0},
  // f 961/1017/822 962/1020/822 921/1016/822
  {-0.42261013, 0.09664342, -0.01973337,1.0},
  {-0.42269701, 0.0984653, -0.01973337,1.0},
  {-0.38011783, 0.097896, -0.01973337,1.0},
  // f 963/1022/822 919/1021/822 920/1019/822
  {-0.42290977, 0.09996384, -0.01973337,1.0},
  {-0.38079774, 0.10733657, -0.01973337,1.0},
  {-0.3802878, 0.10275365, -0.01973337,1.0},
  // f 962/1020/822 963/1022/822 920/1019/822
  {-0.42269701, 0.0984653, -0.01973337,1.0},
  {-0.42290977, 0.09996384, -0.01973337,1.0},
  {-0.3802878, 0.10275365, -0.01973337,1.0},
  // f 963/1022/822 918/1023/822 919/1021/822
  {-0.42290977, 0.09996384, -0.01973337,1.0},
  {-0.38164759, 0.11164479, -0.01973337,1.0},
  {-0.38079774, 0.10733657, -0.01973337,1.0},
  // f 964/1025/822 917/1024/822 918/1023/822
  {-0.42325002, 0.10140404, -0.01973337,1.0},
  {-0.38283733, 0.11567822, -0.01973337,1.0},
  {-0.38164759, 0.11164479, -0.01973337,1.0},
  // f 963/1022/822 964/1025/822 918/1023/822
  {-0.42290977, 0.09996384, -0.01973337,1.0},
  {-0.42325002, 0.10140404, -0.01973337,1.0},
  {-0.38164759, 0.11164479, -0.01973337,1.0},
  // f 965/1027/822 916/1026/822 917/1024/822
  {-0.42371821, 0.10278598, -0.01973337,1.0},
  {-0.38436708, 0.11943695, -0.01973337,1.0},
  {-0.38283733, 0.11567822, -0.01973337,1.0},
  // f 964/1025/822 965/1027/822 917/1024/822
  {-0.42325002, 0.10140404, -0.01973337,1.0},
  {-0.42371821, 0.10278598, -0.01973337,1.0},
  {-0.38283733, 0.11567822, -0.01973337,1.0},
  // f 965/1027/822 915/1028/822 916/1026/822
  {-0.42371821, 0.10278598, -0.01973337,1.0},
  {-0.3859016, 0.1223594, -0.01973337,1.0},
  {-0.38436708, 0.11943695, -0.01973337,1.0},
  // f 965/1027/822 914/1029/822 915/1028/822
  {-0.42371821, 0.10278598, -0.01973337,1.0},
  {-0.38767204, 0.12509105, -0.01973337,1.0},
  {-0.3859016, 0.1223594, -0.01973337,1.0},
  // f 966/1031/822 913/1030/822 914/1029/822
  {-0.4243139, 0.10410966, -0.01973337,1.0},
  {-0.38967857, 0.1276319, -0.01973337,1.0},
  {-0.38767204, 0.12509105, -0.01973337,1.0},
  // f 965/1027/822 966/1031/822 914/1029/822
  {-0.42371821, 0.10278598, -0.01973337,1.0},
  {-0.4243139, 0.10410966, -0.01973337,1.0},
  {-0.38767204, 0.12509105, -0.01973337,1.0},
  // f 966/1031/822 912/1032/822 913/1030/822
  {-0.4243139, 0.10410966, -0.01973337,1.0},
  {-0.39192128, 0.12998199, -0.01973337,1.0},
  {-0.38967857, 0.1276319, -0.01973337,1.0},
  // f 967/1034/822 911/1033/822 912/1032/822
  {-0.42503724, 0.10537505, -0.01973337,1.0},
  {-0.39440098, 0.13212398, -0.01973337,1.0},
  {-0.39192128, 0.12998199, -0.01973337,1.0},
  // f 966/1031/822 967/1034/822 912/1032/822
  {-0.4243139, 0.10410966, -0.01973337,1.0},
  {-0.42503724, 0.10537505, -0.01973337,1.0},
  {-0.39192128, 0.12998199, -0.01973337,1.0},
  // f 967/1034/822 910/1035/822 911/1033/822
  {-0.42503724, 0.10537505, -0.01973337,1.0},
  {-0.39655605, 0.13367519, -0.01973337,1.0},
  {-0.39440098, 0.13212398, -0.01973337,1.0},
  // f 967/1034/822 909/1036/822 910/1035/822
  {-0.42503724, 0.10537505, -0.01973337,1.0},
  {-0.39946407, 0.13541128, -0.01973337,1.0},
  {-0.39655605, 0.13367519, -0.01973337,1.0},
  // f 968/1038/822 908/1037/822 909/1036/822
  {-0.42588824, 0.10658219, -0.01973337,1.0},
  {-0.40326801, 0.13719702, -0.01973337,1.0},
  {-0.39946407, 0.13541128, -0.01973337,1.0},
  // f 967/1034/822 968/1038/822 909/1036/822
  {-0.42503724, 0.10537505, -0.01973337,1.0},
  {-0.42588824, 0.10658219, -0.01973337,1.0},
  {-0.39946407, 0.13541128, -0.01973337,1.0},
  // f 969/1040/822 907/1039/822 908/1037/822
  {-0.42636168, 0.10716394, -0.01973337,1.0},
  {-0.40741441, 0.13865802, -0.01973337,1.0},
  {-0.40326801, 0.13719702, -0.01973337,1.0},
  // f 968/1038/822 969/1040/822 908/1037/822
  {-0.42588824, 0.10658219, -0.01973337,1.0},
  {-0.42636168, 0.10716394, -0.01973337,1.0},
  {-0.40326801, 0.13719702, -0.01973337,1.0},
  // f 970/1042/822 906/1041/822 907/1039/822
  {-0.42704237, 0.10791686, -0.01973337,1.0},
  {-0.4126853, 0.13995226, -0.01973337,1.0},
  {-0.40741441, 0.13865802, -0.01973337,1.0},
  // f 969/1040/822 970/1042/822 907/1039/822
  {-0.42636168, 0.10716394, -0.01973337,1.0},
  {-0.42704237, 0.10791686, -0.01973337,1.0},
  {-0.40741441, 0.13865802, -0.01973337,1.0},
  // f 972/1044/822 905/1043/822 906/1041/822
  {-0.4286606, 0.10927848, -0.01973337,1.0},
  {-0.41928023, 0.14089021, -0.01973337,1.0},
  {-0.4126853, 0.13995226, -0.01973337,1.0},
  // f 971/1045/822 972/1044/822 906/1041/822
  {-0.42780152, 0.10862726, -0.01973337,1.0},
  {-0.4286606, 0.10927848, -0.01973337,1.0},
  {-0.4126853, 0.13995226, -0.01973337,1.0},
  // f 970/1042/822 971/1045/822 906/1041/822
  {-0.42704237, 0.10791686, -0.01973337,1.0},
  {-0.42780152, 0.10862726, -0.01973337,1.0},
  {-0.4126853, 0.13995226, -0.01973337,1.0},
  // f 974/1047/822 904/1046/822 905/1043/822
  {-0.43095866, 0.11052722, -0.01973337,1.0},
  {-0.42742792, 0.14125545, -0.01973337,1.0},
  {-0.41928023, 0.14089021, -0.01973337,1.0},
  // f 973/1048/822 974/1047/822 905/1043/822
  {-0.42961943, 0.10987046, -0.01973337,1.0},
  {-0.43095866, 0.11052722, -0.01973337,1.0},
  {-0.41928023, 0.14089021, -0.01973337,1.0},
  // f 972/1044/822 973/1048/822 905/1043/822
  {-0.4286606, 0.10927848, -0.01973337,1.0},
  {-0.42961943, 0.10987046, -0.01973337,1.0},
  {-0.41928023, 0.14089021, -0.01973337,1.0},
  // f 980/1049/822 903/981/822 904/1046/822
  {-0.45614135, 0.11254187, -0.01973337,1.0},
  {-0.49998266, 0.14125545, -0.01973337,1.0},
  {-0.42742792, 0.14125545, -0.01973337,1.0},
  // f 979/1050/822 980/1049/822 904/1046/822
  {-0.44371176, 0.11254187, -0.01973337,1.0},
  {-0.45614135, 0.11254187, -0.01973337,1.0},
  {-0.42742792, 0.14125545, -0.01973337,1.0},
  // f 978/1051/822 979/1050/822 904/1046/822
  {-0.43999577, 0.11242347, -0.01973337,1.0},
  {-0.44371176, 0.11254187, -0.01973337,1.0},
  {-0.42742792, 0.14125545, -0.01973337,1.0},
  // f 977/1052/822 978/1051/822 904/1046/822
  {-0.43707186, 0.11212561, -0.01973337,1.0},
  {-0.43999577, 0.11242347, -0.01973337,1.0},
  {-0.42742792, 0.14125545, -0.01973337,1.0},
  // f 976/1053/822 977/1052/822 904/1046/822
  {-0.43445414, 0.11164644, -0.01973337,1.0},
  {-0.43707186, 0.11212561, -0.01973337,1.0},
  {-0.42742792, 0.14125545, -0.01973337,1.0},
  // f 975/1054/822 976/1053/822 904/1046/822
  {-0.43245378, 0.11109147, -0.01973337,1.0},
  {-0.43445414, 0.11164644, -0.01973337,1.0},
  {-0.42742792, 0.14125545, -0.01973337,1.0},
  // f 974/1047/822 975/1054/822 904/1046/822
  {-0.43095866, 0.11052722, -0.01973337,1.0},
  {-0.43245378, 0.11109147, -0.01973337,1.0},
  {-0.42742792, 0.14125545, -0.01973337,1.0},
  // f 980/1049/822 942/980/822 903/981/822
  {-0.45614135, 0.11254187, -0.01973337,1.0},
  {-0.45614135, 0.081034, -0.01973337,1.0},
  {-0.49998266, 0.14125545, -0.01973337,1.0},
  // f 982/1057/822 1183/1056/822 1184/1055/822
  {-0.32336509, 0.04933343, -0.01973337,1.0},
  {-0.36350545, 0.03768371, -0.01973337,1.0},
  {-0.36428103, 0.04170631, -0.01973337,1.0},
  // f 981/1058/822 982/1057/822 1184/1055/822
  {-0.36489382, 0.04673185, -0.01973337,1.0},
  {-0.32336509, 0.04933343, -0.01973337,1.0},
  {-0.36428103, 0.04170631, -0.01973337,1.0},
  // f 983/1060/822 1182/1059/822 1183/1056/822
  {-0.32255679, 0.04475466, -0.01973337,1.0},
  {-0.3620148, 0.0323046, -0.01973337,1.0},
  {-0.36350545, 0.03768371, -0.01973337,1.0},
  // f 982/1057/822 983/1060/822 1183/1056/822
  {-0.32336509, 0.04933343, -0.01973337,1.0},
  {-0.32255679, 0.04475466, -0.01973337,1.0},
  {-0.36350545, 0.03768371, -0.01973337,1.0},
  // f 984/1062/822 1181/1061/822 1182/1059/822
  {-0.32163715, 0.04136299, -0.01973337,1.0},
  {-0.36005208, 0.0272201, -0.01973337,1.0},
  {-0.3620148, 0.0323046, -0.01973337,1.0},
  // f 983/1060/822 984/1062/822 1182/1059/822
  {-0.32255679, 0.04475466, -0.01973337,1.0},
  {-0.32163715, 0.04136299, -0.01973337,1.0},
  {-0.3620148, 0.0323046, -0.01973337,1.0},
  // f 984/1062/822 1180/1063/822 1181/1061/822
  {-0.32163715, 0.04136299, -0.01973337,1.0},
  {-0.35799384, 0.02309645, -0.01973337,1.0},
  {-0.36005208, 0.0272201, -0.01973337,1.0},
  // f 985/1065/822 1179/1064/822 1180/1063/822
  {-0.3206183, 0.03864001, -0.01973337,1.0},
  {-0.35558888, 0.01918923, -0.01973337,1.0},
  {-0.35799384, 0.02309645, -0.01973337,1.0},
  // f 984/1062/822 985/1065/822 1180/1063/822
  {-0.32163715, 0.04136299, -0.01973337,1.0},
  {-0.3206183, 0.03864001, -0.01973337,1.0},
  {-0.35799384, 0.02309645, -0.01973337,1.0},
  // f 985/1065/822 1178/1066/822 1179/1064/822
  {-0.3206183, 0.03864001, -0.01973337,1.0},
  {-0.352837, 0.01549848, -0.01973337,1.0},
  {-0.35558888, 0.01918923, -0.01973337,1.0},
  // f 986/1068/822 1177/1067/822 1178/1066/822
  {-0.31956071, 0.03648167, -0.01973337,1.0},
  {-0.34918806, 0.01146617, -0.01973337,1.0},
  {-0.352837, 0.01549848, -0.01973337,1.0},
  // f 985/1065/822 986/1068/822 1178/1066/822
  {-0.3206183, 0.03864001, -0.01973337,1.0},
  {-0.31956071, 0.03648167, -0.01973337,1.0},
  {-0.352837, 0.01549848, -0.01973337,1.0},
  // f 987/1070/822 1176/1069/822 1177/1067/822
  {-0.31882712, 0.03525797, -0.01973337,1.0},
  {-0.34744334, 0.00985111, -0.01973337,1.0},
  {-0.34918806, 0.01146617, -0.01973337,1.0},
  // f 986/1068/822 987/1070/822 1177/1067/822
  {-0.31956071, 0.03648167, -0.01973337,1.0},
  {-0.31882712, 0.03525797, -0.01973337,1.0},
  {-0.34918806, 0.01146617, -0.01973337,1.0},
  // f 987/1070/822 1175/1071/822 1176/1069/822
  {-0.31882712, 0.03525797, -0.01973337,1.0},
  {-0.34619153, 0.00882991, -0.01973337,1.0},
  {-0.34744334, 0.00985111, -0.01973337,1.0},
  // f 988/1073/822 1174/1072/822 1175/1071/822
  {-0.31787291, 0.03391673, -0.01973337,1.0},
  {-0.34347504, 0.00692071, -0.01973337,1.0},
  {-0.34619153, 0.00882991, -0.01973337,1.0},
  // f 987/1070/822 988/1073/822 1175/1071/822
  {-0.31882712, 0.03525797, -0.01973337,1.0},
  {-0.31787291, 0.03391673, -0.01973337,1.0},
  {-0.34619153, 0.00882991, -0.01973337,1.0},
  // f 988/1073/822 1173/1074/822 1174/1072/822
  {-0.31787291, 0.03391673, -0.01973337,1.0},
  {-0.33968067, 0.00478395, -0.01973337,1.0},
  {-0.34347504, 0.00692071, -0.01973337,1.0},
  // f 989/1076/822 1172/1075/822 1173/1074/822
  {-0.31646615, 0.03228737, -0.01973337,1.0},
  {-0.33544296, 0.0029247, -0.01973337,1.0},
  {-0.33968067, 0.00478395, -0.01973337,1.0},
  // f 988/1073/822 989/1076/822 1173/1074/822
  {-0.31787291, 0.03391673, -0.01973337,1.0},
  {-0.31646615, 0.03228737, -0.01973337,1.0},
  {-0.33968067, 0.00478395, -0.01973337,1.0},
  // f 990/1078/822 1171/1077/822 1172/1075/822
  {-0.31494385, 0.03082954, -0.01973337,1.0},
  {-0.32977268, 0.0010599, -0.01973337,1.0},
  {-0.33544296, 0.0029247, -0.01973337,1.0},
  // f 989/1076/822 990/1078/822 1172/1075/822
  {-0.31646615, 0.03228737, -0.01973337,1.0},
  {-0.31494385, 0.03082954, -0.01973337,1.0},
  {-0.33544296, 0.0029247, -0.01973337,1.0},
  // f 991/1080/822 1170/1079/822 1171/1077/822
  {-0.31330577, 0.02954321, -0.01973337,1.0},
  {-0.32235077, -0.00061066, -0.01973337,1.0},
  {-0.32977268, 0.0010599, -0.01973337,1.0},
  // f 990/1078/822 991/1080/822 1171/1077/822
  {-0.31494385, 0.03082954, -0.01973337,1.0},
  {-0.31330577, 0.02954321, -0.01973337,1.0},
  {-0.32977268, 0.0010599, -0.01973337,1.0},
  // f 993/1082/822 1169/1081/822 1170/1079/822
  {-0.30968276, 0.02748508, -0.01973337,1.0},
  {-0.31280473, -0.00185386, -0.01973337,1.0},
  {-0.32235077, -0.00061066, -0.01973337,1.0},
  // f 992/1083/822 993/1082/822 1170/1079/822
  {-0.31155208, 0.02842839, -0.01973337,1.0},
  {-0.30968276, 0.02748508, -0.01973337,1.0},
  {-0.32235077, -0.00061066, -0.01973337,1.0},
  // f 991/1080/822 992/1083/822 1170/1079/822
  {-0.31330577, 0.02954321, -0.01973337,1.0},
  {-0.31155208, 0.02842839, -0.01973337,1.0},
  {-0.32235077, -0.00061066, -0.01973337,1.0},
  // f 996/1085/822 1168/1084/822 1169/1081/822
  {-0.30292422, 0.02561904, -0.01973337,1.0},
  {-0.29927659, -0.00240886, -0.01973337,1.0},
  {-0.31280473, -0.00185386, -0.01973337,1.0},
  // f 995/1086/822 996/1085/822 1169/1081/822
  {-0.30559742, 0.02611299, -0.01973337,1.0},
  {-0.30292422, 0.02561904, -0.01973337,1.0},
  {-0.31280473, -0.00185386, -0.01973337,1.0},
  // f 994/1087/822 995/1086/822 1169/1081/822
  {-0.30769795, 0.02671328, -0.01973337,1.0},
  {-0.30559742, 0.02611299, -0.01973337,1.0},
  {-0.31280473, -0.00185386, -0.01973337,1.0},
  // f 993/1082/822 994/1087/822 1169/1081/822
  {-0.30968276, 0.02748508, -0.01973337,1.0},
  {-0.30769795, 0.02671328, -0.01973337,1.0},
  {-0.31280473, -0.00185386, -0.01973337,1.0},
  // f 999/1089/822 1167/1088/822 1168/1084/822
  {-0.29537636, 0.02548324, -0.01973337,1.0},
  {-0.29134852, -0.00217183, -0.01973337,1.0},
  {-0.29927659, -0.00240886, -0.01973337,1.0},
  // f 998/1090/822 999/1089/822 1168/1084/822
  {-0.29860207, 0.02534119, -0.01973337,1.0},
  {-0.29537636, 0.02548324, -0.01973337,1.0},
  {-0.29927659, -0.00240886, -0.01973337,1.0},
  // f 997/1091/822 998/1090/822 1168/1084/822
  {-0.3010495, 0.02542694, -0.01973337,1.0},
  {-0.29860207, 0.02534119, -0.01973337,1.0},
  {-0.29927659, -0.00240886, -0.01973337,1.0},
  // f 996/1085/822 997/1091/822 1168/1084/822
  {-0.30292422, 0.02561904, -0.01973337,1.0},
  {-0.3010495, 0.02542694, -0.01973337,1.0},
  {-0.29927659, -0.00240886, -0.01973337,1.0},
  // f 1000/1093/822 1166/1092/822 1167/1088/822
  {-0.29274219, 0.025848, -0.01973337,1.0},
  {-0.2839832, -0.00146073, -0.01973337,1.0},
  {-0.29134852, -0.00217183, -0.01973337,1.0},
  // f 999/1089/822 1000/1093/822 1167/1088/822
  {-0.29537636, 0.02548324, -0.01973337,1.0},
  {-0.29274219, 0.025848, -0.01973337,1.0},
  {-0.29134852, -0.00217183, -0.01973337,1.0},
  // f 1002/1095/822 1165/1094/822 1166/1092/822
  {-0.28893006, 0.02691947, -0.01973337,1.0},
  {-0.27653128, -0.00013098, -0.01973337,1.0},
  {-0.2839832, -0.00146073, -0.01973337,1.0},
  // f 1001/1096/822 1002/1095/822 1166/1092/822
  {-0.29061744, 0.02635129, -0.01973337,1.0},
  {-0.28893006, 0.02691947, -0.01973337,1.0},
  {-0.2839832, -0.00146073, -0.01973337,1.0},
  // f 1000/1093/822 1001/1096/822 1166/1092/822
  {-0.29274219, 0.025848, -0.01973337,1.0},
  {-0.29061744, 0.02635129, -0.01973337,1.0},
  {-0.2839832, -0.00146073, -0.01973337,1.0},
  // f 1004/1098/822 1164/1097/822 1165/1094/822
  {-0.2859256, 0.02843463, -0.01973337,1.0},
  {-0.27034768, 0.00157564, -0.01973337,1.0},
  {-0.27653128, -0.00013098, -0.01973337,1.0},
  // f 1003/1099/822 1004/1098/822 1165/1094/822
  {-0.28736609, 0.02761392, -0.01973337,1.0},
  {-0.2859256, 0.02843463, -0.01973337,1.0},
  {-0.27653128, -0.00013098, -0.01973337,1.0},
  // f 1002/1095/822 1003/1099/822 1165/1094/822
  {-0.28893006, 0.02691947, -0.01973337,1.0},
  {-0.28736609, 0.02761392, -0.01973337,1.0},
  {-0.27653128, -0.00013098, -0.01973337,1.0},
  // f 1006/1101/822 1163/1100/822 1164/1097/822
  {-0.28419709, 0.02972531, -0.01973337,1.0},
  {-0.2652635, 0.00351693, -0.01973337,1.0},
  {-0.27034768, 0.00157564, -0.01973337,1.0},
  // f 1005/1102/822 1006/1101/822 1164/1097/822
  {-0.28503397, 0.02905191, -0.01973337,1.0},
  {-0.28419709, 0.02972531, -0.01973337,1.0},
  {-0.27034768, 0.00157564, -0.01973337,1.0},
  // f 1004/1098/822 1005/1102/822 1164/1097/822
  {-0.2859256, 0.02843463, -0.01973337,1.0},
  {-0.28503397, 0.02905191, -0.01973337,1.0},
  {-0.27034768, 0.00157564, -0.01973337,1.0},
  // f 1007/1104/822 1162/1103/822 1163/1100/822
  {-0.28287992, 0.03098026, -0.01973337,1.0},
  {-0.26262525, 0.0047575, -0.01973337,1.0},
  {-0.2652635, 0.00351693, -0.01973337,1.0},
  // f 1006/1101/822 1007/1104/822 1163/1100/822
  {-0.28419709, 0.02972531, -0.01973337,1.0},
  {-0.28287992, 0.03098026, -0.01973337,1.0},
  {-0.2652635, 0.00351693, -0.01973337,1.0},
  // f 1008/1106/822 1161/1105/822 1162/1103/822
  {-0.28190863, 0.0320995, -0.01973337,1.0},
  {-0.26010665, 0.00610887, -0.01973337,1.0},
  {-0.26262525, 0.0047575, -0.01973337,1.0},
  // f 1007/1104/822 1008/1106/822 1162/1103/822
  {-0.28287992, 0.03098026, -0.01973337,1.0},
  {-0.28190863, 0.0320995, -0.01973337,1.0},
  {-0.26262525, 0.0047575, -0.01973337,1.0},
  // f 1008/1106/822 1160/1107/822 1161/1105/822
  {-0.28190863, 0.0320995, -0.01973337,1.0},
  {-0.25632519, 0.00850154, -0.01973337,1.0},
  {-0.26010665, 0.00610887, -0.01973337,1.0},
  // f 1009/1109/822 1159/1108/822 1160/1107/822
  {-0.28108308, 0.033259, -0.01973337,1.0},
  {-0.25284964, 0.01117787, -0.01973337,1.0},
  {-0.25632519, 0.00850154, -0.01973337,1.0},
  // f 1008/1106/822 1009/1109/822 1160/1107/822
  {-0.28190863, 0.0320995, -0.01973337,1.0},
  {-0.28108308, 0.033259, -0.01973337,1.0},
  {-0.25632519, 0.00850154, -0.01973337,1.0},
  // f 1010/1111/822 1158/1110/822 1159/1108/822
  {-0.2805064, 0.03425597, -0.01973337,1.0},
  {-0.24967997, 0.01413788, -0.01973337,1.0},
  {-0.25284964, 0.01117787, -0.01973337,1.0},
  // f 1009/1109/822 1010/1111/822 1159/1108/822
  {-0.28108308, 0.033259, -0.01973337,1.0},
  {-0.2805064, 0.03425597, -0.01973337,1.0},
  {-0.25284964, 0.01117787, -0.01973337,1.0},
  // f 1011/1113/822 1157/1112/822 1158/1110/822
  {-0.28003085, 0.03528089, -0.01973337,1.0},
  {-0.24681623, 0.01738155, -0.01973337,1.0},
  {-0.24967997, 0.01413788, -0.01973337,1.0},
  // f 1010/1111/822 1011/1113/822 1158/1110/822
  {-0.2805064, 0.03425597, -0.01973337,1.0},
  {-0.28003085, 0.03528089, -0.01973337,1.0},
  {-0.24967997, 0.01413788, -0.01973337,1.0},
  // f 1012/1115/822 1156/1114/822 1157/1112/822
  {-0.27965656, 0.03633375, -0.01973337,1.0},
  {-0.24425827, 0.02090889, -0.01973337,1.0},
  {-0.24681623, 0.01738155, -0.01973337,1.0},
  // f 1011/1113/822 1012/1115/822 1157/1112/822
  {-0.28003085, 0.03528089, -0.01973337,1.0},
  {-0.27965656, 0.03633375, -0.01973337,1.0},
  {-0.24681623, 0.01738155, -0.01973337,1.0},
  // f 1013/1117/822 1155/1116/822 1156/1114/822
  {-0.27938333, 0.03741456, -0.01973337,1.0},
  {-0.24283025, 0.02322332, -0.01973337,1.0},
  {-0.24425827, 0.02090889, -0.01973337,1.0},
  // f 1012/1115/822 1013/1117/822 1156/1114/822
  {-0.27965656, 0.03633375, -0.01973337,1.0},
  {-0.27938333, 0.03741456, -0.01973337,1.0},
  {-0.24425827, 0.02090889, -0.01973337,1.0},
  // f 1013/1117/822 1154/1118/822 1155/1116/822
  {-0.27938333, 0.03741456, -0.01973337,1.0},
  {-0.24155261, 0.02558015, -0.01973337,1.0},
  {-0.24283025, 0.02322332, -0.01973337,1.0},
  // f 1014/1120/822 1153/1119/822 1154/1118/822
  {-0.27921134, 0.03852331, -0.01973337,1.0},
  {-0.23982103, 0.02943926, -0.01973337,1.0},
  {-0.24155261, 0.02558015, -0.01973337,1.0},
  // f 1013/1117/822 1014/1120/822 1154/1118/822
  {-0.27938333, 0.03741456, -0.01973337,1.0},
  {-0.27921134, 0.03852331, -0.01973337,1.0},
  {-0.24155261, 0.02558015, -0.01973337,1.0},
  // f 1015/1122/822 1152/1121/822 1153/1119/822
  {-0.27915674, 0.03920197, -0.01973337,1.0},
  {-0.23847421, 0.0334069, -0.01973337,1.0},
  {-0.23982103, 0.02943926, -0.01973337,1.0},
  // f 1014/1120/822 1015/1122/822 1153/1119/822
  {-0.27921134, 0.03852331, -0.01973337,1.0},
  {-0.27915674, 0.03920197, -0.01973337,1.0},
  {-0.23982103, 0.02943926, -0.01973337,1.0},
  // f 1016/1124/822 1151/1123/822 1152/1121/822
  {-0.27913854, 0.0398907, -0.01973337,1.0},
  {-0.2375122, 0.03748308, -0.01973337,1.0},
  {-0.23847421, 0.0334069, -0.01973337,1.0},
  // f 1015/1122/822 1016/1124/822 1152/1121/822
  {-0.27915674, 0.03920197, -0.01973337,1.0},
  {-0.27913854, 0.0398907, -0.01973337,1.0},
  {-0.23847421, 0.0334069, -0.01973337,1.0},
  // f 1017/1126/822 1150/1125/822 1151/1123/822
  {-0.27918667, 0.04097757, -0.01973337,1.0},
  {-0.23693499, 0.04166777, -0.01973337,1.0},
  {-0.2375122, 0.03748308, -0.01973337,1.0},
  // f 1016/1124/822 1017/1126/822 1151/1123/822
  {-0.27913854, 0.0398907, -0.01973337,1.0},
  {-0.27918667, 0.04097757, -0.01973337,1.0},
  {-0.2375122, 0.03748308, -0.01973337,1.0},
  // f 1018/1128/822 1149/1127/822 1150/1125/822
  {-0.27933127, 0.04204133, -0.01973337,1.0},
  {-0.23674259, 0.04596103, -0.01973337,1.0},
  {-0.23693499, 0.04166777, -0.01973337,1.0},
  // f 1017/1126/822 1018/1128/822 1150/1125/822
  {-0.27918667, 0.04097757, -0.01973337,1.0},
  {-0.27933127, 0.04204133, -0.01973337,1.0},
  {-0.23693499, 0.04166777, -0.01973337,1.0},
  // f 1018/1128/822 1148/1129/822 1149/1127/822
  {-0.27933127, 0.04204133, -0.01973337,1.0},
  {-0.23692529, 0.05005173, -0.01973337,1.0},
  {-0.23674259, 0.04596103, -0.01973337,1.0},
  // f 1019/1131/822 1147/1130/822 1148/1129/822
  {-0.2795721, 0.04308196, -0.01973337,1.0},
  {-0.23747315, 0.05399883, -0.01973337,1.0},
  {-0.23692529, 0.05005173, -0.01973337,1.0},
  // f 1018/1128/822 1019/1131/822 1148/1129/822
  {-0.27933127, 0.04204133, -0.01973337,1.0},
  {-0.2795721, 0.04308196, -0.01973337,1.0},
  {-0.23692529, 0.05005173, -0.01973337,1.0},
  // f 1020/1133/822 1146/1132/822 1147/1130/822
  {-0.27990934, 0.04409945, -0.01973337,1.0},
  {-0.23826675, 0.05738679, -0.01973337,1.0},
  {-0.23747315, 0.05399883, -0.01973337,1.0},
  // f 1019/1131/822 1020/1133/822 1147/1130/822
  {-0.2795721, 0.04308196, -0.01973337,1.0},
  {-0.27990934, 0.04409945, -0.01973337,1.0},
  {-0.23747315, 0.05399883, -0.01973337,1.0},
  // f 1020/1133/822 1145/1134/822 1146/1132/822
  {-0.27990934, 0.04409945, -0.01973337,1.0},
  {-0.23934902, 0.0606613, -0.01973337,1.0},
  {-0.23826675, 0.05738679, -0.01973337,1.0},
  // f 1021/1136/822 1144/1135/822 1145/1134/822
  {-0.280343, 0.04509383, -0.01973337,1.0},
  {-0.24071984, 0.06382233, -0.01973337,1.0},
  {-0.23934902, 0.0606613, -0.01973337,1.0},
  // f 1020/1133/822 1021/1136/822 1145/1134/822
  {-0.27990934, 0.04409945, -0.01973337,1.0},
  {-0.280343, 0.04509383, -0.01973337,1.0},
  {-0.23934902, 0.0606613, -0.01973337,1.0},
  // f 1021/1136/822 1143/1137/822 1144/1135/822
  {-0.280343, 0.04509383, -0.01973337,1.0},
  {-0.24237932, 0.06686991, -0.01973337,1.0},
  {-0.24071984, 0.06382233, -0.01973337,1.0},
  // f 1021/1136/822 1142/1138/822 1143/1137/822
  {-0.280343, 0.04509383, -0.01973337,1.0},
  {-0.24332435, 0.0683459, -0.01973337,1.0},
  {-0.24237932, 0.06686991, -0.01973337,1.0},
  // f 1022/1140/822 1141/1139/822 1142/1138/822
  {-0.28099045, 0.04625655, -0.01973337,1.0},
  {-0.24435572, 0.06978304, -0.01973337,1.0},
  {-0.24332435, 0.0683459, -0.01973337,1.0},
  // f 1021/1136/822 1022/1140/822 1142/1138/822
  {-0.280343, 0.04509383, -0.01973337,1.0},
  {-0.28099045, 0.04625655, -0.01973337,1.0},
  {-0.24332435, 0.0683459, -0.01973337,1.0},
  // f 1022/1140/822 1140/1141/822 1141/1139/822
  {-0.28099045, 0.04625655, -0.01973337,1.0},
  {-0.24636841, 0.07220456, -0.01973337,1.0},
  {-0.24435572, 0.06978304, -0.01973337,1.0},
  // f 1022/1140/822 1139/1142/822 1140/1141/822
  {-0.28099045, 0.04625655, -0.01973337,1.0},
  {-0.24899231, 0.07482632, -0.01973337,1.0},
  {-0.24636841, 0.07220456, -0.01973337,1.0},
  // f 1023/1144/822 1138/1143/822 1139/1142/822
  {-0.2817767, 0.04738599, -0.01973337,1.0},
  {-0.25196159, 0.07729269, -0.01973337,1.0},
  {-0.24899231, 0.07482632, -0.01973337,1.0},
  // f 1022/1140/822 1023/1144/822 1139/1142/822
  {-0.28099045, 0.04625655, -0.01973337,1.0},
  {-0.2817767, 0.04738599, -0.01973337,1.0},
  {-0.24899231, 0.07482632, -0.01973337,1.0},
  // f 1023/1144/822 1137/1145/822 1138/1143/822
  {-0.2817767, 0.04738599, -0.01973337,1.0},
  {-0.25571474, 0.07988161, -0.01973337,1.0},
  {-0.25196159, 0.07729269, -0.01973337,1.0},
  // f 1024/1147/822 1136/1146/822 1137/1145/822
  {-0.2827017, 0.04848211, -0.01973337,1.0},
  {-0.26039758, 0.08252749, -0.01973337,1.0},
  {-0.25571474, 0.07988161, -0.01973337,1.0},
  // f 1023/1144/822 1024/1147/822 1137/1145/822
  {-0.2817767, 0.04738599, -0.01973337,1.0},
  {-0.2827017, 0.04848211, -0.01973337,1.0},
  {-0.25571474, 0.07988161, -0.01973337,1.0},
  // f 1026/1149/822 1135/1148/822 1136/1146/822
  {-0.28480846, 0.0504095, -0.01973337,1.0},
  {-0.26727593, 0.08554854, -0.01973337,1.0},
  {-0.26039758, 0.08252749, -0.01973337,1.0},
  // f 1025/1150/822 1026/1149/822 1136/1146/822
  {-0.28395623, 0.04971883, -0.01973337,1.0},
  {-0.28480846, 0.0504095, -0.01973337,1.0},
  {-0.26039758, 0.08252749, -0.01973337,1.0},
  // f 1024/1147/822 1025/1150/822 1136/1146/822
  {-0.2827017, 0.04848211, -0.01973337,1.0},
  {-0.28395623, 0.04971883, -0.01973337,1.0},
  {-0.26039758, 0.08252749, -0.01973337,1.0},
  // f 1027/1152/822 1134/1151/822 1135/1148/822
  {-0.28610268, 0.05126418, -0.01973337,1.0},
  {-0.27602455, 0.08855294, -0.01973337,1.0},
  {-0.26727593, 0.08554854, -0.01973337,1.0},
  // f 1026/1149/822 1027/1152/822 1135/1148/822
  {-0.28480846, 0.0504095, -0.01973337,1.0},
  {-0.28610268, 0.05126418, -0.01973337,1.0},
  {-0.26727593, 0.08554854, -0.01973337,1.0},
  // f 1030/1154/822 1133/1153/822 1134/1151/822
  {-0.2980513, 0.05587222, -0.01973337,1.0},
  {-0.30139637, 0.0950054, -0.01973337,1.0},
  {-0.27602455, 0.08855294, -0.01973337,1.0},
  // f 1029/1155/822 1030/1154/822 1134/1151/822
  {-0.29195365, 0.05393434, -0.01973337,1.0},
  {-0.2980513, 0.05587222, -0.01973337,1.0},
  {-0.27602455, 0.08855294, -0.01973337,1.0},
  // f 1028/1156/822 1029/1155/822 1134/1151/822
  {-0.28834203, 0.0524445, -0.01973337,1.0},
  {-0.29195365, 0.05393434, -0.01973337,1.0},
  {-0.27602455, 0.08855294, -0.01973337,1.0},
  // f 1027/1152/822 1028/1156/822 1134/1151/822
  {-0.28610268, 0.05126418, -0.01973337,1.0},
  {-0.28834203, 0.0524445, -0.01973337,1.0},
  {-0.27602455, 0.08855294, -0.01973337,1.0},
  // f 1031/1158/822 1132/1157/822 1133/1153/822
  {-0.30631042, 0.05790896, -0.01973337,1.0},
  {-0.30683568, 0.09628404, -0.01973337,1.0},
  {-0.30139637, 0.0950054, -0.01973337,1.0},
  // f 1030/1154/822 1031/1158/822 1133/1153/822
  {-0.2980513, 0.05587222, -0.01973337,1.0},
  {-0.30631042, 0.05790896, -0.01973337,1.0},
  {-0.30139637, 0.0950054, -0.01973337,1.0},
  // f 1032/1160/822 1131/1159/822 1132/1157/822
  {-0.3220557, 0.06216396, -0.01973337,1.0},
  {-0.31112865, 0.09761478, -0.01973337,1.0},
  {-0.30683568, 0.09628404, -0.01973337,1.0},
  // f 1031/1158/822 1032/1160/822 1132/1157/822
  {-0.30631042, 0.05790896, -0.01973337,1.0},
  {-0.3220557, 0.06216396, -0.01973337,1.0},
  {-0.30683568, 0.09628404, -0.01973337,1.0},
  // f 1032/1160/822 1130/1161/822 1131/1159/822
  {-0.3220557, 0.06216396, -0.01973337,1.0},
  {-0.31342909, 0.09856663, -0.01973337,1.0},
  {-0.31112865, 0.09761478, -0.01973337,1.0},
  // f 1033/1163/822 1129/1162/822 1130/1161/822
  {-0.33171731, 0.06576931, -0.01973337,1.0},
  {-0.31465751, 0.09921506, -0.01973337,1.0},
  {-0.31342909, 0.09856663, -0.01973337,1.0},
  // f 1032/1160/822 1033/1163/822 1130/1161/822
  {-0.3220557, 0.06216396, -0.01973337,1.0},
  {-0.33171731, 0.06576931, -0.01973337,1.0},
  {-0.31342909, 0.09856663, -0.01973337,1.0},
  // f 1033/1163/822 1128/1164/822 1129/1162/822
  {-0.33171731, 0.06576931, -0.01973337,1.0},
  {-0.31549466, 0.09976391, -0.01973337,1.0},
  {-0.31465751, 0.09921506, -0.01973337,1.0},
  // f 1034/1166/822 1127/1165/822 1128/1164/822
  {-0.33843592, 0.06902439, -0.01973337,1.0},
  {-0.31604224, 0.10020851, -0.01973337,1.0},
  {-0.31549466, 0.09976391, -0.01973337,1.0},
  // f 1033/1163/822 1034/1166/822 1128/1164/822
  {-0.33171731, 0.06576931, -0.01973337,1.0},
  {-0.33843592, 0.06902439, -0.01973337,1.0},
  {-0.31549466, 0.09976391, -0.01973337,1.0},
  // f 1034/1166/822 1126/1167/822 1127/1165/822
  {-0.33843592, 0.06902439, -0.01973337,1.0},
  {-0.31653062, 0.10064447, -0.01973337,1.0},
  {-0.31604224, 0.10020851, -0.01973337,1.0},
  // f 1034/1166/822 1125/1168/822 1126/1167/822
  {-0.33843592, 0.06902439, -0.01973337,1.0},
  {-0.31697828, 0.1010891, -0.01973337,1.0},
  {-0.31653062, 0.10064447, -0.01973337,1.0},
  // f 1035/1170/822 1124/1169/822 1125/1168/822
  {-0.34292337, 0.07174607, -0.01973337,1.0},
  {-0.31757355, 0.10177226, -0.01973337,1.0},
  {-0.31697828, 0.1010891, -0.01973337,1.0},
  // f 1034/1166/822 1035/1170/822 1125/1168/822
  {-0.33843592, 0.06902439, -0.01973337,1.0},
  {-0.34292337, 0.07174607, -0.01973337,1.0},
  {-0.31697828, 0.1010891, -0.01973337,1.0},
  // f 1037/1172/822 1123/1171/822 1124/1169/822
  {-0.34726104, 0.07506003, -0.01973337,1.0},
  {-0.31807724, 0.10247475, -0.01973337,1.0},
  {-0.31757355, 0.10177226, -0.01973337,1.0},
  // f 1036/1173/822 1037/1172/822 1124/1169/822
  {-0.34519771, 0.07337592, -0.01973337,1.0},
  {-0.34726104, 0.07506003, -0.01973337,1.0},
  {-0.31757355, 0.10177226, -0.01973337,1.0},
  // f 1035/1170/822 1036/1173/822 1124/1169/822
  {-0.34292337, 0.07174607, -0.01973337,1.0},
  {-0.34519771, 0.07337592, -0.01973337,1.0},
  {-0.31757355, 0.10177226, -0.01973337,1.0},
  // f 1039/1175/822 1122/1174/822 1123/1171/822
  {-0.35247061, 0.08053235, -0.01973337,1.0},
  {-0.31842694, 0.10307506, -0.01973337,1.0},
  {-0.31807724, 0.10247475, -0.01973337,1.0},
  // f 1038/1176/822 1039/1175/822 1123/1171/822
  {-0.35004342, 0.07771154, -0.01973337,1.0},
  {-0.35247061, 0.08053235, -0.01973337,1.0},
  {-0.31807724, 0.10247475, -0.01973337,1.0},
  // f 1037/1172/822 1038/1176/822 1123/1171/822
  {-0.34726104, 0.07506003, -0.01973337,1.0},
  {-0.35004342, 0.07771154, -0.01973337,1.0},
  {-0.31807724, 0.10247475, -0.01973337,1.0},
  // f 1041/1178/822 1121/1177/822 1122/1174/822
  {-0.35572663, 0.08560983, -0.01973337,1.0},
  {-0.31871322, 0.10368883, -0.01973337,1.0},
  {-0.31842694, 0.10307506, -0.01973337,1.0},
  // f 1040/1179/822 1041/1178/822 1122/1174/822
  {-0.35422194, 0.08301231, -0.01973337,1.0},
  {-0.35572663, 0.08560983, -0.01973337,1.0},
  {-0.31842694, 0.10307506, -0.01973337,1.0},
  // f 1039/1175/822 1040/1179/822 1122/1174/822
  {-0.35247061, 0.08053235, -0.01973337,1.0},
  {-0.35422194, 0.08301231, -0.01973337,1.0},
  {-0.31842694, 0.10307506, -0.01973337,1.0},
  // f 1043/1181/822 1120/1180/822 1121/1177/822
  {-0.35799599, 0.09115754, -0.01973337,1.0},
  {-0.31893578, 0.1043161, -0.01973337,1.0},
  {-0.31871322, 0.10368883, -0.01973337,1.0},
  // f 1042/1182/822 1043/1181/822 1121/1177/822
  {-0.35698465, 0.0883249, -0.01973337,1.0},
  {-0.35799599, 0.09115754, -0.01973337,1.0},
  {-0.31871322, 0.10368883, -0.01973337,1.0},
  // f 1041/1178/822 1042/1182/822 1121/1177/822
  {-0.35572663, 0.08560983, -0.01973337,1.0},
  {-0.35698465, 0.0883249, -0.01973337,1.0},
  {-0.31871322, 0.10368883, -0.01973337,1.0},
  // f 1045/1184/822 1119/1183/822 1120/1180/822
  {-0.35935268, 0.0978031, -0.01973337,1.0},
  {-0.31909472, 0.10495687, -0.01973337,1.0},
  {-0.31893578, 0.1043161, -0.01973337,1.0},
  // f 1044/1185/822 1045/1184/822 1120/1180/822
  {-0.35876068, 0.09410771, -0.01973337,1.0},
  {-0.35935268, 0.0978031, -0.01973337,1.0},
  {-0.31893578, 0.1043161, -0.01973337,1.0},
  // f 1043/1181/822 1044/1185/822 1120/1180/822
  {-0.35799599, 0.09115754, -0.01973337,1.0},
  {-0.35876068, 0.09410771, -0.01973337,1.0},
  {-0.31893578, 0.1043161, -0.01973337,1.0},
  // f 1047/1187/822 1118/1186/822 1119/1183/822
  {-0.35959435, 0.10232831, -0.01973337,1.0},
  {-0.31919011, 0.10561109, -0.01973337,1.0},
  {-0.31909472, 0.10495687, -0.01973337,1.0},
  // f 1046/1188/822 1047/1187/822 1119/1183/822
  {-0.35951549, 0.09971424, -0.01973337,1.0},
  {-0.35959435, 0.10232831, -0.01973337,1.0},
  {-0.31909472, 0.10495687, -0.01973337,1.0},
  // f 1045/1184/822 1046/1188/822 1119/1183/822
  {-0.35935268, 0.0978031, -0.01973337,1.0},
  {-0.35951549, 0.09971424, -0.01973337,1.0},
  {-0.31909472, 0.10495687, -0.01973337,1.0},
  // f 1049/1190/822 1117/1189/822 1118/1186/822
  {-0.35894817, 0.10917331, -0.01973337,1.0},
  {-0.31922197, 0.10627884, -0.01973337,1.0},
  {-0.31919011, 0.10561109, -0.01973337,1.0},
  // f 1048/1191/822 1049/1190/822 1118/1186/822
  {-0.35943285, 0.10578167, -0.01973337,1.0},
  {-0.35894817, 0.10917331, -0.01973337,1.0},
  {-0.31919011, 0.10561109, -0.01973337,1.0},
  // f 1047/1187/822 1048/1191/822 1118/1186/822
  {-0.35959435, 0.10232831, -0.01973337,1.0},
  {-0.35943285, 0.10578167, -0.01973337,1.0},
  {-0.31919011, 0.10561109, -0.01973337,1.0},
  // f 1050/1193/822 1116/1192/822 1117/1189/822
  {-0.35814032, 0.11250332, -0.01973337,1.0},
  {-0.31919718, 0.1070099, -0.01973337,1.0},
  {-0.31922197, 0.10627884, -0.01973337,1.0},
  // f 1049/1190/822 1050/1193/822 1117/1189/822
  {-0.35894817, 0.10917331, -0.01973337,1.0},
  {-0.35814032, 0.11250332, -0.01973337,1.0},
  {-0.31922197, 0.10627884, -0.01973337,1.0},
  // f 1051/1195/822 1115/1194/822 1116/1192/822
  {-0.35700944, 0.11577167, -0.01973337,1.0},
  {-0.31914634, 0.10754646, -0.01973337,1.0},
  {-0.31919718, 0.1070099, -0.01973337,1.0},
  // f 1050/1193/822 1051/1195/822 1116/1192/822
  {-0.35814032, 0.11250332, -0.01973337,1.0},
  {-0.35700944, 0.11577167, -0.01973337,1.0},
  {-0.31919718, 0.1070099, -0.01973337,1.0},
  // f 1053/1197/822 1114/1196/822 1115/1194/822
  {-0.35448223, 0.12095121, -0.01973337,1.0},
  {-0.31899998, 0.10841836, -0.01973337,1.0},
  {-0.31914634, 0.10754646, -0.01973337,1.0},
  // f 1052/1198/822 1053/1197/822 1115/1194/822
  {-0.35555524, 0.11897835, -0.01973337,1.0},
  {-0.35448223, 0.12095121, -0.01973337,1.0},
  {-0.31914634, 0.10754646, -0.01973337,1.0},
  // f 1051/1195/822 1052/1198/822 1115/1194/822
  {-0.35700944, 0.11577167, -0.01973337,1.0},
  {-0.35555524, 0.11897835, -0.01973337,1.0},
  {-0.31914634, 0.10754646, -0.01973337,1.0},
  // f 1055/1200/822 1113/1199/822 1114/1196/822
  {-0.3513914, 0.12553017, -0.01973337,1.0},
  {-0.31877634, 0.10926232, -0.01973337,1.0},
  {-0.31899998, 0.10841836, -0.01973337,1.0},
  // f 1054/1201/822 1055/1200/822 1114/1196/822
  {-0.35328317, 0.12289994, -0.01973337,1.0},
  {-0.3513914, 0.12553017, -0.01973337,1.0},
  {-0.31899998, 0.10841836, -0.01973337,1.0},
  // f 1053/1197/822 1054/1201/822 1114/1196/822
  {-0.35448223, 0.12095121, -0.01973337,1.0},
  {-0.35328317, 0.12289994, -0.01973337,1.0},
  {-0.31899998, 0.10841836, -0.01973337,1.0},
  // f 1056/1203/822 1112/1202/822 1113/1199/822
  {-0.34925041, 0.12799795, -0.01973337,1.0},
  {-0.31847572, 0.11007833, -0.01973337,1.0},
  {-0.31877634, 0.10926232, -0.01973337,1.0},
  // f 1055/1200/822 1056/1203/822 1113/1199/822
  {-0.3513914, 0.12553017, -0.01973337,1.0},
  {-0.34925041, 0.12799795, -0.01973337,1.0},
  {-0.31877634, 0.10926232, -0.01973337,1.0},
  // f 1058/1205/822 1111/1204/822 1112/1202/822
  {-0.34342036, 0.13302872, -0.01973337,1.0},
  {-0.31809804, 0.11086644, -0.01973337,1.0},
  {-0.31847572, 0.11007833, -0.01973337,1.0},
  // f 1057/1206/822 1058/1205/822 1112/1202/822
  {-0.3464981, 0.13061942, -0.01973337,1.0},
  {-0.34342036, 0.13302872, -0.01973337,1.0},
  {-0.31847572, 0.11007833, -0.01973337,1.0},
  // f 1056/1203/822 1057/1206/822 1112/1202/822
  {-0.34925041, 0.12799795, -0.01973337,1.0},
  {-0.3464981, 0.13061942, -0.01973337,1.0},
  {-0.31847572, 0.11007833, -0.01973337,1.0},
  // f 1059/1208/822 1110/1207/822 1111/1204/822
  {-0.34001693, 0.13522591, -0.01973337,1.0},
  {-0.31764328, 0.1116266, -0.01973337,1.0},
  {-0.31809804, 0.11086644, -0.01973337,1.0},
  // f 1058/1205/822 1059/1208/822 1111/1204/822
  {-0.34342036, 0.13302872, -0.01973337,1.0},
  {-0.34001693, 0.13522591, -0.01973337,1.0},
  {-0.31809804, 0.11086644, -0.01973337,1.0},
  // f 1060/1210/822 1109/1209/822 1110/1207/822
  {-0.3372508, 0.13673456, -0.01973337,1.0},
  {-0.31699577, 0.11250187, -0.01973337,1.0},
  {-0.31764328, 0.1116266, -0.01973337,1.0},
  // f 1059/1208/822 1060/1210/822 1110/1207/822
  {-0.34001693, 0.13522591, -0.01973337,1.0},
  {-0.3372508, 0.13673456, -0.01973337,1.0},
  {-0.31764328, 0.1116266, -0.01973337,1.0},
  // f 1061/1212/822 1108/1211/822 1109/1209/822
  {-0.33430135, 0.13812394, -0.01973337,1.0},
  {-0.31623718, 0.1133369, -0.01973337,1.0},
  {-0.31699577, 0.11250187, -0.01973337,1.0},
  // f 1060/1210/822 1061/1212/822 1109/1209/822
  {-0.3372508, 0.13673456, -0.01973337,1.0},
  {-0.33430135, 0.13812394, -0.01973337,1.0},
  {-0.31699577, 0.11250187, -0.01973337,1.0},
  // f 1062/1214/822 1107/1213/822 1108/1211/822
  {-0.33000627, 0.13975506, -0.01973337,1.0},
  {-0.31536773, 0.1141317, -0.01973337,1.0},
  {-0.31623718, 0.1133369, -0.01973337,1.0},
  // f 1061/1212/822 1062/1214/822 1108/1211/822
  {-0.33430135, 0.13812394, -0.01973337,1.0},
  {-0.33000627, 0.13975506, -0.01973337,1.0},
  {-0.31623718, 0.1133369, -0.01973337,1.0},
  // f 1062/1214/822 1106/1215/822 1107/1213/822
  {-0.33000627, 0.13975506, -0.01973337,1.0},
  {-0.31472713, 0.1146275, -0.01973337,1.0},
  {-0.31536773, 0.1141317, -0.01973337,1.0},
  // f 1063/1217/822 1105/1216/822 1106/1215/822
  {-0.32459944, 0.14125092, -0.01973337,1.0},
  {-0.31403819, 0.115082, -0.01973337,1.0},
  {-0.31472713, 0.1146275, -0.01973337,1.0},
  // f 1062/1214/822 1063/1217/822 1106/1215/822
  {-0.33000627, 0.13975506, -0.01973337,1.0},
  {-0.32459944, 0.14125092, -0.01973337,1.0},
  {-0.31472713, 0.1146275, -0.01973337,1.0},
  // f 1063/1217/822 1104/1218/822 1105/1216/822
  {-0.32459944, 0.14125092, -0.01973337,1.0},
  {-0.31310949, 0.11559201, -0.01973337,1.0},
  {-0.31403819, 0.115082, -0.01973337,1.0},
  // f 1064/1220/822 1103/1219/822 1104/1218/822
  {-0.31716415, 0.1425917, -0.01973337,1.0},
  {-0.31189597, 0.11611877, -0.01973337,1.0},
  {-0.31310949, 0.11559201, -0.01973337,1.0},
  // f 1063/1217/822 1064/1220/822 1104/1218/822
  {-0.32459944, 0.14125092, -0.01973337,1.0},
  {-0.31716415, 0.1425917, -0.01973337,1.0},
  {-0.31310949, 0.11559201, -0.01973337,1.0},
  // f 1064/1220/822 1102/1221/822 1103/1219/822
  {-0.31716415, 0.1425917, -0.01973337,1.0},
  {-0.31057414, 0.11655263, -0.01973337,1.0},
  {-0.31189597, 0.11611877, -0.01973337,1.0},
  // f 1064/1220/822 1101/1222/822 1102/1221/822
  {-0.31716415, 0.1425917, -0.01973337,1.0},
  {-0.30889514, 0.11694124, -0.01973337,1.0},
  {-0.31057414, 0.11655263, -0.01973337,1.0},
  // f 1065/1224/822 1100/1223/822 1101/1222/822
  {-0.30881959, 0.14339617, -0.01973337,1.0},
  {-0.30679607, 0.11723047, -0.01973337,1.0},
  {-0.30889514, 0.11694124, -0.01973337,1.0},
  // f 1064/1220/822 1065/1224/822 1101/1222/822
  {-0.31716415, 0.1425917, -0.01973337,1.0},
  {-0.30881959, 0.14339617, -0.01973337,1.0},
  {-0.30889514, 0.11694124, -0.01973337,1.0},
  // f 1065/1224/822 1099/1225/822 1100/1223/822
  {-0.30881959, 0.14339617, -0.01973337,1.0},
  {-0.30390158, 0.11735957, -0.01973337,1.0},
  {-0.30679607, 0.11723047, -0.01973337,1.0},
  // f 1066/1227/822 1098/1226/822 1099/1225/822
  {-0.29956564, 0.1436643, -0.01973337,1.0},
  {-0.30173922, 0.11729716, -0.01973337,1.0},
  {-0.30390158, 0.11735957, -0.01973337,1.0},
  // f 1065/1224/822 1066/1227/822 1099/1225/822
  {-0.30881959, 0.14339617, -0.01973337,1.0},
  {-0.29956564, 0.1436643, -0.01973337,1.0},
  {-0.30390158, 0.11735957, -0.01973337,1.0},
  // f 1066/1227/822 1097/1228/822 1098/1226/822
  {-0.29956564, 0.1436643, -0.01973337,1.0},
  {-0.30002517, 0.11714971, -0.01973337,1.0},
  {-0.30173922, 0.11729716, -0.01973337,1.0},
  // f 1066/1227/822 1096/1229/822 1097/1228/822
  {-0.29956564, 0.1436643, -0.01973337,1.0},
  {-0.29744899, 0.11673349, -0.01973337,1.0},
  {-0.30002517, 0.11714971, -0.01973337,1.0},
  // f 1067/1231/822 1095/1230/822 1096/1229/822
  {-0.28827497, 0.14317694, -0.01973337,1.0},
  {-0.2953628, 0.11618715, -0.01973337,1.0},
  {-0.29744899, 0.11673349, -0.01973337,1.0},
  // f 1066/1227/822 1067/1231/822 1096/1229/822
  {-0.29956564, 0.1436643, -0.01973337,1.0},
  {-0.28827497, 0.14317694, -0.01973337,1.0},
  {-0.29744899, 0.11673349, -0.01973337,1.0},
  // f 1067/1231/822 1094/1232/822 1095/1230/822
  {-0.28827497, 0.14317694, -0.01973337,1.0},
  {-0.29369941, 0.11558357, -0.01973337,1.0},
  {-0.2953628, 0.11618715, -0.01973337,1.0},
  // f 1067/1231/822 1093/1233/822 1094/1232/822
  {-0.28827497, 0.14317694, -0.01973337,1.0},
  {-0.29215118, 0.11485515, -0.01973337,1.0},
  {-0.29369941, 0.11558357, -0.01973337,1.0},
  // f 1068/1235/822 1092/1234/822 1093/1233/822
  {-0.27996331, 0.14205326, -0.01973337,1.0},
  {-0.29071814, 0.11400183, -0.01973337,1.0},
  {-0.29215118, 0.11485515, -0.01973337,1.0},
  // f 1067/1231/822 1068/1235/822 1093/1233/822
  {-0.28827497, 0.14317694, -0.01973337,1.0},
  {-0.27996331, 0.14205326, -0.01973337,1.0},
  {-0.29215118, 0.11485515, -0.01973337,1.0},
  // f 1068/1235/822 1091/1236/822 1092/1234/822
  {-0.27996331, 0.14205326, -0.01973337,1.0},
  {-0.28940025, 0.11302363, -0.01973337,1.0},
  {-0.29071814, 0.11400183, -0.01973337,1.0},
  // f 1069/1238/822 1090/1237/822 1091/1236/822
  {-0.27327496, 0.14050667, -0.01973337,1.0},
  {-0.2885831, 0.11229659, -0.01973337,1.0},
  {-0.28940025, 0.11302363, -0.01973337,1.0},
  // f 1068/1235/822 1069/1238/822 1091/1236/822
  {-0.27996331, 0.14205326, -0.01973337,1.0},
  {-0.27327496, 0.14050667, -0.01973337,1.0},
  {-0.28940025, 0.11302363, -0.01973337,1.0},
  // f 1069/1238/822 1089/1239/822 1090/1237/822
  {-0.27327496, 0.14050667, -0.01973337,1.0},
  {-0.28781238, 0.11150294, -0.01973337,1.0},
  {-0.2885831, 0.11229659, -0.01973337,1.0},
  // f 1070/1241/822 1088/1240/822 1089/1239/822
  {-0.26796278, 0.13873047, -0.01973337,1.0},
  {-0.28674284, 0.11018759, -0.01973337,1.0},
  {-0.28781238, 0.11150294, -0.01973337,1.0},
  // f 1069/1238/822 1070/1241/822 1089/1239/822
  {-0.27327496, 0.14050667, -0.01973337,1.0},
  {-0.26796278, 0.13873047, -0.01973337,1.0},
  {-0.28781238, 0.11150294, -0.01973337,1.0},
  // f 1072/1243/822 1087/1242/822 1088/1240/822
  {-0.26124883, 0.1355084, -0.01973337,1.0},
  {-0.28577727, 0.10872237, -0.01973337,1.0},
  {-0.28674284, 0.11018759, -0.01973337,1.0},
  // f 1071/1244/822 1072/1243/822 1088/1240/822
  {-0.26381099, 0.13689388, -0.01973337,1.0},
  {-0.26124883, 0.1355084, -0.01973337,1.0},
  {-0.28674284, 0.11018759, -0.01973337,1.0},
  // f 1070/1241/822 1071/1244/822 1088/1240/822
  {-0.26796278, 0.13873047, -0.01973337,1.0},
  {-0.26381099, 0.13689388, -0.01973337,1.0},
  {-0.28674284, 0.11018759, -0.01973337,1.0},
  // f 1074/1246/822 1086/1245/822 1087/1242/822
  {-0.2555505, 0.1314625, -0.01973337,1.0},
  {-0.28491592, 0.10710733, -0.01973337,1.0},
  {-0.28577727, 0.10872237, -0.01973337,1.0},
  // f 1073/1247/822 1074/1246/822 1087/1242/822
  {-0.2582778, 0.13359529, -0.01973337,1.0},
  {-0.2555505, 0.1314625, -0.01973337,1.0},
  {-0.28577727, 0.10872237, -0.01973337,1.0},
  // f 1072/1243/822 1073/1247/822 1087/1242/822
  {-0.26124883, 0.1355084, -0.01973337,1.0},
  {-0.2582778, 0.13359529, -0.01973337,1.0},
  {-0.28577727, 0.10872237, -0.01973337,1.0},
  // f 1076/1249/822 1085/1248/822 1086/1245/822
  {-0.25078672, 0.12648289, -0.01973337,1.0},
  {-0.28404248, 0.1050337, -0.01973337,1.0},
  {-0.28491592, 0.10710733, -0.01973337,1.0},
  // f 1075/1250/822 1076/1249/822 1086/1245/822
  {-0.25305346, 0.12909168, -0.01973337,1.0},
  {-0.25078672, 0.12648289, -0.01973337,1.0},
  {-0.28491592, 0.10710733, -0.01973337,1.0},
  // f 1074/1246/822 1075/1250/822 1086/1245/822
  {-0.2555505, 0.1314625, -0.01973337,1.0},
  {-0.25305346, 0.12909168, -0.01973337,1.0},
  {-0.28491592, 0.10710733, -0.01973337,1.0},
  // f 1079/1252/822 1084/1251/822 1085/1248/822
  {-0.24508065, 0.11653544, -0.01973337,1.0},
  {-0.28321767, 0.10241412, -0.01973337,1.0},
  {-0.28404248, 0.1050337, -0.01973337,1.0},
  // f 1078/1253/822 1079/1252/822 1085/1248/822
  {-0.24694414, 0.12055133, -0.01973337,1.0},
  {-0.24508065, 0.11653544, -0.01973337,1.0},
  {-0.28404248, 0.1050337, -0.01973337,1.0},
  // f 1077/1254/822 1078/1253/822 1085/1248/822
  {-0.24875031, 0.1236361, -0.01973337,1.0},
  {-0.24694414, 0.12055133, -0.01973337,1.0},
  {-0.28404248, 0.1050337, -0.01973337,1.0},
  // f 1076/1249/822 1077/1254/822 1085/1248/822
  {-0.25078672, 0.12648289, -0.01973337,1.0},
  {-0.24875031, 0.1236361, -0.01973337,1.0},
  {-0.28404248, 0.1050337, -0.01973337,1.0},
  // f 1080/1256/822 1083/1255/822 1084/1251/822
  {-0.2435489, 0.11217681, -0.01973337,1.0},
  {-0.28251091, 0.09914861, -0.01973337,1.0},
  {-0.28321767, 0.10241412, -0.01973337,1.0},
  // f 1079/1252/822 1080/1256/822 1084/1251/822
  {-0.24508065, 0.11653544, -0.01973337,1.0},
  {-0.2435489, 0.11217681, -0.01973337,1.0},
  {-0.28321767, 0.10241412, -0.01973337,1.0},
  // f 1081/1258/822 1082/1257/822 1083/1255/822
  {-0.24234867, 0.1074755, -0.01973337,1.0},
  {-0.24136762, 0.10155746, -0.01973337,1.0},
  {-0.28251091, 0.09914861, -0.01973337,1.0},
  // f 1080/1256/822 1081/1258/822 1083/1255/822
  {-0.2435489, 0.11217681, -0.01973337,1.0},
  {-0.24234867, 0.1074755, -0.01973337,1.0},
  {-0.28251091, 0.09914861, -0.01973337,1.0},
  // f 1187/1261/822 1189/1260/822 1190/1259/822
  {-0.17112531, 0.03478392, -0.01973337,1.0},
  {-0.10300279, 0, -0.01973337,1.0},
  {-0.21477383, 0, -0.01973337,1.0},
  // f 1185/1262/822 1187/1261/822 1190/1259/822
  {-0.21477383, 0.14125545, -0.01973337,1.0},
  {-0.17112531, 0.03478392, -0.01973337,1.0},
  {-0.21477383, 0, -0.01973337,1.0},
  // f 1187/1261/822 1188/1263/822 1189/1260/822
  {-0.17112531, 0.03478392, -0.01973337,1.0},
  {-0.10300279, 0.03478392, -0.01973337,1.0},
  {-0.10300279, 0, -0.01973337,1.0},
  // f 1185/1262/822 1186/1264/822 1187/1261/822
  {-0.21477383, 0.14125545, -0.01973337,1.0},
  {-0.17112531, 0.14125545, -0.01973337,1.0},
  {-0.17112531, 0.03478392, -0.01973337,1.0},
  // f 1191/1267/822 1209/1266/822 1210/1265/822
  {-0.00057813, 0.1436643, -0.01973337,1.0},
  {-0.03574746, 0.13851516, -0.01973337,1.0},
  {-0.03304954, 0.1436643, -0.01973337,1.0},
  // f 1191/1267/822 1208/1268/822 1209/1266/822
  {-0.00057813, 0.1436643, -0.01973337,1.0},
  {-0.03883079, 0.13370511, -0.01973337,1.0},
  {-0.03574746, 0.13851516, -0.01973337,1.0},
  // f 1194/1270/822 1207/1269/822 1208/1268/822
  {-0.04027612, 0.0941382, -0.01973337,1.0},
  {-0.04229955, 0.12923431, -0.01973337,1.0},
  {-0.03883079, 0.13370511, -0.01973337,1.0},
  // f 1191/1267/822 1194/1270/822 1208/1268/822
  {-0.00057813, 0.1436643, -0.01973337,1.0},
  {-0.04027612, 0.0941382, -0.01973337,1.0},
  {-0.03883079, 0.13370511, -0.01973337,1.0},
  // f 1194/1270/822 1206/1271/822 1207/1269/822
  {-0.04027612, 0.0941382, -0.01973337,1.0},
  {-0.04615372, 0.12510261, -0.01973337,1.0},
  {-0.04229955, 0.12923431, -0.01973337,1.0},
  // f 1194/1270/822 1205/1272/822 1206/1271/822
  {-0.04027612, 0.0941382, -0.01973337,1.0},
  {-0.05039332, 0.12131011, -0.01973337,1.0},
  {-0.04615372, 0.12510261, -0.01973337,1.0},
  // f 1194/1270/822 1204/1273/822 1205/1272/822
  {-0.04027612, 0.0941382, -0.01973337,1.0},
  {-0.05269908, 0.11952468, -0.01973337,1.0},
  {-0.05039332, 0.12131011, -0.01973337,1.0},
  // f 1194/1270/822 1203/1274/822 1204/1273/822
  {-0.04027612, 0.0941382, -0.01973337,1.0},
  {-0.05518406, 0.11779124, -0.01973337,1.0},
  {-0.05269908, 0.11952468, -0.01973337,1.0},
  // f 1194/1270/822 1202/1275/822 1203/1274/822
  {-0.04027612, 0.0941382, -0.01973337,1.0},
  {-0.06128185, 0.11416088, -0.01973337,1.0},
  {-0.05518406, 0.11779124, -0.01973337,1.0},
  // f 1194/1270/822 1195/1276/822 1202/1275/822
  {-0.04027612, 0.0941382, -0.01973337,1.0},
  {-0.05012761, 0.08727793, -0.01973337,1.0},
  {-0.06128185, 0.11416088, -0.01973337,1.0},
  // f 1195/1276/822 1201/1277/822 1202/1275/822
  {-0.05012761, 0.08727793, -0.01973337,1.0},
  {-0.06960668, 0.11019515, -0.01973337,1.0},
  {-0.06128185, 0.11416088, -0.01973337,1.0},
  // f 1197/1279/822 1198/1278/822 1201/1277/822
  {-0.05892068, 0.0822866, -0.01973337,1.0},
  {-0.06907402, 0.0778543, -0.01973337,1.0},
  {-0.06960668, 0.11019515, -0.01973337,1.0},
  // f 1196/1280/822 1197/1279/822 1201/1277/822
  {-0.05456021, 0.08462132, -0.01973337,1.0},
  {-0.05892068, 0.0822866, -0.01973337,1.0},
  {-0.06960668, 0.11019515, -0.01973337,1.0},
  // f 1195/1276/822 1196/1280/822 1201/1277/822
  {-0.05012761, 0.08727793, -0.01973337,1.0},
  {-0.05456021, 0.08462132, -0.01973337,1.0},
  {-0.06960668, 0.11019515, -0.01973337,1.0},
  // f 1198/1278/822 1200/1281/822 1201/1277/822
  {-0.06907402, 0.0778543, -0.01973337,1.0},
  {-0.08151577, 0.10579707, -0.01973337,1.0},
  {-0.06960668, 0.11019515, -0.01973337,1.0},
  // f 1198/1278/822 1199/1282/822 1200/1281/822
  {-0.06907402, 0.0778543, -0.01973337,1.0},
  {-0.08151577, 0.07361472, -0.01973337,1.0},
  {-0.08151577, 0.10579707, -0.01973337,1.0},
  // f 1192/1284/822 1193/1283/822 1194/1270/822
  {-0.00057813, 0, -0.01973337,1.0},
  {-0.04027612, 0, -0.01973337,1.0},
  {-0.04027612, 0.0941382, -0.01973337,1.0},
  // f 1191/1267/822 1192/1284/822 1194/1270/822
  {-0.00057813, 0.1436643, -0.01973337,1.0},
  {-0.00057813, 0, -0.01973337,1.0},
  {-0.04027612, 0.0941382, -0.01973337,1.0},
  // f 1214/1287/822 1332/1286/822 1333/1285/822
  {0.18692741, 0.02033077, -0.01973337,1.0},
  {0.14867473, 0.03786726, -0.01973337,1.0},
  {0.14867473, 0.05116415, -0.01973337,1.0},
  // f 1213/1288/822 1214/1287/822 1333/1285/822
  {0.18692741, 0.08055224, -0.01973337,1.0},
  {0.18692741, 0.02033077, -0.01973337,1.0},
  {0.14867473, 0.05116415, -0.01973337,1.0},
  // f 1212/1289/822 1213/1288/822 1333/1285/822
  {0.11947938, 0.08055224, -0.01973337,1.0},
  {0.18692741, 0.08055224, -0.01973337,1.0},
  {0.14867473, 0.05116415, -0.01973337,1.0},
  // f 1211/1290/822 1212/1289/822 1333/1285/822
  {0.11947938, 0.05116415, -0.01973337,1.0},
  {0.11947938, 0.08055224, -0.01973337,1.0},
  {0.14867473, 0.05116415, -0.01973337,1.0},
  // f 1215/1292/822 1216/1291/822 1332/1286/822
  {0.17215054, 0.01111929, -0.01973337,1.0},
  {0.16200872, 0.0059812, -0.01973337,1.0},
  {0.14867473, 0.03786726, -0.01973337,1.0},
  // f 1214/1287/822 1215/1292/822 1332/1286/822
  {0.18692741, 0.02033077, -0.01973337,1.0},
  {0.17215054, 0.01111929, -0.01973337,1.0},
  {0.14867473, 0.03786726, -0.01973337,1.0},
  // f 1216/1291/822 1331/1293/822 1332/1286/822
  {0.16200872, 0.0059812, -0.01973337,1.0},
  {0.14306615, 0.03473994, -0.01973337,1.0},
  {0.14867473, 0.03786726, -0.01973337,1.0},
  // f 1217/1295/822 1218/1294/822 1331/1293/822
  {0.15755193, 0.00410277, -0.01973337,1.0},
  {0.15267344, 0.00236068, -0.01973337,1.0},
  {0.14306615, 0.03473994, -0.01973337,1.0},
  // f 1216/1291/822 1217/1295/822 1331/1293/822
  {0.16200872, 0.0059812, -0.01973337,1.0},
  {0.15755193, 0.00410277, -0.01973337,1.0},
  {0.14306615, 0.03473994, -0.01973337,1.0},
  // f 1218/1294/822 1330/1296/822 1331/1293/822
  {0.15267344, 0.00236068, -0.01973337,1.0},
  {0.13790309, 0.03230322, -0.01973337,1.0},
  {0.14306615, 0.03473994, -0.01973337,1.0},
  // f 1218/1294/822 1219/1297/822 1330/1296/822
  {0.15267344, 0.00236068, -0.01973337,1.0},
  {0.1452087, 0.00034603, -0.01973337,1.0},
  {0.13790309, 0.03230322, -0.01973337,1.0},
  // f 1219/1297/822 1329/1298/822 1330/1296/822
  {0.1452087, 0.00034603, -0.01973337,1.0},
  {0.13345076, 0.03064069, -0.01973337,1.0},
  {0.13790309, 0.03230322, -0.01973337,1.0},
  // f 1219/1297/822 1220/1299/822 1329/1298/822
  {0.1452087, 0.00034603, -0.01973337,1.0},
  {0.13569312, -0.00130996, -0.01973337,1.0},
  {0.13345076, 0.03064069, -0.01973337,1.0},
  // f 1220/1299/822 1328/1300/822 1329/1298/822
  {0.13569312, -0.00130996, -0.01973337,1.0},
  {0.13076441, 0.02987755, -0.01973337,1.0},
  {0.13345076, 0.03064069, -0.01973337,1.0},
  // f 1220/1299/822 1327/1301/822 1328/1300/822
  {0.13569312, -0.00130996, -0.01973337,1.0},
  {0.12802409, 0.02928401, -0.01973337,1.0},
  {0.13076441, 0.02987755, -0.01973337,1.0},
  // f 1221/1303/822 1326/1302/822 1327/1301/822
  {0.12684548, -0.00213413, -0.01973337,1.0},
  {0.12381244, 0.02871166, -0.01973337,1.0},
  {0.12802409, 0.02928401, -0.01973337,1.0},
  // f 1220/1299/822 1221/1303/822 1327/1301/822
  {0.13569312, -0.00130996, -0.01973337,1.0},
  {0.12684548, -0.00213413, -0.01973337,1.0},
  {0.12802409, 0.02928401, -0.01973337,1.0},
  // f 1221/1303/822 1325/1304/822 1326/1302/822
  {0.12684548, -0.00213413, -0.01973337,1.0},
  {0.11947938, 0.02852089, -0.01973337,1.0},
  {0.12381244, 0.02871166, -0.01973337,1.0},
  // f 1222/1306/822 1324/1305/822 1325/1304/822
  {0.11735957, -0.00240886, -0.01973337,1.0},
  {0.11705618, 0.02858379, -0.01973337,1.0},
  {0.11947938, 0.02852089, -0.01973337,1.0},
  // f 1221/1303/822 1222/1306/822 1325/1304/822
  {0.12684548, -0.00213413, -0.01973337,1.0},
  {0.11735957, -0.00240886, -0.01973337,1.0},
  {0.11947938, 0.02852089, -0.01973337,1.0},
  // f 1222/1306/822 1323/1307/822 1324/1305/822
  {0.11735957, -0.00240886, -0.01973337,1.0},
  {0.11471993, 0.02877249, -0.01973337,1.0},
  {0.11705618, 0.02858379, -0.01973337,1.0},
  // f 1223/1309/822 1322/1308/822 1323/1307/822
  {0.1058239, -0.00191491, -0.01973337,1.0},
  {0.1113786, 0.0292914, -0.01973337,1.0},
  {0.11471993, 0.02877249, -0.01973337,1.0},
  // f 1222/1306/822 1223/1309/822 1323/1307/822
  {0.11735957, -0.00240886, -0.01973337,1.0},
  {0.1058239, -0.00191491, -0.01973337,1.0},
  {0.11471993, 0.02877249, -0.01973337,1.0},
  // f 1223/1309/822 1321/1310/822 1322/1308/822
  {0.1058239, -0.00191491, -0.01973337,1.0},
  {0.10823289, 0.03009339, -0.01973337,1.0},
  {0.1113786, 0.0292914, -0.01973337,1.0},
  // f 1224/1312/822 1320/1311/822 1321/1310/822
  {0.09702906, -0.00074863, -0.01973337,1.0},
  {0.10576094, 0.03097792, -0.01973337,1.0},
  {0.10823289, 0.03009339, -0.01973337,1.0},
  // f 1223/1309/822 1224/1312/822 1321/1310/822
  {0.1058239, -0.00191491, -0.01973337,1.0},
  {0.09702906, -0.00074863, -0.01973337,1.0},
  {0.10823289, 0.03009339, -0.01973337,1.0},
  // f 1224/1312/822 1319/1313/822 1320/1311/822
  {0.09702906, -0.00074863, -0.01973337,1.0},
  {0.10342481, 0.03205901, -0.01973337,1.0},
  {0.10576094, 0.03097792, -0.01973337,1.0},
  // f 1225/1315/822 1318/1314/822 1319/1313/822
  {0.08974966, 0.00088758, -0.01973337,1.0},
  {0.10122457, 0.03333668, -0.01973337,1.0},
  {0.10342481, 0.03205901, -0.01973337,1.0},
  // f 1224/1312/822 1225/1315/822 1319/1313/822
  {0.09702906, -0.00074863, -0.01973337,1.0},
  {0.08974966, 0.00088758, -0.01973337,1.0},
  {0.10342481, 0.03205901, -0.01973337,1.0},
  // f 1226/1317/822 1317/1316/822 1318/1314/822
  {0.08378498, 0.0028085, -0.01973337,1.0},
  {0.09916019, 0.0348109, -0.01973337,1.0},
  {0.10122457, 0.03333668, -0.01973337,1.0},
  // f 1225/1315/822 1226/1317/822 1318/1314/822
  {0.08974966, 0.00088758, -0.01973337,1.0},
  {0.08378498, 0.0028085, -0.01973337,1.0},
  {0.10122457, 0.03333668, -0.01973337,1.0},
  // f 1228/1319/822 1316/1318/822 1317/1316/822
  {0.07636081, 0.00616668, -0.01973337,1.0},
  {0.09723166, 0.03648167, -0.01973337,1.0},
  {0.09916019, 0.0348109, -0.01973337,1.0},
  // f 1227/1320/822 1228/1319/822 1317/1316/822
  {0.08029877, 0.00423204, -0.01973337,1.0},
  {0.07636081, 0.00616668, -0.01973337,1.0},
  {0.09916019, 0.0348109, -0.01973337,1.0},
  // f 1226/1317/822 1227/1320/822 1317/1316/822
  {0.08378498, 0.0028085, -0.01973337,1.0},
  {0.08029877, 0.00423204, -0.01973337,1.0},
  {0.09916019, 0.0348109, -0.01973337,1.0},
  // f 1229/1322/822 1315/1321/822 1316/1318/822
  {0.07147124, 0.00912545, -0.01973337,1.0},
  {0.09543897, 0.03834902, -0.01973337,1.0},
  {0.09723166, 0.03648167, -0.01973337,1.0},
  // f 1228/1319/822 1229/1322/822 1316/1318/822
  {0.07636081, 0.00616668, -0.01973337,1.0},
  {0.07147124, 0.00912545, -0.01973337,1.0},
  {0.09723166, 0.03648167, -0.01973337,1.0},
  // f 1230/1324/822 1314/1323/822 1315/1321/822
  {0.06694677, 0.01251342, -0.01973337,1.0},
  {0.09411437, 0.04000539, -0.01973337,1.0},
  {0.09543897, 0.03834902, -0.01973337,1.0},
  // f 1229/1322/822 1230/1324/822 1315/1321/822
  {0.07147124, 0.00912545, -0.01973337,1.0},
  {0.06694677, 0.01251342, -0.01973337,1.0},
  {0.09543897, 0.03834902, -0.01973337,1.0},
  // f 1231/1326/822 1313/1325/822 1314/1323/822
  {0.06278734, 0.0163306, -0.01973337,1.0},
  {0.09319337, 0.04135774, -0.01973337,1.0},
  {0.09411437, 0.04000539, -0.01973337,1.0},
  // f 1230/1324/822 1231/1326/822 1314/1323/822
  {0.06694677, 0.01251342, -0.01973337,1.0},
  {0.06278734, 0.0163306, -0.01973337,1.0},
  {0.09411437, 0.04000539, -0.01973337,1.0},
  // f 1232/1328/822 1312/1327/822 1313/1325/822
  {0.05899298, 0.02057697, -0.01973337,1.0},
  {0.09179632, 0.04382134, -0.01973337,1.0},
  {0.09319337, 0.04135774, -0.01973337,1.0},
  // f 1231/1326/822 1232/1328/822 1313/1325/822
  {0.06278734, 0.0163306, -0.01973337,1.0},
  {0.05899298, 0.02057697, -0.01973337,1.0},
  {0.09319337, 0.04135774, -0.01973337,1.0},
  // f 1233/1330/822 1311/1329/822 1312/1327/822
  {0.0555637, 0.02525254, -0.01973337,1.0},
  {0.09057174, 0.046547, -0.01973337,1.0},
  {0.09179632, 0.04382134, -0.01973337,1.0},
  // f 1232/1328/822 1233/1330/822 1312/1327/822
  {0.05899298, 0.02057697, -0.01973337,1.0},
  {0.0555637, 0.02525254, -0.01973337,1.0},
  {0.09179632, 0.04382134, -0.01973337,1.0},
  // f 1235/1332/822 1310/1331/822 1311/1329/822
  {0.05179045, 0.03170058, -0.01973337,1.0},
  {0.08932994, 0.05016376, -0.01973337,1.0},
  {0.09057174, 0.046547, -0.01973337,1.0},
  // f 1234/1333/822 1235/1332/822 1311/1329/822
  {0.05360577, 0.02839273, -0.01973337,1.0},
  {0.05179045, 0.03170058, -0.01973337,1.0},
  {0.09057174, 0.046547, -0.01973337,1.0},
  // f 1233/1330/822 1234/1333/822 1311/1329/822
  {0.0555637, 0.02525254, -0.01973337,1.0},
  {0.05360577, 0.02839273, -0.01973337,1.0},
  {0.09057174, 0.046547, -0.01973337,1.0},
  // f 1236/1335/822 1309/1334/822 1310/1331/822
  {0.04893376, 0.03796792, -0.01973337,1.0},
  {0.08819505, 0.0548603, -0.01973337,1.0},
  {0.08932994, 0.05016376, -0.01973337,1.0},
  // f 1235/1332/822 1236/1335/822 1310/1331/822
  {0.05179045, 0.03170058, -0.01973337,1.0},
  {0.04893376, 0.03796792, -0.01973337,1.0},
  {0.08932994, 0.05016376, -0.01973337,1.0},
  // f 1238/1337/822 1308/1336/822 1309/1334/822
  {0.04491555, 0.05147664, -0.01973337,1.0},
  {0.08723264, 0.06165352, -0.01973337,1.0},
  {0.08819505, 0.0548603, -0.01973337,1.0},
  // f 1237/1338/822 1238/1337/822 1309/1334/822
  {0.04664213, 0.04455995, -0.01973337,1.0},
  {0.04491555, 0.05147664, -0.01973337,1.0},
  {0.08819505, 0.0548603, -0.01973337,1.0},
  // f 1236/1335/822 1237/1338/822 1309/1334/822
  {0.04893376, 0.03796792, -0.01973337,1.0},
  {0.04664213, 0.04455995, -0.01973337,1.0},
  {0.08819505, 0.0548603, -0.01973337,1.0},
  // f 1240/1340/822 1307/1339/822 1308/1336/822
  {0.0432413, 0.06457467, -0.01973337,1.0},
  {0.08681526, 0.0711095, -0.01973337,1.0},
  {0.08723264, 0.06165352, -0.01973337,1.0},
  // f 1239/1341/822 1240/1340/822 1308/1336/822
  {0.04365986, 0.05954267, -0.01973337,1.0},
  {0.0432413, 0.06457467, -0.01973337,1.0},
  {0.08723264, 0.06165352, -0.01973337,1.0},
  // f 1238/1337/822 1239/1341/822 1308/1336/822
  {0.04491555, 0.05147664, -0.01973337,1.0},
  {0.04365986, 0.05954267, -0.01973337,1.0},
  {0.08723264, 0.06165352, -0.01973337,1.0},
  // f 1242/1343/822 1306/1342/822 1307/1339/822
  {0.04353208, 0.08051129, -0.01973337,1.0},
  {0.08698238, 0.07692527, -0.01973337,1.0},
  {0.08681526, 0.0711095, -0.01973337,1.0},
  // f 1241/1344/822 1242/1343/822 1307/1339/822
  {0.04307039, 0.07062773, -0.01973337,1.0},
  {0.04353208, 0.08051129, -0.01973337,1.0},
  {0.08681526, 0.0711095, -0.01973337,1.0},
  // f 1240/1340/822 1241/1344/822 1307/1339/822
  {0.0432413, 0.06457467, -0.01973337,1.0},
  {0.04307039, 0.07062773, -0.01973337,1.0},
  {0.08681526, 0.0711095, -0.01973337,1.0},
  // f 1242/1343/822 1305/1345/822 1306/1342/822
  {0.04353208, 0.08051129, -0.01973337,1.0},
  {0.0874838, 0.08228591, -0.01973337,1.0},
  {0.08698238, 0.07692527, -0.01973337,1.0},
  // f 1244/1347/822 1304/1346/822 1305/1345/822
  {0.04650446, 0.09620979, -0.01973337,1.0},
  {0.08846615, 0.08785504, -0.01973337,1.0},
  {0.0874838, 0.08228591, -0.01973337,1.0},
  // f 1243/1348/822 1244/1347/822 1305/1345/822
  {0.04475309, 0.08899045, -0.01973337,1.0},
  {0.04650446, 0.09620979, -0.01973337,1.0},
  {0.0874838, 0.08228591, -0.01973337,1.0},
  // f 1242/1343/822 1243/1348/822 1305/1345/822
  {0.04353208, 0.08051129, -0.01973337,1.0},
  {0.04475309, 0.08899045, -0.01973337,1.0},
  {0.0874838, 0.08228591, -0.01973337,1.0},
  // f 1245/1350/822 1303/1349/822 1304/1346/822
  {0.04858016, 0.10229938, -0.01973337,1.0},
  {0.08968385, 0.09224036, -0.01973337,1.0},
  {0.08846615, 0.08785504, -0.01973337,1.0},
  // f 1244/1347/822 1245/1350/822 1304/1346/822
  {0.04650446, 0.09620979, -0.01973337,1.0},
  {0.04858016, 0.10229938, -0.01973337,1.0},
  {0.08846615, 0.08785504, -0.01973337,1.0},
  // f 1247/1352/822 1302/1351/822 1303/1349/822
  {0.05260947, 0.11085567, -0.01973337,1.0},
  {0.09099369, 0.09563697, -0.01973337,1.0},
  {0.08968385, 0.09224036, -0.01973337,1.0},
  // f 1246/1353/822 1247/1352/822 1303/1349/822
  {0.05045747, 0.10666426, -0.01973337,1.0},
  {0.05260947, 0.11085567, -0.01973337,1.0},
  {0.08968385, 0.09224036, -0.01973337,1.0},
  // f 1245/1350/822 1246/1353/822 1303/1349/822
  {0.04858016, 0.10229938, -0.01973337,1.0},
  {0.05045747, 0.10666426, -0.01973337,1.0},
  {0.08968385, 0.09224036, -0.01973337,1.0},
  // f 1248/1355/822 1301/1354/822 1302/1351/822
  {0.05588891, 0.11611768, -0.01973337,1.0},
  {0.09227277, 0.09821206, -0.01973337,1.0},
  {0.09099369, 0.09563697, -0.01973337,1.0},
  // f 1247/1352/822 1248/1355/822 1302/1351/822
  {0.05260947, 0.11085567, -0.01973337,1.0},
  {0.05588891, 0.11611768, -0.01973337,1.0},
  {0.09099369, 0.09563697, -0.01973337,1.0},
  // f 1250/1357/822 1300/1356/822 1301/1354/822
  {0.06272005, 0.12431127, -0.01973337,1.0},
  {0.09372242, 0.1005549, -0.01973337,1.0},
  {0.09227277, 0.09821206, -0.01973337,1.0},
  // f 1249/1358/822 1250/1357/822 1301/1354/822
  {0.05913073, 0.12037592, -0.01973337,1.0},
  {0.06272005, 0.12431127, -0.01973337,1.0},
  {0.09227277, 0.09821206, -0.01973337,1.0},
  // f 1248/1355/822 1249/1358/822 1301/1354/822
  {0.05588891, 0.11611768, -0.01973337,1.0},
  {0.05913073, 0.12037592, -0.01973337,1.0},
  {0.09227277, 0.09821206, -0.01973337,1.0},
  // f 1251/1360/822 1299/1359/822 1300/1356/822
  {0.06665686, 0.12792365, -0.01973337,1.0},
  {0.09534263, 0.10266554, -0.01973337,1.0},
  {0.09372242, 0.1005549, -0.01973337,1.0},
  // f 1250/1357/822 1251/1360/822 1300/1356/822
  {0.06272005, 0.12431127, -0.01973337,1.0},
  {0.06665686, 0.12792365, -0.01973337,1.0},
  {0.09372242, 0.1005549, -0.01973337,1.0},
  // f 1252/1362/822 1298/1361/822 1299/1359/822
  {0.07158156, 0.13165666, -0.01973337,1.0},
  {0.09674428, 0.10419735, -0.01973337,1.0},
  {0.09534263, 0.10266554, -0.01973337,1.0},
  // f 1251/1360/822 1252/1362/822 1299/1359/822
  {0.06665686, 0.12792365, -0.01973337,1.0},
  {0.07158156, 0.13165666, -0.01973337,1.0},
  {0.09534263, 0.10266554, -0.01973337,1.0},
  // f 1253/1364/822 1297/1363/822 1298/1361/822
  {0.07557292, 0.13417959, -0.01973337,1.0},
  {0.09784453, 0.10526244, -0.01973337,1.0},
  {0.09674428, 0.10419735, -0.01973337,1.0},
  // f 1252/1362/822 1253/1364/822 1298/1361/822
  {0.07158156, 0.13165666, -0.01973337,1.0},
  {0.07557292, 0.13417959, -0.01973337,1.0},
  {0.09674428, 0.10419735, -0.01973337,1.0},
  // f 1254/1366/822 1296/1365/822 1297/1363/822
  {0.08055224, 0.13682316, -0.01973337,1.0},
  {0.09977153, 0.106878, -0.01973337,1.0},
  {0.09784453, 0.10526244, -0.01973337,1.0},
  // f 1253/1364/822 1254/1366/822 1297/1363/822
  {0.07557292, 0.13417959, -0.01973337,1.0},
  {0.08055224, 0.13682316, -0.01973337,1.0},
  {0.09784453, 0.10526244, -0.01973337,1.0},
  // f 1255/1368/822 1295/1367/822 1296/1365/822
  {0.08540048, 0.1388372, -0.01973337,1.0},
  {0.1018151, 0.10829411, -0.01973337,1.0},
  {0.09977153, 0.106878, -0.01973337,1.0},
  // f 1254/1366/822 1255/1368/822 1296/1365/822
  {0.08055224, 0.13682316, -0.01973337,1.0},
  {0.08540048, 0.1388372, -0.01973337,1.0},
  {0.09977153, 0.106878, -0.01973337,1.0},
  // f 1256/1370/822 1294/1369/822 1295/1367/822
  {0.09146211, 0.14068431, -0.01973337,1.0},
  {0.10397528, 0.10951076, -0.01973337,1.0},
  {0.1018151, 0.10829411, -0.01973337,1.0},
  // f 1255/1368/822 1256/1370/822 1295/1367/822
  {0.08540048, 0.1388372, -0.01973337,1.0},
  {0.09146211, 0.14068431, -0.01973337,1.0},
  {0.1018151, 0.10829411, -0.01973337,1.0},
  // f 1256/1370/822 1293/1371/822 1294/1369/822
  {0.09146211, 0.14068431, -0.01973337,1.0},
  {0.10625204, 0.11052801, -0.01973337,1.0},
  {0.10397528, 0.10951076, -0.01973337,1.0},
  // f 1257/1373/822 1292/1372/822 1293/1371/822
  {0.09895155, 0.14221673, -0.01973337,1.0},
  {0.10864536, 0.11134578, -0.01973337,1.0},
  {0.10625204, 0.11052801, -0.01973337,1.0},
  // f 1256/1370/822 1257/1373/822 1293/1371/822
  {0.09146211, 0.14068431, -0.01973337,1.0},
  {0.09895155, 0.14221673, -0.01973337,1.0},
  {0.10625204, 0.11052801, -0.01973337,1.0},
  // f 1258/1375/822 1291/1374/822 1292/1372/822
  {0.10810695, 0.14327027, -0.01973337,1.0},
  {0.11167126, 0.11206378, -0.01973337,1.0},
  {0.10864536, 0.11134578, -0.01973337,1.0},
  // f 1257/1373/822 1258/1375/822 1292/1372/822
  {0.09895155, 0.14221673, -0.01973337,1.0},
  {0.10810695, 0.14327027, -0.01973337,1.0},
  {0.10864536, 0.11134578, -0.01973337,1.0},
  // f 1258/1375/822 1290/1376/822 1291/1374/822
  {0.10810695, 0.14327027, -0.01973337,1.0},
  {0.11486509, 0.11249463, -0.01973337,1.0},
  {0.11167126, 0.11206378, -0.01973337,1.0},
  // f 1259/1378/822 1289/1377/822 1290/1376/822
  {0.11919031, 0.1436643, -0.01973337,1.0},
  {0.11822677, 0.11263821, -0.01973337,1.0},
  {0.11486509, 0.11249463, -0.01973337,1.0},
  // f 1258/1375/822 1259/1378/822 1290/1376/822
  {0.10810695, 0.14327027, -0.01973337,1.0},
  {0.11919031, 0.1436643, -0.01973337,1.0},
  {0.11486509, 0.11249463, -0.01973337,1.0},
  // f 1259/1378/822 1288/1379/822 1289/1377/822
  {0.11919031, 0.1436643, -0.01973337,1.0},
  {0.12085765, 0.11255514, -0.01973337,1.0},
  {0.11822677, 0.11263821, -0.01973337,1.0},
  // f 1259/1378/822 1287/1380/822 1288/1379/822
  {0.11919031, 0.1436643, -0.01973337,1.0},
  {0.12300596, 0.1123516, -0.01973337,1.0},
  {0.12085765, 0.11255514, -0.01973337,1.0},
  // f 1260/1382/822 1286/1381/822 1287/1380/822
  {0.13603304, 0.14298595, -0.01973337,1.0},
  {0.12571715, 0.11189037, -0.01973337,1.0},
  {0.12300596, 0.1123516, -0.01973337,1.0},
  // f 1259/1378/822 1260/1382/822 1287/1380/822
  {0.11919031, 0.1436643, -0.01973337,1.0},
  {0.13603304, 0.14298595, -0.01973337,1.0},
  {0.12300596, 0.1123516, -0.01973337,1.0},
  // f 1260/1382/822 1285/1383/822 1286/1381/822
  {0.13603304, 0.14298595, -0.01973337,1.0},
  {0.12825319, 0.11121202, -0.01973337,1.0},
  {0.12571715, 0.11189037, -0.01973337,1.0},
  // f 1260/1382/822 1284/1384/822 1285/1383/822
  {0.13603304, 0.14298595, -0.01973337,1.0},
  {0.13032858, 0.1104404, -0.01973337,1.0},
  {0.12825319, 0.11121202, -0.01973337,1.0},
  // f 1261/1386/822 1283/1385/822 1284/1384/822
  {0.1455756, 0.14170393, -0.01973337,1.0},
  {0.13226987, 0.10950262, -0.01973337,1.0},
  {0.13032858, 0.1104404, -0.01973337,1.0},
  // f 1260/1382/822 1261/1386/822 1284/1384/822
  {0.13603304, 0.14298595, -0.01973337,1.0},
  {0.1455756, 0.14170393, -0.01973337,1.0},
  {0.13032858, 0.1104404, -0.01973337,1.0},
  // f 1261/1386/822 1282/1387/822 1283/1385/822
  {0.1455756, 0.14170393, -0.01973337,1.0},
  {0.13407704, 0.10839865, -0.01973337,1.0},
  {0.13226987, 0.10950262, -0.01973337,1.0},
  // f 1262/1389/822 1281/1388/822 1282/1387/822
  {0.15019752, 0.14067285, -0.01973337,1.0},
  {0.13574637, 0.1071379, -0.01973337,1.0},
  {0.13407704, 0.10839865, -0.01973337,1.0},
  // f 1261/1386/822 1262/1389/822 1282/1387/822
  {0.1455756, 0.14170393, -0.01973337,1.0},
  {0.15019752, 0.14067285, -0.01973337,1.0},
  {0.13407704, 0.10839865, -0.01973337,1.0},
  // f 1264/1391/822 1280/1390/822 1281/1388/822
  {0.15870138, 0.13754196, -0.01973337,1.0},
  {0.13706453, 0.10594001, -0.01973337,1.0},
  {0.13574637, 0.1071379, -0.01973337,1.0},
  // f 1263/1392/822 1264/1391/822 1281/1388/822
  {0.15421513, 0.13942471, -0.01973337,1.0},
  {0.15870138, 0.13754196, -0.01973337,1.0},
  {0.13574637, 0.1071379, -0.01973337,1.0},
  // f 1262/1389/822 1263/1392/822 1281/1388/822
  {0.15019752, 0.14067285, -0.01973337,1.0},
  {0.15421513, 0.13942471, -0.01973337,1.0},
  {0.13574637, 0.1071379, -0.01973337,1.0},
  // f 1265/1394/822 1279/1393/822 1280/1390/822
  {0.16246878, 0.13552782, -0.01973337,1.0},
  {0.1384708, 0.10440568, -0.01973337,1.0},
  {0.13706453, 0.10594001, -0.01973337,1.0},
  // f 1264/1391/822 1265/1394/822 1280/1390/822
  {0.15870138, 0.13754196, -0.01973337,1.0},
  {0.16246878, 0.13552782, -0.01973337,1.0},
  {0.13706453, 0.10594001, -0.01973337,1.0},
  // f 1267/1396/822 1278/1395/822 1279/1393/822
  {0.16923565, 0.13059105, -0.01973337,1.0},
  {0.13973548, 0.10272409, -0.01973337,1.0},
  {0.1384708, 0.10440568, -0.01973337,1.0},
  // f 1266/1397/822 1267/1396/822 1279/1393/822
  {0.16598019, 0.13321085, -0.01973337,1.0},
  {0.16923565, 0.13059105, -0.01973337,1.0},
  {0.1384708, 0.10440568, -0.01973337,1.0},
  // f 1265/1394/822 1266/1397/822 1279/1393/822
  {0.16246878, 0.13552782, -0.01973337,1.0},
  {0.16598019, 0.13321085, -0.01973337,1.0},
  {0.1384708, 0.10440568, -0.01973337,1.0},
  // f 1269/1399/822 1277/1398/822 1278/1395/822
  {0.17348598, 0.12627235, -0.01973337,1.0},
  {0.14085847, 0.10089514, -0.01973337,1.0},
  {0.13973548, 0.10272409, -0.01973337,1.0},
  // f 1268/1400/822 1269/1399/822 1278/1395/822
  {0.17126371, 0.12867633, -0.01973337,1.0},
  {0.17348598, 0.12627235, -0.01973337,1.0},
  {0.13973548, 0.10272409, -0.01973337,1.0},
  // f 1267/1396/822 1268/1400/822 1278/1395/822
  {0.16923565, 0.13059105, -0.01973337,1.0},
  {0.17126371, 0.12867633, -0.01973337,1.0},
  {0.13973548, 0.10272409, -0.01973337,1.0},
  // f 1271/1402/822 1276/1401/822 1277/1398/822
  {0.17875814, 0.11880992, -0.01973337,1.0},
  {0.14196849, 0.09862459, -0.01973337,1.0},
  {0.14085847, 0.10089514, -0.01973337,1.0},
  // f 1270/1403/822 1271/1402/822 1277/1398/822
  {0.17612036, 0.12290923, -0.01973337,1.0},
  {0.17875814, 0.11880992, -0.01973337,1.0},
  {0.14085847, 0.10089514, -0.01973337,1.0},
  // f 1269/1399/822 1270/1403/822 1277/1398/822
  {0.17348598, 0.12627235, -0.01973337,1.0},
  {0.17612036, 0.12290923, -0.01973337,1.0},
  {0.14085847, 0.10089514, -0.01973337,1.0},
  // f 1272/1405/822 1275/1404/822 1276/1401/822
  {0.18109146, 0.11432907, -0.01973337,1.0},
  {0.14289349, 0.09616163, -0.01973337,1.0},
  {0.14196849, 0.09862459, -0.01973337,1.0},
  // f 1271/1402/822 1272/1405/822 1276/1401/822
  {0.17875814, 0.11880992, -0.01973337,1.0},
  {0.18109146, 0.11432907, -0.01973337,1.0},
  {0.14196849, 0.09862459, -0.01973337,1.0},
  // f 1273/1407/822 1274/1406/822 1275/1404/822
  {0.1831203, 0.10946664, -0.01973337,1.0},
  {0.18500036, 0.10367728, -0.01973337,1.0},
  {0.14289349, 0.09616163, -0.01973337,1.0},
  // f 1272/1405/822 1273/1407/822 1275/1404/822
  {0.18109146, 0.11432907, -0.01973337,1.0},
  {0.1831203, 0.10946664, -0.01973337,1.0},
  {0.14289349, 0.09616163, -0.01973337,1.0},
  // f 1343/1410/822 1344/1409/822 1345/1408/822
  {0.25668794, 0.05713812, -0.01973337,1.0},
  {0.25668794, 0, -0.01973337,1.0},
  {0.21303943, 0, -0.01973337,1.0},
  // f 1334/1411/822 1343/1410/822 1345/1408/822
  {0.21303943, 0.14125545, -0.01973337,1.0},
  {0.25668794, 0.05713812, -0.01973337,1.0},
  {0.21303943, 0, -0.01973337,1.0},
  // f 1337/1413/822 1342/1412/822 1343/1410/822
  {0.30438331, 0.09182569, -0.01973337,1.0},
  {0.30438331, 0.05713812, -0.01973337,1.0},
  {0.25668794, 0.05713812, -0.01973337,1.0},
  // f 1336/1414/822 1337/1413/822 1343/1410/822
  {0.25668794, 0.09182569, -0.01973337,1.0},
  {0.30438331, 0.09182569, -0.01973337,1.0},
  {0.25668794, 0.05713812, -0.01973337,1.0},
  // f 1334/1411/822 1336/1414/822 1343/1410/822
  {0.21303943, 0.14125545, -0.01973337,1.0},
  {0.25668794, 0.09182569, -0.01973337,1.0},
  {0.25668794, 0.05713812, -0.01973337,1.0},
  // f 1340/1416/822 1341/1415/822 1342/1412/822
  {0.34822458, 0, -0.01973337,1.0},
  {0.30438331, 0, -0.01973337,1.0},
  {0.30438331, 0.05713812, -0.01973337,1.0},
  // f 1339/1417/822 1340/1416/822 1342/1412/822
  {0.34822458, 0.14125545, -0.01973337,1.0},
  {0.34822458, 0, -0.01973337,1.0},
  {0.30438331, 0.05713812, -0.01973337,1.0},
  // f 1337/1413/822 1339/1417/822 1342/1412/822
  {0.30438331, 0.09182569, -0.01973337,1.0},
  {0.34822458, 0.14125545, -0.01973337,1.0},
  {0.30438331, 0.05713812, -0.01973337,1.0},
  // f 1337/1413/822 1338/1418/822 1339/1417/822
  {0.30438331, 0.09182569, -0.01973337,1.0},
  {0.30438331, 0.14125545, -0.01973337,1.0},
  {0.34822458, 0.14125545, -0.01973337,1.0},
  // f 1334/1411/822 1335/1419/822 1336/1414/822
  {0.21303943, 0.14125545, -0.01973337,1.0},
  {0.25668794, 0.14125545, -0.01973337,1.0},
  {0.25668794, 0.09182569, -0.01973337,1.0},
  // f 1346/1422/822 1352/1421/822 1353/1420/822
  {0.36730275, 0.14125545, -0.01973337,1.0},
  {0.41181844, 0.10637518, -0.01973337,1.0},
  {0.36730275, 0.10637518, -0.01973337,1.0},
  // f 1349/1424/822 1351/1423/822 1352/1421/822
  {0.45546687, 0.10637518, -0.01973337,1.0},
  {0.41181844, 0, -0.01973337,1.0},
  {0.41181844, 0.10637518, -0.01973337,1.0},
  // f 1346/1422/822 1349/1424/822 1352/1421/822
  {0.36730275, 0.14125545, -0.01973337,1.0},
  {0.45546687, 0.10637518, -0.01973337,1.0},
  {0.41181844, 0.10637518, -0.01973337,1.0},
  // f 1349/1424/822 1350/1425/822 1351/1423/822
  {0.45546687, 0.10637518, -0.01973337,1.0},
  {0.45546687, 0, -0.01973337,1.0},
  {0.41181844, 0, -0.01973337,1.0},
  // f 1347/1427/822 1348/1426/822 1349/1424/822
  {0.49998266, 0.14125545, -0.01973337,1.0},
  {0.49998266, 0.10637518, -0.01973337,1.0},
  {0.45546687, 0.10637518, -0.01973337,1.0},
  // f 1346/1422/822 1347/1427/822 1349/1424/822
  {0.36730275, 0.14125545, -0.01973337,1.0},
  {0.49998266, 0.14125545, -0.01973337,1.0},
  {0.45546687, 0.10637518, -0.01973337,1.0},
  // f 1431/1430/823 1355/1429/823 1354/1428/823
  {-0.45614135, 0.11254187, 0,1.0},
  {-0.42742792, 0.14125545, 0,1.0},
  {-0.49998266, 0.14125545, 0,1.0},
  // f 1393/1431/823 1431/1430/823 1354/1428/823
  {-0.45614135, 0.081034, 0,1.0},
  {-0.45614135, 0.11254187, 0,1.0},
  {-0.49998266, 0.14125545, 0,1.0},
  // f 1392/1432/823 1393/1431/823 1354/1428/823
  {-0.49998266, 0, 0,1.0},
  {-0.45614135, 0.081034, 0,1.0},
  {-0.49998266, 0.14125545, 0,1.0},
  // f 1425/1434/823 1356/1433/823 1355/1429/823
  {-0.43095866, 0.11052722, 0,1.0},
  {-0.41928023, 0.14089021, 0,1.0},
  {-0.42742792, 0.14125545, 0,1.0},
  // f 1426/1435/823 1425/1434/823 1355/1429/823
  {-0.43245378, 0.11109147, 0,1.0},
  {-0.43095866, 0.11052722, 0,1.0},
  {-0.42742792, 0.14125545, 0,1.0},
  // f 1427/1436/823 1426/1435/823 1355/1429/823
  {-0.43445414, 0.11164644, 0,1.0},
  {-0.43245378, 0.11109147, 0,1.0},
  {-0.42742792, 0.14125545, 0,1.0},
  // f 1428/1437/823 1427/1436/823 1355/1429/823
  {-0.43707186, 0.11212561, 0,1.0},
  {-0.43445414, 0.11164644, 0,1.0},
  {-0.42742792, 0.14125545, 0,1.0},
  // f 1429/1438/823 1428/1437/823 1355/1429/823
  {-0.43999577, 0.11242347, 0,1.0},
  {-0.43707186, 0.11212561, 0,1.0},
  {-0.42742792, 0.14125545, 0,1.0},
  // f 1430/1439/823 1429/1438/823 1355/1429/823
  {-0.44371176, 0.11254187, 0,1.0},
  {-0.43999577, 0.11242347, 0,1.0},
  {-0.42742792, 0.14125545, 0,1.0},
  // f 1431/1430/823 1430/1439/823 1355/1429/823
  {-0.45614135, 0.11254187, 0,1.0},
  {-0.44371176, 0.11254187, 0,1.0},
  {-0.42742792, 0.14125545, 0,1.0},
  // f 1423/1441/823 1357/1440/823 1356/1433/823
  {-0.4286606, 0.10927848, 0,1.0},
  {-0.4126853, 0.13995226, 0,1.0},
  {-0.41928023, 0.14089021, 0,1.0},
  // f 1424/1442/823 1423/1441/823 1356/1433/823
  {-0.42961943, 0.10987046, 0,1.0},
  {-0.4286606, 0.10927848, 0,1.0},
  {-0.41928023, 0.14089021, 0,1.0},
  // f 1425/1434/823 1424/1442/823 1356/1433/823
  {-0.43095866, 0.11052722, 0,1.0},
  {-0.42961943, 0.10987046, 0,1.0},
  {-0.41928023, 0.14089021, 0,1.0},
  // f 1421/1444/823 1358/1443/823 1357/1440/823
  {-0.42704237, 0.10791686, 0,1.0},
  {-0.40741441, 0.13865802, 0,1.0},
  {-0.4126853, 0.13995226, 0,1.0},
  // f 1422/1445/823 1421/1444/823 1357/1440/823
  {-0.42780152, 0.10862726, 0,1.0},
  {-0.42704237, 0.10791686, 0,1.0},
  {-0.4126853, 0.13995226, 0,1.0},
  // f 1423/1441/823 1422/1445/823 1357/1440/823
  {-0.4286606, 0.10927848, 0,1.0},
  {-0.42780152, 0.10862726, 0,1.0},
  {-0.4126853, 0.13995226, 0,1.0},
  // f 1420/1447/823 1359/1446/823 1358/1443/823
  {-0.42636168, 0.10716394, 0,1.0},
  {-0.40326801, 0.13719702, 0,1.0},
  {-0.40741441, 0.13865802, 0,1.0},
  // f 1421/1444/823 1420/1447/823 1358/1443/823
  {-0.42704237, 0.10791686, 0,1.0},
  {-0.42636168, 0.10716394, 0,1.0},
  {-0.40741441, 0.13865802, 0,1.0},
  // f 1419/1449/823 1360/1448/823 1359/1446/823
  {-0.42588824, 0.10658219, 0,1.0},
  {-0.39946407, 0.13541128, 0,1.0},
  {-0.40326801, 0.13719702, 0,1.0},
  // f 1420/1447/823 1419/1449/823 1359/1446/823
  {-0.42636168, 0.10716394, 0,1.0},
  {-0.42588824, 0.10658219, 0,1.0},
  {-0.40326801, 0.13719702, 0,1.0},
  // f 1418/1451/823 1361/1450/823 1360/1448/823
  {-0.42503724, 0.10537505, 0,1.0},
  {-0.39655605, 0.13367519, 0,1.0},
  {-0.39946407, 0.13541128, 0,1.0},
  // f 1419/1449/823 1418/1451/823 1360/1448/823
  {-0.42588824, 0.10658219, 0,1.0},
  {-0.42503724, 0.10537505, 0,1.0},
  {-0.39946407, 0.13541128, 0,1.0},
  // f 1418/1451/823 1362/1452/823 1361/1450/823
  {-0.42503724, 0.10537505, 0,1.0},
  {-0.39440098, 0.13212398, 0,1.0},
  {-0.39655605, 0.13367519, 0,1.0},
  // f 1418/1451/823 1363/1453/823 1362/1452/823
  {-0.42503724, 0.10537505, 0,1.0},
  {-0.39192128, 0.12998199, 0,1.0},
  {-0.39440098, 0.13212398, 0,1.0},
  // f 1417/1455/823 1364/1454/823 1363/1453/823
  {-0.4243139, 0.10410966, 0,1.0},
  {-0.38967857, 0.1276319, 0,1.0},
  {-0.39192128, 0.12998199, 0,1.0},
  // f 1418/1451/823 1417/1455/823 1363/1453/823
  {-0.42503724, 0.10537505, 0,1.0},
  {-0.4243139, 0.10410966, 0,1.0},
  {-0.39192128, 0.12998199, 0,1.0},
  // f 1417/1455/823 1365/1456/823 1364/1454/823
  {-0.4243139, 0.10410966, 0,1.0},
  {-0.38767204, 0.12509105, 0,1.0},
  {-0.38967857, 0.1276319, 0,1.0},
  // f 1416/1458/823 1366/1457/823 1365/1456/823
  {-0.42371821, 0.10278598, 0,1.0},
  {-0.3859016, 0.1223594, 0,1.0},
  {-0.38767204, 0.12509105, 0,1.0},
  // f 1417/1455/823 1416/1458/823 1365/1456/823
  {-0.4243139, 0.10410966, 0,1.0},
  {-0.42371821, 0.10278598, 0,1.0},
  {-0.38767204, 0.12509105, 0,1.0},
  // f 1416/1458/823 1367/1459/823 1366/1457/823
  {-0.42371821, 0.10278598, 0,1.0},
  {-0.38436708, 0.11943695, 0,1.0},
  {-0.3859016, 0.1223594, 0,1.0},
  // f 1416/1458/823 1368/1460/823 1367/1459/823
  {-0.42371821, 0.10278598, 0,1.0},
  {-0.38283733, 0.11567822, 0,1.0},
  {-0.38436708, 0.11943695, 0,1.0},
  // f 1415/1462/823 1369/1461/823 1368/1460/823
  {-0.42325002, 0.10140404, 0,1.0},
  {-0.38164759, 0.11164479, 0,1.0},
  {-0.38283733, 0.11567822, 0,1.0},
  // f 1416/1458/823 1415/1462/823 1368/1460/823
  {-0.42371821, 0.10278598, 0,1.0},
  {-0.42325002, 0.10140404, 0,1.0},
  {-0.38283733, 0.11567822, 0,1.0},
  // f 1414/1464/823 1370/1463/823 1369/1461/823
  {-0.42290977, 0.09996384, 0,1.0},
  {-0.38079774, 0.10733657, 0,1.0},
  {-0.38164759, 0.11164479, 0,1.0},
  // f 1415/1462/823 1414/1464/823 1369/1461/823
  {-0.42325002, 0.10140404, 0,1.0},
  {-0.42290977, 0.09996384, 0,1.0},
  {-0.38164759, 0.11164479, 0,1.0},
  // f 1414/1464/823 1371/1465/823 1370/1463/823
  {-0.42290977, 0.09996384, 0,1.0},
  {-0.3802878, 0.10275365, 0,1.0},
  {-0.38079774, 0.10733657, 0,1.0},
  // f 1413/1467/823 1372/1466/823 1371/1465/823
  {-0.42269701, 0.0984653, 0,1.0},
  {-0.38011783, 0.097896, 0,1.0},
  {-0.3802878, 0.10275365, 0,1.0},
  // f 1414/1464/823 1413/1467/823 1371/1465/823
  {-0.42290977, 0.09996384, 0,1.0},
  {-0.42269701, 0.0984653, 0,1.0},
  {-0.3802878, 0.10275365, 0,1.0},
  // f 1412/1469/823 1373/1468/823 1372/1466/823
  {-0.42261013, 0.09664342, 0,1.0},
  {-0.38036993, 0.09208981, 0,1.0},
  {-0.38011783, 0.097896, 0,1.0},
  // f 1413/1467/823 1412/1469/823 1372/1466/823
  {-0.42269701, 0.0984653, 0,1.0},
  {-0.42261013, 0.09664342, 0,1.0},
  {-0.38011783, 0.097896, 0,1.0},
  // f 1410/1471/823 1374/1470/823 1373/1468/823
  {-0.42271024, 0.09477603, 0,1.0},
  {-0.38112643, 0.08665, 0,1.0},
  {-0.38036993, 0.09208981, 0,1.0},
  // f 1411/1472/823 1410/1471/823 1373/1468/823
  {-0.42264283, 0.09556455, 0,1.0},
  {-0.42271024, 0.09477603, 0,1.0},
  {-0.38036993, 0.09208981, 0,1.0},
  // f 1412/1469/823 1411/1472/823 1373/1468/823
  {-0.42261013, 0.09664342, 0,1.0},
  {-0.42264283, 0.09556455, 0,1.0},
  {-0.38036993, 0.09208981, 0,1.0},
  // f 1409/1474/823 1375/1473/823 1374/1470/823
  {-0.42295536, 0.09325211, 0,1.0},
  {-0.38217598, 0.08227888, 0,1.0},
  {-0.38112643, 0.08665, 0,1.0},
  // f 1410/1471/823 1409/1474/823 1374/1470/823
  {-0.42271024, 0.09477603, 0,1.0},
  {-0.42295536, 0.09325211, 0,1.0},
  {-0.38112643, 0.08665, 0,1.0},
  // f 1408/1476/823 1376/1475/823 1375/1473/823
  {-0.42327198, 0.09203622, 0,1.0},
  {-0.38359609, 0.07817698, 0,1.0},
  {-0.38217598, 0.08227888, 0,1.0},
  // f 1409/1474/823 1408/1476/823 1375/1473/823
  {-0.42295536, 0.09325211, 0,1.0},
  {-0.42327198, 0.09203622, 0,1.0},
  {-0.38217598, 0.08227888, 0,1.0},
  // f 1408/1476/823 1377/1477/823 1376/1475/823
  {-0.42327198, 0.09203622, 0,1.0},
  {-0.38506255, 0.07496432, 0,1.0},
  {-0.38359609, 0.07817698, 0,1.0},
  // f 1407/1479/823 1378/1478/823 1377/1477/823
  {-0.42369071, 0.09086943, 0,1.0},
  {-0.38678622, 0.07193859, 0,1.0},
  {-0.38506255, 0.07496432, 0,1.0},
  // f 1408/1476/823 1407/1479/823 1377/1477/823
  {-0.42327198, 0.09203622, 0,1.0},
  {-0.42369071, 0.09086943, 0,1.0},
  {-0.38506255, 0.07496432, 0,1.0},
  // f 1407/1479/823 1379/1480/823 1378/1478/823
  {-0.42369071, 0.09086943, 0,1.0},
  {-0.38876721, 0.06909983, 0,1.0},
  {-0.38678622, 0.07193859, 0,1.0},
  // f 1406/1482/823 1380/1481/823 1379/1480/823
  {-0.42421165, 0.08975184, 0,1.0},
  {-0.39053714, 0.06696337, 0,1.0},
  {-0.38876721, 0.06909983, 0,1.0},
  // f 1407/1479/823 1406/1482/823 1379/1480/823
  {-0.42369071, 0.09086943, 0,1.0},
  {-0.42421165, 0.08975184, 0,1.0},
  {-0.38876721, 0.06909983, 0,1.0},
  // f 1406/1482/823 1381/1483/823 1380/1481/823
  {-0.42421165, 0.08975184, 0,1.0},
  {-0.39298108, 0.06446107, 0,1.0},
  {-0.39053714, 0.06696337, 0,1.0},
  // f 1405/1485/823 1382/1484/823 1381/1483/823
  {-0.42483458, 0.08868335, 0,1.0},
  {-0.39568928, 0.06217263, 0,1.0},
  {-0.39298108, 0.06446107, 0,1.0},
  // f 1406/1482/823 1405/1485/823 1381/1483/823
  {-0.42421165, 0.08975184, 0,1.0},
  {-0.42483458, 0.08868335, 0,1.0},
  {-0.39298108, 0.06446107, 0,1.0},
  // f 1405/1485/823 1383/1486/823 1382/1484/823
  {-0.42483458, 0.08868335, 0,1.0},
  {-0.39866793, 0.06012511, 0,1.0},
  {-0.39568928, 0.06217263, 0,1.0},
  // f 1404/1488/823 1384/1487/823 1383/1486/823
  {-0.42555979, 0.08766402, 0,1.0},
  {-0.40191758, 0.05831847, 0,1.0},
  {-0.39866793, 0.06012511, 0,1.0},
  // f 1405/1485/823 1404/1488/823 1383/1486/823
  {-0.42483458, 0.08868335, 0,1.0},
  {-0.42555979, 0.08766402, 0,1.0},
  {-0.39866793, 0.06012511, 0,1.0},
  // f 1404/1488/823 1385/1489/823 1384/1487/823
  {-0.42555979, 0.08766402, 0,1.0},
  {-0.40617442, 0.05646847, 0,1.0},
  {-0.40191758, 0.05831847, 0,1.0},
  // f 1403/1491/823 1386/1490/823 1385/1489/823
  {-0.42656487, 0.0865057, 0,1.0},
  {-0.4108212, 0.05496534, 0,1.0},
  {-0.40617442, 0.05646847, 0,1.0},
  // f 1404/1488/823 1403/1491/823 1385/1489/823
  {-0.42555979, 0.08766402, 0,1.0},
  {-0.42656487, 0.0865057, 0,1.0},
  {-0.40617442, 0.05646847, 0,1.0},
  // f 1401/1493/823 1387/1492/823 1386/1490/823
  {-0.42858201, 0.0847447, 0,1.0},
  {-0.41673514, 0.0536501, 0,1.0},
  {-0.4108212, 0.05496534, 0,1.0},
  // f 1402/1494/823 1401/1493/823 1386/1490/823
  {-0.42771694, 0.08541812, 0,1.0},
  {-0.42858201, 0.0847447, 0,1.0},
  {-0.4108212, 0.05496534, 0,1.0},
  // f 1403/1491/823 1402/1494/823 1386/1490/823
  {-0.42656487, 0.0865057, 0,1.0},
  {-0.42771694, 0.08541812, 0,1.0},
  {-0.4108212, 0.05496534, 0,1.0},
  // f 1399/1496/823 1388/1495/823 1387/1492/823
  {-0.43060106, 0.08356626, 0,1.0},
  {-0.42414385, 0.05272509, 0,1.0},
  {-0.41673514, 0.0536501, 0,1.0},
  // f 1400/1497/823 1399/1496/823 1387/1492/823
  {-0.42929417, 0.08427649, 0,1.0},
  {-0.43060106, 0.08356626, 0,1.0},
  {-0.41673514, 0.0536501, 0,1.0},
  // f 1401/1493/823 1400/1497/823 1387/1492/823
  {-0.42858201, 0.0847447, 0,1.0},
  {-0.42929417, 0.08427649, 0,1.0},
  {-0.41673514, 0.0536501, 0,1.0},
  // f 1397/1499/823 1389/1498/823 1388/1495/823
  {-0.43400544, 0.0823124, 0,1.0},
  {-0.43224552, 0.05241676, 0,1.0},
  {-0.42414385, 0.05272509, 0,1.0},
  // f 1398/1500/823 1397/1499/823 1388/1495/823
  {-0.43205836, 0.08294371, 0,1.0},
  {-0.43400544, 0.0823124, 0,1.0},
  {-0.42414385, 0.05272509, 0,1.0},
  // f 1399/1496/823 1398/1500/823 1388/1495/823
  {-0.43060106, 0.08356626, 0,1.0},
  {-0.43205836, 0.08294371, 0,1.0},
  {-0.42414385, 0.05272509, 0,1.0},
  // f 1394/1502/823 1390/1501/823 1389/1498/823
  {-0.4454461, 0.081034, 0,1.0},
  {-0.45614135, 0.05241676, 0,1.0},
  {-0.43224552, 0.05241676, 0,1.0},
  // f 1395/1503/823 1394/1502/823 1389/1498/823
  {-0.440256, 0.08124619, 0,1.0},
  {-0.4454461, 0.081034, 0,1.0},
  {-0.43224552, 0.05241676, 0,1.0},
  // f 1396/1504/823 1395/1503/823 1389/1498/823
  {-0.43655068, 0.08173544, 0,1.0},
  {-0.440256, 0.08124619, 0,1.0},
  {-0.43224552, 0.05241676, 0,1.0},
  // f 1397/1499/823 1396/1504/823 1389/1498/823
  {-0.43400544, 0.0823124, 0,1.0},
  {-0.43655068, 0.08173544, 0,1.0},
  {-0.43224552, 0.05241676, 0,1.0},
  // f 1394/1502/823 1393/1431/823 1390/1501/823
  {-0.4454461, 0.081034, 0,1.0},
  {-0.45614135, 0.081034, 0,1.0},
  {-0.45614135, 0.05241676, 0,1.0},
  // f 1392/1432/823 1391/1505/823 1390/1501/823
  {-0.49998266, 0, 0,1.0},
  {-0.45614135, 0, 0,1.0},
  {-0.45614135, 0.05241676, 0,1.0},
  // f 1393/1431/823 1392/1432/823 1390/1501/823
  {-0.45614135, 0.081034, 0,1.0},
  {-0.49998266, 0, 0,1.0},
  {-0.45614135, 0.05241676, 0,1.0},
  // f 1635/1508/823 1433/1507/823 1432/1506/823
  {-0.36428103, 0.04170631, 0,1.0},
  {-0.32336509, 0.04933343, 0,1.0},
  {-0.36489382, 0.04673185, 0,1.0},
  // f 1634/1510/823 1434/1509/823 1433/1507/823
  {-0.36350545, 0.03768371, 0,1.0},
  {-0.32255679, 0.04475466, 0,1.0},
  {-0.32336509, 0.04933343, 0,1.0},
  // f 1635/1508/823 1634/1510/823 1433/1507/823
  {-0.36428103, 0.04170631, 0,1.0},
  {-0.36350545, 0.03768371, 0,1.0},
  {-0.32336509, 0.04933343, 0,1.0},
  // f 1633/1512/823 1435/1511/823 1434/1509/823
  {-0.3620148, 0.0323046, 0,1.0},
  {-0.32163715, 0.04136299, 0,1.0},
  {-0.32255679, 0.04475466, 0,1.0},
  // f 1634/1510/823 1633/1512/823 1434/1509/823
  {-0.36350545, 0.03768371, 0,1.0},
  {-0.3620148, 0.0323046, 0,1.0},
  {-0.32255679, 0.04475466, 0,1.0},
  // f 1631/1514/823 1436/1513/823 1435/1511/823
  {-0.35799384, 0.02309645, 0,1.0},
  {-0.3206183, 0.03864001, 0,1.0},
  {-0.32163715, 0.04136299, 0,1.0},
  // f 1632/1515/823 1631/1514/823 1435/1511/823
  {-0.36005208, 0.0272201, 0,1.0},
  {-0.35799384, 0.02309645, 0,1.0},
  {-0.32163715, 0.04136299, 0,1.0},
  // f 1633/1512/823 1632/1515/823 1435/1511/823
  {-0.3620148, 0.0323046, 0,1.0},
  {-0.36005208, 0.0272201, 0,1.0},
  {-0.32163715, 0.04136299, 0,1.0},
  // f 1629/1517/823 1437/1516/823 1436/1513/823
  {-0.352837, 0.01549848, 0,1.0},
  {-0.31956071, 0.03648167, 0,1.0},
  {-0.3206183, 0.03864001, 0,1.0},
  // f 1630/1518/823 1629/1517/823 1436/1513/823
  {-0.35558888, 0.01918923, 0,1.0},
  {-0.352837, 0.01549848, 0,1.0},
  {-0.3206183, 0.03864001, 0,1.0},
  // f 1631/1514/823 1630/1518/823 1436/1513/823
  {-0.35799384, 0.02309645, 0,1.0},
  {-0.35558888, 0.01918923, 0,1.0},
  {-0.3206183, 0.03864001, 0,1.0},
  // f 1628/1520/823 1438/1519/823 1437/1516/823
  {-0.34918806, 0.01146617, 0,1.0},
  {-0.31882712, 0.03525797, 0,1.0},
  {-0.31956071, 0.03648167, 0,1.0},
  // f 1629/1517/823 1628/1520/823 1437/1516/823
  {-0.352837, 0.01549848, 0,1.0},
  {-0.34918806, 0.01146617, 0,1.0},
  {-0.31956071, 0.03648167, 0,1.0},
  // f 1626/1522/823 1439/1521/823 1438/1519/823
  {-0.34619153, 0.00882991, 0,1.0},
  {-0.31787291, 0.03391673, 0,1.0},
  {-0.31882712, 0.03525797, 0,1.0},
  // f 1627/1523/823 1626/1522/823 1438/1519/823
  {-0.34744334, 0.00985111, 0,1.0},
  {-0.34619153, 0.00882991, 0,1.0},
  {-0.31882712, 0.03525797, 0,1.0},
  // f 1628/1520/823 1627/1523/823 1438/1519/823
  {-0.34918806, 0.01146617, 0,1.0},
  {-0.34744334, 0.00985111, 0,1.0},
  {-0.31882712, 0.03525797, 0,1.0},
  // f 1624/1525/823 1440/1524/823 1439/1521/823
  {-0.33968067, 0.00478395, 0,1.0},
  {-0.31646615, 0.03228737, 0,1.0},
  {-0.31787291, 0.03391673, 0,1.0},
  // f 1625/1526/823 1624/1525/823 1439/1521/823
  {-0.34347504, 0.00692071, 0,1.0},
  {-0.33968067, 0.00478395, 0,1.0},
  {-0.31787291, 0.03391673, 0,1.0},
  // f 1626/1522/823 1625/1526/823 1439/1521/823
  {-0.34619153, 0.00882991, 0,1.0},
  {-0.34347504, 0.00692071, 0,1.0},
  {-0.31787291, 0.03391673, 0,1.0},
  // f 1623/1528/823 1441/1527/823 1440/1524/823
  {-0.33544296, 0.0029247, 0,1.0},
  {-0.31494385, 0.03082954, 0,1.0},
  {-0.31646615, 0.03228737, 0,1.0},
  // f 1624/1525/823 1623/1528/823 1440/1524/823
  {-0.33968067, 0.00478395, 0,1.0},
  {-0.33544296, 0.0029247, 0,1.0},
  {-0.31646615, 0.03228737, 0,1.0},
  // f 1622/1530/823 1442/1529/823 1441/1527/823
  {-0.32977268, 0.0010599, 0,1.0},
  {-0.31330577, 0.02954321, 0,1.0},
  {-0.31494385, 0.03082954, 0,1.0},
  // f 1623/1528/823 1622/1530/823 1441/1527/823
  {-0.33544296, 0.0029247, 0,1.0},
  {-0.32977268, 0.0010599, 0,1.0},
  {-0.31494385, 0.03082954, 0,1.0},
  // f 1621/1532/823 1443/1531/823 1442/1529/823
  {-0.32235077, -0.00061066, 0,1.0},
  {-0.31155208, 0.02842839, 0,1.0},
  {-0.31330577, 0.02954321, 0,1.0},
  // f 1622/1530/823 1621/1532/823 1442/1529/823
  {-0.32977268, 0.0010599, 0,1.0},
  {-0.32235077, -0.00061066, 0,1.0},
  {-0.31330577, 0.02954321, 0,1.0},
  // f 1621/1532/823 1444/1533/823 1443/1531/823
  {-0.32235077, -0.00061066, 0,1.0},
  {-0.30968276, 0.02748508, 0,1.0},
  {-0.31155208, 0.02842839, 0,1.0},
  // f 1620/1535/823 1445/1534/823 1444/1533/823
  {-0.31280473, -0.00185386, 0,1.0},
  {-0.30769795, 0.02671328, 0,1.0},
  {-0.30968276, 0.02748508, 0,1.0},
  // f 1621/1532/823 1620/1535/823 1444/1533/823
  {-0.32235077, -0.00061066, 0,1.0},
  {-0.31280473, -0.00185386, 0,1.0},
  {-0.30968276, 0.02748508, 0,1.0},
  // f 1620/1535/823 1446/1536/823 1445/1534/823
  {-0.31280473, -0.00185386, 0,1.0},
  {-0.30559742, 0.02611299, 0,1.0},
  {-0.30769795, 0.02671328, 0,1.0},
  // f 1620/1535/823 1447/1537/823 1446/1536/823
  {-0.31280473, -0.00185386, 0,1.0},
  {-0.30292422, 0.02561904, 0,1.0},
  {-0.30559742, 0.02611299, 0,1.0},
  // f 1619/1539/823 1448/1538/823 1447/1537/823
  {-0.29927659, -0.00240886, 0,1.0},
  {-0.3010495, 0.02542694, 0,1.0},
  {-0.30292422, 0.02561904, 0,1.0},
  // f 1620/1535/823 1619/1539/823 1447/1537/823
  {-0.31280473, -0.00185386, 0,1.0},
  {-0.29927659, -0.00240886, 0,1.0},
  {-0.30292422, 0.02561904, 0,1.0},
  // f 1619/1539/823 1449/1540/823 1448/1538/823
  {-0.29927659, -0.00240886, 0,1.0},
  {-0.29860207, 0.02534119, 0,1.0},
  {-0.3010495, 0.02542694, 0,1.0},
  // f 1619/1539/823 1450/1541/823 1449/1540/823
  {-0.29927659, -0.00240886, 0,1.0},
  {-0.29537636, 0.02548324, 0,1.0},
  {-0.29860207, 0.02534119, 0,1.0},
  // f 1618/1543/823 1451/1542/823 1450/1541/823
  {-0.29134852, -0.00217183, 0,1.0},
  {-0.29274219, 0.025848, 0,1.0},
  {-0.29537636, 0.02548324, 0,1.0},
  // f 1619/1539/823 1618/1543/823 1450/1541/823
  {-0.29927659, -0.00240886, 0,1.0},
  {-0.29134852, -0.00217183, 0,1.0},
  {-0.29537636, 0.02548324, 0,1.0},
  // f 1617/1545/823 1452/1544/823 1451/1542/823
  {-0.2839832, -0.00146073, 0,1.0},
  {-0.29061744, 0.02635129, 0,1.0},
  {-0.29274219, 0.025848, 0,1.0},
  // f 1618/1543/823 1617/1545/823 1451/1542/823
  {-0.29134852, -0.00217183, 0,1.0},
  {-0.2839832, -0.00146073, 0,1.0},
  {-0.29274219, 0.025848, 0,1.0},
  // f 1617/1545/823 1453/1546/823 1452/1544/823
  {-0.2839832, -0.00146073, 0,1.0},
  {-0.28893006, 0.02691947, 0,1.0},
  {-0.29061744, 0.02635129, 0,1.0},
  // f 1616/1548/823 1454/1547/823 1453/1546/823
  {-0.27653128, -0.00013098, 0,1.0},
  {-0.28736609, 0.02761392, 0,1.0},
  {-0.28893006, 0.02691947, 0,1.0},
  // f 1617/1545/823 1616/1548/823 1453/1546/823
  {-0.2839832, -0.00146073, 0,1.0},
  {-0.27653128, -0.00013098, 0,1.0},
  {-0.28893006, 0.02691947, 0,1.0},
  // f 1616/1548/823 1455/1549/823 1454/1547/823
  {-0.27653128, -0.00013098, 0,1.0},
  {-0.2859256, 0.02843463, 0,1.0},
  {-0.28736609, 0.02761392, 0,1.0},
  // f 1615/1551/823 1456/1550/823 1455/1549/823
  {-0.27034768, 0.00157564, 0,1.0},
  {-0.28503397, 0.02905191, 0,1.0},
  {-0.2859256, 0.02843463, 0,1.0},
  // f 1616/1548/823 1615/1551/823 1455/1549/823
  {-0.27653128, -0.00013098, 0,1.0},
  {-0.27034768, 0.00157564, 0,1.0},
  {-0.2859256, 0.02843463, 0,1.0},
  // f 1615/1551/823 1457/1552/823 1456/1550/823
  {-0.27034768, 0.00157564, 0,1.0},
  {-0.28419709, 0.02972531, 0,1.0},
  {-0.28503397, 0.02905191, 0,1.0},
  // f 1614/1554/823 1458/1553/823 1457/1552/823
  {-0.2652635, 0.00351693, 0,1.0},
  {-0.28287992, 0.03098026, 0,1.0},
  {-0.28419709, 0.02972531, 0,1.0},
  // f 1615/1551/823 1614/1554/823 1457/1552/823
  {-0.27034768, 0.00157564, 0,1.0},
  {-0.2652635, 0.00351693, 0,1.0},
  {-0.28419709, 0.02972531, 0,1.0},
  // f 1613/1556/823 1459/1555/823 1458/1553/823
  {-0.26262525, 0.0047575, 0,1.0},
  {-0.28190863, 0.0320995, 0,1.0},
  {-0.28287992, 0.03098026, 0,1.0},
  // f 1614/1554/823 1613/1556/823 1458/1553/823
  {-0.2652635, 0.00351693, 0,1.0},
  {-0.26262525, 0.0047575, 0,1.0},
  {-0.28287992, 0.03098026, 0,1.0},
  // f 1611/1558/823 1460/1557/823 1459/1555/823
  {-0.25632519, 0.00850154, 0,1.0},
  {-0.28108308, 0.033259, 0,1.0},
  {-0.28190863, 0.0320995, 0,1.0},
  // f 1612/1559/823 1611/1558/823 1459/1555/823
  {-0.26010665, 0.00610887, 0,1.0},
  {-0.25632519, 0.00850154, 0,1.0},
  {-0.28190863, 0.0320995, 0,1.0},
  // f 1613/1556/823 1612/1559/823 1459/1555/823
  {-0.26262525, 0.0047575, 0,1.0},
  {-0.26010665, 0.00610887, 0,1.0},
  {-0.28190863, 0.0320995, 0,1.0},
  // f 1610/1561/823 1461/1560/823 1460/1557/823
  {-0.25284964, 0.01117787, 0,1.0},
  {-0.2805064, 0.03425597, 0,1.0},
  {-0.28108308, 0.033259, 0,1.0},
  // f 1611/1558/823 1610/1561/823 1460/1557/823
  {-0.25632519, 0.00850154, 0,1.0},
  {-0.25284964, 0.01117787, 0,1.0},
  {-0.28108308, 0.033259, 0,1.0},
  // f 1609/1563/823 1462/1562/823 1461/1560/823
  {-0.24967997, 0.01413788, 0,1.0},
  {-0.28003085, 0.03528089, 0,1.0},
  {-0.2805064, 0.03425597, 0,1.0},
  // f 1610/1561/823 1609/1563/823 1461/1560/823
  {-0.25284964, 0.01117787, 0,1.0},
  {-0.24967997, 0.01413788, 0,1.0},
  {-0.2805064, 0.03425597, 0,1.0},
  // f 1608/1565/823 1463/1564/823 1462/1562/823
  {-0.24681623, 0.01738155, 0,1.0},
  {-0.27965656, 0.03633375, 0,1.0},
  {-0.28003085, 0.03528089, 0,1.0},
  // f 1609/1563/823 1608/1565/823 1462/1562/823
  {-0.24967997, 0.01413788, 0,1.0},
  {-0.24681623, 0.01738155, 0,1.0},
  {-0.28003085, 0.03528089, 0,1.0},
  // f 1607/1567/823 1464/1566/823 1463/1564/823
  {-0.24425827, 0.02090889, 0,1.0},
  {-0.27938333, 0.03741456, 0,1.0},
  {-0.27965656, 0.03633375, 0,1.0},
  // f 1608/1565/823 1607/1567/823 1463/1564/823
  {-0.24681623, 0.01738155, 0,1.0},
  {-0.24425827, 0.02090889, 0,1.0},
  {-0.27965656, 0.03633375, 0,1.0},
  // f 1605/1569/823 1465/1568/823 1464/1566/823
  {-0.24155261, 0.02558015, 0,1.0},
  {-0.27921134, 0.03852331, 0,1.0},
  {-0.27938333, 0.03741456, 0,1.0},
  // f 1606/1570/823 1605/1569/823 1464/1566/823
  {-0.24283025, 0.02322332, 0,1.0},
  {-0.24155261, 0.02558015, 0,1.0},
  {-0.27938333, 0.03741456, 0,1.0},
  // f 1607/1567/823 1606/1570/823 1464/1566/823
  {-0.24425827, 0.02090889, 0,1.0},
  {-0.24283025, 0.02322332, 0,1.0},
  {-0.27938333, 0.03741456, 0,1.0},
  // f 1604/1572/823 1466/1571/823 1465/1568/823
  {-0.23982103, 0.02943926, 0,1.0},
  {-0.27915674, 0.03920197, 0,1.0},
  {-0.27921134, 0.03852331, 0,1.0},
  // f 1605/1569/823 1604/1572/823 1465/1568/823
  {-0.24155261, 0.02558015, 0,1.0},
  {-0.23982103, 0.02943926, 0,1.0},
  {-0.27921134, 0.03852331, 0,1.0},
  // f 1603/1574/823 1467/1573/823 1466/1571/823
  {-0.23847421, 0.0334069, 0,1.0},
  {-0.27913854, 0.0398907, 0,1.0},
  {-0.27915674, 0.03920197, 0,1.0},
  // f 1604/1572/823 1603/1574/823 1466/1571/823
  {-0.23982103, 0.02943926, 0,1.0},
  {-0.23847421, 0.0334069, 0,1.0},
  {-0.27915674, 0.03920197, 0,1.0},
  // f 1603/1574/823 1602/1575/823 1467/1573/823
  {-0.23847421, 0.0334069, 0,1.0},
  {-0.2375122, 0.03748308, 0,1.0},
  {-0.27913854, 0.0398907, 0,1.0},
  // f 1602/1575/823 1468/1576/823 1467/1573/823
  {-0.2375122, 0.03748308, 0,1.0},
  {-0.27918667, 0.04097757, 0,1.0},
  {-0.27913854, 0.0398907, 0,1.0},
  // f 1602/1575/823 1601/1577/823 1468/1576/823
  {-0.2375122, 0.03748308, 0,1.0},
  {-0.23693499, 0.04166777, 0,1.0},
  {-0.27918667, 0.04097757, 0,1.0},
  // f 1601/1577/823 1469/1578/823 1468/1576/823
  {-0.23693499, 0.04166777, 0,1.0},
  {-0.27933127, 0.04204133, 0,1.0},
  {-0.27918667, 0.04097757, 0,1.0},
  // f 1600/1580/823 1599/1579/823 1469/1578/823
  {-0.23674259, 0.04596103, 0,1.0},
  {-0.23692529, 0.05005173, 0,1.0},
  {-0.27933127, 0.04204133, 0,1.0},
  // f 1601/1577/823 1600/1580/823 1469/1578/823
  {-0.23693499, 0.04166777, 0,1.0},
  {-0.23674259, 0.04596103, 0,1.0},
  {-0.27933127, 0.04204133, 0,1.0},
  // f 1599/1579/823 1470/1581/823 1469/1578/823
  {-0.23692529, 0.05005173, 0,1.0},
  {-0.2795721, 0.04308196, 0,1.0},
  {-0.27933127, 0.04204133, 0,1.0},
  // f 1599/1579/823 1598/1582/823 1470/1581/823
  {-0.23692529, 0.05005173, 0,1.0},
  {-0.23747315, 0.05399883, 0,1.0},
  {-0.2795721, 0.04308196, 0,1.0},
  // f 1598/1582/823 1471/1583/823 1470/1581/823
  {-0.23747315, 0.05399883, 0,1.0},
  {-0.27990934, 0.04409945, 0,1.0},
  {-0.2795721, 0.04308196, 0,1.0},
  // f 1597/1585/823 1596/1584/823 1471/1583/823
  {-0.23826675, 0.05738679, 0,1.0},
  {-0.23934902, 0.0606613, 0,1.0},
  {-0.27990934, 0.04409945, 0,1.0},
  // f 1598/1582/823 1597/1585/823 1471/1583/823
  {-0.23747315, 0.05399883, 0,1.0},
  {-0.23826675, 0.05738679, 0,1.0},
  {-0.27990934, 0.04409945, 0,1.0},
  // f 1596/1584/823 1472/1586/823 1471/1583/823
  {-0.23934902, 0.0606613, 0,1.0},
  {-0.280343, 0.04509383, 0,1.0},
  {-0.27990934, 0.04409945, 0,1.0},
  // f 1594/1588/823 1593/1587/823 1472/1586/823
  {-0.24237932, 0.06686991, 0,1.0},
  {-0.24332435, 0.0683459, 0,1.0},
  {-0.280343, 0.04509383, 0,1.0},
  // f 1595/1589/823 1594/1588/823 1472/1586/823
  {-0.24071984, 0.06382233, 0,1.0},
  {-0.24237932, 0.06686991, 0,1.0},
  {-0.280343, 0.04509383, 0,1.0},
  // f 1596/1584/823 1595/1589/823 1472/1586/823
  {-0.23934902, 0.0606613, 0,1.0},
  {-0.24071984, 0.06382233, 0,1.0},
  {-0.280343, 0.04509383, 0,1.0},
  // f 1593/1587/823 1473/1590/823 1472/1586/823
  {-0.24332435, 0.0683459, 0,1.0},
  {-0.28099045, 0.04625655, 0,1.0},
  {-0.280343, 0.04509383, 0,1.0},
  // f 1591/1592/823 1590/1591/823 1473/1590/823
  {-0.24636841, 0.07220456, 0,1.0},
  {-0.24899231, 0.07482632, 0,1.0},
  {-0.28099045, 0.04625655, 0,1.0},
  // f 1592/1593/823 1591/1592/823 1473/1590/823
  {-0.24435572, 0.06978304, 0,1.0},
  {-0.24636841, 0.07220456, 0,1.0},
  {-0.28099045, 0.04625655, 0,1.0},
  // f 1593/1587/823 1592/1593/823 1473/1590/823
  {-0.24332435, 0.0683459, 0,1.0},
  {-0.24435572, 0.06978304, 0,1.0},
  {-0.28099045, 0.04625655, 0,1.0},
  // f 1590/1591/823 1474/1594/823 1473/1590/823
  {-0.24899231, 0.07482632, 0,1.0},
  {-0.2817767, 0.04738599, 0,1.0},
  {-0.28099045, 0.04625655, 0,1.0},
  // f 1589/1596/823 1588/1595/823 1474/1594/823
  {-0.25196159, 0.07729269, 0,1.0},
  {-0.25571474, 0.07988161, 0,1.0},
  {-0.2817767, 0.04738599, 0,1.0},
  // f 1590/1591/823 1589/1596/823 1474/1594/823
  {-0.24899231, 0.07482632, 0,1.0},
  {-0.25196159, 0.07729269, 0,1.0},
  {-0.2817767, 0.04738599, 0,1.0},
  // f 1588/1595/823 1475/1597/823 1474/1594/823
  {-0.25571474, 0.07988161, 0,1.0},
  {-0.2827017, 0.04848211, 0,1.0},
  {-0.2817767, 0.04738599, 0,1.0},
  // f 1588/1595/823 1587/1598/823 1475/1597/823
  {-0.25571474, 0.07988161, 0,1.0},
  {-0.26039758, 0.08252749, 0,1.0},
  {-0.2827017, 0.04848211, 0,1.0},
  // f 1587/1598/823 1476/1599/823 1475/1597/823
  {-0.26039758, 0.08252749, 0,1.0},
  {-0.28395623, 0.04971883, 0,1.0},
  {-0.2827017, 0.04848211, 0,1.0},
  // f 1587/1598/823 1477/1600/823 1476/1599/823
  {-0.26039758, 0.08252749, 0,1.0},
  {-0.28480846, 0.0504095, 0,1.0},
  {-0.28395623, 0.04971883, 0,1.0},
  // f 1587/1598/823 1586/1601/823 1477/1600/823
  {-0.26039758, 0.08252749, 0,1.0},
  {-0.26727593, 0.08554854, 0,1.0},
  {-0.28480846, 0.0504095, 0,1.0},
  // f 1586/1601/823 1478/1602/823 1477/1600/823
  {-0.26727593, 0.08554854, 0,1.0},
  {-0.28610268, 0.05126418, 0,1.0},
  {-0.28480846, 0.0504095, 0,1.0},
  // f 1586/1601/823 1585/1603/823 1478/1602/823
  {-0.26727593, 0.08554854, 0,1.0},
  {-0.27602455, 0.08855294, 0,1.0},
  {-0.28610268, 0.05126418, 0,1.0},
  // f 1585/1603/823 1479/1604/823 1478/1602/823
  {-0.27602455, 0.08855294, 0,1.0},
  {-0.28834203, 0.0524445, 0,1.0},
  {-0.28610268, 0.05126418, 0,1.0},
  // f 1585/1603/823 1480/1605/823 1479/1604/823
  {-0.27602455, 0.08855294, 0,1.0},
  {-0.29195365, 0.05393434, 0,1.0},
  {-0.28834203, 0.0524445, 0,1.0},
  // f 1585/1603/823 1481/1606/823 1480/1605/823
  {-0.27602455, 0.08855294, 0,1.0},
  {-0.2980513, 0.05587222, 0,1.0},
  {-0.29195365, 0.05393434, 0,1.0},
  // f 1584/1608/823 1482/1607/823 1481/1606/823
  {-0.30139637, 0.0950054, 0,1.0},
  {-0.30631042, 0.05790896, 0,1.0},
  {-0.2980513, 0.05587222, 0,1.0},
  // f 1585/1603/823 1584/1608/823 1481/1606/823
  {-0.27602455, 0.08855294, 0,1.0},
  {-0.30139637, 0.0950054, 0,1.0},
  {-0.2980513, 0.05587222, 0,1.0},
  // f 1583/1610/823 1483/1609/823 1482/1607/823
  {-0.30683568, 0.09628404, 0,1.0},
  {-0.3220557, 0.06216396, 0,1.0},
  {-0.30631042, 0.05790896, 0,1.0},
  // f 1584/1608/823 1583/1610/823 1482/1607/823
  {-0.30139637, 0.0950054, 0,1.0},
  {-0.30683568, 0.09628404, 0,1.0},
  {-0.30631042, 0.05790896, 0,1.0},
  // f 1582/1612/823 1581/1611/823 1483/1609/823
  {-0.31112865, 0.09761478, 0,1.0},
  {-0.31342909, 0.09856663, 0,1.0},
  {-0.3220557, 0.06216396, 0,1.0},
  // f 1583/1610/823 1582/1612/823 1483/1609/823
  {-0.30683568, 0.09628404, 0,1.0},
  {-0.31112865, 0.09761478, 0,1.0},
  {-0.3220557, 0.06216396, 0,1.0},
  // f 1581/1611/823 1484/1613/823 1483/1609/823
  {-0.31342909, 0.09856663, 0,1.0},
  {-0.33171731, 0.06576931, 0,1.0},
  {-0.3220557, 0.06216396, 0,1.0},
  // f 1580/1615/823 1579/1614/823 1484/1613/823
  {-0.31465751, 0.09921506, 0,1.0},
  {-0.31549466, 0.09976391, 0,1.0},
  {-0.33171731, 0.06576931, 0,1.0},
  // f 1581/1611/823 1580/1615/823 1484/1613/823
  {-0.31342909, 0.09856663, 0,1.0},
  {-0.31465751, 0.09921506, 0,1.0},
  {-0.33171731, 0.06576931, 0,1.0},
  // f 1579/1614/823 1485/1616/823 1484/1613/823
  {-0.31549466, 0.09976391, 0,1.0},
  {-0.33843592, 0.06902439, 0,1.0},
  {-0.33171731, 0.06576931, 0,1.0},
  // f 1577/1618/823 1576/1617/823 1485/1616/823
  {-0.31653062, 0.10064447, 0,1.0},
  {-0.31697828, 0.1010891, 0,1.0},
  {-0.33843592, 0.06902439, 0,1.0},
  // f 1578/1619/823 1577/1618/823 1485/1616/823
  {-0.31604224, 0.10020851, 0,1.0},
  {-0.31653062, 0.10064447, 0,1.0},
  {-0.33843592, 0.06902439, 0,1.0},
  // f 1579/1614/823 1578/1619/823 1485/1616/823
  {-0.31549466, 0.09976391, 0,1.0},
  {-0.31604224, 0.10020851, 0,1.0},
  {-0.33843592, 0.06902439, 0,1.0},
  // f 1576/1617/823 1486/1620/823 1485/1616/823
  {-0.31697828, 0.1010891, 0,1.0},
  {-0.34292337, 0.07174607, 0,1.0},
  {-0.33843592, 0.06902439, 0,1.0},
  // f 1576/1617/823 1575/1621/823 1486/1620/823
  {-0.31697828, 0.1010891, 0,1.0},
  {-0.31757355, 0.10177226, 0,1.0},
  {-0.34292337, 0.07174607, 0,1.0},
  // f 1575/1621/823 1487/1622/823 1486/1620/823
  {-0.31757355, 0.10177226, 0,1.0},
  {-0.34519771, 0.07337592, 0,1.0},
  {-0.34292337, 0.07174607, 0,1.0},
  // f 1575/1621/823 1488/1623/823 1487/1622/823
  {-0.31757355, 0.10177226, 0,1.0},
  {-0.34726104, 0.07506003, 0,1.0},
  {-0.34519771, 0.07337592, 0,1.0},
  // f 1575/1621/823 1574/1624/823 1488/1623/823
  {-0.31757355, 0.10177226, 0,1.0},
  {-0.31807724, 0.10247475, 0,1.0},
  {-0.34726104, 0.07506003, 0,1.0},
  // f 1574/1624/823 1489/1625/823 1488/1623/823
  {-0.31807724, 0.10247475, 0,1.0},
  {-0.35004342, 0.07771154, 0,1.0},
  {-0.34726104, 0.07506003, 0,1.0},
  // f 1574/1624/823 1490/1626/823 1489/1625/823
  {-0.31807724, 0.10247475, 0,1.0},
  {-0.35247061, 0.08053235, 0,1.0},
  {-0.35004342, 0.07771154, 0,1.0},
  // f 1574/1624/823 1573/1627/823 1490/1626/823
  {-0.31807724, 0.10247475, 0,1.0},
  {-0.31842694, 0.10307506, 0,1.0},
  {-0.35247061, 0.08053235, 0,1.0},
  // f 1573/1627/823 1491/1628/823 1490/1626/823
  {-0.31842694, 0.10307506, 0,1.0},
  {-0.35422194, 0.08301231, 0,1.0},
  {-0.35247061, 0.08053235, 0,1.0},
  // f 1573/1627/823 1492/1629/823 1491/1628/823
  {-0.31842694, 0.10307506, 0,1.0},
  {-0.35572663, 0.08560983, 0,1.0},
  {-0.35422194, 0.08301231, 0,1.0},
  // f 1573/1627/823 1572/1630/823 1492/1629/823
  {-0.31842694, 0.10307506, 0,1.0},
  {-0.31871322, 0.10368883, 0,1.0},
  {-0.35572663, 0.08560983, 0,1.0},
  // f 1572/1630/823 1493/1631/823 1492/1629/823
  {-0.31871322, 0.10368883, 0,1.0},
  {-0.35698465, 0.0883249, 0,1.0},
  {-0.35572663, 0.08560983, 0,1.0},
  // f 1572/1630/823 1494/1632/823 1493/1631/823
  {-0.31871322, 0.10368883, 0,1.0},
  {-0.35799599, 0.09115754, 0,1.0},
  {-0.35698465, 0.0883249, 0,1.0},
  // f 1572/1630/823 1571/1633/823 1494/1632/823
  {-0.31871322, 0.10368883, 0,1.0},
  {-0.31893578, 0.1043161, 0,1.0},
  {-0.35799599, 0.09115754, 0,1.0},
  // f 1571/1633/823 1495/1634/823 1494/1632/823
  {-0.31893578, 0.1043161, 0,1.0},
  {-0.35876068, 0.09410771, 0,1.0},
  {-0.35799599, 0.09115754, 0,1.0},
  // f 1571/1633/823 1496/1635/823 1495/1634/823
  {-0.31893578, 0.1043161, 0,1.0},
  {-0.35935268, 0.0978031, 0,1.0},
  {-0.35876068, 0.09410771, 0,1.0},
  // f 1571/1633/823 1570/1636/823 1496/1635/823
  {-0.31893578, 0.1043161, 0,1.0},
  {-0.31909472, 0.10495687, 0,1.0},
  {-0.35935268, 0.0978031, 0,1.0},
  // f 1570/1636/823 1497/1637/823 1496/1635/823
  {-0.31909472, 0.10495687, 0,1.0},
  {-0.35951549, 0.09971424, 0,1.0},
  {-0.35935268, 0.0978031, 0,1.0},
  // f 1570/1636/823 1498/1638/823 1497/1637/823
  {-0.31909472, 0.10495687, 0,1.0},
  {-0.35959435, 0.10232831, 0,1.0},
  {-0.35951549, 0.09971424, 0,1.0},
  // f 1569/1640/823 1499/1639/823 1498/1638/823
  {-0.31919011, 0.10561109, 0,1.0},
  {-0.35943285, 0.10578167, 0,1.0},
  {-0.35959435, 0.10232831, 0,1.0},
  // f 1570/1636/823 1569/1640/823 1498/1638/823
  {-0.31909472, 0.10495687, 0,1.0},
  {-0.31919011, 0.10561109, 0,1.0},
  {-0.35959435, 0.10232831, 0,1.0},
  // f 1569/1640/823 1500/1641/823 1499/1639/823
  {-0.31919011, 0.10561109, 0,1.0},
  {-0.35894817, 0.10917331, 0,1.0},
  {-0.35943285, 0.10578167, 0,1.0},
  // f 1568/1643/823 1501/1642/823 1500/1641/823
  {-0.31922197, 0.10627884, 0,1.0},
  {-0.35814032, 0.11250332, 0,1.0},
  {-0.35894817, 0.10917331, 0,1.0},
  // f 1569/1640/823 1568/1643/823 1500/1641/823
  {-0.31919011, 0.10561109, 0,1.0},
  {-0.31922197, 0.10627884, 0,1.0},
  {-0.35894817, 0.10917331, 0,1.0},
  // f 1567/1645/823 1502/1644/823 1501/1642/823
  {-0.31919718, 0.1070099, 0,1.0},
  {-0.35700944, 0.11577167, 0,1.0},
  {-0.35814032, 0.11250332, 0,1.0},
  // f 1568/1643/823 1567/1645/823 1501/1642/823
  {-0.31922197, 0.10627884, 0,1.0},
  {-0.31919718, 0.1070099, 0,1.0},
  {-0.35814032, 0.11250332, 0,1.0},
  // f 1566/1647/823 1503/1646/823 1502/1644/823
  {-0.31914634, 0.10754646, 0,1.0},
  {-0.35555524, 0.11897835, 0,1.0},
  {-0.35700944, 0.11577167, 0,1.0},
  // f 1567/1645/823 1566/1647/823 1502/1644/823
  {-0.31919718, 0.1070099, 0,1.0},
  {-0.31914634, 0.10754646, 0,1.0},
  {-0.35700944, 0.11577167, 0,1.0},
  // f 1566/1647/823 1504/1648/823 1503/1646/823
  {-0.31914634, 0.10754646, 0,1.0},
  {-0.35448223, 0.12095121, 0,1.0},
  {-0.35555524, 0.11897835, 0,1.0},
  // f 1565/1650/823 1505/1649/823 1504/1648/823
  {-0.31899998, 0.10841836, 0,1.0},
  {-0.35328317, 0.12289994, 0,1.0},
  {-0.35448223, 0.12095121, 0,1.0},
  // f 1566/1647/823 1565/1650/823 1504/1648/823
  {-0.31914634, 0.10754646, 0,1.0},
  {-0.31899998, 0.10841836, 0,1.0},
  {-0.35448223, 0.12095121, 0,1.0},
  // f 1565/1650/823 1506/1651/823 1505/1649/823
  {-0.31899998, 0.10841836, 0,1.0},
  {-0.3513914, 0.12553017, 0,1.0},
  {-0.35328317, 0.12289994, 0,1.0},
  // f 1564/1653/823 1507/1652/823 1506/1651/823
  {-0.31877634, 0.10926232, 0,1.0},
  {-0.34925041, 0.12799795, 0,1.0},
  {-0.3513914, 0.12553017, 0,1.0},
  // f 1565/1650/823 1564/1653/823 1506/1651/823
  {-0.31899998, 0.10841836, 0,1.0},
  {-0.31877634, 0.10926232, 0,1.0},
  {-0.3513914, 0.12553017, 0,1.0},
  // f 1563/1655/823 1508/1654/823 1507/1652/823
  {-0.31847572, 0.11007833, 0,1.0},
  {-0.3464981, 0.13061942, 0,1.0},
  {-0.34925041, 0.12799795, 0,1.0},
  // f 1564/1653/823 1563/1655/823 1507/1652/823
  {-0.31877634, 0.10926232, 0,1.0},
  {-0.31847572, 0.11007833, 0,1.0},
  {-0.34925041, 0.12799795, 0,1.0},
  // f 1563/1655/823 1509/1656/823 1508/1654/823
  {-0.31847572, 0.11007833, 0,1.0},
  {-0.34342036, 0.13302872, 0,1.0},
  {-0.3464981, 0.13061942, 0,1.0},
  // f 1562/1658/823 1510/1657/823 1509/1656/823
  {-0.31809804, 0.11086644, 0,1.0},
  {-0.34001693, 0.13522591, 0,1.0},
  {-0.34342036, 0.13302872, 0,1.0},
  // f 1563/1655/823 1562/1658/823 1509/1656/823
  {-0.31847572, 0.11007833, 0,1.0},
  {-0.31809804, 0.11086644, 0,1.0},
  {-0.34342036, 0.13302872, 0,1.0},
  // f 1561/1660/823 1511/1659/823 1510/1657/823
  {-0.31764328, 0.1116266, 0,1.0},
  {-0.3372508, 0.13673456, 0,1.0},
  {-0.34001693, 0.13522591, 0,1.0},
  // f 1562/1658/823 1561/1660/823 1510/1657/823
  {-0.31809804, 0.11086644, 0,1.0},
  {-0.31764328, 0.1116266, 0,1.0},
  {-0.34001693, 0.13522591, 0,1.0},
  // f 1560/1662/823 1512/1661/823 1511/1659/823
  {-0.31699577, 0.11250187, 0,1.0},
  {-0.33430135, 0.13812394, 0,1.0},
  {-0.3372508, 0.13673456, 0,1.0},
  // f 1561/1660/823 1560/1662/823 1511/1659/823
  {-0.31764328, 0.1116266, 0,1.0},
  {-0.31699577, 0.11250187, 0,1.0},
  {-0.3372508, 0.13673456, 0,1.0},
  // f 1559/1664/823 1513/1663/823 1512/1661/823
  {-0.31623718, 0.1133369, 0,1.0},
  {-0.33000627, 0.13975506, 0,1.0},
  {-0.33430135, 0.13812394, 0,1.0},
  // f 1560/1662/823 1559/1664/823 1512/1661/823
  {-0.31699577, 0.11250187, 0,1.0},
  {-0.31623718, 0.1133369, 0,1.0},
  {-0.33430135, 0.13812394, 0,1.0},
  // f 1557/1666/823 1514/1665/823 1513/1663/823
  {-0.31472713, 0.1146275, 0,1.0},
  {-0.32459944, 0.14125092, 0,1.0},
  {-0.33000627, 0.13975506, 0,1.0},
  // f 1558/1667/823 1557/1666/823 1513/1663/823
  {-0.31536773, 0.1141317, 0,1.0},
  {-0.31472713, 0.1146275, 0,1.0},
  {-0.33000627, 0.13975506, 0,1.0},
  // f 1559/1664/823 1558/1667/823 1513/1663/823
  {-0.31623718, 0.1133369, 0,1.0},
  {-0.31536773, 0.1141317, 0,1.0},
  {-0.33000627, 0.13975506, 0,1.0},
  // f 1555/1669/823 1515/1668/823 1514/1665/823
  {-0.31310949, 0.11559201, 0,1.0},
  {-0.31716415, 0.1425917, 0,1.0},
  {-0.32459944, 0.14125092, 0,1.0},
  // f 1556/1670/823 1555/1669/823 1514/1665/823
  {-0.31403819, 0.115082, 0,1.0},
  {-0.31310949, 0.11559201, 0,1.0},
  {-0.32459944, 0.14125092, 0,1.0},
  // f 1557/1666/823 1556/1670/823 1514/1665/823
  {-0.31472713, 0.1146275, 0,1.0},
  {-0.31403819, 0.115082, 0,1.0},
  {-0.32459944, 0.14125092, 0,1.0},
  // f 1552/1672/823 1516/1671/823 1515/1668/823
  {-0.30889514, 0.11694124, 0,1.0},
  {-0.30881959, 0.14339617, 0,1.0},
  {-0.31716415, 0.1425917, 0,1.0},
  // f 1553/1673/823 1552/1672/823 1515/1668/823
  {-0.31057414, 0.11655263, 0,1.0},
  {-0.30889514, 0.11694124, 0,1.0},
  {-0.31716415, 0.1425917, 0,1.0},
  // f 1554/1674/823 1553/1673/823 1515/1668/823
  {-0.31189597, 0.11611877, 0,1.0},
  {-0.31057414, 0.11655263, 0,1.0},
  {-0.31716415, 0.1425917, 0,1.0},
  // f 1555/1669/823 1554/1674/823 1515/1668/823
  {-0.31310949, 0.11559201, 0,1.0},
  {-0.31189597, 0.11611877, 0,1.0},
  {-0.31716415, 0.1425917, 0,1.0},
  // f 1550/1676/823 1517/1675/823 1516/1671/823
  {-0.30390158, 0.11735957, 0,1.0},
  {-0.29956564, 0.1436643, 0,1.0},
  {-0.30881959, 0.14339617, 0,1.0},
  // f 1551/1677/823 1550/1676/823 1516/1671/823
  {-0.30679607, 0.11723047, 0,1.0},
  {-0.30390158, 0.11735957, 0,1.0},
  {-0.30881959, 0.14339617, 0,1.0},
  // f 1552/1672/823 1551/1677/823 1516/1671/823
  {-0.30889514, 0.11694124, 0,1.0},
  {-0.30679607, 0.11723047, 0,1.0},
  {-0.30881959, 0.14339617, 0,1.0},
  // f 1547/1679/823 1518/1678/823 1517/1675/823
  {-0.29744899, 0.11673349, 0,1.0},
  {-0.28827497, 0.14317694, 0,1.0},
  {-0.29956564, 0.1436643, 0,1.0},
  // f 1548/1680/823 1547/1679/823 1517/1675/823
  {-0.30002517, 0.11714971, 0,1.0},
  {-0.29744899, 0.11673349, 0,1.0},
  {-0.29956564, 0.1436643, 0,1.0},
  // f 1549/1681/823 1548/1680/823 1517/1675/823
  {-0.30173922, 0.11729716, 0,1.0},
  {-0.30002517, 0.11714971, 0,1.0},
  {-0.29956564, 0.1436643, 0,1.0},
  // f 1550/1676/823 1549/1681/823 1517/1675/823
  {-0.30390158, 0.11735957, 0,1.0},
  {-0.30173922, 0.11729716, 0,1.0},
  {-0.29956564, 0.1436643, 0,1.0},
  // f 1544/1683/823 1519/1682/823 1518/1678/823
  {-0.29215118, 0.11485515, 0,1.0},
  {-0.27996331, 0.14205326, 0,1.0},
  {-0.28827497, 0.14317694, 0,1.0},
  // f 1545/1684/823 1544/1683/823 1518/1678/823
  {-0.29369941, 0.11558357, 0,1.0},
  {-0.29215118, 0.11485515, 0,1.0},
  {-0.28827497, 0.14317694, 0,1.0},
  // f 1546/1685/823 1545/1684/823 1518/1678/823
  {-0.2953628, 0.11618715, 0,1.0},
  {-0.29369941, 0.11558357, 0,1.0},
  {-0.28827497, 0.14317694, 0,1.0},
  // f 1547/1679/823 1546/1685/823 1518/1678/823
  {-0.29744899, 0.11673349, 0,1.0},
  {-0.2953628, 0.11618715, 0,1.0},
  {-0.28827497, 0.14317694, 0,1.0},
  // f 1542/1687/823 1520/1686/823 1519/1682/823
  {-0.28940025, 0.11302363, 0,1.0},
  {-0.27327496, 0.14050667, 0,1.0},
  {-0.27996331, 0.14205326, 0,1.0},
  // f 1543/1688/823 1542/1687/823 1519/1682/823
  {-0.29071814, 0.11400183, 0,1.0},
  {-0.28940025, 0.11302363, 0,1.0},
  {-0.27996331, 0.14205326, 0,1.0},
  // f 1544/1683/823 1543/1688/823 1519/1682/823
  {-0.29215118, 0.11485515, 0,1.0},
  {-0.29071814, 0.11400183, 0,1.0},
  {-0.27996331, 0.14205326, 0,1.0},
  // f 1540/1690/823 1521/1689/823 1520/1686/823
  {-0.28781238, 0.11150294, 0,1.0},
  {-0.26796278, 0.13873047, 0,1.0},
  {-0.27327496, 0.14050667, 0,1.0},
  // f 1541/1691/823 1540/1690/823 1520/1686/823
  {-0.2885831, 0.11229659, 0,1.0},
  {-0.28781238, 0.11150294, 0,1.0},
  {-0.27327496, 0.14050667, 0,1.0},
  // f 1542/1687/823 1541/1691/823 1520/1686/823
  {-0.28940025, 0.11302363, 0,1.0},
  {-0.2885831, 0.11229659, 0,1.0},
  {-0.27327496, 0.14050667, 0,1.0},
  // f 1539/1693/823 1522/1692/823 1521/1689/823
  {-0.28674284, 0.11018759, 0,1.0},
  {-0.26381099, 0.13689388, 0,1.0},
  {-0.26796278, 0.13873047, 0,1.0},
  // f 1540/1690/823 1539/1693/823 1521/1689/823
  {-0.28781238, 0.11150294, 0,1.0},
  {-0.28674284, 0.11018759, 0,1.0},
  {-0.26796278, 0.13873047, 0,1.0},
  // f 1539/1693/823 1523/1694/823 1522/1692/823
  {-0.28674284, 0.11018759, 0,1.0},
  {-0.26124883, 0.1355084, 0,1.0},
  {-0.26381099, 0.13689388, 0,1.0},
  // f 1538/1696/823 1524/1695/823 1523/1694/823
  {-0.28577727, 0.10872237, 0,1.0},
  {-0.2582778, 0.13359529, 0,1.0},
  {-0.26124883, 0.1355084, 0,1.0},
  // f 1539/1693/823 1538/1696/823 1523/1694/823
  {-0.28674284, 0.11018759, 0,1.0},
  {-0.28577727, 0.10872237, 0,1.0},
  {-0.26124883, 0.1355084, 0,1.0},
  // f 1538/1696/823 1525/1697/823 1524/1695/823
  {-0.28577727, 0.10872237, 0,1.0},
  {-0.2555505, 0.1314625, 0,1.0},
  {-0.2582778, 0.13359529, 0,1.0},
  // f 1537/1699/823 1526/1698/823 1525/1697/823
  {-0.28491592, 0.10710733, 0,1.0},
  {-0.25305346, 0.12909168, 0,1.0},
  {-0.2555505, 0.1314625, 0,1.0},
  // f 1538/1696/823 1537/1699/823 1525/1697/823
  {-0.28577727, 0.10872237, 0,1.0},
  {-0.28491592, 0.10710733, 0,1.0},
  {-0.2555505, 0.1314625, 0,1.0},
  // f 1537/1699/823 1527/1700/823 1526/1698/823
  {-0.28491592, 0.10710733, 0,1.0},
  {-0.25078672, 0.12648289, 0,1.0},
  {-0.25305346, 0.12909168, 0,1.0},
  // f 1536/1702/823 1528/1701/823 1527/1700/823
  {-0.28404248, 0.1050337, 0,1.0},
  {-0.24875031, 0.1236361, 0,1.0},
  {-0.25078672, 0.12648289, 0,1.0},
  // f 1537/1699/823 1536/1702/823 1527/1700/823
  {-0.28491592, 0.10710733, 0,1.0},
  {-0.28404248, 0.1050337, 0,1.0},
  {-0.25078672, 0.12648289, 0,1.0},
  // f 1536/1702/823 1529/1703/823 1528/1701/823
  {-0.28404248, 0.1050337, 0,1.0},
  {-0.24694414, 0.12055133, 0,1.0},
  {-0.24875031, 0.1236361, 0,1.0},
  // f 1536/1702/823 1530/1704/823 1529/1703/823
  {-0.28404248, 0.1050337, 0,1.0},
  {-0.24508065, 0.11653544, 0,1.0},
  {-0.24694414, 0.12055133, 0,1.0},
  // f 1535/1706/823 1531/1705/823 1530/1704/823
  {-0.28321767, 0.10241412, 0,1.0},
  {-0.2435489, 0.11217681, 0,1.0},
  {-0.24508065, 0.11653544, 0,1.0},
  // f 1536/1702/823 1535/1706/823 1530/1704/823
  {-0.28404248, 0.1050337, 0,1.0},
  {-0.28321767, 0.10241412, 0,1.0},
  {-0.24508065, 0.11653544, 0,1.0},
  // f 1534/1708/823 1532/1707/823 1531/1705/823
  {-0.28251091, 0.09914861, 0,1.0},
  {-0.24234867, 0.1074755, 0,1.0},
  {-0.2435489, 0.11217681, 0,1.0},
  // f 1535/1706/823 1534/1708/823 1531/1705/823
  {-0.28321767, 0.10241412, 0,1.0},
  {-0.28251091, 0.09914861, 0,1.0},
  {-0.2435489, 0.11217681, 0,1.0},
  // f 1534/1708/823 1533/1709/823 1532/1707/823
  {-0.28251091, 0.09914861, 0,1.0},
  {-0.24136762, 0.10155746, 0,1.0},
  {-0.24234867, 0.1074755, 0,1.0},
  // f 1638/1712/823 1637/1711/823 1636/1710/823
  {-0.17112531, 0.03478392, 0,1.0},
  {-0.17112531, 0.14125545, 0,1.0},
  {-0.21477383, 0.14125545, 0,1.0},
  // f 1641/1713/823 1638/1712/823 1636/1710/823
  {-0.21477383, 0, 0,1.0},
  {-0.17112531, 0.03478392, 0,1.0},
  {-0.21477383, 0.14125545, 0,1.0},
  // f 1640/1715/823 1639/1714/823 1638/1712/823
  {-0.10300279, 0, 0,1.0},
  {-0.10300279, 0.03478392, 0,1.0},
  {-0.17112531, 0.03478392, 0,1.0},
  // f 1641/1713/823 1640/1715/823 1638/1712/823
  {-0.21477383, 0, 0,1.0},
  {-0.10300279, 0, 0,1.0},
  {-0.17112531, 0.03478392, 0,1.0},
  // f 1645/1718/823 1643/1717/823 1642/1716/823
  {-0.04027612, 0.0941382, 0,1.0},
  {-0.00057813, 0, 0,1.0},
  {-0.00057813, 0.1436643, 0,1.0},
  // f 1659/1719/823 1645/1718/823 1642/1716/823
  {-0.03883079, 0.13370511, 0,1.0},
  {-0.04027612, 0.0941382, 0,1.0},
  {-0.00057813, 0.1436643, 0,1.0},
  // f 1660/1720/823 1659/1719/823 1642/1716/823
  {-0.03574746, 0.13851516, 0,1.0},
  {-0.03883079, 0.13370511, 0,1.0},
  {-0.00057813, 0.1436643, 0,1.0},
  // f 1661/1721/823 1660/1720/823 1642/1716/823
  {-0.03304954, 0.1436643, 0,1.0},
  {-0.03574746, 0.13851516, 0,1.0},
  {-0.00057813, 0.1436643, 0,1.0},
  // f 1645/1718/823 1644/1722/823 1643/1717/823
  {-0.04027612, 0.0941382, 0,1.0},
  {-0.04027612, 0, 0,1.0},
  {-0.00057813, 0, 0,1.0},
  // f 1653/1724/823 1646/1723/823 1645/1718/823
  {-0.06128185, 0.11416088, 0,1.0},
  {-0.05012761, 0.08727793, 0,1.0},
  {-0.04027612, 0.0941382, 0,1.0},
  // f 1654/1725/823 1653/1724/823 1645/1718/823
  {-0.05518406, 0.11779124, 0,1.0},
  {-0.06128185, 0.11416088, 0,1.0},
  {-0.04027612, 0.0941382, 0,1.0},
  // f 1655/1726/823 1654/1725/823 1645/1718/823
  {-0.05269908, 0.11952468, 0,1.0},
  {-0.05518406, 0.11779124, 0,1.0},
  {-0.04027612, 0.0941382, 0,1.0},
  // f 1656/1727/823 1655/1726/823 1645/1718/823
  {-0.05039332, 0.12131011, 0,1.0},
  {-0.05269908, 0.11952468, 0,1.0},
  {-0.04027612, 0.0941382, 0,1.0},
  // f 1657/1728/823 1656/1727/823 1645/1718/823
  {-0.04615372, 0.12510261, 0,1.0},
  {-0.05039332, 0.12131011, 0,1.0},
  {-0.04027612, 0.0941382, 0,1.0},
  // f 1658/1729/823 1657/1728/823 1645/1718/823
  {-0.04229955, 0.12923431, 0,1.0},
  {-0.04615372, 0.12510261, 0,1.0},
  {-0.04027612, 0.0941382, 0,1.0},
  // f 1659/1719/823 1658/1729/823 1645/1718/823
  {-0.03883079, 0.13370511, 0,1.0},
  {-0.04229955, 0.12923431, 0,1.0},
  {-0.04027612, 0.0941382, 0,1.0},
  // f 1652/1731/823 1647/1730/823 1646/1723/823
  {-0.06960668, 0.11019515, 0,1.0},
  {-0.05456021, 0.08462132, 0,1.0},
  {-0.05012761, 0.08727793, 0,1.0},
  // f 1653/1724/823 1652/1731/823 1646/1723/823
  {-0.06128185, 0.11416088, 0,1.0},
  {-0.06960668, 0.11019515, 0,1.0},
  {-0.05012761, 0.08727793, 0,1.0},
  // f 1652/1731/823 1648/1732/823 1647/1730/823
  {-0.06960668, 0.11019515, 0,1.0},
  {-0.05892068, 0.0822866, 0,1.0},
  {-0.05456021, 0.08462132, 0,1.0},
  // f 1652/1731/823 1649/1733/823 1648/1732/823
  {-0.06960668, 0.11019515, 0,1.0},
  {-0.06907402, 0.0778543, 0,1.0},
  {-0.05892068, 0.0822866, 0,1.0},
  // f 1651/1735/823 1650/1734/823 1649/1733/823
  {-0.08151577, 0.10579707, 0,1.0},
  {-0.08151577, 0.07361472, 0,1.0},
  {-0.06907402, 0.0778543, 0,1.0},
  // f 1652/1731/823 1651/1735/823 1649/1733/823
  {-0.06960668, 0.11019515, 0,1.0},
  {-0.08151577, 0.10579707, 0,1.0},
  {-0.06907402, 0.0778543, 0,1.0},
  // f 1784/1738/823 1663/1737/823 1662/1736/823
  {0.14867473, 0.05116415, 0,1.0},
  {0.11947938, 0.08055224, 0,1.0},
  {0.11947938, 0.05116415, 0,1.0},
  // f 1784/1738/823 1664/1739/823 1663/1737/823
  {0.14867473, 0.05116415, 0,1.0},
  {0.18692741, 0.08055224, 0,1.0},
  {0.11947938, 0.08055224, 0,1.0},
  // f 1784/1738/823 1665/1740/823 1664/1739/823
  {0.14867473, 0.05116415, 0,1.0},
  {0.18692741, 0.02033077, 0,1.0},
  {0.18692741, 0.08055224, 0,1.0},
  // f 1783/1742/823 1666/1741/823 1665/1740/823
  {0.14867473, 0.03786726, 0,1.0},
  {0.17215054, 0.01111929, 0,1.0},
  {0.18692741, 0.02033077, 0,1.0},
  // f 1784/1738/823 1783/1742/823 1665/1740/823
  {0.14867473, 0.05116415, 0,1.0},
  {0.14867473, 0.03786726, 0,1.0},
  {0.18692741, 0.02033077, 0,1.0},
  // f 1783/1742/823 1667/1743/823 1666/1741/823
  {0.14867473, 0.03786726, 0,1.0},
  {0.16200872, 0.0059812, 0,1.0},
  {0.17215054, 0.01111929, 0,1.0},
  // f 1782/1745/823 1668/1744/823 1667/1743/823
  {0.14306615, 0.03473994, 0,1.0},
  {0.15755193, 0.00410277, 0,1.0},
  {0.16200872, 0.0059812, 0,1.0},
  // f 1783/1742/823 1782/1745/823 1667/1743/823
  {0.14867473, 0.03786726, 0,1.0},
  {0.14306615, 0.03473994, 0,1.0},
  {0.16200872, 0.0059812, 0,1.0},
  // f 1782/1745/823 1669/1746/823 1668/1744/823
  {0.14306615, 0.03473994, 0,1.0},
  {0.15267344, 0.00236068, 0,1.0},
  {0.15755193, 0.00410277, 0,1.0},
  // f 1781/1748/823 1670/1747/823 1669/1746/823
  {0.13790309, 0.03230322, 0,1.0},
  {0.1452087, 0.00034603, 0,1.0},
  {0.15267344, 0.00236068, 0,1.0},
  // f 1782/1745/823 1781/1748/823 1669/1746/823
  {0.14306615, 0.03473994, 0,1.0},
  {0.13790309, 0.03230322, 0,1.0},
  {0.15267344, 0.00236068, 0,1.0},
  // f 1780/1750/823 1671/1749/823 1670/1747/823
  {0.13345076, 0.03064069, 0,1.0},
  {0.13569312, -0.00130996, 0,1.0},
  {0.1452087, 0.00034603, 0,1.0},
  // f 1781/1748/823 1780/1750/823 1670/1747/823
  {0.13790309, 0.03230322, 0,1.0},
  {0.13345076, 0.03064069, 0,1.0},
  {0.1452087, 0.00034603, 0,1.0},
  // f 1778/1752/823 1672/1751/823 1671/1749/823
  {0.12802409, 0.02928401, 0,1.0},
  {0.12684548, -0.00213413, 0,1.0},
  {0.13569312, -0.00130996, 0,1.0},
  // f 1779/1753/823 1778/1752/823 1671/1749/823
  {0.13076441, 0.02987755, 0,1.0},
  {0.12802409, 0.02928401, 0,1.0},
  {0.13569312, -0.00130996, 0,1.0},
  // f 1780/1750/823 1779/1753/823 1671/1749/823
  {0.13345076, 0.03064069, 0,1.0},
  {0.13076441, 0.02987755, 0,1.0},
  {0.13569312, -0.00130996, 0,1.0},
  // f 1776/1755/823 1673/1754/823 1672/1751/823
  {0.11947938, 0.02852089, 0,1.0},
  {0.11735957, -0.00240886, 0,1.0},
  {0.12684548, -0.00213413, 0,1.0},
  // f 1777/1756/823 1776/1755/823 1672/1751/823
  {0.12381244, 0.02871166, 0,1.0},
  {0.11947938, 0.02852089, 0,1.0},
  {0.12684548, -0.00213413, 0,1.0},
  // f 1778/1752/823 1777/1756/823 1672/1751/823
  {0.12802409, 0.02928401, 0,1.0},
  {0.12381244, 0.02871166, 0,1.0},
  {0.12684548, -0.00213413, 0,1.0},
  // f 1774/1758/823 1674/1757/823 1673/1754/823
  {0.11471993, 0.02877249, 0,1.0},
  {0.1058239, -0.00191491, 0,1.0},
  {0.11735957, -0.00240886, 0,1.0},
  // f 1775/1759/823 1774/1758/823 1673/1754/823
  {0.11705618, 0.02858379, 0,1.0},
  {0.11471993, 0.02877249, 0,1.0},
  {0.11735957, -0.00240886, 0,1.0},
  // f 1776/1755/823 1775/1759/823 1673/1754/823
  {0.11947938, 0.02852089, 0,1.0},
  {0.11705618, 0.02858379, 0,1.0},
  {0.11735957, -0.00240886, 0,1.0},
  // f 1773/1761/823 1772/1760/823 1674/1757/823
  {0.1113786, 0.0292914, 0,1.0},
  {0.10823289, 0.03009339, 0,1.0},
  {0.1058239, -0.00191491, 0,1.0},
  // f 1774/1758/823 1773/1761/823 1674/1757/823
  {0.11471993, 0.02877249, 0,1.0},
  {0.1113786, 0.0292914, 0,1.0},
  {0.1058239, -0.00191491, 0,1.0},
  // f 1772/1760/823 1675/1762/823 1674/1757/823
  {0.10823289, 0.03009339, 0,1.0},
  {0.09702906, -0.00074863, 0,1.0},
  {0.1058239, -0.00191491, 0,1.0},
  // f 1771/1764/823 1770/1763/823 1675/1762/823
  {0.10576094, 0.03097792, 0,1.0},
  {0.10342481, 0.03205901, 0,1.0},
  {0.09702906, -0.00074863, 0,1.0},
  // f 1772/1760/823 1771/1764/823 1675/1762/823
  {0.10823289, 0.03009339, 0,1.0},
  {0.10576094, 0.03097792, 0,1.0},
  {0.09702906, -0.00074863, 0,1.0},
  // f 1770/1763/823 1676/1765/823 1675/1762/823
  {0.10342481, 0.03205901, 0,1.0},
  {0.08974966, 0.00088758, 0,1.0},
  {0.09702906, -0.00074863, 0,1.0},
  // f 1770/1763/823 1769/1766/823 1676/1765/823
  {0.10342481, 0.03205901, 0,1.0},
  {0.10122457, 0.03333668, 0,1.0},
  {0.08974966, 0.00088758, 0,1.0},
  // f 1769/1766/823 1677/1767/823 1676/1765/823
  {0.10122457, 0.03333668, 0,1.0},
  {0.08378498, 0.0028085, 0,1.0},
  {0.08974966, 0.00088758, 0,1.0},
  // f 1769/1766/823 1768/1768/823 1677/1767/823
  {0.10122457, 0.03333668, 0,1.0},
  {0.09916019, 0.0348109, 0,1.0},
  {0.08378498, 0.0028085, 0,1.0},
  // f 1768/1768/823 1678/1769/823 1677/1767/823
  {0.09916019, 0.0348109, 0,1.0},
  {0.08029877, 0.00423204, 0,1.0},
  {0.08378498, 0.0028085, 0,1.0},
  // f 1768/1768/823 1679/1770/823 1678/1769/823
  {0.09916019, 0.0348109, 0,1.0},
  {0.07636081, 0.00616668, 0,1.0},
  {0.08029877, 0.00423204, 0,1.0},
  // f 1768/1768/823 1767/1771/823 1679/1770/823
  {0.09916019, 0.0348109, 0,1.0},
  {0.09723166, 0.03648167, 0,1.0},
  {0.07636081, 0.00616668, 0,1.0},
  // f 1767/1771/823 1680/1772/823 1679/1770/823
  {0.09723166, 0.03648167, 0,1.0},
  {0.07147124, 0.00912545, 0,1.0},
  {0.07636081, 0.00616668, 0,1.0},
  // f 1767/1771/823 1766/1773/823 1680/1772/823
  {0.09723166, 0.03648167, 0,1.0},
  {0.09543897, 0.03834902, 0,1.0},
  {0.07147124, 0.00912545, 0,1.0},
  // f 1766/1773/823 1681/1774/823 1680/1772/823
  {0.09543897, 0.03834902, 0,1.0},
  {0.06694677, 0.01251342, 0,1.0},
  {0.07147124, 0.00912545, 0,1.0},
  // f 1766/1773/823 1765/1775/823 1681/1774/823
  {0.09543897, 0.03834902, 0,1.0},
  {0.09411437, 0.04000539, 0,1.0},
  {0.06694677, 0.01251342, 0,1.0},
  // f 1765/1775/823 1682/1776/823 1681/1774/823
  {0.09411437, 0.04000539, 0,1.0},
  {0.06278734, 0.0163306, 0,1.0},
  {0.06694677, 0.01251342, 0,1.0},
  // f 1765/1775/823 1764/1777/823 1682/1776/823
  {0.09411437, 0.04000539, 0,1.0},
  {0.09319337, 0.04135774, 0,1.0},
  {0.06278734, 0.0163306, 0,1.0},
  // f 1764/1777/823 1683/1778/823 1682/1776/823
  {0.09319337, 0.04135774, 0,1.0},
  {0.05899298, 0.02057697, 0,1.0},
  {0.06278734, 0.0163306, 0,1.0},
  // f 1764/1777/823 1763/1779/823 1683/1778/823
  {0.09319337, 0.04135774, 0,1.0},
  {0.09179632, 0.04382134, 0,1.0},
  {0.05899298, 0.02057697, 0,1.0},
  // f 1763/1779/823 1684/1780/823 1683/1778/823
  {0.09179632, 0.04382134, 0,1.0},
  {0.0555637, 0.02525254, 0,1.0},
  {0.05899298, 0.02057697, 0,1.0},
  // f 1763/1779/823 1762/1781/823 1684/1780/823
  {0.09179632, 0.04382134, 0,1.0},
  {0.09057174, 0.046547, 0,1.0},
  {0.0555637, 0.02525254, 0,1.0},
  // f 1762/1781/823 1685/1782/823 1684/1780/823
  {0.09057174, 0.046547, 0,1.0},
  {0.05360577, 0.02839273, 0,1.0},
  {0.0555637, 0.02525254, 0,1.0},
  // f 1762/1781/823 1686/1783/823 1685/1782/823
  {0.09057174, 0.046547, 0,1.0},
  {0.05179045, 0.03170058, 0,1.0},
  {0.05360577, 0.02839273, 0,1.0},
  // f 1762/1781/823 1761/1784/823 1686/1783/823
  {0.09057174, 0.046547, 0,1.0},
  {0.08932994, 0.05016376, 0,1.0},
  {0.05179045, 0.03170058, 0,1.0},
  // f 1761/1784/823 1687/1785/823 1686/1783/823
  {0.08932994, 0.05016376, 0,1.0},
  {0.04893376, 0.03796792, 0,1.0},
  {0.05179045, 0.03170058, 0,1.0},
  // f 1761/1784/823 1760/1786/823 1687/1785/823
  {0.08932994, 0.05016376, 0,1.0},
  {0.08819505, 0.0548603, 0,1.0},
  {0.04893376, 0.03796792, 0,1.0},
  // f 1760/1786/823 1688/1787/823 1687/1785/823
  {0.08819505, 0.0548603, 0,1.0},
  {0.04664213, 0.04455995, 0,1.0},
  {0.04893376, 0.03796792, 0,1.0},
  // f 1760/1786/823 1689/1788/823 1688/1787/823
  {0.08819505, 0.0548603, 0,1.0},
  {0.04491555, 0.05147664, 0,1.0},
  {0.04664213, 0.04455995, 0,1.0},
  // f 1760/1786/823 1759/1789/823 1689/1788/823
  {0.08819505, 0.0548603, 0,1.0},
  {0.08723264, 0.06165352, 0,1.0},
  {0.04491555, 0.05147664, 0,1.0},
  // f 1759/1789/823 1690/1790/823 1689/1788/823
  {0.08723264, 0.06165352, 0,1.0},
  {0.04365986, 0.05954267, 0,1.0},
  {0.04491555, 0.05147664, 0,1.0},
  // f 1759/1789/823 1691/1791/823 1690/1790/823
  {0.08723264, 0.06165352, 0,1.0},
  {0.0432413, 0.06457467, 0,1.0},
  {0.04365986, 0.05954267, 0,1.0},
  // f 1759/1789/823 1758/1792/823 1691/1791/823
  {0.08723264, 0.06165352, 0,1.0},
  {0.08681526, 0.0711095, 0,1.0},
  {0.0432413, 0.06457467, 0,1.0},
  // f 1758/1792/823 1692/1793/823 1691/1791/823
  {0.08681526, 0.0711095, 0,1.0},
  {0.04307039, 0.07062773, 0,1.0},
  {0.0432413, 0.06457467, 0,1.0},
  // f 1758/1792/823 1693/1794/823 1692/1793/823
  {0.08681526, 0.0711095, 0,1.0},
  {0.04353208, 0.08051129, 0,1.0},
  {0.04307039, 0.07062773, 0,1.0},
  // f 1756/1796/823 1694/1795/823 1693/1794/823
  {0.0874838, 0.08228591, 0,1.0},
  {0.04475309, 0.08899045, 0,1.0},
  {0.04353208, 0.08051129, 0,1.0},
  // f 1757/1797/823 1756/1796/823 1693/1794/823
  {0.08698238, 0.07692527, 0,1.0},
  {0.0874838, 0.08228591, 0,1.0},
  {0.04353208, 0.08051129, 0,1.0},
  // f 1758/1792/823 1757/1797/823 1693/1794/823
  {0.08681526, 0.0711095, 0,1.0},
  {0.08698238, 0.07692527, 0,1.0},
  {0.04353208, 0.08051129, 0,1.0},
  // f 1756/1796/823 1695/1798/823 1694/1795/823
  {0.0874838, 0.08228591, 0,1.0},
  {0.04650446, 0.09620979, 0,1.0},
  {0.04475309, 0.08899045, 0,1.0},
  // f 1755/1800/823 1696/1799/823 1695/1798/823
  {0.08846615, 0.08785504, 0,1.0},
  {0.04858016, 0.10229938, 0,1.0},
  {0.04650446, 0.09620979, 0,1.0},
  // f 1756/1796/823 1755/1800/823 1695/1798/823
  {0.0874838, 0.08228591, 0,1.0},
  {0.08846615, 0.08785504, 0,1.0},
  {0.04650446, 0.09620979, 0,1.0},
  // f 1754/1802/823 1697/1801/823 1696/1799/823
  {0.08968385, 0.09224036, 0,1.0},
  {0.05045747, 0.10666426, 0,1.0},
  {0.04858016, 0.10229938, 0,1.0},
  // f 1755/1800/823 1754/1802/823 1696/1799/823
  {0.08846615, 0.08785504, 0,1.0},
  {0.08968385, 0.09224036, 0,1.0},
  {0.04858016, 0.10229938, 0,1.0},
  // f 1754/1802/823 1698/1803/823 1697/1801/823
  {0.08968385, 0.09224036, 0,1.0},
  {0.05260947, 0.11085567, 0,1.0},
  {0.05045747, 0.10666426, 0,1.0},
  // f 1753/1805/823 1699/1804/823 1698/1803/823
  {0.09099369, 0.09563697, 0,1.0},
  {0.05588891, 0.11611768, 0,1.0},
  {0.05260947, 0.11085567, 0,1.0},
  // f 1754/1802/823 1753/1805/823 1698/1803/823
  {0.08968385, 0.09224036, 0,1.0},
  {0.09099369, 0.09563697, 0,1.0},
  {0.05260947, 0.11085567, 0,1.0},
  // f 1752/1807/823 1700/1806/823 1699/1804/823
  {0.09227277, 0.09821206, 0,1.0},
  {0.05913073, 0.12037592, 0,1.0},
  {0.05588891, 0.11611768, 0,1.0},
  // f 1753/1805/823 1752/1807/823 1699/1804/823
  {0.09099369, 0.09563697, 0,1.0},
  {0.09227277, 0.09821206, 0,1.0},
  {0.05588891, 0.11611768, 0,1.0},
  // f 1752/1807/823 1701/1808/823 1700/1806/823
  {0.09227277, 0.09821206, 0,1.0},
  {0.06272005, 0.12431127, 0,1.0},
  {0.05913073, 0.12037592, 0,1.0},
  // f 1751/1810/823 1702/1809/823 1701/1808/823
  {0.09372242, 0.1005549, 0,1.0},
  {0.06665686, 0.12792365, 0,1.0},
  {0.06272005, 0.12431127, 0,1.0},
  // f 1752/1807/823 1751/1810/823 1701/1808/823
  {0.09227277, 0.09821206, 0,1.0},
  {0.09372242, 0.1005549, 0,1.0},
  {0.06272005, 0.12431127, 0,1.0},
  // f 1750/1812/823 1703/1811/823 1702/1809/823
  {0.09534263, 0.10266554, 0,1.0},
  {0.07158156, 0.13165666, 0,1.0},
  {0.06665686, 0.12792365, 0,1.0},
  // f 1751/1810/823 1750/1812/823 1702/1809/823
  {0.09372242, 0.1005549, 0,1.0},
  {0.09534263, 0.10266554, 0,1.0},
  {0.06665686, 0.12792365, 0,1.0},
  // f 1749/1814/823 1704/1813/823 1703/1811/823
  {0.09674428, 0.10419735, 0,1.0},
  {0.07557292, 0.13417959, 0,1.0},
  {0.07158156, 0.13165666, 0,1.0},
  // f 1750/1812/823 1749/1814/823 1703/1811/823
  {0.09534263, 0.10266554, 0,1.0},
  {0.09674428, 0.10419735, 0,1.0},
  {0.07158156, 0.13165666, 0,1.0},
  // f 1748/1816/823 1705/1815/823 1704/1813/823
  {0.09784453, 0.10526244, 0,1.0},
  {0.08055224, 0.13682316, 0,1.0},
  {0.07557292, 0.13417959, 0,1.0},
  // f 1749/1814/823 1748/1816/823 1704/1813/823
  {0.09674428, 0.10419735, 0,1.0},
  {0.09784453, 0.10526244, 0,1.0},
  {0.07557292, 0.13417959, 0,1.0},
  // f 1747/1818/823 1706/1817/823 1705/1815/823
  {0.09977153, 0.106878, 0,1.0},
  {0.08540048, 0.1388372, 0,1.0},
  {0.08055224, 0.13682316, 0,1.0},
  // f 1748/1816/823 1747/1818/823 1705/1815/823
  {0.09784453, 0.10526244, 0,1.0},
  {0.09977153, 0.106878, 0,1.0},
  {0.08055224, 0.13682316, 0,1.0},
  // f 1746/1820/823 1707/1819/823 1706/1817/823
  {0.1018151, 0.10829411, 0,1.0},
  {0.09146211, 0.14068431, 0,1.0},
  {0.08540048, 0.1388372, 0,1.0},
  // f 1747/1818/823 1746/1820/823 1706/1817/823
  {0.09977153, 0.106878, 0,1.0},
  {0.1018151, 0.10829411, 0,1.0},
  {0.08540048, 0.1388372, 0,1.0},
  // f 1744/1822/823 1708/1821/823 1707/1819/823
  {0.10625204, 0.11052801, 0,1.0},
  {0.09895155, 0.14221673, 0,1.0},
  {0.09146211, 0.14068431, 0,1.0},
  // f 1745/1823/823 1744/1822/823 1707/1819/823
  {0.10397528, 0.10951076, 0,1.0},
  {0.10625204, 0.11052801, 0,1.0},
  {0.09146211, 0.14068431, 0,1.0},
  // f 1746/1820/823 1745/1823/823 1707/1819/823
  {0.1018151, 0.10829411, 0,1.0},
  {0.10397528, 0.10951076, 0,1.0},
  {0.09146211, 0.14068431, 0,1.0},
  // f 1743/1825/823 1709/1824/823 1708/1821/823
  {0.10864536, 0.11134578, 0,1.0},
  {0.10810695, 0.14327027, 0,1.0},
  {0.09895155, 0.14221673, 0,1.0},
  // f 1744/1822/823 1743/1825/823 1708/1821/823
  {0.10625204, 0.11052801, 0,1.0},
  {0.10864536, 0.11134578, 0,1.0},
  {0.09895155, 0.14221673, 0,1.0},
  // f 1741/1827/823 1710/1826/823 1709/1824/823
  {0.11486509, 0.11249463, 0,1.0},
  {0.11919031, 0.1436643, 0,1.0},
  {0.10810695, 0.14327027, 0,1.0},
  // f 1742/1828/823 1741/1827/823 1709/1824/823
  {0.11167126, 0.11206378, 0,1.0},
  {0.11486509, 0.11249463, 0,1.0},
  {0.10810695, 0.14327027, 0,1.0},
  // f 1743/1825/823 1742/1828/823 1709/1824/823
  {0.10864536, 0.11134578, 0,1.0},
  {0.11167126, 0.11206378, 0,1.0},
  {0.10810695, 0.14327027, 0,1.0},
  // f 1738/1830/823 1711/1829/823 1710/1826/823
  {0.12300596, 0.1123516, 0,1.0},
  {0.13603304, 0.14298595, 0,1.0},
  {0.11919031, 0.1436643, 0,1.0},
  // f 1739/1831/823 1738/1830/823 1710/1826/823
  {0.12085765, 0.11255514, 0,1.0},
  {0.12300596, 0.1123516, 0,1.0},
  {0.11919031, 0.1436643, 0,1.0},
  // f 1740/1832/823 1739/1831/823 1710/1826/823
  {0.11822677, 0.11263821, 0,1.0},
  {0.12085765, 0.11255514, 0,1.0},
  {0.11919031, 0.1436643, 0,1.0},
  // f 1741/1827/823 1740/1832/823 1710/1826/823
  {0.11486509, 0.11249463, 0,1.0},
  {0.11822677, 0.11263821, 0,1.0},
  {0.11919031, 0.1436643, 0,1.0},
  // f 1735/1834/823 1712/1833/823 1711/1829/823
  {0.13032858, 0.1104404, 0,1.0},
  {0.1455756, 0.14170393, 0,1.0},
  {0.13603304, 0.14298595, 0,1.0},
  // f 1736/1835/823 1735/1834/823 1711/1829/823
  {0.12825319, 0.11121202, 0,1.0},
  {0.13032858, 0.1104404, 0,1.0},
  {0.13603304, 0.14298595, 0,1.0},
  // f 1737/1836/823 1736/1835/823 1711/1829/823
  {0.12571715, 0.11189037, 0,1.0},
  {0.12825319, 0.11121202, 0,1.0},
  {0.13603304, 0.14298595, 0,1.0},
  // f 1738/1830/823 1737/1836/823 1711/1829/823
  {0.12300596, 0.1123516, 0,1.0},
  {0.12571715, 0.11189037, 0,1.0},
  {0.13603304, 0.14298595, 0,1.0},
  // f 1733/1838/823 1713/1837/823 1712/1833/823
  {0.13407704, 0.10839865, 0,1.0},
  {0.15019752, 0.14067285, 0,1.0},
  {0.1455756, 0.14170393, 0,1.0},
  // f 1734/1839/823 1733/1838/823 1712/1833/823
  {0.13226987, 0.10950262, 0,1.0},
  {0.13407704, 0.10839865, 0,1.0},
  {0.1455756, 0.14170393, 0,1.0},
  // f 1735/1834/823 1734/1839/823 1712/1833/823
  {0.13032858, 0.1104404, 0,1.0},
  {0.13226987, 0.10950262, 0,1.0},
  {0.1455756, 0.14170393, 0,1.0},
  // f 1732/1841/823 1714/1840/823 1713/1837/823
  {0.13574637, 0.1071379, 0,1.0},
  {0.15421513, 0.13942471, 0,1.0},
  {0.15019752, 0.14067285, 0,1.0},
  // f 1733/1838/823 1732/1841/823 1713/1837/823
  {0.13407704, 0.10839865, 0,1.0},
  {0.13574637, 0.1071379, 0,1.0},
  {0.15019752, 0.14067285, 0,1.0},
  // f 1732/1841/823 1715/1842/823 1714/1840/823
  {0.13574637, 0.1071379, 0,1.0},
  {0.15870138, 0.13754196, 0,1.0},
  {0.15421513, 0.13942471, 0,1.0},
  // f 1731/1844/823 1716/1843/823 1715/1842/823
  {0.13706453, 0.10594001, 0,1.0},
  {0.16246878, 0.13552782, 0,1.0},
  {0.15870138, 0.13754196, 0,1.0},
  // f 1732/1841/823 1731/1844/823 1715/1842/823
  {0.13574637, 0.1071379, 0,1.0},
  {0.13706453, 0.10594001, 0,1.0},
  {0.15870138, 0.13754196, 0,1.0},
  // f 1730/1846/823 1717/1845/823 1716/1843/823
  {0.1384708, 0.10440568, 0,1.0},
  {0.16598019, 0.13321085, 0,1.0},
  {0.16246878, 0.13552782, 0,1.0},
  // f 1731/1844/823 1730/1846/823 1716/1843/823
  {0.13706453, 0.10594001, 0,1.0},
  {0.1384708, 0.10440568, 0,1.0},
  {0.16246878, 0.13552782, 0,1.0},
  // f 1730/1846/823 1718/1847/823 1717/1845/823
  {0.1384708, 0.10440568, 0,1.0},
  {0.16923565, 0.13059105, 0,1.0},
  {0.16598019, 0.13321085, 0,1.0},
  // f 1729/1849/823 1719/1848/823 1718/1847/823
  {0.13973548, 0.10272409, 0,1.0},
  {0.17126371, 0.12867633, 0,1.0},
  {0.16923565, 0.13059105, 0,1.0},
  // f 1730/1846/823 1729/1849/823 1718/1847/823
  {0.1384708, 0.10440568, 0,1.0},
  {0.13973548, 0.10272409, 0,1.0},
  {0.16923565, 0.13059105, 0,1.0},
  // f 1729/1849/823 1720/1850/823 1719/1848/823
  {0.13973548, 0.10272409, 0,1.0},
  {0.17348598, 0.12627235, 0,1.0},
  {0.17126371, 0.12867633, 0,1.0},
  // f 1728/1852/823 1721/1851/823 1720/1850/823
  {0.14085847, 0.10089514, 0,1.0},
  {0.17612036, 0.12290923, 0,1.0},
  {0.17348598, 0.12627235, 0,1.0},
  // f 1729/1849/823 1728/1852/823 1720/1850/823
  {0.13973548, 0.10272409, 0,1.0},
  {0.14085847, 0.10089514, 0,1.0},
  {0.17348598, 0.12627235, 0,1.0},
  // f 1728/1852/823 1722/1853/823 1721/1851/823
  {0.14085847, 0.10089514, 0,1.0},
  {0.17875814, 0.11880992, 0,1.0},
  {0.17612036, 0.12290923, 0,1.0},
  // f 1727/1855/823 1723/1854/823 1722/1853/823
  {0.14196849, 0.09862459, 0,1.0},
  {0.18109146, 0.11432907, 0,1.0},
  {0.17875814, 0.11880992, 0,1.0},
  // f 1728/1852/823 1727/1855/823 1722/1853/823
  {0.14085847, 0.10089514, 0,1.0},
  {0.14196849, 0.09862459, 0,1.0},
  {0.17875814, 0.11880992, 0,1.0},
  // f 1726/1857/823 1724/1856/823 1723/1854/823
  {0.14289349, 0.09616163, 0,1.0},
  {0.1831203, 0.10946664, 0,1.0},
  {0.18109146, 0.11432907, 0,1.0},
  // f 1727/1855/823 1726/1857/823 1723/1854/823
  {0.14196849, 0.09862459, 0,1.0},
  {0.14289349, 0.09616163, 0,1.0},
  {0.18109146, 0.11432907, 0,1.0},
  // f 1726/1857/823 1725/1858/823 1724/1856/823
  {0.14289349, 0.09616163, 0,1.0},
  {0.18500036, 0.10367728, 0,1.0},
  {0.1831203, 0.10946664, 0,1.0},
  // f 1787/1861/823 1786/1860/823 1785/1859/823
  {0.25668794, 0.09182569, 0,1.0},
  {0.25668794, 0.14125545, 0,1.0},
  {0.21303943, 0.14125545, 0,1.0},
  // f 1794/1862/823 1787/1861/823 1785/1859/823
  {0.25668794, 0.05713812, 0,1.0},
  {0.25668794, 0.09182569, 0,1.0},
  {0.21303943, 0.14125545, 0,1.0},
  // f 1796/1863/823 1794/1862/823 1785/1859/823
  {0.21303943, 0, 0,1.0},
  {0.25668794, 0.05713812, 0,1.0},
  {0.21303943, 0.14125545, 0,1.0},
  // f 1794/1862/823 1788/1864/823 1787/1861/823
  {0.25668794, 0.05713812, 0,1.0},
  {0.30438331, 0.09182569, 0,1.0},
  {0.25668794, 0.09182569, 0,1.0},
  // f 1790/1866/823 1789/1865/823 1788/1864/823
  {0.34822458, 0.14125545, 0,1.0},
  {0.30438331, 0.14125545, 0,1.0},
  {0.30438331, 0.09182569, 0,1.0},
  // f 1793/1867/823 1790/1866/823 1788/1864/823
  {0.30438331, 0.05713812, 0,1.0},
  {0.34822458, 0.14125545, 0,1.0},
  {0.30438331, 0.09182569, 0,1.0},
  // f 1794/1862/823 1793/1867/823 1788/1864/823
  {0.25668794, 0.05713812, 0,1.0},
  {0.30438331, 0.05713812, 0,1.0},
  {0.30438331, 0.09182569, 0,1.0},
  // f 1793/1867/823 1791/1868/823 1790/1866/823
  {0.30438331, 0.05713812, 0,1.0},
  {0.34822458, 0, 0,1.0},
  {0.34822458, 0.14125545, 0,1.0},
  // f 1793/1867/823 1792/1869/823 1791/1868/823
  {0.30438331, 0.05713812, 0,1.0},
  {0.30438331, 0, 0,1.0},
  {0.34822458, 0, 0,1.0},
  // f 1796/1863/823 1795/1870/823 1794/1862/823
  {0.21303943, 0, 0,1.0},
  {0.25668794, 0, 0,1.0},
  {0.25668794, 0.05713812, 0,1.0},
  // f 1803/1873/823 1798/1872/823 1797/1871/823
  {0.41181844, 0.10637518, 0,1.0},
  {0.49998266, 0.14125545, 0,1.0},
  {0.36730275, 0.14125545, 0,1.0},
  // f 1804/1874/823 1803/1873/823 1797/1871/823
  {0.36730275, 0.10637518, 0,1.0},
  {0.41181844, 0.10637518, 0,1.0},
  {0.36730275, 0.14125545, 0,1.0},
  // f 1800/1876/823 1799/1875/823 1798/1872/823
  {0.45546687, 0.10637518, 0,1.0},
  {0.49998266, 0.10637518, 0,1.0},
  {0.49998266, 0.14125545, 0,1.0},
  // f 1803/1873/823 1800/1876/823 1798/1872/823
  {0.41181844, 0.10637518, 0,1.0},
  {0.45546687, 0.10637518, 0,1.0},
  {0.49998266, 0.14125545, 0,1.0},
  // f 1802/1878/823 1801/1877/823 1800/1876/823
  {0.41181844, 0, 0,1.0},
  {0.45546687, 0, 0,1.0},
  {0.45546687, 0.10637518, 0,1.0},
  // f 1803/1873/823 1802/1878/823 1800/1876/823
  {0.41181844, 0.10637518, 0,1.0},
  {0.41181844, 0, 0,1.0},
  {0.45546687, 0.10637518, 0,1.0},
  // f 2/4/1 453/3/1 1/1/1
  {-0.42742792, 0.14125545, 0,1.0},
  {-0.42742792, 0.14125545, -0.01973337,1.0},
  {-0.49998266, 0.14125545, 0,1.0},
  // f 3/6/4 454/5/3 2/4/2
  {-0.41928023, 0.14089021, 0,1.0},
  {-0.41928023, 0.14089021, -0.01973337,1.0},
  {-0.42742792, 0.14125545, 0,1.0},
  // f 4/8/6 455/7/5 3/6/4
  {-0.4126853, 0.13995226, 0,1.0},
  {-0.4126853, 0.13995226, -0.01973337,1.0},
  {-0.41928023, 0.14089021, 0,1.0},
  // f 5/10/8 456/9/7 4/8/6
  {-0.40741441, 0.13865802, 0,1.0},
  {-0.40741441, 0.13865802, -0.01973337,1.0},
  {-0.4126853, 0.13995226, 0,1.0},
  // f 6/12/10 457/11/9 5/10/824
  {-0.40326801, 0.13719702, 0,1.0},
  {-0.40326801, 0.13719702, -0.01973337,1.0},
  {-0.40741441, 0.13865802, 0,1.0},
  // f 7/14/12 458/13/11 6/12/825
  {-0.39946407, 0.13541128, 0,1.0},
  {-0.39946407, 0.13541128, -0.01973337,1.0},
  {-0.40326801, 0.13719702, 0,1.0},
  // f 8/16/14 459/15/13 7/14/826
  {-0.39655605, 0.13367519, 0,1.0},
  {-0.39655605, 0.13367519, -0.01973337,1.0},
  {-0.39946407, 0.13541128, 0,1.0},
  // f 9/18/16 460/17/15 8/16/14
  {-0.39440098, 0.13212398, 0,1.0},
  {-0.39440098, 0.13212398, -0.01973337,1.0},
  {-0.39655605, 0.13367519, 0,1.0},
  // f 10/1879/18 461/19/17 9/18/16
  {-0.39192128, 0.12998199, 0,1.0},
  {-0.39192128, 0.12998199, -0.01973337,1.0},
  {-0.39440098, 0.13212398, 0,1.0},
  // f 11/23/20 462/22/19 10/20/827
  {-0.38967857, 0.1276319, 0,1.0},
  {-0.38967857, 0.1276319, -0.01973337,1.0},
  {-0.39192128, 0.12998199, 0,1.0},
  // f 12/25/22 463/24/21 11/23/828
  {-0.38767204, 0.12509105, 0,1.0},
  {-0.38767204, 0.12509105, -0.01973337,1.0},
  {-0.38967857, 0.1276319, 0,1.0},
  // f 13/27/24 464/26/23 12/25/829
  {-0.3859016, 0.1223594, 0,1.0},
  {-0.3859016, 0.1223594, -0.01973337,1.0},
  {-0.38767204, 0.12509105, 0,1.0},
  // f 14/29/26 465/28/25 13/27/830
  {-0.38436708, 0.11943695, 0,1.0},
  {-0.38436708, 0.11943695, -0.01973337,1.0},
  {-0.3859016, 0.1223594, 0,1.0},
  // f 15/31/28 466/30/27 14/29/831
  {-0.38283733, 0.11567822, 0,1.0},
  {-0.38283733, 0.11567822, -0.01973337,1.0},
  {-0.38436708, 0.11943695, 0,1.0},
  // f 16/33/30 467/32/29 15/31/28
  {-0.38164759, 0.11164479, 0,1.0},
  {-0.38164759, 0.11164479, -0.01973337,1.0},
  {-0.38283733, 0.11567822, 0,1.0},
  // f 17/35/32 468/34/31 16/33/832
  {-0.38079774, 0.10733657, 0,1.0},
  {-0.38079774, 0.10733657, -0.01973337,1.0},
  {-0.38164759, 0.11164479, 0,1.0},
  // f 18/37/34 469/36/33 17/35/833
  {-0.3802878, 0.10275365, 0,1.0},
  {-0.3802878, 0.10275365, -0.01973337,1.0},
  {-0.38079774, 0.10733657, 0,1.0},
  // f 19/39/36 470/38/35 18/37/34
  {-0.38011783, 0.097896, 0,1.0},
  {-0.38011783, 0.097896, -0.01973337,1.0},
  {-0.3802878, 0.10275365, 0,1.0},
  // f 20/41/38 471/40/37 19/39/834
  {-0.38036993, 0.09208981, 0,1.0},
  {-0.38036993, 0.09208981, -0.01973337,1.0},
  {-0.38011783, 0.097896, 0,1.0},
  // f 21/43/40 472/42/39 20/41/835
  {-0.38112643, 0.08665, 0,1.0},
  {-0.38112643, 0.08665, -0.01973337,1.0},
  {-0.38036993, 0.09208981, 0,1.0},
  // f 22/45/42 473/44/41 21/43/836
  {-0.38217598, 0.08227888, 0,1.0},
  {-0.38217598, 0.08227888, -0.01973337,1.0},
  {-0.38112643, 0.08665, 0,1.0},
  // f 23/47/44 474/46/43 22/45/42
  {-0.38359609, 0.07817698, 0,1.0},
  {-0.38359609, 0.07817698, -0.01973337,1.0},
  {-0.38217598, 0.08227888, 0,1.0},
  // f 24/49/46 475/48/45 23/47/837
  {-0.38506255, 0.07496432, 0,1.0},
  {-0.38506255, 0.07496432, -0.01973337,1.0},
  {-0.38359609, 0.07817698, 0,1.0},
  // f 25/51/48 476/50/47 24/49/838
  {-0.38678622, 0.07193859, 0,1.0},
  {-0.38678622, 0.07193859, -0.01973337,1.0},
  {-0.38506255, 0.07496432, 0,1.0},
  // f 26/53/50 477/52/49 25/51/839
  {-0.38876721, 0.06909983, 0,1.0},
  {-0.38876721, 0.06909983, -0.01973337,1.0},
  {-0.38678622, 0.07193859, 0,1.0},
  // f 27/55/52 478/54/51 26/53/50
  {-0.39053714, 0.06696337, 0,1.0},
  {-0.39053714, 0.06696337, -0.01973337,1.0},
  {-0.38876721, 0.06909983, 0,1.0},
  // f 28/1880/54 479/56/53 27/55/52
  {-0.39298108, 0.06446107, 0,1.0},
  {-0.39298108, 0.06446107, -0.01973337,1.0},
  {-0.39053714, 0.06696337, 0,1.0},
  // f 29/60/56 480/59/55 28/57/840
  {-0.39568928, 0.06217263, 0,1.0},
  {-0.39568928, 0.06217263, -0.01973337,1.0},
  {-0.39298108, 0.06446107, 0,1.0},
  // f 30/62/58 481/61/57 29/60/56
  {-0.39866793, 0.06012511, 0,1.0},
  {-0.39866793, 0.06012511, -0.01973337,1.0},
  {-0.39568928, 0.06217263, 0,1.0},
  // f 31/64/60 482/63/59 30/62/58
  {-0.40191758, 0.05831847, 0,1.0},
  {-0.40191758, 0.05831847, -0.01973337,1.0},
  {-0.39866793, 0.06012511, 0,1.0},
  // f 32/66/62 483/65/61 31/64/60
  {-0.40617442, 0.05646847, 0,1.0},
  {-0.40617442, 0.05646847, -0.01973337,1.0},
  {-0.40191758, 0.05831847, 0,1.0},
  // f 33/68/64 484/67/63 32/66/62
  {-0.4108212, 0.05496534, 0,1.0},
  {-0.4108212, 0.05496534, -0.01973337,1.0},
  {-0.40617442, 0.05646847, 0,1.0},
  // f 34/70/66 485/69/65 33/68/64
  {-0.41673514, 0.0536501, 0,1.0},
  {-0.41673514, 0.0536501, -0.01973337,1.0},
  {-0.4108212, 0.05496534, 0,1.0},
  // f 35/72/68 486/71/67 34/70/66
  {-0.42414385, 0.05272509, 0,1.0},
  {-0.42414385, 0.05272509, -0.01973337,1.0},
  {-0.41673514, 0.0536501, 0,1.0},
  // f 36/74/69 487/73/69 35/72/68
  {-0.43224552, 0.05241676, 0,1.0},
  {-0.43224552, 0.05241676, -0.01973337,1.0},
  {-0.42414385, 0.05272509, 0,1.0},
  // f 37/1881/70 488/75/70 36/74/70
  {-0.45614135, 0.05241676, 0,1.0},
  {-0.45614135, 0.05241676, -0.01973337,1.0},
  {-0.43224552, 0.05241676, 0,1.0},
  // f 38/1882/71 489/78/71 37/76/71
  {-0.45614135, 0, 0,1.0},
  {-0.45614135, 0, -0.01973337,1.0},
  {-0.45614135, 0.05241676, 0,1.0},
  // f 39/1883/70 490/81/70 38/79/70
  {-0.49998266, 0, 0,1.0},
  {-0.49998266, 0, -0.01973337,1.0},
  {-0.45614135, 0, 0,1.0},
  // f 1/1884/72 452/84/72 39/82/72
  {-0.49998266, 0.14125545, 0,1.0},
  {-0.49998266, 0.14125545, -0.01973337,1.0},
  {-0.49998266, 0, 0,1.0},
  // f 41/88/1 492/87/1 40/85/1
  {-0.4454461, 0.081034, 0,1.0},
  {-0.4454461, 0.081034, -0.01973337,1.0},
  {-0.45614135, 0.081034, 0,1.0},
  // f 42/90/75 493/89/74 41/88/73
  {-0.440256, 0.08124619, 0,1.0},
  {-0.440256, 0.08124619, -0.01973337,1.0},
  {-0.4454461, 0.081034, 0,1.0},
  // f 43/92/77 494/91/76 42/90/75
  {-0.43655068, 0.08173544, 0,1.0},
  {-0.43655068, 0.08173544, -0.01973337,1.0},
  {-0.440256, 0.08124619, 0,1.0},
  // f 44/94/79 495/93/78 43/92/77
  {-0.43400544, 0.0823124, 0,1.0},
  {-0.43400544, 0.0823124, -0.01973337,1.0},
  {-0.43655068, 0.08173544, 0,1.0},
  // f 45/96/81 496/95/80 44/94/79
  {-0.43205836, 0.08294371, 0,1.0},
  {-0.43205836, 0.08294371, -0.01973337,1.0},
  {-0.43400544, 0.0823124, 0,1.0},
  // f 46/98/83 497/97/82 45/96/81
  {-0.43060106, 0.08356626, 0,1.0},
  {-0.43060106, 0.08356626, -0.01973337,1.0},
  {-0.43205836, 0.08294371, 0,1.0},
  // f 47/100/85 498/99/84 46/98/83
  {-0.42929417, 0.08427649, 0,1.0},
  {-0.42929417, 0.08427649, -0.01973337,1.0},
  {-0.43060106, 0.08356626, 0,1.0},
  // f 48/102/87 499/101/86 47/100/85
  {-0.42858201, 0.0847447, 0,1.0},
  {-0.42858201, 0.0847447, -0.01973337,1.0},
  {-0.42929417, 0.08427649, 0,1.0},
  // f 49/104/89 500/103/88 48/102/841
  {-0.42771694, 0.08541812, 0,1.0},
  {-0.42771694, 0.08541812, -0.01973337,1.0},
  {-0.42858201, 0.0847447, 0,1.0},
  // f 50/1885/91 501/105/90 49/104/842
  {-0.42656487, 0.0865057, 0,1.0},
  {-0.42656487, 0.0865057, -0.01973337,1.0},
  {-0.42771694, 0.08541812, 0,1.0},
  // f 51/109/93 502/108/92 50/106/843
  {-0.42555979, 0.08766402, 0,1.0},
  {-0.42555979, 0.08766402, -0.01973337,1.0},
  {-0.42656487, 0.0865057, 0,1.0},
  // f 52/111/95 503/110/94 51/109/93
  {-0.42483458, 0.08868335, 0,1.0},
  {-0.42483458, 0.08868335, -0.01973337,1.0},
  {-0.42555979, 0.08766402, 0,1.0},
  // f 53/113/97 504/112/96 52/111/95
  {-0.42421165, 0.08975184, 0,1.0},
  {-0.42421165, 0.08975184, -0.01973337,1.0},
  {-0.42483458, 0.08868335, 0,1.0},
  // f 54/115/99 505/114/98 53/113/844
  {-0.42369071, 0.09086943, 0,1.0},
  {-0.42369071, 0.09086943, -0.01973337,1.0},
  {-0.42421165, 0.08975184, 0,1.0},
  // f 55/117/101 506/116/100 54/115/845
  {-0.42327198, 0.09203622, 0,1.0},
  {-0.42327198, 0.09203622, -0.01973337,1.0},
  {-0.42369071, 0.09086943, 0,1.0},
  // f 56/119/103 507/118/102 55/117/101
  {-0.42295536, 0.09325211, 0,1.0},
  {-0.42295536, 0.09325211, -0.01973337,1.0},
  {-0.42327198, 0.09203622, 0,1.0},
  // f 57/121/105 508/120/104 56/119/846
  {-0.42271024, 0.09477603, 0,1.0},
  {-0.42271024, 0.09477603, -0.01973337,1.0},
  {-0.42295536, 0.09325211, 0,1.0},
  // f 58/123/107 509/122/106 57/121/105
  {-0.42264283, 0.09556455, 0,1.0},
  {-0.42264283, 0.09556455, -0.01973337,1.0},
  {-0.42271024, 0.09477603, 0,1.0},
  // f 59/125/109 510/124/108 58/123/107
  {-0.42261013, 0.09664342, 0,1.0},
  {-0.42261013, 0.09664342, -0.01973337,1.0},
  {-0.42264283, 0.09556455, 0,1.0},
  // f 60/127/111 511/126/110 59/125/847
  {-0.42269701, 0.0984653, 0,1.0},
  {-0.42269701, 0.0984653, -0.01973337,1.0},
  {-0.42261013, 0.09664342, 0,1.0},
  // f 61/129/113 512/128/112 60/127/111
  {-0.42290977, 0.09996384, 0,1.0},
  {-0.42290977, 0.09996384, -0.01973337,1.0},
  {-0.42269701, 0.0984653, 0,1.0},
  // f 62/131/115 513/130/114 61/129/113
  {-0.42325002, 0.10140404, 0,1.0},
  {-0.42325002, 0.10140404, -0.01973337,1.0},
  {-0.42290977, 0.09996384, 0,1.0},
  // f 63/133/117 514/132/116 62/131/848
  {-0.42371821, 0.10278598, 0,1.0},
  {-0.42371821, 0.10278598, -0.01973337,1.0},
  {-0.42325002, 0.10140404, 0,1.0},
  // f 64/135/119 515/134/118 63/133/117
  {-0.4243139, 0.10410966, 0,1.0},
  {-0.4243139, 0.10410966, -0.01973337,1.0},
  {-0.42371821, 0.10278598, 0,1.0},
  // f 65/137/121 516/136/120 64/135/849
  {-0.42503724, 0.10537505, 0,1.0},
  {-0.42503724, 0.10537505, -0.01973337,1.0},
  {-0.4243139, 0.10410966, 0,1.0},
  // f 66/139/123 517/138/122 65/137/850
  {-0.42588824, 0.10658219, 0,1.0},
  {-0.42588824, 0.10658219, -0.01973337,1.0},
  {-0.42503724, 0.10537505, 0,1.0},
  // f 67/141/125 518/140/124 66/139/851
  {-0.42636168, 0.10716394, 0,1.0},
  {-0.42636168, 0.10716394, -0.01973337,1.0},
  {-0.42588824, 0.10658219, 0,1.0},
  // f 68/1886/127 519/142/126 67/141/852
  {-0.42704237, 0.10791686, 0,1.0},
  {-0.42704237, 0.10791686, -0.01973337,1.0},
  {-0.42636168, 0.10716394, 0,1.0},
  // f 69/146/129 520/145/128 68/143/853
  {-0.42780152, 0.10862726, 0,1.0},
  {-0.42780152, 0.10862726, -0.01973337,1.0},
  {-0.42704237, 0.10791686, 0,1.0},
  // f 70/148/131 521/147/130 69/146/854
  {-0.4286606, 0.10927848, 0,1.0},
  {-0.4286606, 0.10927848, -0.01973337,1.0},
  {-0.42780152, 0.10862726, 0,1.0},
  // f 71/150/133 522/149/132 70/148/855
  {-0.42961943, 0.10987046, 0,1.0},
  {-0.42961943, 0.10987046, -0.01973337,1.0},
  {-0.4286606, 0.10927848, 0,1.0},
  // f 72/152/135 523/151/134 71/150/133
  {-0.43095866, 0.11052722, 0,1.0},
  {-0.43095866, 0.11052722, -0.01973337,1.0},
  {-0.42961943, 0.10987046, 0,1.0},
  // f 73/154/137 524/153/136 72/152/135
  {-0.43245378, 0.11109147, 0,1.0},
  {-0.43245378, 0.11109147, -0.01973337,1.0},
  {-0.43095866, 0.11052722, 0,1.0},
  // f 74/156/139 525/155/138 73/154/137
  {-0.43445414, 0.11164644, 0,1.0},
  {-0.43445414, 0.11164644, -0.01973337,1.0},
  {-0.43245378, 0.11109147, 0,1.0},
  // f 75/158/141 526/157/140 74/156/139
  {-0.43707186, 0.11212561, 0,1.0},
  {-0.43707186, 0.11212561, -0.01973337,1.0},
  {-0.43445414, 0.11164644, 0,1.0},
  // f 76/160/143 527/159/142 75/158/141
  {-0.43999577, 0.11242347, 0,1.0},
  {-0.43999577, 0.11242347, -0.01973337,1.0},
  {-0.43707186, 0.11212561, 0,1.0},
  // f 77/162/144 528/161/144 76/160/143
  {-0.44371176, 0.11254187, 0,1.0},
  {-0.44371176, 0.11254187, -0.01973337,1.0},
  {-0.43999577, 0.11242347, 0,1.0},
  // f 78/1887/70 529/163/70 77/162/70
  {-0.45614135, 0.11254187, 0,1.0},
  {-0.45614135, 0.11254187, -0.01973337,1.0},
  {-0.44371176, 0.11254187, 0,1.0},
  // f 40/1888/71 491/166/71 78/164/71
  {-0.45614135, 0.081034, 0,1.0},
  {-0.45614135, 0.081034, -0.01973337,1.0},
  {-0.45614135, 0.11254187, 0,1.0},
  // f 80/1889/145 531/169/145 79/167/145
  {-0.32336509, 0.04933343, 0,1.0},
  {-0.32336509, 0.04933343, -0.01973337,1.0},
  {-0.36489382, 0.04673185, 0,1.0},
  // f 81/173/149 532/172/148 80/170/856
  {-0.32255679, 0.04475466, 0,1.0},
  {-0.32255679, 0.04475466, -0.01973337,1.0},
  {-0.32336509, 0.04933343, 0,1.0},
  // f 82/175/151 533/174/150 81/173/149
  {-0.32163715, 0.04136299, 0,1.0},
  {-0.32163715, 0.04136299, -0.01973337,1.0},
  {-0.32255679, 0.04475466, 0,1.0},
  // f 83/177/153 534/176/152 82/175/857
  {-0.3206183, 0.03864001, 0,1.0},
  {-0.3206183, 0.03864001, -0.01973337,1.0},
  {-0.32163715, 0.04136299, 0,1.0},
  // f 84/179/155 535/178/154 83/177/858
  {-0.31956071, 0.03648167, 0,1.0},
  {-0.31956071, 0.03648167, -0.01973337,1.0},
  {-0.3206183, 0.03864001, 0,1.0},
  // f 85/181/157 536/180/156 84/179/859
  {-0.31882712, 0.03525797, 0,1.0},
  {-0.31882712, 0.03525797, -0.01973337,1.0},
  {-0.31956071, 0.03648167, 0,1.0},
  // f 86/183/159 537/182/158 85/181/157
  {-0.31787291, 0.03391673, 0,1.0},
  {-0.31787291, 0.03391673, -0.01973337,1.0},
  {-0.31882712, 0.03525797, 0,1.0},
  // f 87/1890/161 538/184/160 86/183/860
  {-0.31646615, 0.03228737, 0,1.0},
  {-0.31646615, 0.03228737, -0.01973337,1.0},
  {-0.31787291, 0.03391673, 0,1.0},
  // f 88/188/163 539/187/162 87/185/161
  {-0.31494385, 0.03082954, 0,1.0},
  {-0.31494385, 0.03082954, -0.01973337,1.0},
  {-0.31646615, 0.03228737, 0,1.0},
  // f 89/190/165 540/189/164 88/188/861
  {-0.31330577, 0.02954321, 0,1.0},
  {-0.31330577, 0.02954321, -0.01973337,1.0},
  {-0.31494385, 0.03082954, 0,1.0},
  // f 90/192/167 541/191/166 89/190/165
  {-0.31155208, 0.02842839, 0,1.0},
  {-0.31155208, 0.02842839, -0.01973337,1.0},
  {-0.31330577, 0.02954321, 0,1.0},
  // f 91/194/169 542/193/168 90/192/862
  {-0.30968276, 0.02748508, 0,1.0},
  {-0.30968276, 0.02748508, -0.01973337,1.0},
  {-0.31155208, 0.02842839, 0,1.0},
  // f 92/196/171 543/195/170 91/194/169
  {-0.30769795, 0.02671328, 0,1.0},
  {-0.30769795, 0.02671328, -0.01973337,1.0},
  {-0.30968276, 0.02748508, 0,1.0},
  // f 93/198/173 544/197/172 92/196/863
  {-0.30559742, 0.02611299, 0,1.0},
  {-0.30559742, 0.02611299, -0.01973337,1.0},
  {-0.30769795, 0.02671328, 0,1.0},
  // f 94/200/175 545/199/174 93/198/173
  {-0.30292422, 0.02561904, 0,1.0},
  {-0.30292422, 0.02561904, -0.01973337,1.0},
  {-0.30559742, 0.02611299, 0,1.0},
  // f 95/202/177 546/201/176 94/200/175
  {-0.3010495, 0.02542694, 0,1.0},
  {-0.3010495, 0.02542694, -0.01973337,1.0},
  {-0.30292422, 0.02561904, 0,1.0},
  // f 96/204/179 547/203/178 95/202/177
  {-0.29860207, 0.02534119, 0,1.0},
  {-0.29860207, 0.02534119, -0.01973337,1.0},
  {-0.3010495, 0.02542694, 0,1.0},
  // f 97/206/181 548/205/180 96/204/179
  {-0.29537636, 0.02548324, 0,1.0},
  {-0.29537636, 0.02548324, -0.01973337,1.0},
  {-0.29860207, 0.02534119, 0,1.0},
  // f 98/208/183 549/207/182 97/206/181
  {-0.29274219, 0.025848, 0,1.0},
  {-0.29274219, 0.025848, -0.01973337,1.0},
  {-0.29537636, 0.02548324, 0,1.0},
  // f 99/210/185 550/209/184 98/208/183
  {-0.29061744, 0.02635129, 0,1.0},
  {-0.29061744, 0.02635129, -0.01973337,1.0},
  {-0.29274219, 0.025848, 0,1.0},
  // f 100/212/187 551/211/186 99/210/864
  {-0.28893006, 0.02691947, 0,1.0},
  {-0.28893006, 0.02691947, -0.01973337,1.0},
  {-0.29061744, 0.02635129, 0,1.0},
  // f 101/214/189 552/213/188 100/212/187
  {-0.28736609, 0.02761392, 0,1.0},
  {-0.28736609, 0.02761392, -0.01973337,1.0},
  {-0.28893006, 0.02691947, 0,1.0},
  // f 102/216/191 553/215/190 101/214/189
  {-0.2859256, 0.02843463, 0,1.0},
  {-0.2859256, 0.02843463, -0.01973337,1.0},
  {-0.28736609, 0.02761392, 0,1.0},
  // f 103/218/193 554/217/192 102/216/191
  {-0.28503397, 0.02905191, 0,1.0},
  {-0.28503397, 0.02905191, -0.01973337,1.0},
  {-0.2859256, 0.02843463, 0,1.0},
  // f 104/220/195 555/219/194 103/218/865
  {-0.28419709, 0.02972531, 0,1.0},
  {-0.28419709, 0.02972531, -0.01973337,1.0},
  {-0.28503397, 0.02905191, 0,1.0},
  // f 105/1891/197 556/221/196 104/220/866
  {-0.28287992, 0.03098026, 0,1.0},
  {-0.28287992, 0.03098026, -0.01973337,1.0},
  {-0.28419709, 0.02972531, 0,1.0},
  // f 106/225/199 557/224/198 105/222/197
  {-0.28190863, 0.0320995, 0,1.0},
  {-0.28190863, 0.0320995, -0.01973337,1.0},
  {-0.28287992, 0.03098026, 0,1.0},
  // f 107/227/201 558/226/200 106/225/867
  {-0.28108308, 0.033259, 0,1.0},
  {-0.28108308, 0.033259, -0.01973337,1.0},
  {-0.28190863, 0.0320995, 0,1.0},
  // f 108/229/203 559/228/202 107/227/868
  {-0.2805064, 0.03425597, 0,1.0},
  {-0.2805064, 0.03425597, -0.01973337,1.0},
  {-0.28108308, 0.033259, 0,1.0},
  // f 109/231/205 560/230/204 108/229/869
  {-0.28003085, 0.03528089, 0,1.0},
  {-0.28003085, 0.03528089, -0.01973337,1.0},
  {-0.2805064, 0.03425597, 0,1.0},
  // f 110/233/207 561/232/206 109/231/205
  {-0.27965656, 0.03633375, 0,1.0},
  {-0.27965656, 0.03633375, -0.01973337,1.0},
  {-0.28003085, 0.03528089, 0,1.0},
  // f 111/235/209 562/234/208 110/233/870
  {-0.27938333, 0.03741456, 0,1.0},
  {-0.27938333, 0.03741456, -0.01973337,1.0},
  {-0.27965656, 0.03633375, 0,1.0},
  // f 112/237/211 563/236/210 111/235/871
  {-0.27921134, 0.03852331, 0,1.0},
  {-0.27921134, 0.03852331, -0.01973337,1.0},
  {-0.27938333, 0.03741456, 0,1.0},
  // f 113/239/213 564/238/212 112/237/872
  {-0.27915674, 0.03920197, 0,1.0},
  {-0.27915674, 0.03920197, -0.01973337,1.0},
  {-0.27921134, 0.03852331, 0,1.0},
  // f 114/241/215 565/240/214 113/239/873
  {-0.27913854, 0.0398907, 0,1.0},
  {-0.27913854, 0.0398907, -0.01973337,1.0},
  {-0.27915674, 0.03920197, 0,1.0},
  // f 115/243/217 566/242/216 114/241/215
  {-0.27918667, 0.04097757, 0,1.0},
  {-0.27918667, 0.04097757, -0.01973337,1.0},
  {-0.27913854, 0.0398907, 0,1.0},
  // f 116/245/219 567/244/218 115/243/874
  {-0.27933127, 0.04204133, 0,1.0},
  {-0.27933127, 0.04204133, -0.01973337,1.0},
  {-0.27918667, 0.04097757, 0,1.0},
  // f 117/247/221 568/246/220 116/245/875
  {-0.2795721, 0.04308196, 0,1.0},
  {-0.2795721, 0.04308196, -0.01973337,1.0},
  {-0.27933127, 0.04204133, 0,1.0},
  // f 118/249/223 569/248/222 117/247/876
  {-0.27990934, 0.04409945, 0,1.0},
  {-0.27990934, 0.04409945, -0.01973337,1.0},
  {-0.2795721, 0.04308196, 0,1.0},
  // f 119/251/225 570/250/224 118/249/877
  {-0.280343, 0.04509383, 0,1.0},
  {-0.280343, 0.04509383, -0.01973337,1.0},
  {-0.27990934, 0.04409945, 0,1.0},
  // f 120/253/227 571/252/226 119/251/225
  {-0.28099045, 0.04625655, 0,1.0},
  {-0.28099045, 0.04625655, -0.01973337,1.0},
  {-0.280343, 0.04509383, 0,1.0},
  // f 121/255/229 572/254/228 120/253/878
  {-0.2817767, 0.04738599, 0,1.0},
  {-0.2817767, 0.04738599, -0.01973337,1.0},
  {-0.28099045, 0.04625655, 0,1.0},
  // f 122/1892/231 573/256/230 121/255/879
  {-0.2827017, 0.04848211, 0,1.0},
  {-0.2827017, 0.04848211, -0.01973337,1.0},
  {-0.2817767, 0.04738599, 0,1.0},
  // f 123/260/233 574/259/232 122/257/880
  {-0.28395623, 0.04971883, 0,1.0},
  {-0.28395623, 0.04971883, -0.01973337,1.0},
  {-0.2827017, 0.04848211, 0,1.0},
  // f 124/262/235 575/261/234 123/260/881
  {-0.28480846, 0.0504095, 0,1.0},
  {-0.28480846, 0.0504095, -0.01973337,1.0},
  {-0.28395623, 0.04971883, 0,1.0},
  // f 125/264/237 576/263/236 124/262/882
  {-0.28610268, 0.05126418, 0,1.0},
  {-0.28610268, 0.05126418, -0.01973337,1.0},
  {-0.28480846, 0.0504095, 0,1.0},
  // f 126/266/239 577/265/238 125/264/237
  {-0.28834203, 0.0524445, 0,1.0},
  {-0.28834203, 0.0524445, -0.01973337,1.0},
  {-0.28610268, 0.05126418, 0,1.0},
  // f 127/268/241 578/267/240 126/266/239
  {-0.29195365, 0.05393434, 0,1.0},
  {-0.29195365, 0.05393434, -0.01973337,1.0},
  {-0.28834203, 0.0524445, 0,1.0},
  // f 128/270/243 579/269/242 127/268/241
  {-0.2980513, 0.05587222, 0,1.0},
  {-0.2980513, 0.05587222, -0.01973337,1.0},
  {-0.29195365, 0.05393434, 0,1.0},
  // f 129/272/245 580/271/244 128/270/243
  {-0.30631042, 0.05790896, 0,1.0},
  {-0.30631042, 0.05790896, -0.01973337,1.0},
  {-0.2980513, 0.05587222, 0,1.0},
  // f 130/274/247 581/273/246 129/272/883
  {-0.3220557, 0.06216396, 0,1.0},
  {-0.3220557, 0.06216396, -0.01973337,1.0},
  {-0.30631042, 0.05790896, 0,1.0},
  // f 131/276/249 582/275/248 130/274/884
  {-0.33171731, 0.06576931, 0,1.0},
  {-0.33171731, 0.06576931, -0.01973337,1.0},
  {-0.3220557, 0.06216396, 0,1.0},
  // f 132/278/251 583/277/250 131/276/249
  {-0.33843592, 0.06902439, 0,1.0},
  {-0.33843592, 0.06902439, -0.01973337,1.0},
  {-0.33171731, 0.06576931, 0,1.0},
  // f 133/280/253 584/279/252 132/278/251
  {-0.34292337, 0.07174607, 0,1.0},
  {-0.34292337, 0.07174607, -0.01973337,1.0},
  {-0.33843592, 0.06902439, 0,1.0},
  // f 134/282/255 585/281/254 133/280/253
  {-0.34519771, 0.07337592, 0,1.0},
  {-0.34519771, 0.07337592, -0.01973337,1.0},
  {-0.34292337, 0.07174607, 0,1.0},
  // f 135/284/257 586/283/256 134/282/885
  {-0.34726104, 0.07506003, 0,1.0},
  {-0.34726104, 0.07506003, -0.01973337,1.0},
  {-0.34519771, 0.07337592, 0,1.0},
  // f 136/1893/259 587/285/258 135/284/257
  {-0.35004342, 0.07771154, 0,1.0},
  {-0.35004342, 0.07771154, -0.01973337,1.0},
  {-0.34726104, 0.07506003, 0,1.0},
  // f 137/289/261 588/288/260 136/286/259
  {-0.35247061, 0.08053235, 0,1.0},
  {-0.35247061, 0.08053235, -0.01973337,1.0},
  {-0.35004342, 0.07771154, 0,1.0},
  // f 138/291/263 589/290/262 137/289/886
  {-0.35422194, 0.08301231, 0,1.0},
  {-0.35422194, 0.08301231, -0.01973337,1.0},
  {-0.35247061, 0.08053235, 0,1.0},
  // f 139/293/265 590/292/264 138/291/887
  {-0.35572663, 0.08560983, 0,1.0},
  {-0.35572663, 0.08560983, -0.01973337,1.0},
  {-0.35422194, 0.08301231, 0,1.0},
  // f 140/295/267 591/294/266 139/293/888
  {-0.35698465, 0.0883249, 0,1.0},
  {-0.35698465, 0.0883249, -0.01973337,1.0},
  {-0.35572663, 0.08560983, 0,1.0},
  // f 141/297/269 592/296/268 140/295/889
  {-0.35799599, 0.09115754, 0,1.0},
  {-0.35799599, 0.09115754, -0.01973337,1.0},
  {-0.35698465, 0.0883249, 0,1.0},
  // f 142/299/271 593/298/270 141/297/890
  {-0.35876068, 0.09410771, 0,1.0},
  {-0.35876068, 0.09410771, -0.01973337,1.0},
  {-0.35799599, 0.09115754, 0,1.0},
  // f 143/301/273 594/300/272 142/299/891
  {-0.35935268, 0.0978031, 0,1.0},
  {-0.35935268, 0.0978031, -0.01973337,1.0},
  {-0.35876068, 0.09410771, 0,1.0},
  // f 144/303/275 595/302/274 143/301/892
  {-0.35951549, 0.09971424, 0,1.0},
  {-0.35951549, 0.09971424, -0.01973337,1.0},
  {-0.35935268, 0.0978031, 0,1.0},
  // f 145/305/277 596/304/276 144/303/275
  {-0.35959435, 0.10232831, 0,1.0},
  {-0.35959435, 0.10232831, -0.01973337,1.0},
  {-0.35951549, 0.09971424, 0,1.0},
  // f 146/307/279 597/306/278 145/305/893
  {-0.35943285, 0.10578167, 0,1.0},
  {-0.35943285, 0.10578167, -0.01973337,1.0},
  {-0.35959435, 0.10232831, 0,1.0},
  // f 147/309/281 598/308/280 146/307/279
  {-0.35894817, 0.10917331, 0,1.0},
  {-0.35894817, 0.10917331, -0.01973337,1.0},
  {-0.35943285, 0.10578167, 0,1.0},
  // f 148/311/283 599/310/282 147/309/281
  {-0.35814032, 0.11250332, 0,1.0},
  {-0.35814032, 0.11250332, -0.01973337,1.0},
  {-0.35894817, 0.10917331, 0,1.0},
  // f 149/313/285 600/312/284 148/311/894
  {-0.35700944, 0.11577167, 0,1.0},
  {-0.35700944, 0.11577167, -0.01973337,1.0},
  {-0.35814032, 0.11250332, 0,1.0},
  // f 150/315/287 601/314/286 149/313/895
  {-0.35555524, 0.11897835, 0,1.0},
  {-0.35555524, 0.11897835, -0.01973337,1.0},
  {-0.35700944, 0.11577167, 0,1.0},
  // f 151/317/289 602/316/288 150/315/896
  {-0.35448223, 0.12095121, 0,1.0},
  {-0.35448223, 0.12095121, -0.01973337,1.0},
  {-0.35555524, 0.11897835, 0,1.0},
  // f 152/319/291 603/318/290 151/317/897
  {-0.35328317, 0.12289994, 0,1.0},
  {-0.35328317, 0.12289994, -0.01973337,1.0},
  {-0.35448223, 0.12095121, 0,1.0},
  // f 153/321/293 604/320/292 152/319/898
  {-0.3513914, 0.12553017, 0,1.0},
  {-0.3513914, 0.12553017, -0.01973337,1.0},
  {-0.35328317, 0.12289994, 0,1.0},
  // f 154/1894/295 605/322/294 153/321/899
  {-0.34925041, 0.12799795, 0,1.0},
  {-0.34925041, 0.12799795, -0.01973337,1.0},
  {-0.3513914, 0.12553017, 0,1.0},
  // f 155/326/297 606/325/296 154/323/900
  {-0.3464981, 0.13061942, 0,1.0},
  {-0.3464981, 0.13061942, -0.01973337,1.0},
  {-0.34925041, 0.12799795, 0,1.0},
  // f 156/328/299 607/327/298 155/326/297
  {-0.34342036, 0.13302872, 0,1.0},
  {-0.34342036, 0.13302872, -0.01973337,1.0},
  {-0.3464981, 0.13061942, 0,1.0},
  // f 157/330/301 608/329/300 156/328/901
  {-0.34001693, 0.13522591, 0,1.0},
  {-0.34001693, 0.13522591, -0.01973337,1.0},
  {-0.34342036, 0.13302872, 0,1.0},
  // f 158/332/303 609/331/302 157/330/902
  {-0.3372508, 0.13673456, 0,1.0},
  {-0.3372508, 0.13673456, -0.01973337,1.0},
  {-0.34001693, 0.13522591, 0,1.0},
  // f 159/334/305 610/333/304 158/332/903
  {-0.33430135, 0.13812394, 0,1.0},
  {-0.33430135, 0.13812394, -0.01973337,1.0},
  {-0.3372508, 0.13673456, 0,1.0},
  // f 160/336/307 611/335/306 159/334/305
  {-0.33000627, 0.13975506, 0,1.0},
  {-0.33000627, 0.13975506, -0.01973337,1.0},
  {-0.33430135, 0.13812394, 0,1.0},
  // f 161/338/309 612/337/308 160/336/307
  {-0.32459944, 0.14125092, 0,1.0},
  {-0.32459944, 0.14125092, -0.01973337,1.0},
  {-0.33000627, 0.13975506, 0,1.0},
  // f 162/340/311 613/339/310 161/338/309
  {-0.31716415, 0.1425917, 0,1.0},
  {-0.31716415, 0.1425917, -0.01973337,1.0},
  {-0.32459944, 0.14125092, 0,1.0},
  // f 163/342/313 614/341/312 162/340/311
  {-0.30881959, 0.14339617, 0,1.0},
  {-0.30881959, 0.14339617, -0.01973337,1.0},
  {-0.31716415, 0.1425917, 0,1.0},
  // f 164/344/315 615/343/314 163/342/313
  {-0.29956564, 0.1436643, 0,1.0},
  {-0.29956564, 0.1436643, -0.01973337,1.0},
  {-0.30881959, 0.14339617, 0,1.0},
  // f 165/346/317 616/345/316 164/344/315
  {-0.28827497, 0.14317694, 0,1.0},
  {-0.28827497, 0.14317694, -0.01973337,1.0},
  {-0.29956564, 0.1436643, 0,1.0},
  // f 166/348/319 617/347/318 165/346/317
  {-0.27996331, 0.14205326, 0,1.0},
  {-0.27996331, 0.14205326, -0.01973337,1.0},
  {-0.28827497, 0.14317694, 0,1.0},
  // f 167/350/321 618/349/320 166/348/319
  {-0.27327496, 0.14050667, 0,1.0},
  {-0.27327496, 0.14050667, -0.01973337,1.0},
  {-0.27996331, 0.14205326, 0,1.0},
  // f 168/352/323 619/351/322 167/350/321
  {-0.26796278, 0.13873047, 0,1.0},
  {-0.26796278, 0.13873047, -0.01973337,1.0},
  {-0.27327496, 0.14050667, 0,1.0},
  // f 169/354/325 620/353/324 168/352/323
  {-0.26381099, 0.13689388, 0,1.0},
  {-0.26381099, 0.13689388, -0.01973337,1.0},
  {-0.26796278, 0.13873047, 0,1.0},
  // f 170/356/327 621/355/326 169/354/325
  {-0.26124883, 0.1355084, 0,1.0},
  {-0.26124883, 0.1355084, -0.01973337,1.0},
  {-0.26381099, 0.13689388, 0,1.0},
  // f 171/358/329 622/357/328 170/356/904
  {-0.2582778, 0.13359529, 0,1.0},
  {-0.2582778, 0.13359529, -0.01973337,1.0},
  {-0.26124883, 0.1355084, 0,1.0},
  // f 172/360/331 623/359/330 171/358/329
  {-0.2555505, 0.1314625, 0,1.0},
  {-0.2555505, 0.1314625, -0.01973337,1.0},
  {-0.2582778, 0.13359529, 0,1.0},
  // f 173/1895/333 624/361/332 172/360/331
  {-0.25305346, 0.12909168, 0,1.0},
  {-0.25305346, 0.12909168, -0.01973337,1.0},
  {-0.2555505, 0.1314625, 0,1.0},
  // f 174/365/335 625/364/334 173/362/333
  {-0.25078672, 0.12648289, 0,1.0},
  {-0.25078672, 0.12648289, -0.01973337,1.0},
  {-0.25305346, 0.12909168, 0,1.0},
  // f 175/367/337 626/366/336 174/365/905
  {-0.24875031, 0.1236361, 0,1.0},
  {-0.24875031, 0.1236361, -0.01973337,1.0},
  {-0.25078672, 0.12648289, 0,1.0},
  // f 176/369/339 627/368/338 175/367/906
  {-0.24694414, 0.12055133, 0,1.0},
  {-0.24694414, 0.12055133, -0.01973337,1.0},
  {-0.24875031, 0.1236361, 0,1.0},
  // f 177/371/341 628/370/340 176/369/907
  {-0.24508065, 0.11653544, 0,1.0},
  {-0.24508065, 0.11653544, -0.01973337,1.0},
  {-0.24694414, 0.12055133, 0,1.0},
  // f 178/373/343 629/372/342 177/371/341
  {-0.2435489, 0.11217681, 0,1.0},
  {-0.2435489, 0.11217681, -0.01973337,1.0},
  {-0.24508065, 0.11653544, 0,1.0},
  // f 179/375/345 630/374/344 178/373/343
  {-0.24234867, 0.1074755, 0,1.0},
  {-0.24234867, 0.1074755, -0.01973337,1.0},
  {-0.2435489, 0.11217681, 0,1.0},
  // f 180/1896/346 631/376/346 179/375/345
  {-0.24136762, 0.10155746, 0,1.0},
  {-0.24136762, 0.10155746, -0.01973337,1.0},
  {-0.24234867, 0.1074755, 0,1.0},
  // f 181/1897/347 632/379/347 180/377/347
  {-0.28251091, 0.09914861, 0,1.0},
  {-0.28251091, 0.09914861, -0.01973337,1.0},
  {-0.24136762, 0.10155746, 0,1.0},
  // f 182/383/351 633/382/350 181/380/908
  {-0.28321767, 0.10241412, 0,1.0},
  {-0.28321767, 0.10241412, -0.01973337,1.0},
  {-0.28251091, 0.09914861, 0,1.0},
  // f 183/385/353 634/384/352 182/383/351
  {-0.28404248, 0.1050337, 0,1.0},
  {-0.28404248, 0.1050337, -0.01973337,1.0},
  {-0.28321767, 0.10241412, 0,1.0},
  // f 184/387/355 635/386/354 183/385/909
  {-0.28491592, 0.10710733, 0,1.0},
  {-0.28491592, 0.10710733, -0.01973337,1.0},
  {-0.28404248, 0.1050337, 0,1.0},
  // f 185/389/357 636/388/356 184/387/355
  {-0.28577727, 0.10872237, 0,1.0},
  {-0.28577727, 0.10872237, -0.01973337,1.0},
  {-0.28491592, 0.10710733, 0,1.0},
  // f 186/391/359 637/390/358 185/389/910
  {-0.28674284, 0.11018759, 0,1.0},
  {-0.28674284, 0.11018759, -0.01973337,1.0},
  {-0.28577727, 0.10872237, 0,1.0},
  // f 187/393/361 638/392/360 186/391/359
  {-0.28781238, 0.11150294, 0,1.0},
  {-0.28781238, 0.11150294, -0.01973337,1.0},
  {-0.28674284, 0.11018759, 0,1.0},
  // f 188/1898/363 639/394/362 187/393/911
  {-0.2885831, 0.11229659, 0,1.0},
  {-0.2885831, 0.11229659, -0.01973337,1.0},
  {-0.28781238, 0.11150294, 0,1.0},
  // f 189/398/365 640/397/364 188/395/912
  {-0.28940025, 0.11302363, 0,1.0},
  {-0.28940025, 0.11302363, -0.01973337,1.0},
  {-0.2885831, 0.11229659, 0,1.0},
  // f 190/400/367 641/399/366 189/398/913
  {-0.29071814, 0.11400183, 0,1.0},
  {-0.29071814, 0.11400183, -0.01973337,1.0},
  {-0.28940025, 0.11302363, 0,1.0},
  // f 191/402/369 642/401/368 190/400/914
  {-0.29215118, 0.11485515, 0,1.0},
  {-0.29215118, 0.11485515, -0.01973337,1.0},
  {-0.29071814, 0.11400183, 0,1.0},
  // f 192/404/371 643/403/370 191/402/369
  {-0.29369941, 0.11558357, 0,1.0},
  {-0.29369941, 0.11558357, -0.01973337,1.0},
  {-0.29215118, 0.11485515, 0,1.0},
  // f 193/406/373 644/405/372 192/404/371
  {-0.2953628, 0.11618715, 0,1.0},
  {-0.2953628, 0.11618715, -0.01973337,1.0},
  {-0.29369941, 0.11558357, 0,1.0},
  // f 194/408/375 645/407/374 193/406/915
  {-0.29744899, 0.11673349, 0,1.0},
  {-0.29744899, 0.11673349, -0.01973337,1.0},
  {-0.2953628, 0.11618715, 0,1.0},
  // f 195/410/377 646/409/376 194/408/375
  {-0.30002517, 0.11714971, 0,1.0},
  {-0.30002517, 0.11714971, -0.01973337,1.0},
  {-0.29744899, 0.11673349, 0,1.0},
  // f 196/412/379 647/411/378 195/410/377
  {-0.30173922, 0.11729716, 0,1.0},
  {-0.30173922, 0.11729716, -0.01973337,1.0},
  {-0.30002517, 0.11714971, 0,1.0},
  // f 197/414/381 648/413/380 196/412/379
  {-0.30390158, 0.11735957, 0,1.0},
  {-0.30390158, 0.11735957, -0.01973337,1.0},
  {-0.30173922, 0.11729716, 0,1.0},
  // f 198/416/383 649/415/382 197/414/381
  {-0.30679607, 0.11723047, 0,1.0},
  {-0.30679607, 0.11723047, -0.01973337,1.0},
  {-0.30390158, 0.11735957, 0,1.0},
  // f 199/418/385 650/417/384 198/416/383
  {-0.30889514, 0.11694124, 0,1.0},
  {-0.30889514, 0.11694124, -0.01973337,1.0},
  {-0.30679607, 0.11723047, 0,1.0},
  // f 200/420/387 651/419/386 199/418/385
  {-0.31057414, 0.11655263, 0,1.0},
  {-0.31057414, 0.11655263, -0.01973337,1.0},
  {-0.30889514, 0.11694124, 0,1.0},
  // f 201/422/389 652/421/388 200/420/387
  {-0.31189597, 0.11611877, 0,1.0},
  {-0.31189597, 0.11611877, -0.01973337,1.0},
  {-0.31057414, 0.11655263, 0,1.0},
  // f 202/424/391 653/423/390 201/422/389
  {-0.31310949, 0.11559201, 0,1.0},
  {-0.31310949, 0.11559201, -0.01973337,1.0},
  {-0.31189597, 0.11611877, 0,1.0},
  // f 203/426/393 654/425/392 202/424/391
  {-0.31403819, 0.115082, 0,1.0},
  {-0.31403819, 0.115082, -0.01973337,1.0},
  {-0.31310949, 0.11559201, 0,1.0},
  // f 204/428/395 655/427/394 203/426/916
  {-0.31472713, 0.1146275, 0,1.0},
  {-0.31472713, 0.1146275, -0.01973337,1.0},
  {-0.31403819, 0.115082, 0,1.0},
  // f 205/430/397 656/429/396 204/428/917
  {-0.31536773, 0.1141317, 0,1.0},
  {-0.31536773, 0.1141317, -0.01973337,1.0},
  {-0.31472713, 0.1146275, 0,1.0},
  // f 206/1899/399 657/431/398 205/430/397
  {-0.31623718, 0.1133369, 0,1.0},
  {-0.31623718, 0.1133369, -0.01973337,1.0},
  {-0.31536773, 0.1141317, 0,1.0},
  // f 207/435/401 658/434/400 206/432/918
  {-0.31699577, 0.11250187, 0,1.0},
  {-0.31699577, 0.11250187, -0.01973337,1.0},
  {-0.31623718, 0.1133369, 0,1.0},
  // f 208/437/403 659/436/402 207/435/401
  {-0.31764328, 0.1116266, 0,1.0},
  {-0.31764328, 0.1116266, -0.01973337,1.0},
  {-0.31699577, 0.11250187, 0,1.0},
  // f 209/439/405 660/438/404 208/437/919
  {-0.31809804, 0.11086644, 0,1.0},
  {-0.31809804, 0.11086644, -0.01973337,1.0},
  {-0.31764328, 0.1116266, 0,1.0},
  // f 210/441/407 661/440/406 209/439/920
  {-0.31847572, 0.11007833, 0,1.0},
  {-0.31847572, 0.11007833, -0.01973337,1.0},
  {-0.31809804, 0.11086644, 0,1.0},
  // f 211/443/409 662/442/408 210/441/921
  {-0.31877634, 0.10926232, 0,1.0},
  {-0.31877634, 0.10926232, -0.01973337,1.0},
  {-0.31847572, 0.11007833, 0,1.0},
  // f 212/445/411 663/444/410 211/443/409
  {-0.31899998, 0.10841836, 0,1.0},
  {-0.31899998, 0.10841836, -0.01973337,1.0},
  {-0.31877634, 0.10926232, 0,1.0},
  // f 213/447/413 664/446/412 212/445/411
  {-0.31914634, 0.10754646, 0,1.0},
  {-0.31914634, 0.10754646, -0.01973337,1.0},
  {-0.31899998, 0.10841836, 0,1.0},
  // f 214/449/415 665/448/414 213/447/922
  {-0.31919718, 0.1070099, 0,1.0},
  {-0.31919718, 0.1070099, -0.01973337,1.0},
  {-0.31914634, 0.10754646, 0,1.0},
  // f 215/451/417 666/450/416 214/449/923
  {-0.31922197, 0.10627884, 0,1.0},
  {-0.31922197, 0.10627884, -0.01973337,1.0},
  {-0.31919718, 0.1070099, 0,1.0},
  // f 216/453/419 667/452/418 215/451/417
  {-0.31919011, 0.10561109, 0,1.0},
  {-0.31919011, 0.10561109, -0.01973337,1.0},
  {-0.31922197, 0.10627884, 0,1.0},
  // f 217/455/421 668/454/420 216/453/924
  {-0.31909472, 0.10495687, 0,1.0},
  {-0.31909472, 0.10495687, -0.01973337,1.0},
  {-0.31919011, 0.10561109, 0,1.0},
  // f 218/457/423 669/456/422 217/455/925
  {-0.31893578, 0.1043161, 0,1.0},
  {-0.31893578, 0.1043161, -0.01973337,1.0},
  {-0.31909472, 0.10495687, 0,1.0},
  // f 219/459/425 670/458/424 218/457/423
  {-0.31871322, 0.10368883, 0,1.0},
  {-0.31871322, 0.10368883, -0.01973337,1.0},
  {-0.31893578, 0.1043161, 0,1.0},
  // f 220/461/427 671/460/426 219/459/926
  {-0.31842694, 0.10307506, 0,1.0},
  {-0.31842694, 0.10307506, -0.01973337,1.0},
  {-0.31871322, 0.10368883, 0,1.0},
  // f 221/463/336 672/462/428 220/461/927
  {-0.31807724, 0.10247475, 0,1.0},
  {-0.31807724, 0.10247475, -0.01973337,1.0},
  {-0.31842694, 0.10307506, 0,1.0},
  // f 222/465/430 673/464/429 221/463/336
  {-0.31757355, 0.10177226, 0,1.0},
  {-0.31757355, 0.10177226, -0.01973337,1.0},
  {-0.31807724, 0.10247475, 0,1.0},
  // f 223/1900/432 674/466/431 222/465/928
  {-0.31697828, 0.1010891, 0,1.0},
  {-0.31697828, 0.1010891, -0.01973337,1.0},
  {-0.31757355, 0.10177226, 0,1.0},
  // f 224/470/434 675/469/433 223/467/929
  {-0.31653062, 0.10064447, 0,1.0},
  {-0.31653062, 0.10064447, -0.01973337,1.0},
  {-0.31697828, 0.1010891, 0,1.0},
  // f 225/472/436 676/471/435 224/470/930
  {-0.31604224, 0.10020851, 0,1.0},
  {-0.31604224, 0.10020851, -0.01973337,1.0},
  {-0.31653062, 0.10064447, 0,1.0},
  // f 226/474/438 677/473/437 225/472/436
  {-0.31549466, 0.09976391, 0,1.0},
  {-0.31549466, 0.09976391, -0.01973337,1.0},
  {-0.31604224, 0.10020851, 0,1.0},
  // f 227/476/440 678/475/439 226/474/438
  {-0.31465751, 0.09921506, 0,1.0},
  {-0.31465751, 0.09921506, -0.01973337,1.0},
  {-0.31549466, 0.09976391, 0,1.0},
  // f 228/478/442 679/477/441 227/476/931
  {-0.31342909, 0.09856663, 0,1.0},
  {-0.31342909, 0.09856663, -0.01973337,1.0},
  {-0.31465751, 0.09921506, 0,1.0},
  // f 229/480/444 680/479/443 228/478/442
  {-0.31112865, 0.09761478, 0,1.0},
  {-0.31112865, 0.09761478, -0.01973337,1.0},
  {-0.31342909, 0.09856663, 0,1.0},
  // f 230/482/446 681/481/445 229/480/444
  {-0.30683568, 0.09628404, 0,1.0},
  {-0.30683568, 0.09628404, -0.01973337,1.0},
  {-0.31112865, 0.09761478, 0,1.0},
  // f 231/484/448 682/483/447 230/482/932
  {-0.30139637, 0.0950054, 0,1.0},
  {-0.30139637, 0.0950054, -0.01973337,1.0},
  {-0.30683568, 0.09628404, 0,1.0},
  // f 232/486/450 683/485/449 231/484/448
  {-0.27602455, 0.08855294, 0,1.0},
  {-0.27602455, 0.08855294, -0.01973337,1.0},
  {-0.30139637, 0.0950054, 0,1.0},
  // f 233/488/452 684/487/451 232/486/933
  {-0.26727593, 0.08554854, 0,1.0},
  {-0.26727593, 0.08554854, -0.01973337,1.0},
  {-0.27602455, 0.08855294, 0,1.0},
  // f 234/490/454 685/489/453 233/488/452
  {-0.26039758, 0.08252749, 0,1.0},
  {-0.26039758, 0.08252749, -0.01973337,1.0},
  {-0.26727593, 0.08554854, 0,1.0},
  // f 235/492/456 686/491/455 234/490/454
  {-0.25571474, 0.07988161, 0,1.0},
  {-0.25571474, 0.07988161, -0.01973337,1.0},
  {-0.26039758, 0.08252749, 0,1.0},
  // f 236/494/458 687/493/457 235/492/934
  {-0.25196159, 0.07729269, 0,1.0},
  {-0.25196159, 0.07729269, -0.01973337,1.0},
  {-0.25571474, 0.07988161, 0,1.0},
  // f 237/496/460 688/495/459 236/494/458
  {-0.24899231, 0.07482632, 0,1.0},
  {-0.24899231, 0.07482632, -0.01973337,1.0},
  {-0.25196159, 0.07729269, 0,1.0},
  // f 238/1901/462 689/497/461 237/496/935
  {-0.24636841, 0.07220456, 0,1.0},
  {-0.24636841, 0.07220456, -0.01973337,1.0},
  {-0.24899231, 0.07482632, 0,1.0},
  // f 239/501/464 690/500/463 238/498/936
  {-0.24435572, 0.06978304, 0,1.0},
  {-0.24435572, 0.06978304, -0.01973337,1.0},
  {-0.24636841, 0.07220456, 0,1.0},
  // f 240/503/466 691/502/465 239/501/464
  {-0.24332435, 0.0683459, 0,1.0},
  {-0.24332435, 0.0683459, -0.01973337,1.0},
  {-0.24435572, 0.06978304, 0,1.0},
  // f 241/505/468 692/504/467 240/503/937
  {-0.24237932, 0.06686991, 0,1.0},
  {-0.24237932, 0.06686991, -0.01973337,1.0},
  {-0.24332435, 0.0683459, 0,1.0},
  // f 242/507/470 693/506/469 241/505/938
  {-0.24071984, 0.06382233, 0,1.0},
  {-0.24071984, 0.06382233, -0.01973337,1.0},
  {-0.24237932, 0.06686991, 0,1.0},
  // f 243/509/472 694/508/471 242/507/470
  {-0.23934902, 0.0606613, 0,1.0},
  {-0.23934902, 0.0606613, -0.01973337,1.0},
  {-0.24071984, 0.06382233, 0,1.0},
  // f 244/511/474 695/510/473 243/509/939
  {-0.23826675, 0.05738679, 0,1.0},
  {-0.23826675, 0.05738679, -0.01973337,1.0},
  {-0.23934902, 0.0606613, 0,1.0},
  // f 245/513/476 696/512/475 244/511/940
  {-0.23747315, 0.05399883, 0,1.0},
  {-0.23747315, 0.05399883, -0.01973337,1.0},
  {-0.23826675, 0.05738679, 0,1.0},
  // f 246/515/478 697/514/477 245/513/941
  {-0.23692529, 0.05005173, 0,1.0},
  {-0.23692529, 0.05005173, -0.01973337,1.0},
  {-0.23747315, 0.05399883, 0,1.0},
  // f 247/517/480 698/516/479 246/515/942
  {-0.23674259, 0.04596103, 0,1.0},
  {-0.23674259, 0.04596103, -0.01973337,1.0},
  {-0.23692529, 0.05005173, 0,1.0},
  // f 248/519/482 699/518/481 247/517/480
  {-0.23693499, 0.04166777, 0,1.0},
  {-0.23693499, 0.04166777, -0.01973337,1.0},
  {-0.23674259, 0.04596103, 0,1.0},
  // f 249/521/484 700/520/483 248/519/482
  {-0.2375122, 0.03748308, 0,1.0},
  {-0.2375122, 0.03748308, -0.01973337,1.0},
  {-0.23693499, 0.04166777, 0,1.0},
  // f 250/523/486 701/522/485 249/521/943
  {-0.23847421, 0.0334069, 0,1.0},
  {-0.23847421, 0.0334069, -0.01973337,1.0},
  {-0.2375122, 0.03748308, 0,1.0},
  // f 251/525/488 702/524/487 250/523/944
  {-0.23982103, 0.02943926, 0,1.0},
  {-0.23982103, 0.02943926, -0.01973337,1.0},
  {-0.23847421, 0.0334069, 0,1.0},
  // f 252/527/490 703/526/489 251/525/945
  {-0.24155261, 0.02558015, 0,1.0},
  {-0.24155261, 0.02558015, -0.01973337,1.0},
  {-0.23982103, 0.02943926, 0,1.0},
  // f 253/529/492 704/528/491 252/527/490
  {-0.24283025, 0.02322332, 0,1.0},
  {-0.24283025, 0.02322332, -0.01973337,1.0},
  {-0.24155261, 0.02558015, 0,1.0},
  // f 254/531/494 705/530/493 253/529/492
  {-0.24425827, 0.02090889, 0,1.0},
  {-0.24425827, 0.02090889, -0.01973337,1.0},
  {-0.24283025, 0.02322332, 0,1.0},
  // f 255/533/496 706/532/495 254/531/946
  {-0.24681623, 0.01738155, 0,1.0},
  {-0.24681623, 0.01738155, -0.01973337,1.0},
  {-0.24425827, 0.02090889, 0,1.0},
  // f 256/1902/498 707/534/497 255/533/947
  {-0.24967997, 0.01413788, 0,1.0},
  {-0.24967997, 0.01413788, -0.01973337,1.0},
  {-0.24681623, 0.01738155, 0,1.0},
  // f 257/538/500 708/537/499 256/535/948
  {-0.25284964, 0.01117787, 0,1.0},
  {-0.25284964, 0.01117787, -0.01973337,1.0},
  {-0.24967997, 0.01413788, 0,1.0},
  // f 258/540/502 709/539/501 257/538/500
  {-0.25632519, 0.00850154, 0,1.0},
  {-0.25632519, 0.00850154, -0.01973337,1.0},
  {-0.25284964, 0.01117787, 0,1.0},
  // f 259/542/504 710/541/503 258/540/502
  {-0.26010665, 0.00610887, 0,1.0},
  {-0.26010665, 0.00610887, -0.01973337,1.0},
  {-0.25632519, 0.00850154, 0,1.0},
  // f 260/544/506 711/543/505 259/542/949
  {-0.26262525, 0.0047575, 0,1.0},
  {-0.26262525, 0.0047575, -0.01973337,1.0},
  {-0.26010665, 0.00610887, 0,1.0},
  // f 261/546/508 712/545/507 260/544/506
  {-0.2652635, 0.00351693, 0,1.0},
  {-0.2652635, 0.00351693, -0.01973337,1.0},
  {-0.26262525, 0.0047575, 0,1.0},
  // f 262/548/510 713/547/509 261/546/508
  {-0.27034768, 0.00157564, 0,1.0},
  {-0.27034768, 0.00157564, -0.01973337,1.0},
  {-0.2652635, 0.00351693, 0,1.0},
  // f 263/550/512 714/549/511 262/548/510
  {-0.27653128, -0.00013098, 0,1.0},
  {-0.27653128, -0.00013098, -0.01973337,1.0},
  {-0.27034768, 0.00157564, 0,1.0},
  // f 264/552/514 715/551/513 263/550/512
  {-0.2839832, -0.00146073, 0,1.0},
  {-0.2839832, -0.00146073, -0.01973337,1.0},
  {-0.27653128, -0.00013098, 0,1.0},
  // f 265/554/516 716/553/515 264/552/514
  {-0.29134852, -0.00217183, 0,1.0},
  {-0.29134852, -0.00217183, -0.01973337,1.0},
  {-0.2839832, -0.00146073, 0,1.0},
  // f 266/556/518 717/555/517 265/554/516
  {-0.29927659, -0.00240886, 0,1.0},
  {-0.29927659, -0.00240886, -0.01973337,1.0},
  {-0.29134852, -0.00217183, 0,1.0},
  // f 267/558/520 718/557/519 266/556/518
  {-0.31280473, -0.00185386, 0,1.0},
  {-0.31280473, -0.00185386, -0.01973337,1.0},
  {-0.29927659, -0.00240886, 0,1.0},
  // f 268/560/522 719/559/521 267/558/520
  {-0.32235077, -0.00061066, 0,1.0},
  {-0.32235077, -0.00061066, -0.01973337,1.0},
  {-0.31280473, -0.00185386, 0,1.0},
  // f 269/562/524 720/561/523 268/560/522
  {-0.32977268, 0.0010599, 0,1.0},
  {-0.32977268, 0.0010599, -0.01973337,1.0},
  {-0.32235077, -0.00061066, 0,1.0},
  // f 270/564/526 721/563/525 269/562/524
  {-0.33544296, 0.0029247, 0,1.0},
  {-0.33544296, 0.0029247, -0.01973337,1.0},
  {-0.32977268, 0.0010599, 0,1.0},
  // f 271/566/528 722/565/527 270/564/526
  {-0.33968067, 0.00478395, 0,1.0},
  {-0.33968067, 0.00478395, -0.01973337,1.0},
  {-0.33544296, 0.0029247, 0,1.0},
  // f 272/568/530 723/567/529 271/566/528
  {-0.34347504, 0.00692071, 0,1.0},
  {-0.34347504, 0.00692071, -0.01973337,1.0},
  {-0.33968067, 0.00478395, 0,1.0},
  // f 273/570/532 724/569/531 272/568/950
  {-0.34619153, 0.00882991, 0,1.0},
  {-0.34619153, 0.00882991, -0.01973337,1.0},
  {-0.34347504, 0.00692071, 0,1.0},
  // f 274/572/534 725/571/533 273/570/532
  {-0.34744334, 0.00985111, 0,1.0},
  {-0.34744334, 0.00985111, -0.01973337,1.0},
  {-0.34619153, 0.00882991, 0,1.0},
  // f 275/1903/536 726/573/535 274/572/534
  {-0.34918806, 0.01146617, 0,1.0},
  {-0.34918806, 0.01146617, -0.01973337,1.0},
  {-0.34744334, 0.00985111, 0,1.0},
  // f 276/577/538 727/576/537 275/574/951
  {-0.352837, 0.01549848, 0,1.0},
  {-0.352837, 0.01549848, -0.01973337,1.0},
  {-0.34918806, 0.01146617, 0,1.0},
  // f 277/579/540 728/578/539 276/577/952
  {-0.35558888, 0.01918923, 0,1.0},
  {-0.35558888, 0.01918923, -0.01973337,1.0},
  {-0.352837, 0.01549848, 0,1.0},
  // f 278/581/542 729/580/541 277/579/953
  {-0.35799384, 0.02309645, 0,1.0},
  {-0.35799384, 0.02309645, -0.01973337,1.0},
  {-0.35558888, 0.01918923, 0,1.0},
  // f 279/583/544 730/582/543 278/581/954
  {-0.36005208, 0.0272201, 0,1.0},
  {-0.36005208, 0.0272201, -0.01973337,1.0},
  {-0.35799384, 0.02309645, 0,1.0},
  // f 280/585/546 731/584/545 279/583/955
  {-0.3620148, 0.0323046, 0,1.0},
  {-0.3620148, 0.0323046, -0.01973337,1.0},
  {-0.36005208, 0.0272201, 0,1.0},
  // f 281/587/548 732/586/547 280/585/546
  {-0.36350545, 0.03768371, 0,1.0},
  {-0.36350545, 0.03768371, -0.01973337,1.0},
  {-0.3620148, 0.0323046, 0,1.0},
  // f 282/589/550 733/588/549 281/587/956
  {-0.36428103, 0.04170631, 0,1.0},
  {-0.36428103, 0.04170631, -0.01973337,1.0},
  {-0.36350545, 0.03768371, 0,1.0},
  // f 79/1904/551 530/590/551 282/589/550
  {-0.36489382, 0.04673185, 0,1.0},
  {-0.36489382, 0.04673185, -0.01973337,1.0},
  {-0.36428103, 0.04170631, 0,1.0},
  // f 284/1905/1 735/593/1 283/591/1
  {-0.17112531, 0.14125545, 0,1.0},
  {-0.17112531, 0.14125545, -0.01973337,1.0},
  {-0.21477383, 0.14125545, 0,1.0},
  // f 285/1906/71 736/596/71 284/594/71
  {-0.17112531, 0.03478392, 0,1.0},
  {-0.17112531, 0.03478392, -0.01973337,1.0},
  {-0.17112531, 0.14125545, 0,1.0},
  // f 286/1907/1 737/599/1 285/597/1
  {-0.10300279, 0.03478392, 0,1.0},
  {-0.10300279, 0.03478392, -0.01973337,1.0},
  {-0.17112531, 0.03478392, 0,1.0},
  // f 287/1908/71 738/602/71 286/600/71
  {-0.10300279, 0, 0,1.0},
  {-0.10300279, 0, -0.01973337,1.0},
  {-0.10300279, 0.03478392, 0,1.0},
  // f 288/1909/70 739/605/70 287/603/70
  {-0.21477383, 0, 0,1.0},
  {-0.21477383, 0, -0.01973337,1.0},
  {-0.10300279, 0, 0,1.0},
  // f 283/1910/72 734/608/72 288/606/72
  {-0.21477383, 0.14125545, 0,1.0},
  {-0.21477383, 0.14125545, -0.01973337,1.0},
  {-0.21477383, 0, 0,1.0},
  // f 290/1911/71 741/611/71 289/609/71
  {-0.00057813, 0, 0,1.0},
  {-0.00057813, 0, -0.01973337,1.0},
  {-0.00057813, 0.1436643, 0,1.0},
  // f 291/1912/70 742/614/70 290/612/70
  {-0.04027612, 0, 0,1.0},
  {-0.04027612, 0, -0.01973337,1.0},
  {-0.00057813, 0, 0,1.0},
  // f 292/1913/72 743/617/72 291/615/72
  {-0.04027612, 0.0941382, 0,1.0},
  {-0.04027612, 0.0941382, -0.01973337,1.0},
  {-0.04027612, 0, 0,1.0},
  // f 293/621/555 744/620/554 292/618/957
  {-0.05012761, 0.08727793, 0,1.0},
  {-0.05012761, 0.08727793, -0.01973337,1.0},
  {-0.04027612, 0.0941382, 0,1.0},
  // f 294/623/557 745/622/556 293/621/958
  {-0.05456021, 0.08462132, 0,1.0},
  {-0.05456021, 0.08462132, -0.01973337,1.0},
  {-0.05012761, 0.08727793, 0,1.0},
  // f 295/625/559 746/624/558 294/623/557
  {-0.05892068, 0.0822866, 0,1.0},
  {-0.05892068, 0.0822866, -0.01973337,1.0},
  {-0.05456021, 0.08462132, 0,1.0},
  // f 296/627/561 747/626/560 295/625/959
  {-0.06907402, 0.0778543, 0,1.0},
  {-0.06907402, 0.0778543, -0.01973337,1.0},
  {-0.05892068, 0.0822866, 0,1.0},
  // f 297/1914/562 748/628/562 296/627/561
  {-0.08151577, 0.07361472, 0,1.0},
  {-0.08151577, 0.07361472, -0.01973337,1.0},
  {-0.06907402, 0.0778543, 0,1.0},
  // f 298/1915/72 749/631/72 297/629/72
  {-0.08151577, 0.10579707, 0,1.0},
  {-0.08151577, 0.10579707, -0.01973337,1.0},
  {-0.08151577, 0.07361472, 0,1.0},
  // f 299/635/565 750/634/564 298/632/563
  {-0.06960668, 0.11019515, 0,1.0},
  {-0.06960668, 0.11019515, -0.01973337,1.0},
  {-0.08151577, 0.10579707, 0,1.0},
  // f 300/637/567 751/636/566 299/635/960
  {-0.06128185, 0.11416088, 0,1.0},
  {-0.06128185, 0.11416088, -0.01973337,1.0},
  {-0.06960668, 0.11019515, 0,1.0},
  // f 301/639/569 752/638/568 300/637/961
  {-0.05518406, 0.11779124, 0,1.0},
  {-0.05518406, 0.11779124, -0.01973337,1.0},
  {-0.06128185, 0.11416088, 0,1.0},
  // f 302/641/571 753/640/570 301/639/962
  {-0.05269908, 0.11952468, 0,1.0},
  {-0.05269908, 0.11952468, -0.01973337,1.0},
  {-0.05518406, 0.11779124, 0,1.0},
  // f 303/643/573 754/642/572 302/641/963
  {-0.05039332, 0.12131011, 0,1.0},
  {-0.05039332, 0.12131011, -0.01973337,1.0},
  {-0.05269908, 0.11952468, 0,1.0},
  // f 304/1916/575 755/644/574 303/643/964
  {-0.04615372, 0.12510261, 0,1.0},
  {-0.04615372, 0.12510261, -0.01973337,1.0},
  {-0.05039332, 0.12131011, 0,1.0},
  // f 305/648/577 756/647/576 304/645/965
  {-0.04229955, 0.12923431, 0,1.0},
  {-0.04229955, 0.12923431, -0.01973337,1.0},
  {-0.04615372, 0.12510261, 0,1.0},
  // f 306/650/579 757/649/578 305/648/577
  {-0.03883079, 0.13370511, 0,1.0},
  {-0.03883079, 0.13370511, -0.01973337,1.0},
  {-0.04229955, 0.12923431, 0,1.0},
  // f 307/652/581 758/651/580 306/650/966
  {-0.03574746, 0.13851516, 0,1.0},
  {-0.03574746, 0.13851516, -0.01973337,1.0},
  {-0.03883079, 0.13370511, 0,1.0},
  // f 308/1917/582 759/653/582 307/652/967
  {-0.03304954, 0.1436643, 0,1.0},
  {-0.03304954, 0.1436643, -0.01973337,1.0},
  {-0.03574746, 0.13851516, 0,1.0},
  // f 289/1918/1 740/656/1 308/654/1
  {-0.00057813, 0.1436643, 0,1.0},
  {-0.00057813, 0.1436643, -0.01973337,1.0},
  {-0.03304954, 0.1436643, 0,1.0},
  // f 310/1919/72 761/659/72 309/657/72
  {0.11947938, 0.08055224, 0,1.0},
  {0.11947938, 0.08055224, -0.01973337,1.0},
  {0.11947938, 0.05116415, 0,1.0},
  // f 311/1920/1 762/662/1 310/660/1
  {0.18692741, 0.08055224, 0,1.0},
  {0.18692741, 0.08055224, -0.01973337,1.0},
  {0.11947938, 0.08055224, 0,1.0},
  // f 312/1921/71 763/665/71 311/663/71
  {0.18692741, 0.02033077, 0,1.0},
  {0.18692741, 0.02033077, -0.01973337,1.0},
  {0.18692741, 0.08055224, 0,1.0},
  // f 313/669/586 764/668/585 312/666/968
  {0.17215054, 0.01111929, 0,1.0},
  {0.17215054, 0.01111929, -0.01973337,1.0},
  {0.18692741, 0.02033077, 0,1.0},
  // f 314/671/588 765/670/587 313/669/586
  {0.16200872, 0.0059812, 0,1.0},
  {0.16200872, 0.0059812, -0.01973337,1.0},
  {0.17215054, 0.01111929, 0,1.0},
  // f 315/673/590 766/672/589 314/671/588
  {0.15755193, 0.00410277, 0,1.0},
  {0.15755193, 0.00410277, -0.01973337,1.0},
  {0.16200872, 0.0059812, 0,1.0},
  // f 316/675/592 767/674/591 315/673/590
  {0.15267344, 0.00236068, 0,1.0},
  {0.15267344, 0.00236068, -0.01973337,1.0},
  {0.15755193, 0.00410277, 0,1.0},
  // f 317/677/594 768/676/593 316/675/969
  {0.1452087, 0.00034603, 0,1.0},
  {0.1452087, 0.00034603, -0.01973337,1.0},
  {0.15267344, 0.00236068, 0,1.0},
  // f 318/679/596 769/678/595 317/677/594
  {0.13569312, -0.00130996, 0,1.0},
  {0.13569312, -0.00130996, -0.01973337,1.0},
  {0.1452087, 0.00034603, 0,1.0},
  // f 319/681/598 770/680/597 318/679/596
  {0.12684548, -0.00213413, 0,1.0},
  {0.12684548, -0.00213413, -0.01973337,1.0},
  {0.13569312, -0.00130996, 0,1.0},
  // f 320/683/600 771/682/599 319/681/598
  {0.11735957, -0.00240886, 0,1.0},
  {0.11735957, -0.00240886, -0.01973337,1.0},
  {0.12684548, -0.00213413, 0,1.0},
  // f 321/685/602 772/684/601 320/683/600
  {0.1058239, -0.00191491, 0,1.0},
  {0.1058239, -0.00191491, -0.01973337,1.0},
  {0.11735957, -0.00240886, 0,1.0},
  // f 322/687/604 773/686/603 321/685/602
  {0.09702906, -0.00074863, 0,1.0},
  {0.09702906, -0.00074863, -0.01973337,1.0},
  {0.1058239, -0.00191491, 0,1.0},
  // f 323/689/606 774/688/605 322/687/604
  {0.08974966, 0.00088758, 0,1.0},
  {0.08974966, 0.00088758, -0.01973337,1.0},
  {0.09702906, -0.00074863, 0,1.0},
  // f 324/691/608 775/690/607 323/689/970
  {0.08378498, 0.0028085, 0,1.0},
  {0.08378498, 0.0028085, -0.01973337,1.0},
  {0.08974966, 0.00088758, 0,1.0},
  // f 325/693/610 776/692/609 324/691/971
  {0.08029877, 0.00423204, 0,1.0},
  {0.08029877, 0.00423204, -0.01973337,1.0},
  {0.08378498, 0.0028085, 0,1.0},
  // f 326/695/612 777/694/611 325/693/610
  {0.07636081, 0.00616668, 0,1.0},
  {0.07636081, 0.00616668, -0.01973337,1.0},
  {0.08029877, 0.00423204, 0,1.0},
  // f 327/697/614 778/696/613 326/695/972
  {0.07147124, 0.00912545, 0,1.0},
  {0.07147124, 0.00912545, -0.01973337,1.0},
  {0.07636081, 0.00616668, 0,1.0},
  // f 328/699/616 779/698/615 327/697/973
  {0.06694677, 0.01251342, 0,1.0},
  {0.06694677, 0.01251342, -0.01973337,1.0},
  {0.07147124, 0.00912545, 0,1.0},
  // f 329/1922/618 780/700/617 328/699/616
  {0.06278734, 0.0163306, 0,1.0},
  {0.06278734, 0.0163306, -0.01973337,1.0},
  {0.06694677, 0.01251342, 0,1.0},
  // f 330/704/620 781/703/619 329/701/974
  {0.05899298, 0.02057697, 0,1.0},
  {0.05899298, 0.02057697, -0.01973337,1.0},
  {0.06278734, 0.0163306, 0,1.0},
  // f 331/706/622 782/705/621 330/704/975
  {0.0555637, 0.02525254, 0,1.0},
  {0.0555637, 0.02525254, -0.01973337,1.0},
  {0.05899298, 0.02057697, 0,1.0},
  // f 332/708/624 783/707/623 331/706/976
  {0.05360577, 0.02839273, 0,1.0},
  {0.05360577, 0.02839273, -0.01973337,1.0},
  {0.0555637, 0.02525254, 0,1.0},
  // f 333/710/626 784/709/625 332/708/977
  {0.05179045, 0.03170058, 0,1.0},
  {0.05179045, 0.03170058, -0.01973337,1.0},
  {0.05360577, 0.02839273, 0,1.0},
  // f 334/712/628 785/711/627 333/710/978
  {0.04893376, 0.03796792, 0,1.0},
  {0.04893376, 0.03796792, -0.01973337,1.0},
  {0.05179045, 0.03170058, 0,1.0},
  // f 335/714/630 786/713/629 334/712/628
  {0.04664213, 0.04455995, 0,1.0},
  {0.04664213, 0.04455995, -0.01973337,1.0},
  {0.04893376, 0.03796792, 0,1.0},
  // f 336/716/632 787/715/631 335/714/979
  {0.04491555, 0.05147664, 0,1.0},
  {0.04491555, 0.05147664, -0.01973337,1.0},
  {0.04664213, 0.04455995, 0,1.0},
  // f 337/718/634 788/717/633 336/716/980
  {0.04365986, 0.05954267, 0,1.0},
  {0.04365986, 0.05954267, -0.01973337,1.0},
  {0.04491555, 0.05147664, 0,1.0},
  // f 338/720/636 789/719/635 337/718/981
  {0.0432413, 0.06457467, 0,1.0},
  {0.0432413, 0.06457467, -0.01973337,1.0},
  {0.04365986, 0.05954267, 0,1.0},
  // f 339/722/638 790/721/637 338/720/636
  {0.04307039, 0.07062773, 0,1.0},
  {0.04307039, 0.07062773, -0.01973337,1.0},
  {0.0432413, 0.06457467, 0,1.0},
  // f 340/724/640 791/723/639 339/722/982
  {0.04353208, 0.08051129, 0,1.0},
  {0.04353208, 0.08051129, -0.01973337,1.0},
  {0.04307039, 0.07062773, 0,1.0},
  // f 341/726/642 792/725/641 340/724/983
  {0.04475309, 0.08899045, 0,1.0},
  {0.04475309, 0.08899045, -0.01973337,1.0},
  {0.04353208, 0.08051129, 0,1.0},
  // f 342/728/644 793/727/643 341/726/984
  {0.04650446, 0.09620979, 0,1.0},
  {0.04650446, 0.09620979, -0.01973337,1.0},
  {0.04475309, 0.08899045, 0,1.0},
  // f 343/730/646 794/729/645 342/728/985
  {0.04858016, 0.10229938, 0,1.0},
  {0.04858016, 0.10229938, -0.01973337,1.0},
  {0.04650446, 0.09620979, 0,1.0},
  // f 344/732/648 795/731/647 343/730/986
  {0.05045747, 0.10666426, 0,1.0},
  {0.05045747, 0.10666426, -0.01973337,1.0},
  {0.04858016, 0.10229938, 0,1.0},
  // f 345/734/650 796/733/649 344/732/987
  {0.05260947, 0.11085567, 0,1.0},
  {0.05260947, 0.11085567, -0.01973337,1.0},
  {0.05045747, 0.10666426, 0,1.0},
  // f 346/736/652 797/735/651 345/734/650
  {0.05588891, 0.11611768, 0,1.0},
  {0.05588891, 0.11611768, -0.01973337,1.0},
  {0.05260947, 0.11085567, 0,1.0},
  // f 347/738/654 798/737/653 346/736/988
  {0.05913073, 0.12037592, 0,1.0},
  {0.05913073, 0.12037592, -0.01973337,1.0},
  {0.05588891, 0.11611768, 0,1.0},
  // f 348/1923/656 799/739/655 347/738/989
  {0.06272005, 0.12431127, 0,1.0},
  {0.06272005, 0.12431127, -0.01973337,1.0},
  {0.05913073, 0.12037592, 0,1.0},
  // f 349/743/658 800/742/657 348/740/656
  {0.06665686, 0.12792365, 0,1.0},
  {0.06665686, 0.12792365, -0.01973337,1.0},
  {0.06272005, 0.12431127, 0,1.0},
  // f 350/745/660 801/744/659 349/743/658
  {0.07158156, 0.13165666, 0,1.0},
  {0.07158156, 0.13165666, -0.01973337,1.0},
  {0.06665686, 0.12792365, 0,1.0},
  // f 351/747/662 802/746/661 350/745/990
  {0.07557292, 0.13417959, 0,1.0},
  {0.07557292, 0.13417959, -0.01973337,1.0},
  {0.07158156, 0.13165666, 0,1.0},
  // f 352/749/664 803/748/663 351/747/662
  {0.08055224, 0.13682316, 0,1.0},
  {0.08055224, 0.13682316, -0.01973337,1.0},
  {0.07557292, 0.13417959, 0,1.0},
  // f 353/751/666 804/750/665 352/749/991
  {0.08540048, 0.1388372, 0,1.0},
  {0.08540048, 0.1388372, -0.01973337,1.0},
  {0.08055224, 0.13682316, 0,1.0},
  // f 354/753/668 805/752/667 353/751/666
  {0.09146211, 0.14068431, 0,1.0},
  {0.09146211, 0.14068431, -0.01973337,1.0},
  {0.08540048, 0.1388372, 0,1.0},
  // f 355/755/670 806/754/669 354/753/668
  {0.09895155, 0.14221673, 0,1.0},
  {0.09895155, 0.14221673, -0.01973337,1.0},
  {0.09146211, 0.14068431, 0,1.0},
  // f 356/757/672 807/756/671 355/755/670
  {0.10810695, 0.14327027, 0,1.0},
  {0.10810695, 0.14327027, -0.01973337,1.0},
  {0.09895155, 0.14221673, 0,1.0},
  // f 357/759/674 808/758/673 356/757/672
  {0.11919031, 0.1436643, 0,1.0},
  {0.11919031, 0.1436643, -0.01973337,1.0},
  {0.10810695, 0.14327027, 0,1.0},
  // f 358/761/676 809/760/675 357/759/674
  {0.13603304, 0.14298595, 0,1.0},
  {0.13603304, 0.14298595, -0.01973337,1.0},
  {0.11919031, 0.1436643, 0,1.0},
  // f 359/763/678 810/762/677 358/761/676
  {0.1455756, 0.14170393, 0,1.0},
  {0.1455756, 0.14170393, -0.01973337,1.0},
  {0.13603304, 0.14298595, 0,1.0},
  // f 360/765/680 811/764/679 359/763/678
  {0.15019752, 0.14067285, 0,1.0},
  {0.15019752, 0.14067285, -0.01973337,1.0},
  {0.1455756, 0.14170393, 0,1.0},
  // f 361/767/682 812/766/681 360/765/680
  {0.15421513, 0.13942471, 0,1.0},
  {0.15421513, 0.13942471, -0.01973337,1.0},
  {0.15019752, 0.14067285, 0,1.0},
  // f 362/769/684 813/768/683 361/767/682
  {0.15870138, 0.13754196, 0,1.0},
  {0.15870138, 0.13754196, -0.01973337,1.0},
  {0.15421513, 0.13942471, 0,1.0},
  // f 363/771/686 814/770/685 362/769/992
  {0.16246878, 0.13552782, 0,1.0},
  {0.16246878, 0.13552782, -0.01973337,1.0},
  {0.15870138, 0.13754196, 0,1.0},
  // f 364/773/688 815/772/687 363/771/993
  {0.16598019, 0.13321085, 0,1.0},
  {0.16598019, 0.13321085, -0.01973337,1.0},
  {0.16246878, 0.13552782, 0,1.0},
  // f 365/775/690 816/774/689 364/773/994
  {0.16923565, 0.13059105, 0,1.0},
  {0.16923565, 0.13059105, -0.01973337,1.0},
  {0.16598019, 0.13321085, 0,1.0},
  // f 366/1924/692 817/776/691 365/775/690
  {0.17126371, 0.12867633, 0,1.0},
  {0.17126371, 0.12867633, -0.01973337,1.0},
  {0.16923565, 0.13059105, 0,1.0},
  // f 367/780/694 818/779/693 366/777/995
  {0.17348598, 0.12627235, 0,1.0},
  {0.17348598, 0.12627235, -0.01973337,1.0},
  {0.17126371, 0.12867633, 0,1.0},
  // f 368/782/696 819/781/695 367/780/996
  {0.17612036, 0.12290923, 0,1.0},
  {0.17612036, 0.12290923, -0.01973337,1.0},
  {0.17348598, 0.12627235, 0,1.0},
  // f 369/784/698 820/783/697 368/782/997
  {0.17875814, 0.11880992, 0,1.0},
  {0.17875814, 0.11880992, -0.01973337,1.0},
  {0.17612036, 0.12290923, 0,1.0},
  // f 370/786/700 821/785/699 369/784/698
  {0.18109146, 0.11432907, 0,1.0},
  {0.18109146, 0.11432907, -0.01973337,1.0},
  {0.17875814, 0.11880992, 0,1.0},
  // f 371/788/702 822/787/701 370/786/998
  {0.1831203, 0.10946664, 0,1.0},
  {0.1831203, 0.10946664, -0.01973337,1.0},
  {0.18109146, 0.11432907, 0,1.0},
  // f 372/1925/703 823/789/703 371/788/702
  {0.18500036, 0.10367728, 0,1.0},
  {0.18500036, 0.10367728, -0.01973337,1.0},
  {0.1831203, 0.10946664, 0,1.0},
  // f 373/1926/704 824/792/704 372/790/704
  {0.14289349, 0.09616163, 0,1.0},
  {0.14289349, 0.09616163, -0.01973337,1.0},
  {0.18500036, 0.10367728, 0,1.0},
  // f 374/796/708 825/795/707 373/793/999
  {0.14196849, 0.09862459, 0,1.0},
  {0.14196849, 0.09862459, -0.01973337,1.0},
  {0.14289349, 0.09616163, 0,1.0},
  // f 375/798/710 826/797/709 374/796/708
  {0.14085847, 0.10089514, 0,1.0},
  {0.14085847, 0.10089514, -0.01973337,1.0},
  {0.14196849, 0.09862459, 0,1.0},
  // f 376/800/712 827/799/711 375/798/1000
  {0.13973548, 0.10272409, 0,1.0},
  {0.13973548, 0.10272409, -0.01973337,1.0},
  {0.14085847, 0.10089514, 0,1.0},
  // f 377/802/714 828/801/713 376/800/1001
  {0.1384708, 0.10440568, 0,1.0},
  {0.1384708, 0.10440568, -0.01973337,1.0},
  {0.13973548, 0.10272409, 0,1.0},
  // f 378/1927/716 829/803/715 377/802/1002
  {0.13706453, 0.10594001, 0,1.0},
  {0.13706453, 0.10594001, -0.01973337,1.0},
  {0.1384708, 0.10440568, 0,1.0},
  // f 379/807/718 830/806/717 378/804/1003
  {0.13574637, 0.1071379, 0,1.0},
  {0.13574637, 0.1071379, -0.01973337,1.0},
  {0.13706453, 0.10594001, 0,1.0},
  // f 380/809/720 831/808/719 379/807/718
  {0.13407704, 0.10839865, 0,1.0},
  {0.13407704, 0.10839865, -0.01973337,1.0},
  {0.13574637, 0.1071379, 0,1.0},
  // f 381/811/722 832/810/721 380/809/1004
  {0.13226987, 0.10950262, 0,1.0},
  {0.13226987, 0.10950262, -0.01973337,1.0},
  {0.13407704, 0.10839865, 0,1.0},
  // f 382/813/724 833/812/723 381/811/722
  {0.13032858, 0.1104404, 0,1.0},
  {0.13032858, 0.1104404, -0.01973337,1.0},
  {0.13226987, 0.10950262, 0,1.0},
  // f 383/815/726 834/814/725 382/813/724
  {0.12825319, 0.11121202, 0,1.0},
  {0.12825319, 0.11121202, -0.01973337,1.0},
  {0.13032858, 0.1104404, 0,1.0},
  // f 384/817/728 835/816/727 383/815/726
  {0.12571715, 0.11189037, 0,1.0},
  {0.12571715, 0.11189037, -0.01973337,1.0},
  {0.12825319, 0.11121202, 0,1.0},
  // f 385/819/730 836/818/729 384/817/728
  {0.12300596, 0.1123516, 0,1.0},
  {0.12300596, 0.1123516, -0.01973337,1.0},
  {0.12571715, 0.11189037, 0,1.0},
  // f 386/821/732 837/820/731 385/819/730
  {0.12085765, 0.11255514, 0,1.0},
  {0.12085765, 0.11255514, -0.01973337,1.0},
  {0.12300596, 0.1123516, 0,1.0},
  // f 387/823/734 838/822/733 386/821/732
  {0.11822677, 0.11263821, 0,1.0},
  {0.11822677, 0.11263821, -0.01973337,1.0},
  {0.12085765, 0.11255514, 0,1.0},
  // f 388/825/736 839/824/735 387/823/734
  {0.11486509, 0.11249463, 0,1.0},
  {0.11486509, 0.11249463, -0.01973337,1.0},
  {0.11822677, 0.11263821, 0,1.0},
  // f 389/827/738 840/826/737 388/825/736
  {0.11167126, 0.11206378, 0,1.0},
  {0.11167126, 0.11206378, -0.01973337,1.0},
  {0.11486509, 0.11249463, 0,1.0},
  // f 390/829/740 841/828/739 389/827/738
  {0.10864536, 0.11134578, 0,1.0},
  {0.10864536, 0.11134578, -0.01973337,1.0},
  {0.11167126, 0.11206378, 0,1.0},
  // f 391/831/742 842/830/741 390/829/740
  {0.10625204, 0.11052801, 0,1.0},
  {0.10625204, 0.11052801, -0.01973337,1.0},
  {0.10864536, 0.11134578, 0,1.0},
  // f 392/833/744 843/832/743 391/831/742
  {0.10397528, 0.10951076, 0,1.0},
  {0.10397528, 0.10951076, -0.01973337,1.0},
  {0.10625204, 0.11052801, 0,1.0},
  // f 393/835/746 844/834/745 392/833/1005
  {0.1018151, 0.10829411, 0,1.0},
  {0.1018151, 0.10829411, -0.01973337,1.0},
  {0.10397528, 0.10951076, 0,1.0},
  // f 394/837/748 845/836/747 393/835/1006
  {0.09977153, 0.106878, 0,1.0},
  {0.09977153, 0.106878, -0.01973337,1.0},
  {0.1018151, 0.10829411, 0,1.0},
  // f 395/839/750 846/838/749 394/837/748
  {0.09784453, 0.10526244, 0,1.0},
  {0.09784453, 0.10526244, -0.01973337,1.0},
  {0.09977153, 0.106878, 0,1.0},
  // f 396/1928/752 847/840/751 395/839/1007
  {0.09674428, 0.10419735, 0,1.0},
  {0.09674428, 0.10419735, -0.01973337,1.0},
  {0.09784453, 0.10526244, 0,1.0},
  // f 397/844/754 848/843/753 396/841/752
  {0.09534263, 0.10266554, 0,1.0},
  {0.09534263, 0.10266554, -0.01973337,1.0},
  {0.09674428, 0.10419735, 0,1.0},
  // f 398/846/756 849/845/755 397/844/754
  {0.09372242, 0.1005549, 0,1.0},
  {0.09372242, 0.1005549, -0.01973337,1.0},
  {0.09534263, 0.10266554, 0,1.0},
  // f 399/848/758 850/847/757 398/846/756
  {0.09227277, 0.09821206, 0,1.0},
  {0.09227277, 0.09821206, -0.01973337,1.0},
  {0.09372242, 0.1005549, 0,1.0},
  // f 400/850/760 851/849/759 399/848/1008
  {0.09099369, 0.09563697, 0,1.0},
  {0.09099369, 0.09563697, -0.01973337,1.0},
  {0.09227277, 0.09821206, 0,1.0},
  // f 401/852/762 852/851/761 400/850/1009
  {0.08968385, 0.09224036, 0,1.0},
  {0.08968385, 0.09224036, -0.01973337,1.0},
  {0.09099369, 0.09563697, 0,1.0},
  // f 402/854/764 853/853/763 401/852/1010
  {0.08846615, 0.08785504, 0,1.0},
  {0.08846615, 0.08785504, -0.01973337,1.0},
  {0.08968385, 0.09224036, 0,1.0},
  // f 403/856/766 854/855/765 402/854/764
  {0.0874838, 0.08228591, 0,1.0},
  {0.0874838, 0.08228591, -0.01973337,1.0},
  {0.08846615, 0.08785504, 0,1.0},
  // f 404/858/768 855/857/767 403/856/1011
  {0.08698238, 0.07692527, 0,1.0},
  {0.08698238, 0.07692527, -0.01973337,1.0},
  {0.0874838, 0.08228591, 0,1.0},
  // f 405/860/770 856/859/769 404/858/1012
  {0.08681526, 0.0711095, 0,1.0},
  {0.08681526, 0.0711095, -0.01973337,1.0},
  {0.08698238, 0.07692527, 0,1.0},
  // f 406/862/772 857/861/771 405/860/770
  {0.08723264, 0.06165352, 0,1.0},
  {0.08723264, 0.06165352, -0.01973337,1.0},
  {0.08681526, 0.0711095, 0,1.0},
  // f 407/864/774 858/863/773 406/862/1013
  {0.08819505, 0.0548603, 0,1.0},
  {0.08819505, 0.0548603, -0.01973337,1.0},
  {0.08723264, 0.06165352, 0,1.0},
  // f 408/866/776 859/865/775 407/864/774
  {0.08932994, 0.05016376, 0,1.0},
  {0.08932994, 0.05016376, -0.01973337,1.0},
  {0.08819505, 0.0548603, 0,1.0},
  // f 409/868/778 860/867/777 408/866/1014
  {0.09057174, 0.046547, 0,1.0},
  {0.09057174, 0.046547, -0.01973337,1.0},
  {0.08932994, 0.05016376, 0,1.0},
  // f 410/870/780 861/869/779 409/868/1015
  {0.09179632, 0.04382134, 0,1.0},
  {0.09179632, 0.04382134, -0.01973337,1.0},
  {0.09057174, 0.046547, 0,1.0},
  // f 411/872/782 862/871/781 410/870/1016
  {0.09319337, 0.04135774, 0,1.0},
  {0.09319337, 0.04135774, -0.01973337,1.0},
  {0.09179632, 0.04382134, 0,1.0},
  // f 412/874/784 863/873/783 411/872/782
  {0.09411437, 0.04000539, 0,1.0},
  {0.09411437, 0.04000539, -0.01973337,1.0},
  {0.09319337, 0.04135774, 0,1.0},
  // f 413/876/786 864/875/785 412/874/784
  {0.09543897, 0.03834902, 0,1.0},
  {0.09543897, 0.03834902, -0.01973337,1.0},
  {0.09411437, 0.04000539, 0,1.0},
  // f 414/1929/788 865/877/787 413/876/1017
  {0.09723166, 0.03648167, 0,1.0},
  {0.09723166, 0.03648167, -0.01973337,1.0},
  {0.09543897, 0.03834902, 0,1.0},
  // f 415/881/790 866/880/789 414/878/1018
  {0.09916019, 0.0348109, 0,1.0},
  {0.09916019, 0.0348109, -0.01973337,1.0},
  {0.09723166, 0.03648167, 0,1.0},
  // f 416/883/792 867/882/791 415/881/1019
  {0.10122457, 0.03333668, 0,1.0},
  {0.10122457, 0.03333668, -0.01973337,1.0},
  {0.09916019, 0.0348109, 0,1.0},
  // f 417/885/794 868/884/793 416/883/1020
  {0.10342481, 0.03205901, 0,1.0},
  {0.10342481, 0.03205901, -0.01973337,1.0},
  {0.10122457, 0.03333668, 0,1.0},
  // f 418/887/796 869/886/795 417/885/1021
  {0.10576094, 0.03097792, 0,1.0},
  {0.10576094, 0.03097792, -0.01973337,1.0},
  {0.10342481, 0.03205901, 0,1.0},
  // f 419/889/798 870/888/797 418/887/796
  {0.10823289, 0.03009339, 0,1.0},
  {0.10823289, 0.03009339, -0.01973337,1.0},
  {0.10576094, 0.03097792, 0,1.0},
  // f 420/891/800 871/890/799 419/889/798
  {0.1113786, 0.0292914, 0,1.0},
  {0.1113786, 0.0292914, -0.01973337,1.0},
  {0.10823289, 0.03009339, 0,1.0},
  // f 421/893/802 872/892/801 420/891/800
  {0.11471993, 0.02877249, 0,1.0},
  {0.11471993, 0.02877249, -0.01973337,1.0},
  {0.1113786, 0.0292914, 0,1.0},
  // f 422/895/804 873/894/803 421/893/802
  {0.11705618, 0.02858379, 0,1.0},
  {0.11705618, 0.02858379, -0.01973337,1.0},
  {0.11471993, 0.02877249, 0,1.0},
  // f 423/897/806 874/896/805 422/895/804
  {0.11947938, 0.02852089, 0,1.0},
  {0.11947938, 0.02852089, -0.01973337,1.0},
  {0.11705618, 0.02858379, 0,1.0},
  // f 424/899/808 875/898/807 423/897/806
  {0.12381244, 0.02871166, 0,1.0},
  {0.12381244, 0.02871166, -0.01973337,1.0},
  {0.11947938, 0.02852089, 0,1.0},
  // f 425/901/810 876/900/809 424/899/808
  {0.12802409, 0.02928401, 0,1.0},
  {0.12802409, 0.02928401, -0.01973337,1.0},
  {0.12381244, 0.02871166, 0,1.0},
  // f 426/903/812 877/902/811 425/901/810
  {0.13076441, 0.02987755, 0,1.0},
  {0.13076441, 0.02987755, -0.01973337,1.0},
  {0.12802409, 0.02928401, 0,1.0},
  // f 427/905/814 878/904/813 426/903/1022
  {0.13345076, 0.03064069, 0,1.0},
  {0.13345076, 0.03064069, -0.01973337,1.0},
  {0.13076441, 0.02987755, 0,1.0},
  // f 428/907/816 879/906/815 427/905/814
  {0.13790309, 0.03230322, 0,1.0},
  {0.13790309, 0.03230322, -0.01973337,1.0},
  {0.13345076, 0.03064069, 0,1.0},
  // f 429/909/818 880/908/817 428/907/816
  {0.14306615, 0.03473994, 0,1.0},
  {0.14306615, 0.03473994, -0.01973337,1.0},
  {0.13790309, 0.03230322, 0,1.0},
  // f 430/1930/819 881/910/819 429/909/1023
  {0.14867473, 0.03786726, 0,1.0},
  {0.14867473, 0.03786726, -0.01973337,1.0},
  {0.14306615, 0.03473994, 0,1.0},
  // f 431/1931/72 882/913/72 430/911/72
  {0.14867473, 0.05116415, 0,1.0},
  {0.14867473, 0.05116415, -0.01973337,1.0},
  {0.14867473, 0.03786726, 0,1.0},
  // f 309/1932/70 760/916/70 431/914/70
  {0.11947938, 0.05116415, 0,1.0},
  {0.11947938, 0.05116415, -0.01973337,1.0},
  {0.14867473, 0.05116415, 0,1.0},
  // f 433/1933/1 884/919/1 432/917/1
  {0.25668794, 0.14125545, 0,1.0},
  {0.25668794, 0.14125545, -0.01973337,1.0},
  {0.21303943, 0.14125545, 0,1.0},
  // f 434/1934/71 885/922/71 433/920/71
  {0.25668794, 0.09182569, 0,1.0},
  {0.25668794, 0.09182569, -0.01973337,1.0},
  {0.25668794, 0.14125545, 0,1.0},
  // f 435/1935/1 886/925/1 434/923/1
  {0.30438331, 0.09182569, 0,1.0},
  {0.30438331, 0.09182569, -0.01973337,1.0},
  {0.25668794, 0.09182569, 0,1.0},
  // f 436/1936/72 887/928/72 435/926/72
  {0.30438331, 0.14125545, 0,1.0},
  {0.30438331, 0.14125545, -0.01973337,1.0},
  {0.30438331, 0.09182569, 0,1.0},
  // f 437/1937/1 888/931/1 436/929/1
  {0.34822458, 0.14125545, 0,1.0},
  {0.34822458, 0.14125545, -0.01973337,1.0},
  {0.30438331, 0.14125545, 0,1.0},
  // f 438/1938/71 889/934/71 437/932/71
  {0.34822458, 0, 0,1.0},
  {0.34822458, 0, -0.01973337,1.0},
  {0.34822458, 0.14125545, 0,1.0},
  // f 439/1939/70 890/937/70 438/935/70
  {0.30438331, 0, 0,1.0},
  {0.30438331, 0, -0.01973337,1.0},
  {0.34822458, 0, 0,1.0},
  // f 440/1940/72 891/940/72 439/938/72
  {0.30438331, 0.05713812, 0,1.0},
  {0.30438331, 0.05713812, -0.01973337,1.0},
  {0.30438331, 0, 0,1.0},
  // f 441/1941/70 892/943/70 440/941/70
  {0.25668794, 0.05713812, 0,1.0},
  {0.25668794, 0.05713812, -0.01973337,1.0},
  {0.30438331, 0.05713812, 0,1.0},
  // f 442/1942/71 893/946/71 441/944/71
  {0.25668794, 0, 0,1.0},
  {0.25668794, 0, -0.01973337,1.0},
  {0.25668794, 0.05713812, 0,1.0},
  // f 443/1943/70 894/949/70 442/947/70
  {0.21303943, 0, 0,1.0},
  {0.21303943, 0, -0.01973337,1.0},
  {0.25668794, 0, 0,1.0},
  // f 432/1944/72 883/952/72 443/950/72
  {0.21303943, 0.14125545, 0,1.0},
  {0.21303943, 0.14125545, -0.01973337,1.0},
  {0.21303943, 0, 0,1.0},
  // f 445/1945/1 896/955/1 444/953/1
  {0.49998266, 0.14125545, 0,1.0},
  {0.49998266, 0.14125545, -0.01973337,1.0},
  {0.36730275, 0.14125545, 0,1.0},
  // f 446/1946/820 897/958/71 445/956/71
  {0.49998266, 0.10637518, 0,1.0},
  {0.49998266, 0.10637518, -0.01973337,1.0},
  {0.49998266, 0.14125545, 0,1.0},
  // f 447/1947/70 898/961/70 446/959/70
  {0.45546687, 0.10637518, 0,1.0},
  {0.45546687, 0.10637518, -0.01973337,1.0},
  {0.49998266, 0.10637518, 0,1.0},
  // f 448/1948/71 899/964/71 447/962/71
  {0.45546687, 0, 0,1.0},
  {0.45546687, 0, -0.01973337,1.0},
  {0.45546687, 0.10637518, 0,1.0},
  // f 449/1949/70 900/967/70 448/965/70
  {0.41181844, 0, 0,1.0},
  {0.41181844, 0, -0.01973337,1.0},
  {0.45546687, 0, 0,1.0},
  // f 450/1950/72 901/970/72 449/968/72
  {0.41181844, 0.10637518, 0,1.0},
  {0.41181844, 0.10637518, -0.01973337,1.0},
  {0.41181844, 0, 0,1.0},
  // f 451/1951/70 902/973/70 450/971/70
  {0.36730275, 0.10637518, 0,1.0},
  {0.36730275, 0.10637518, -0.01973337,1.0},
  {0.41181844, 0.10637518, 0,1.0},
  // f 444/1952/821 895/976/72 451/974/72
  {0.36730275, 0.14125545, 0,1.0},
  {0.36730275, 0.14125545, -0.01973337,1.0},
  {0.36730275, 0.10637518, 0,1.0}
  
  
//	{600.0,300.0,0.0,1.0},
//	{1400.0,300.0,0.0,1.0},
//	{1400.0,900.0,0.0,1.0},
//	{600.0,900.0,0.0,1.0}
};

const struct TexST
{
float s,t;
} TexCoordBufferMain[]=
{
    // f 453/3/1 452/2/1 1/1/1
  {0.23816000, 0.84061944},
  {0.13693565, 0.84061944},
  {0.13693565, 0.86834013},
  // f 454/5/3 453/3/2 2/4/2
  {0.24952707, 0.84061944},
  {0.23816000, 0.84061944},
  {0.23816000, 0.86834013},
  // f 455/7/5 454/5/3 3/6/4
  {0.25872794, 0.84061944},
  {0.24952707, 0.84061944},
  {0.24952707, 0.86834013},
  // f 456/9/7 455/7/5 4/8/6
  {0.26608145, 0.84061944},
  {0.25872794, 0.84061944},
  {0.25872794, 0.86834013},
  // f 457/11/9 456/9/7 5/10/8
  {0.27186647, 0.84061944},
  {0.26608145, 0.84061944},
  {0.26608145, 0.86834013},
  // f 458/13/11 457/11/9 6/12/10
  {0.27717340, 0.84061944},
  {0.27186647, 0.84061944},
  {0.27186647, 0.86834013},
  // f 459/15/13 458/13/11 7/14/12
  {0.28123057, 0.84061944},
  {0.27717340, 0.84061944},
  {0.27717340, 0.86834013},
  // f 460/17/15 459/15/13 8/16/14
  {0.28423724, 0.84061944},
  {0.28123057, 0.84061944},
  {0.28123057, 0.86834013},
  // f 461/19/17 460/17/15 9/18/16
  {0.28769672, 0.84061944},
  {0.28423724, 0.84061944},
  {0.28423724, 0.86834013},
  // f 462/22/19 461/21/17 10/20/18
  {0.88556695, 0.12260771},
  {0.88556695, 0.11930633},
  {0.85803604, 0.11930633},
  // f 463/24/21 462/22/19 11/23/20
  {0.88556695, 0.12617695},
  {0.88556695, 0.12260771},
  {0.85803604, 0.12260771},
  // f 464/26/23 463/24/21 12/25/22
  {0.88556695, 0.1300143},
  {0.88556695, 0.12617695},
  {0.85803604, 0.12617695},
  // f 465/28/25 464/26/23 13/27/24
  {0.88556695, 0.13411963},
  {0.88556695, 0.1300143},
  {0.85803604, 0.1300143},
  // f 466/30/27 465/28/25 14/29/26
  {0.88556695, 0.13939977},
  {0.88556695, 0.13411963},
  {0.85803604, 0.13411963},
  // f 467/32/29 466/30/27 15/31/28
  {0.88556695, 0.14506578},
  {0.88556695, 0.13939977},
  {0.85803604, 0.13939977},
  // f 468/34/31 467/32/29 16/33/30
  {0.88556695, 0.1511178},
  {0.88556695, 0.14506578},
  {0.85803604, 0.14506578},
  // f 469/36/33 468/34/31 17/35/32
  {0.88556695, 0.15755576},
  {0.88556695, 0.1511178},
  {0.85803604, 0.1511178},
  // f 470/38/35 469/36/33 18/37/34
  {0.88556695, 0.1643796},
  {0.88556695, 0.15755576},
  {0.85803604, 0.15755576},
  // f 471/40/37 470/38/35 19/39/36
  {0.88556695, 0.17253596},
  {0.88556695, 0.1643796},
  {0.85803604, 0.1643796},
  // f 472/42/39 471/40/37 20/41/38
  {0.88556695, 0.18017763},
  {0.88556695, 0.17253596},
  {0.85803604, 0.17253596},
  // f 473/44/41 472/42/39 21/43/40
  {0.88556695, 0.18631804},
  {0.88556695, 0.18017763},
  {0.85803604, 0.18017763},
  // f 474/46/43 473/44/41 22/45/42
  {0.88556695, 0.19208026},
  {0.88556695, 0.18631804},
  {0.85803604, 0.18631804},
  // f 475/48/45 474/46/43 23/47/44
  {0.88556695, 0.19659328},
  {0.88556695, 0.19208026},
  {0.85803604, 0.19208026},
  // f 476/50/47 475/48/45 24/49/46
  {0.88556695, 0.20084369},
  {0.88556695, 0.19659328},
  {0.85803604, 0.19659328},
  // f 477/52/49 476/50/47 25/51/48
  {0.88556695, 0.20483148},
  {0.88556695, 0.20084369},
  {0.85803604, 0.20084369},
  // f 478/54/51 477/52/49 26/53/50
  {0.88556695, 0.20783269},
  {0.88556695, 0.20483148},
  {0.85803604, 0.20483148},
  // f 479/56/53 478/54/51 27/55/52
  {0.88556695, 0.21134788},
  {0.88556695, 0.20783269},
  {0.85803604, 0.20783269},
  // f 480/59/55 479/58/53 28/57/54
  {0.37149429, 0.91218209},
  {0.37527239, 0.91218209},
  {0.37527239, 0.88446128},
  // f 481/61/57 480/59/55 29/60/56
  {0.36733842, 0.91218209},
  {0.37149429, 0.91218209},
  {0.37149429, 0.88446128},
  // f 482/63/59 481/61/57 30/62/58
  {0.36280486, 0.91218209},
  {0.36733842, 0.91218209},
  {0.36733842, 0.88446128},
  // f 483/65/61 482/63/59 31/64/60
  {0.35686591, 0.91218209},
  {0.36280486, 0.91218209},
  {0.36280486, 0.88446128},
  // f 484/67/63 483/65/61 32/66/62
  {0.35038304, 0.91218209},
  {0.35686591, 0.91218209},
  {0.35686591, 0.88446128},
  // f 485/69/65 484/67/63 33/68/64
  {0.34213221, 0.91218209},
  {0.35038304, 0.91218209},
  {0.35038304, 0.88446128},
  // f 486/71/67 485/69/65 34/70/66
  {0.33179602, 0.91218209},
  {0.34213221, 0.91218209},
  {0.34213221, 0.88446128},
  // f 487/73/69 486/71/67 35/72/68
  {0.32049292, 0.91218209},
  {0.33179602, 0.91218209},
  {0.33179602, 0.88446128},
  // f 488/75/70 487/73/70 36/74/70
  {0.28715476, 0.91218209},
  {0.32049292, 0.91218209},
  {0.32049292, 0.88446128},
  // f 489/78/71 488/77/71 37/76/71
  {0.90491146, 0.66633463},
  {0.90491146, 0.59270144},
  {0.87738061, 0.59270144},
  // f 490/81/70 489/80/70 38/79/70
  {0.20997930, 0.95602393},
  {0.27114406, 0.95602393},
  {0.27114406, 0.92830324},
  // f 452/84/72 490/83/72 39/82/72
  {0.56404120, 0.65848398},
  {0.56404120, 0.85691476},
  {0.59157205, 0.85691476},
  // f 492/87/1 491/86/1 40/85/1
  {0.94630802, 0.12977266},
  {0.93138647, 0.12977266},
  {0.93138647, 0.15749347},
  // f 493/89/74 492/87/73 41/88/73
  {0.95354897, 0.12977266},
  {0.94630802, 0.12977266},
  {0.94630802, 0.15749347},
  // f 494/91/76 493/89/74 42/90/75
  {0.95871848, 0.12977266},
  {0.95354897, 0.12977266},
  {0.95354897, 0.15749347},
  // f 495/93/78 494/91/76 43/92/77
  {0.96226937, 0.12977266},
  {0.95871848, 0.12977266},
  {0.95871848, 0.15749347},
  // f 496/95/80 495/93/78 44/94/79
  {0.96498585, 0.12977266},
  {0.96226937, 0.12977266},
  {0.96226937, 0.15749347},
  // f 497/97/82 496/95/80 45/96/81
  {0.96701896, 0.12977266},
  {0.96498585, 0.12977266},
  {0.96498585, 0.15749347},
  // f 498/99/84 497/97/82 46/98/83
  {0.96884233, 0.12977266},
  {0.96701896, 0.12977266},
  {0.96701896, 0.15749347},
  // f 499/101/86 498/99/84 47/100/85
  {0.96983576, 0.12977266},
  {0.96884233, 0.12977266},
  {0.96884233, 0.15749347},
  // f 500/103/88 499/101/86 48/102/87
  {0.97104275, 0.12977266},
  {0.96983576, 0.12977266},
  {0.96983576, 0.15749347},
  // f 501/105/90 500/103/88 49/104/89
  {0.97265011, 0.12977266},
  {0.97104275, 0.12977266},
  {0.97104275, 0.15749347},
  // f 502/108/92 501/107/90 50/106/91
  {0.96446377, 0.6211518},
  {0.96446377, 0.62277901},
  {0.99199462, 0.62277901},
  // f 503/110/94 502/108/92 51/109/93
  {0.96446377, 0.61971998},
  {0.96446377, 0.6211518},
  {0.99199462, 0.6211518},
  // f 504/112/96 503/110/94 52/111/95
  {0.96446377, 0.61821902},
  {0.96446377, 0.61971998},
  {0.99199462, 0.61971998},
  // f 505/114/98 504/112/96 53/113/97
  {0.96446377, 0.61664903},
  {0.96446377, 0.61821902},
  {0.99199462, 0.61821902},
  // f 506/116/100 505/114/98 54/115/99
  {0.96446377, 0.61501002},
  {0.96446377, 0.61664903},
  {0.99199462, 0.61664903},
  // f 507/118/102 506/116/100 55/117/101
  {0.96446377, 0.61330199},
  {0.96446377, 0.61501002},
  {0.99199462, 0.61501002},
  // f 508/120/104 507/118/102 56/119/103
  {0.96446377, 0.61116123},
  {0.96446377, 0.61330199},
  {0.99199462, 0.61330199},
  // f 509/122/106 508/120/104 57/121/105
  {0.96446377, 0.61005354},
  {0.96446377, 0.61116123},
  {0.99199462, 0.61116123},
  // f 510/124/108 509/122/106 58/123/107
  {0.96446377, 0.60853791},
  {0.96446377, 0.61005354},
  {0.99199462, 0.61005354},
  // f 511/126/110 510/124/108 59/125/109
  {0.96446377, 0.60597861},
  {0.96446377, 0.60853791},
  {0.99199462, 0.60853791},
  // f 512/128/112 511/126/110 60/127/111
  {0.96446377, 0.60387349},
  {0.96446377, 0.60597861},
  {0.99199462, 0.60597861},
  // f 513/130/114 512/128/112 61/129/113
  {0.96446377, 0.60185039},
  {0.96446377, 0.60387349},
  {0.99199462, 0.60387349},
  // f 514/132/116 513/130/114 62/131/115
  {0.96446377, 0.59990907},
  {0.96446377, 0.60185039},
  {0.99199462, 0.60185039},
  // f 515/134/118 514/132/116 63/133/117
  {0.96446377, 0.59804964},
  {0.96446377, 0.59990907},
  {0.99199462, 0.59990907},
  // f 516/136/120 515/134/118 64/135/119
  {0.96446377, 0.59627199},
  {0.96446377, 0.59804964},
  {0.99199462, 0.59804964},
  // f 517/138/122 516/136/120 65/137/121
  {0.96446377, 0.59457624},
  {0.96446377, 0.59627199},
  {0.99199462, 0.59627199},
  // f 518/140/124 517/138/122 66/139/123
  {0.96446377, 0.59375906},
  {0.96446377, 0.59457624},
  {0.99199462, 0.59457624},
  // f 519/142/126 518/140/124 67/141/125
  {0.96446377, 0.59270144},
  {0.96446377, 0.59375906},
  {0.99199462, 0.59375906},
  // f 520/145/128 519/144/126 68/143/127
  {0.70665878, 0.99182439},
  {0.70771784, 0.99182439},
  {0.70771784, 0.9641037},
  // f 521/147/130 520/145/128 69/146/129
  {0.70546031, 0.99182439},
  {0.70665878, 0.99182439},
  {0.70665878, 0.9641037},
  // f 522/149/132 521/147/130 70/148/131
  {0.70412248, 0.99182439},
  {0.70546031, 0.99182439},
  {0.70546031, 0.9641037},
  // f 523/151/134 522/149/132 71/150/133
  {0.70225424, 0.99182439},
  {0.70412248, 0.99182439},
  {0.70412248, 0.9641037},
  // f 524/153/136 523/151/134 72/152/135
  {0.70016825, 0.99182439},
  {0.70225424, 0.99182439},
  {0.70225424, 0.9641037},
  // f 525/155/138 524/153/136 73/154/137
  {0.69737750, 0.99182439},
  {0.70016825, 0.99182439},
  {0.70016825, 0.9641037},
  // f 526/157/140 525/155/138 74/156/139
  {0.69372529, 0.99182439},
  {0.69737750, 0.99182439},
  {0.69737750, 0.9641037},
  // f 527/159/142 526/157/140 75/158/141
  {0.68964624, 0.99182439},
  {0.69372529, 0.99182439},
  {0.69372529, 0.9641037},
  // f 528/161/144 527/159/142 76/160/143
  {0.68446183, 0.99182439},
  {0.68964624, 0.99182439},
  {0.68964624, 0.9641037},
  // f 529/163/70 528/161/70 77/162/70
  {0.66712064, 0.99182439},
  {0.68446183, 0.99182439},
  {0.68446183, 0.9641037},
  // f 491/166/71 529/165/71 78/164/71
  {0.97265011, 0.11365151},
  {0.97265011, 0.06939036},
  {0.94511920, 0.06939036},
  // f 531/169/145 530/168/145 79/167/145
  {0.93531907, 0.95682156},
  {0.87738061, 0.95682156},
  {0.87738061, 0.98454225},
  // f 532/172/148 531/171/147 80/170/146
  {0.75125951, 0.97431076},
  {0.75125951, 0.9678787},
  {0.72372866, 0.9678787},
  // f 533/174/150 532/172/148 81/173/149
  {0.75125951, 0.97907531},
  {0.75125951, 0.97431076},
  {0.72372866, 0.97431076},
  // f 534/176/152 533/174/150 82/175/151
  {0.75125951, 0.98290038},
  {0.75125951, 0.97907531},
  {0.72372866, 0.97907531},
  // f 535/178/154 534/176/152 83/177/153
  {0.75125951, 0.98593235},
  {0.75125951, 0.98290038},
  {0.72372866, 0.98290038},
  // f 536/180/156 535/178/154 84/179/155
  {0.75125951, 0.98765135},
  {0.75125951, 0.98593235},
  {0.72372866, 0.98593235},
  // f 537/182/158 536/180/156 85/181/157
  {0.75125951, 0.98953557},
  {0.75125951, 0.98765135},
  {0.72372866, 0.98765135},
  // f 538/184/160 537/182/158 86/183/159
  {0.75125951, 0.99182439},
  {0.75125951, 0.98953557},
  {0.72372866, 0.98953557},
  // f 539/187/162 538/186/160 87/185/161
  {0.50329673, 0.9607197},
  {0.50117278, 0.9607197},
  {0.50117278, 0.98844051},
  // f 540/189/164 539/187/162 88/188/163
  {0.50558197, 0.9607197},
  {0.50329673, 0.9607197},
  {0.50329673, 0.98844051},
  // f 541/191/166 540/189/164 89/190/165
  {0.50802857, 0.9607197},
  {0.50558197, 0.9607197},
  {0.50558197, 0.98844051},
  // f 542/193/168 541/191/166 90/192/167
  {0.51063651, 0.9607197},
  {0.50802857, 0.9607197},
  {0.50802857, 0.98844051},
  // f 543/195/170 542/193/168 91/194/169
  {0.51340568, 0.9607197},
  {0.51063651, 0.9607197},
  {0.51063651, 0.98844051},
  // f 544/197/172 543/195/170 92/196/171
  {0.51633626, 0.9607197},
  {0.51340568, 0.9607197},
  {0.51340568, 0.98844051},
  // f 545/199/174 544/197/172 93/198/173
  {0.52006572, 0.9607197},
  {0.51633626, 0.9607197},
  {0.51633626, 0.98844051},
  // f 546/201/176 545/199/174 94/200/175
  {0.52268124, 0.9607197},
  {0.52006572, 0.9607197},
  {0.52006572, 0.98844051},
  // f 547/203/178 546/201/176 95/202/177
  {0.52609569, 0.9607197},
  {0.52268124, 0.9607197},
  {0.52268124, 0.98844051},
  // f 548/205/180 547/203/178 96/204/179
  {0.53059614, 0.9607197},
  {0.52609569, 0.9607197},
  {0.52609569, 0.98844051},
  // f 549/207/182 548/205/180 97/206/181
  {0.53427106, 0.9607197},
  {0.53059614, 0.9607197},
  {0.53059614, 0.98844051},
  // f 550/209/184 549/207/182 98/208/183
  {0.53723544, 0.9607197},
  {0.53427106, 0.9607197},
  {0.53427106, 0.98844051},
  // f 551/211/186 550/209/184 99/210/185
  {0.53958958, 0.9607197},
  {0.53723544, 0.9607197},
  {0.53723544, 0.98844051},
  // f 552/213/188 551/211/186 100/212/187
  {0.54177153, 0.9607197},
  {0.53958958, 0.9607197},
  {0.53958958, 0.98844051},
  // f 553/215/190 552/213/188 101/214/189
  {0.54378122, 0.9607197},
  {0.54177153, 0.9607197},
  {0.54177153, 0.98844051},
  // f 554/217/192 553/215/190 102/216/191
  {0.54502523, 0.9607197},
  {0.54378122, 0.9607197},
  {0.54378122, 0.98844051},
  // f 555/219/194 554/217/192 103/218/193
  {0.54619271, 0.9607197},
  {0.54502523, 0.9607197},
  {0.54502523, 0.98844051},
  // f 556/221/196 555/219/194 104/220/195
  {0.54803050, 0.9607197},
  {0.54619271, 0.9607197},
  {0.54619271, 0.98844051},
  // f 557/224/198 556/223/196 105/222/197
  {0.45763114, 0.98373342},
  {0.45763114, 0.98530579},
  {0.48516199, 0.98530579},
  // f 558/226/200 557/224/198 106/225/199
  {0.45763114, 0.98210466},
  {0.45763114, 0.98373342},
  {0.48516199, 0.98373342},
  // f 559/228/202 558/226/200 107/227/201
  {0.45763114, 0.98070407},
  {0.45763114, 0.98210466},
  {0.48516199, 0.98210466},
  // f 560/230/204 559/228/202 108/229/203
  {0.45763114, 0.97926438},
  {0.45763114, 0.98070407},
  {0.48516199, 0.98070407},
  // f 561/232/206 560/230/204 109/231/205
  {0.45763114, 0.97778535},
  {0.45763114, 0.97926438},
  {0.48516199, 0.97926438},
  // f 562/234/208 561/232/206 110/233/207
  {0.45763114, 0.9762671},
  {0.45763114, 0.97778535},
  {0.48516199, 0.97778535},
  // f 563/236/210 562/234/208 111/235/209
  {0.45763114, 0.97470951},
  {0.45763114, 0.9762671},
  {0.48516199, 0.9762671},
  // f 564/238/212 563/236/210 112/237/211
  {0.45763114, 0.97375619},
  {0.45763114, 0.97470951},
  {0.48516199, 0.97470951},
  // f 565/240/214 564/238/212 113/239/213
  {0.45763114, 0.97278869},
  {0.45763114, 0.97375619},
  {0.48516199, 0.97375619},
  // f 566/242/216 565/240/214 114/241/215
  {0.45763114, 0.97126186},
  {0.45763114, 0.97278869},
  {0.48516199, 0.97278869},
  // f 567/244/218 566/242/216 115/243/217
  {0.45763114, 0.96976757},
  {0.45763114, 0.97126186},
  {0.48516199, 0.97126186},
  // f 568/246/220 567/244/218 116/245/219
  {0.45763114, 0.96830571},
  {0.45763114, 0.96976757},
  {0.48516199, 0.96976757},
  // f 569/248/222 568/246/220 117/247/221
  {0.45763114, 0.96687639},
  {0.45763114, 0.96830571},
  {0.48516199, 0.96830571},
  // f 570/250/224 569/248/222 118/249/223
  {0.45763114, 0.96547949},
  {0.45763114, 0.96687639},
  {0.48516199, 0.96687639},
  // f 571/252/226 570/250/224 119/251/225
  {0.45763114, 0.96384609},
  {0.45763114, 0.96547949},
  {0.48516199, 0.96547949},
  // f 572/254/228 571/252/226 120/253/227
  {0.45763114, 0.96225965},
  {0.45763114, 0.96384609},
  {0.48516199, 0.96384609},
  // f 573/256/230 572/254/228 121/255/229
  {0.45763114, 0.9607197},
  {0.45763114, 0.96225965},
  {0.48516199, 0.96225965},
  // f 574/259/232 573/258/230 122/257/231
  {0.10020646, 0.91090834},
  {0.10195672, 0.91090834},
  {0.10195672, 0.88318765},
  // f 575/261/234 574/259/232 123/260/233
  {0.09901745, 0.91090834},
  {0.10020646, 0.91090834},
  {0.10020646, 0.88318765},
  // f 576/263/236 575/261/234 124/262/235
  {0.09721183, 0.91090834},
  {0.09901745, 0.91090834},
  {0.09901745, 0.88318765},
  // f 577/265/238 576/263/236 125/264/237
  {0.09408768, 0.91090834},
  {0.09721183, 0.91090834},
  {0.09721183, 0.88318765},
  // f 578/267/240 577/265/238 126/266/239
  {0.08904894, 0.91090834},
  {0.09408768, 0.91090834},
  {0.09408768, 0.88318765},
  // f 579/269/242 578/267/240 127/268/241
  {0.08054190, 0.91090834},
  {0.08904894, 0.91090834},
  {0.08904894, 0.88318765},
  // f 580/271/244 579/269/242 128/270/243
  {0.06901915, 0.91090834},
  {0.08054190, 0.91090834},
  {0.08054190, 0.88318765},
  // f 581/273/246 580/271/244 129/272/245
  {0.04705226, 0.91090834},
  {0.06901915, 0.91090834},
  {0.06901915, 0.88318765},
  // f 582/275/248 581/273/246 130/274/247
  {0.03357295, 0.91090834},
  {0.04705226, 0.91090834},
  {0.04705226, 0.88318765},
  // f 583/277/250 582/275/248 131/276/249
  {0.02419953, 0.91090834},
  {0.03357295, 0.91090834},
  {0.03357295, 0.88318765},
  // f 584/279/252 583/277/250 132/278/251
  {0.01793888, 0.91090834},
  {0.02419953, 0.91090834},
  {0.02419953, 0.88318765},
  // f 585/281/254 584/279/252 133/280/253
  {0.01476588, 0.91090834},
  {0.01793888, 0.91090834},
  {0.01793888, 0.88318765},
  // f 586/283/256 585/281/254 134/282/255
  {0.01188723, 0.91090834},
  {0.01476588, 0.91090834},
  {0.01476588, 0.88318765},
  // f 587/285/258 586/283/256 135/284/257
  {0.00800536, 0.91090834},
  {0.01188723, 0.91090834},
  {0.01188723, 0.88318765},
  // f 588/288/260 587/287/258 136/286/259
  {0.87738061, 0.80733752},
  {0.87738061, 0.81130004},
  {0.90491146, 0.81130004},
  // f 589/290/262 588/288/260 137/289/261
  {0.87738061, 0.80385375},
  {0.87738061, 0.80733752},
  {0.90491146, 0.80733752},
  // f 590/292/264 589/290/262 138/291/263
  {0.87738061, 0.80020487},
  {0.87738061, 0.80385375},
  {0.90491146, 0.80385375},
  // f 591/294/266 590/292/264 139/293/265
  {0.87738061, 0.79639077},
  {0.87738061, 0.80020487},
  {0.90491146, 0.80020487},
  // f 592/296/268 591/294/266 140/295/267
  {0.87738061, 0.79241168},
  {0.87738061, 0.79639077},
  {0.90491146, 0.79639077},
  // f 593/298/270 592/296/268 141/297/269
  {0.87738061, 0.78826737},
  {0.87738061, 0.79241168},
  {0.90491146, 0.79241168},
  // f 594/300/272 593/298/270 142/299/271
  {0.87738061, 0.78307617},
  {0.87738061, 0.78826737},
  {0.90491146, 0.78826737},
  // f 595/302/274 594/300/272 143/301/273
  {0.87738061, 0.78039145},
  {0.87738061, 0.78307617},
  {0.90491146, 0.78307617},
  // f 596/304/276 595/302/274 144/303/275
  {0.87738061, 0.77671933},
  {0.87738061, 0.78039145},
  {0.90491146, 0.78039145},
  // f 597/306/278 596/304/276 145/305/277
  {0.87738061, 0.77186823},
  {0.87738061, 0.77671933},
  {0.90491146, 0.77671933},
  // f 598/308/280 597/306/278 146/307/279
  {0.87738061, 0.76710367},
  {0.87738061, 0.77186823},
  {0.90491146, 0.77186823},
  // f 599/310/282 598/308/280 147/309/281
  {0.87738061, 0.76242578},
  {0.87738061, 0.76710367},
  {0.90491146, 0.76710367},
  // f 600/312/284 599/310/282 148/311/283
  {0.87738061, 0.75783455},
  {0.87738061, 0.76242578},
  {0.90491146, 0.76242578},
  // f 601/314/286 600/312/284 149/313/285
  {0.87738061, 0.75332999},
  {0.87738061, 0.75783455},
  {0.90491146, 0.75783455},
  // f 602/316/288 601/314/286 150/315/287
  {0.87738061, 0.7505585},
  {0.87738061, 0.75332999},
  {0.90491146, 0.75332999},
  // f 603/318/290 602/316/288 151/317/289
  {0.87738061, 0.74782097},
  {0.87738061, 0.7505585},
  {0.90491146, 0.7505585},
  // f 604/320/292 603/318/290 152/319/291
  {0.87738061, 0.74412608},
  {0.87738061, 0.74782097},
  {0.90491146, 0.74782097},
  // f 605/322/294 604/320/292 153/321/293
  {0.87738061, 0.74065948},
  {0.87738061, 0.74412608},
  {0.90491146, 0.74412608},
  // f 606/325/296 605/324/294 154/323/295
  {0.14077538, 0.88446128},
  {0.13693565, 0.88446128},
  {0.13693565, 0.91218209},
  // f 607/327/298 606/325/296 155/326/297
  {0.14506935, 0.88446128},
  {0.14077538, 0.88446128},
  {0.14077538, 0.91218209},
  // f 608/329/300 607/327/298 156/328/299
  {0.14981756, 0.88446128},
  {0.14506935, 0.88446128},
  {0.14506935, 0.91218209},
  // f 609/331/302 608/329/300 157/330/301
  {0.15367682, 0.88446128},
  {0.14981756, 0.88446128},
  {0.14981756, 0.91218209},
  // f 610/333/304 609/331/302 158/332/303
  {0.15779163, 0.88446128},
  {0.15367682, 0.88446128},
  {0.15367682, 0.91218209},
  // f 611/335/306 610/333/304 159/334/305
  {0.16378388, 0.88446128},
  {0.15779163, 0.88446128},
  {0.15779163, 0.91218209},
  // f 612/337/308 611/335/306 160/336/307
  {0.17132731, 0.88446128},
  {0.16378388, 0.88446128},
  {0.16378388, 0.91218209},
  // f 613/339/310 612/337/308 161/338/309
  {0.18170050, 0.88446128},
  {0.17132731, 0.88446128},
  {0.17132731, 0.91218209},
  // f 614/341/312 613/339/310 162/340/311
  {0.19334236, 0.88446128},
  {0.18170050, 0.88446128},
  {0.18170050, 0.91218209},
  // f 615/343/314 614/341/312 163/342/313
  {0.20625295, 0.88446128},
  {0.19334236, 0.88446128},
  {0.19334236, 0.91218209},
  // f 616/345/316 615/343/314 164/344/315
  {0.22200502, 0.88446128},
  {0.20625295, 0.88446128},
  {0.20625295, 0.91218209},
  // f 617/347/318 616/345/316 165/346/317
  {0.23360099, 0.88446128},
  {0.22200502, 0.88446128},
  {0.22200502, 0.91218209},
  // f 618/349/320 617/347/318 166/348/319
  {0.24293223, 0.88446128},
  {0.23360099, 0.88446128},
  {0.23360099, 0.91218209},
  // f 619/351/322 618/349/320 167/350/321
  {0.25034332, 0.88446128},
  {0.24293223, 0.88446128},
  {0.24293223, 0.91218209},
  // f 620/353/324 619/351/322 168/352/323
  {0.25613579, 0.88446128},
  {0.25034332, 0.88446128},
  {0.25034332, 0.91218209},
  // f 621/355/326 620/353/324 169/354/325
  {0.25971031, 0.88446128},
  {0.25613579, 0.88446128},
  {0.25613579, 0.91218209},
  // f 622/357/328 621/355/326 170/356/327
  {0.26385537, 0.88446128},
  {0.25971031, 0.88446128},
  {0.25971031, 0.91218209},
  // f 623/359/330 622/357/328 171/358/329
  {0.26766035, 0.88446128},
  {0.26385537, 0.88446128},
  {0.26385537, 0.91218209},
  // f 624/361/332 623/359/330 172/360/331
  {0.27114406, 0.88446128},
  {0.26766035, 0.88446128},
  {0.26766035, 0.91218209},
  // f 625/364/334 624/363/332 173/362/333
  {0.98208696, 0.90568602},
  {0.98208696, 0.90202129},
  {0.95455611, 0.90202129},
  // f 626/366/336 625/364/334 174/365/335
  {0.98208696, 0.90968513},
  {0.98208696, 0.90568602},
  {0.95455611, 0.90568602},
  // f 627/368/338 626/366/336 175/367/337
  {0.98208696, 0.91401851},
  {0.98208696, 0.90968513},
  {0.95455611, 0.90968513},
  // f 628/370/340 627/368/338 176/369/339
  {0.98208696, 0.91965985},
  {0.98208696, 0.91401851},
  {0.95455611, 0.91401851},
  // f 629/372/342 628/370/340 177/371/341
  {0.98208696, 0.92578268},
  {0.98208696, 0.91965985},
  {0.95455611, 0.91965985},
  // f 630/374/344 629/372/342 178/373/343
  {0.98208696, 0.93238699},
  {0.98208696, 0.92578268},
  {0.95455611, 0.92578268},
  // f 631/376/346 630/374/344 179/375/345
  {0.98208696, 0.94070041},
  {0.98208696, 0.93238699},
  {0.95455611, 0.93238699},
  // f 632/379/347 631/378/347 180/377/347
  {0.92092216, 0.3356334},
  {0.97832298, 0.3356334},
  {0.97832298, 0.30791265},
  // f 633/382/350 632/381/349 181/380/348
  {0.92092216, 0.67224514},
  {0.92092216, 0.67683244},
  {0.94845307, 0.67683244},
  // f 634/384/352 633/382/350 182/383/351
  {0.92092216, 0.66856515},
  {0.92092216, 0.67224514},
  {0.94845307, 0.67224514},
  // f 635/386/354 634/384/352 183/385/353
  {0.92092216, 0.66565228},
  {0.92092216, 0.66856515},
  {0.94845307, 0.66856515},
  // f 636/388/356 635/386/354 184/387/355
  {0.92092216, 0.66338348},
  {0.92092216, 0.66565228},
  {0.94845307, 0.66565228},
  // f 637/390/358 636/388/356 185/389/357
  {0.92092216, 0.66132522},
  {0.92092216, 0.66338348},
  {0.94845307, 0.66338348},
  // f 638/392/360 637/390/358 186/391/359
  {0.92092216, 0.65947747},
  {0.92092216, 0.66132522},
  {0.94845307, 0.66132522},
  // f 639/394/362 638/392/360 187/393/361
  {0.92092216, 0.65836251},
  {0.92092216, 0.65947747},
  {0.94845307, 0.65947747},
  // f 640/397/364 639/396/362 188/395/363
  {0.68403524, 0.94459867},
  {0.68517518, 0.94459867},
  {0.68517518, 0.91687787},
  // f 641/399/366 640/397/364 189/398/365
  {0.68219662, 0.94459867},
  {0.68403524, 0.94459867},
  {0.68403524, 0.91687787},
  // f 642/401/368 641/399/366 190/400/367
  {0.68019724, 0.94459867},
  {0.68219662, 0.94459867},
  {0.68219662, 0.91687787},
  // f 643/403/370 642/401/368 191/402/369
  {0.67803729, 0.94459867},
  {0.68019724, 0.94459867},
  {0.68019724, 0.91687787},
  // f 644/405/372 643/403/370 192/404/371
  {0.67571664, 0.94459867},
  {0.67803729, 0.94459867},
  {0.67803729, 0.91687787},
  // f 645/407/374 644/405/372 193/406/373
  {0.67280614, 0.94459867},
  {0.67571664, 0.94459867},
  {0.67571664, 0.91687787},
  // f 646/409/376 645/407/374 194/408/375
  {0.66921204, 0.94459867},
  {0.67280614, 0.94459867},
  {0.67280614, 0.91687787},
  // f 647/411/378 646/409/376 195/410/377
  {0.66682070, 0.94459867},
  {0.66921204, 0.94459867},
  {0.66921204, 0.91687787},
  // f 648/413/380 647/411/378 196/412/379
  {0.66380382, 0.94459867},
  {0.66682070, 0.94459867},
  {0.66682070, 0.91687787},
  // f 649/415/382 648/413/380 197/414/381
  {0.65976560, 0.94459867},
  {0.66380382, 0.94459867},
  {0.66380382, 0.91687787},
  // f 650/417/384 649/415/382 198/416/383
  {0.65683705, 0.94459867},
  {0.65976560, 0.94459867},
  {0.65976560, 0.91687787},
  // f 651/419/386 650/417/384 199/418/385
  {0.65449470, 0.94459867},
  {0.65683705, 0.94459867},
  {0.65683705, 0.91687787},
  // f 652/421/388 651/419/386 200/420/387
  {0.65265054, 0.94459867},
  {0.65449470, 0.94459867},
  {0.65449470, 0.91687787},
  // f 653/423/390 652/421/388 201/422/389
  {0.65095741, 0.94459867},
  {0.65265054, 0.94459867},
  {0.65265054, 0.91687787},
  // f 654/425/392 653/423/390 202/424/391
  {0.64966178, 0.94459867},
  {0.65095741, 0.94459867},
  {0.65095741, 0.91687787},
  // f 655/427/394 654/425/392 203/426/393
  {0.64870071, 0.94459867},
  {0.64966178, 0.94459867},
  {0.64966178, 0.91687787},
  // f 656/429/396 655/427/394 204/428/395
  {0.64780688, 0.94459867},
  {0.64870071, 0.94459867},
  {0.64870071, 0.91687787},
  // f 657/431/398 656/429/396 205/430/397
  {0.64659381, 0.94459867},
  {0.64780688, 0.94459867},
  {0.64780688, 0.91687787},
  // f 658/434/400 657/433/398 206/432/399
  {0.11271173, 0.97039855},
  {0.11271173, 0.96922553},
  {0.08518086, 0.96922553},
  // f 659/436/402 658/434/400 207/435/401
  {0.11271173, 0.97162807},
  {0.11271173, 0.97039855},
  {0.08518086, 0.97039855},
  // f 660/438/404 659/436/402 208/437/403
  {0.11271173, 0.97269595},
  {0.11271173, 0.97162807},
  {0.08518086, 0.97162807},
  // f 661/440/406 660/438/404 209/439/405
  {0.11271173, 0.97380292},
  {0.11271173, 0.97269595},
  {0.08518086, 0.97269595},
  // f 662/442/408 661/440/406 210/441/407
  {0.11271173, 0.97494936},
  {0.11271173, 0.97380292},
  {0.08518086, 0.97380292},
  // f 663/444/410 662/442/408 211/443/409
  {0.11271173, 0.9761349},
  {0.11271173, 0.97494936},
  {0.08518086, 0.97494936},
  // f 664/446/412 663/444/410 212/445/411
  {0.11271173, 0.97735977},
  {0.11271173, 0.9761349},
  {0.08518086, 0.9761349},
  // f 665/448/414 664/446/412 213/447/413
  {0.11271173, 0.97811341},
  {0.11271173, 0.97735977},
  {0.08518086, 0.97735977},
  // f 666/450/416 665/448/414 214/449/415
  {0.11271173, 0.9791404},
  {0.11271173, 0.97811341},
  {0.08518086, 0.97811341},
  // f 667/452/418 666/450/416 215/451/417
  {0.11271173, 0.98007846},
  {0.11271173, 0.9791404},
  {0.08518086, 0.9791404},
  // f 668/454/420 667/452/418 216/453/419
  {0.11271173, 0.98099744},
  {0.11271173, 0.98007846},
  {0.08518086, 0.98007846},
  // f 669/456/422 668/454/420 217/455/421
  {0.11271173, 0.98189759},
  {0.11271173, 0.98099744},
  {0.08518086, 0.98099744},
  // f 670/458/424 669/456/422 218/457/423
  {0.11271173, 0.98277879},
  {0.11271173, 0.98189759},
  {0.08518086, 0.98189759},
  // f 671/460/426 670/458/424 219/459/425
  {0.11271173, 0.98364103},
  {0.11271173, 0.98277879},
  {0.08518086, 0.98277879},
  // f 672/462/428 671/460/426 220/461/427
  {0.11271173, 0.9844842},
  {0.11271173, 0.98364103},
  {0.08518086, 0.98364103},
  // f 673/464/429 672/462/428 221/463/336
  {0.11271173, 0.98547113},
  {0.11271173, 0.9844842},
  {0.08518086, 0.9844842},
  // f 674/466/431 673/464/429 222/465/430
  {0.11271173, 0.98643076},
  {0.11271173, 0.98547113},
  {0.08518086, 0.98547113},
  // f 675/469/433 674/468/431 223/467/432
  {0.64852452, 0.68083084},
  {0.64789993, 0.68083084},
  {0.64789993, 0.70855165},
  // f 676/471/435 675/469/433 224/470/434
  {0.64920592, 0.68083084},
  {0.64852452, 0.68083084},
  {0.64852452, 0.70855165},
  // f 677/473/437 676/471/435 225/472/436
  {0.64996994, 0.68083084},
  {0.64920592, 0.68083084},
  {0.64920592, 0.70855165},
  // f 678/475/439 677/473/437 226/474/438
  {0.65113783, 0.68083084},
  {0.64996994, 0.68083084},
  {0.64996994, 0.70855165},
  // f 679/477/441 678/475/439 227/476/440
  {0.65285164, 0.68083084},
  {0.65113783, 0.68083084},
  {0.65113783, 0.70855165},
  // f 680/479/443 679/477/441 228/478/442
  {0.65606111, 0.68083084},
  {0.65285164, 0.68083084},
  {0.65285164, 0.70855165},
  // f 681/481/445 680/479/443 229/480/444
  {0.66205037, 0.68083084},
  {0.65606111, 0.68083084},
  {0.65606111, 0.70855165},
  // f 682/483/447 681/481/445 230/482/446
  {0.66963899, 0.68083084},
  {0.66205037, 0.68083084},
  {0.66205037, 0.70855165},
  // f 683/485/449 682/483/447 231/484/448
  {0.70503628, 0.68083084},
  {0.66963899, 0.68083084},
  {0.66963899, 0.70855165},
  // f 684/487/451 683/485/449 232/486/450
  {0.71724188, 0.68083084},
  {0.70503628, 0.68083084},
  {0.70503628, 0.70855165},
  // f 685/489/453 684/487/451 233/488/452
  {0.72683811, 0.68083084},
  {0.71724188, 0.68083084},
  {0.71724188, 0.70855165},
  // f 686/491/455 685/489/453 234/490/454
  {0.73337132, 0.68083084},
  {0.72683811, 0.68083084},
  {0.72683811, 0.70855165},
  // f 687/493/457 686/491/455 235/492/456
  {0.73860759, 0.68083084},
  {0.73337132, 0.68083084},
  {0.73337132, 0.70855165},
  // f 688/495/459 687/493/457 236/494/458
  {0.74275011, 0.68083084},
  {0.73860759, 0.68083084},
  {0.73860759, 0.70855165},
  // f 689/497/461 688/495/459 237/496/460
  {0.74641085, 0.68083084},
  {0.74275011, 0.68083084},
  {0.74275011, 0.70855165},
  // f 690/500/463 689/499/461 238/498/462
  {0.86136985, 0.83082294},
  {0.86136985, 0.82742131},
  {0.83383900, 0.82742131},
  // f 691/502/465 690/500/463 239/501/464
  {0.86136985, 0.83284175},
  {0.86136985, 0.83082294},
  {0.83383900, 0.83082294},
  // f 692/504/467 691/502/465 240/503/466
  {0.86136985, 0.83491516},
  {0.86136985, 0.83284175},
  {0.83383900, 0.83284175},
  // f 693/506/469 692/504/467 241/505/468
  {0.86136985, 0.83919632},
  {0.86136985, 0.83491516},
  {0.83383900, 0.83491516},
  // f 694/508/471 693/506/469 242/507/470
  {0.86136985, 0.84363687},
  {0.86136985, 0.83919632},
  {0.83383900, 0.83919632},
  // f 695/510/473 694/508/471 243/509/472
  {0.86136985, 0.84823668},
  {0.86136985, 0.84363687},
  {0.83383900, 0.84363687},
  // f 696/512/475 695/510/473 244/511/474
  {0.86136985, 0.85299599},
  {0.86136985, 0.84823668},
  {0.83383900, 0.84823668},
  // f 697/514/477 696/512/475 245/513/476
  {0.86136985, 0.85854077},
  {0.86136985, 0.85299599},
  {0.83383900, 0.85299599},
  // f 698/516/479 697/514/477 246/515/478
  {0.86136985, 0.86428726},
  {0.86136985, 0.85854077},
  {0.83383900, 0.85854077},
  // f 699/518/481 698/516/479 247/517/480
  {0.86136985, 0.87031829},
  {0.86136985, 0.86428726},
  {0.83383900, 0.86428726},
  // f 700/520/483 699/518/481 248/519/482
  {0.86136985, 0.87619674},
  {0.86136985, 0.87031829},
  {0.83383900, 0.87031829},
  // f 701/522/485 700/520/483 249/521/484
  {0.86136985, 0.88192284},
  {0.86136985, 0.87619674},
  {0.83383900, 0.87619674},
  // f 702/524/487 701/522/485 250/523/486
  {0.86136985, 0.88749647},
  {0.86136985, 0.88192284},
  {0.83383900, 0.88192284},
  // f 703/526/489 702/524/487 251/525/488
  {0.86136985, 0.89291763},
  {0.86136985, 0.88749647},
  {0.83383900, 0.88749647},
  // f 704/528/491 703/526/489 252/527/490
  {0.86136985, 0.89622843},
  {0.86136985, 0.89291763},
  {0.83383900, 0.89291763},
  // f 705/530/493 704/528/491 253/529/492
  {0.86136985, 0.89947963},
  {0.86136985, 0.89622843},
  {0.83383900, 0.89622843},
  // f 706/532/495 705/530/493 254/531/494
  {0.86136985, 0.90443468},
  {0.86136985, 0.89947963},
  {0.83383900, 0.89947963},
  // f 707/534/497 706/532/495 255/533/496
  {0.86136985, 0.90899134},
  {0.86136985, 0.90443468},
  {0.83383900, 0.90443468},
  // f 708/537/499 707/536/497 256/535/498
  {0.69844711, 0.90075672},
  {0.70286930, 0.90075672},
  {0.70286930, 0.87303603},
  // f 709/539/501 708/537/499 257/538/500
  {0.69359821, 0.90075672},
  {0.69844711, 0.90075672},
  {0.69844711, 0.87303603},
  // f 710/541/503 709/539/501 258/540/502
  {0.68832260, 0.90075672},
  {0.69359821, 0.90075672},
  {0.69359821, 0.87303603},
  // f 711/543/505 710/541/503 259/542/504
  {0.68480867, 0.90075672},
  {0.68832260, 0.90075672},
  {0.68832260, 0.87303603},
  // f 712/545/507 711/543/505 260/544/506
  {0.68112803, 0.90075672},
  {0.68480867, 0.90075672},
  {0.68480867, 0.87303603},
  // f 713/547/509 712/545/507 261/546/508
  {0.67403489, 0.90075672},
  {0.68112803, 0.90075672},
  {0.68112803, 0.87303603},
  // f 714/549/511 713/547/509 262/548/510
  {0.66540784, 0.90075672},
  {0.67403489, 0.90075672},
  {0.67403489, 0.87303603},
  // f 715/551/513 714/549/511 263/550/512
  {0.65501136, 0.90075672},
  {0.66540784, 0.90075672},
  {0.66540784, 0.87303603},
  // f 716/553/515 715/551/513 264/552/514
  {0.64473563, 0.90075672},
  {0.65501136, 0.90075672},
  {0.65501136, 0.87303603},
  // f 717/555/517 716/553/515 265/554/516
  {0.63367492, 0.90075672},
  {0.64473563, 0.90075672},
  {0.64473563, 0.87303603},
  // f 718/557/519 717/555/517 266/556/518
  {0.61480123, 0.90075672},
  {0.63367492, 0.90075672},
  {0.63367492, 0.87303603},
  // f 719/559/521 718/557/519 267/558/520
  {0.60148317, 0.90075672},
  {0.61480123, 0.90075672},
  {0.61480123, 0.87303603},
  // f 720/561/523 719/559/521 268/560/522
  {0.59112841, 0.90075672},
  {0.60148317, 0.90075672},
  {0.60148317, 0.87303603},
  // f 721/563/525 720/561/523 269/562/524
  {0.58321762, 0.90075672},
  {0.59112841, 0.90075672},
  {0.59112841, 0.87303603},
  // f 722/565/527 721/563/525 270/564/526
  {0.57730544, 0.90075672},
  {0.58321762, 0.90075672},
  {0.58321762, 0.87303603},
  // f 723/567/529 722/565/527 271/566/528
  {0.57201171, 0.90075672},
  {0.57730544, 0.90075672},
  {0.57730544, 0.87303603},
  // f 724/569/531 723/567/529 272/568/530
  {0.56822187, 0.90075672},
  {0.57201171, 0.90075672},
  {0.57201171, 0.87303603},
  // f 725/571/533 724/569/531 273/570/532
  {0.56647545, 0.90075672},
  {0.56822187, 0.90075672},
  {0.56822187, 0.87303603},
  // f 726/573/535 725/571/533 274/572/534
  {0.56404120, 0.90075672},
  {0.56647545, 0.90075672},
  {0.56647545, 0.87303603},
  // f 727/576/537 726/575/535 275/574/536
  {0.92092216, 0.63657701},
  {0.92092216, 0.64224136},
  {0.94845307, 0.64224136},
  // f 728/578/539 727/576/537 276/577/538
  {0.92092216, 0.63139236},
  {0.92092216, 0.63657701},
  {0.94845307, 0.63657701},
  // f 729/580/541 728/578/539 277/579/540
  {0.92092216, 0.62590361},
  {0.92092216, 0.63139236},
  {0.94845307, 0.63139236},
  // f 730/582/543 729/580/541 278/581/542
  {0.92092216, 0.62011087},
  {0.92092216, 0.62590361},
  {0.94845307, 0.62590361},
  // f 731/584/545 730/582/543 279/583/544
  {0.92092216, 0.61296833},
  {0.92092216, 0.62011087},
  {0.94845307, 0.62011087},
  // f 732/586/547 731/584/545 280/585/546
  {0.92092216, 0.60541189},
  {0.92092216, 0.61296833},
  {0.94845307, 0.61296833},
  // f 733/588/549 732/586/547 281/587/548
  {0.92092216, 0.59976113},
  {0.92092216, 0.60541189},
  {0.94845307, 0.60541189},
  // f 530/590/551 733/588/549 282/589/550
  {0.92092216, 0.59270144},
  {0.92092216, 0.59976113},
  {0.94845307, 0.59976113},
  // f 735/593/1 734/592/1 283/591/1
  {0.34805068, 0.92830324},
  {0.28715476, 0.92830324},
  {0.28715476, 0.95602393},
  // f 736/596/71 735/595/71 284/594/71
  {0.40912628, 0.81481934},
  {0.40912628, 0.66525185},
  {0.38159543, 0.66525185},
  // f 737/599/1 736/598/1 285/597/1
  {0.39874810, 0.84061944},
  {0.30370745, 0.84061944},
  {0.30370745, 0.86834013},
  // f 738/602/71 737/601/71 286/600/71
  {0.94845307, 0.87628448},
  {0.94845307, 0.82742131},
  {0.92092216, 0.82742131},
  // f 739/605/70 738/604/70 287/603/70
  {0.13693565, 0.78065634},
  {0.29287222, 0.78065634},
  {0.29287222, 0.75293565},
  // f 734/608/72 739/607/72 288/606/72
  {0.81449449, 0.00806057000000004},
  {0.81449449, 0.20649135},
  {0.84202534, 0.20649135},
  // f 741/611/71 740/610/71 289/609/71
  {0.81782830, 0.64574683},
  {0.81782830, 0.44393218},
  {0.79029745, 0.44393218},
  // f 742/614/70 741/613/70 290/612/70
  {0.92092216, 0.46715909},
  {0.97630656, 0.46715909},
  {0.97630656, 0.43943834},
  // f 743/617/72 742/616/72 291/615/72
  {0.71888000, 0.72467268},
  {0.71888000, 0.85691488},
  {0.74641085, 0.85691488},
  // f 744/620/554 743/619/553 292/618/552
  {0.96471310, 0.3794753},
  {0.97845739, 0.3794753},
  {0.97845739, 0.35175455},
  // f 745/622/556 744/620/554 293/621/555
  {0.95852900, 0.3794753},
  {0.96471310, 0.3794753},
  {0.96471310, 0.35175455},
  // f 746/624/558 745/622/556 294/623/557
  {0.95244557, 0.3794753},
  {0.95852900, 0.3794753},
  {0.95852900, 0.35175455},
  // f 747/626/560 746/624/558 295/625/559
  {0.93828022, 0.3794753},
  {0.95244557, 0.3794753},
  {0.95244557, 0.35175455},
  // f 748/628/562 747/626/560 296/627/561
  {0.92092216, 0.3794753},
  {0.93828022, 0.3794753},
  {0.93828022, 0.35175455},
  // f 749/631/72 748/630/72 297/629/72
  {0.94511920, 0.00806057000000004},
  {0.94511920, 0.05326921},
  {0.97265011, 0.05326921},
  // f 750/634/564 749/633/563 298/632/563
  {0.93753707, 0.54885948},
  {0.92092216, 0.54885948},
  {0.92092216, 0.57658029},
  // f 751/636/566 750/634/564 299/635/565
  {0.94915140, 0.54885948},
  {0.93753707, 0.54885948},
  {0.93753707, 0.57658029},
  // f 752/638/568 751/636/566 300/637/567
  {0.95765865, 0.54885948},
  {0.94915140, 0.54885948},
  {0.94915140, 0.57658029},
  // f 753/640/570 752/638/568 301/639/569
  {0.96112555, 0.54885948},
  {0.95765865, 0.54885948},
  {0.95765865, 0.57658029},
  // f 754/642/572 753/640/570 302/641/571
  {0.96434242, 0.54885948},
  {0.96112555, 0.54885948},
  {0.96112555, 0.57658029},
  // f 755/644/574 754/642/572 303/643/573
  {0.97025728, 0.54885948},
  {0.96434242, 0.54885948},
  {0.96434242, 0.57658029},
  // f 756/647/576 755/646/574 304/645/575
  {0.08518086, 0.94730031},
  {0.08518086, 0.95310438},
  {0.11271173, 0.95310438},
  // f 757/649/578 756/647/576 305/648/577
  {0.08518086, 0.94101977},
  {0.08518086, 0.94730031},
  {0.11271173, 0.94730031},
  // f 758/651/580 757/649/578 306/650/579
  {0.08518086, 0.93426287},
  {0.08518086, 0.94101977},
  {0.11271173, 0.94101977},
  // f 759/653/582 758/651/580 307/652/581
  {0.08518086, 0.92702949},
  {0.08518086, 0.93426287},
  {0.11271173, 0.93426287},
  // f 740/656/1 759/655/1 308/654/1
  {0.60934347, 0.9607197},
  {0.56404120, 0.9607197},
  {0.56404120, 0.98844051},
  // f 761/659/72 760/658/72 309/657/72
  {0.83383900, 0.43891007},
  {0.83383900, 0.48019344},
  {0.86136985, 0.48019344},
  // f 762/662/1 761/661/1 310/660/1
  {0.81782830, 0.8764199},
  {0.72372866, 0.8764199},
  {0.72372866, 0.90414059},
  // f 763/665/71 762/664/71 311/663/71
  {0.90491146, 0.4801935},
  {0.90491146, 0.39559644},
  {0.87738061, 0.39559644},
  // f 764/668/585 763/667/584 312/666/583
  {0.28951290, 0.7368145},
  {0.31012878, 0.7368145},
  {0.31012878, 0.70909369},
  // f 765/670/587 764/668/585 313/669/586
  {0.27536368, 0.7368145},
  {0.28951290, 0.7368145},
  {0.28951290, 0.70909369},
  // f 766/672/589 765/670/587 314/671/588
  {0.26914576, 0.7368145},
  {0.27536368, 0.7368145},
  {0.27536368, 0.70909369},
  // f 767/674/591 766/672/589 315/673/590
  {0.26233962, 0.7368145},
  {0.26914576, 0.7368145},
  {0.26914576, 0.70909369},
  // f 768/676/593 767/674/591 316/675/592
  {0.25192523, 0.7368145},
  {0.26233962, 0.7368145},
  {0.26233962, 0.70909369},
  // f 769/678/595 768/676/593 317/677/594
  {0.23864965, 0.7368145},
  {0.25192523, 0.7368145},
  {0.25192523, 0.70909369},
  // f 770/680/597 769/678/595 318/679/596
  {0.22630590, 0.7368145},
  {0.23864965, 0.7368145},
  {0.23864965, 0.70909369},
  // f 771/682/599 770/680/597 319/681/598
  {0.21307173, 0.7368145},
  {0.22630590, 0.7368145},
  {0.22630590, 0.70909369},
  // f 772/684/601 771/682/599 320/683/600
  {0.19697779, 0.7368145},
  {0.21307173, 0.7368145},
  {0.21307173, 0.70909369},
  // f 773/686/603 772/684/601 321/685/602
  {0.18470772, 0.7368145},
  {0.19697779, 0.7368145},
  {0.19697779, 0.70909369},
  // f 774/688/605 773/686/603 322/687/604
  {0.17455199, 0.7368145},
  {0.18470772, 0.7368145},
  {0.18470772, 0.70909369},
  // f 775/690/607 774/688/605 323/689/606
  {0.16623035, 0.7368145},
  {0.17455199, 0.7368145},
  {0.17455199, 0.70909369},
  // f 776/692/609 775/690/607 324/691/608
  {0.16136661, 0.7368145},
  {0.16623035, 0.7368145},
  {0.16623035, 0.70909369},
  // f 777/694/611 776/692/609 325/693/610
  {0.15587258, 0.7368145},
  {0.16136661, 0.7368145},
  {0.16136661, 0.70909369},
  // f 778/696/613 777/694/611 326/695/612
  {0.14905095, 0.7368145},
  {0.15587258, 0.7368145},
  {0.15587258, 0.70909369},
  // f 779/698/615 778/696/613 327/697/614
  {0.14273864, 0.7368145},
  {0.14905095, 0.7368145},
  {0.14905095, 0.70909369},
  // f 780/700/617 779/698/615 328/699/616
  {0.13693565, 0.7368145},
  {0.14273864, 0.7368145},
  {0.14273864, 0.70909369},
  // f 781/703/619 780/702/617 329/701/618
  {0.33805382, 0.81097412},
  {0.33805382, 0.81693923},
  {0.36558470, 0.81693923},
  // f 782/705/621 781/703/619 330/704/620
  {0.33805382, 0.80440605},
  {0.33805382, 0.81097412},
  {0.36558470, 0.81097412},
  // f 783/707/623 782/705/621 331/706/622
  {0.33805382, 0.79999483},
  {0.33805382, 0.80440605},
  {0.36558470, 0.80440605},
  // f 784/709/625 783/707/623 332/708/624
  {0.33805382, 0.79534805},
  {0.33805382, 0.79999483},
  {0.36558470, 0.79999483},
  // f 785/711/627 784/709/625 333/710/626
  {0.33805382, 0.78654385},
  {0.33805382, 0.79534805},
  {0.36558470, 0.79534805},
  // f 786/713/629 785/711/627 334/712/628
  {0.33805382, 0.77728367},
  {0.33805382, 0.78654385},
  {0.36558470, 0.78654385},
  // f 787/715/631 786/713/629 335/714/630
  {0.33805382, 0.7675674},
  {0.33805382, 0.77728367},
  {0.36558470, 0.77728367},
  // f 788/717/633 787/715/631 336/716/632
  {0.33805382, 0.75623655},
  {0.33805382, 0.7675674},
  {0.36558470, 0.7675674},
  // f 789/719/635 788/717/633 337/718/634
  {0.33805382, 0.74916768},
  {0.33805382, 0.75623655},
  {0.36558470, 0.75623655},
  // f 790/721/637 789/719/635 338/720/636
  {0.33805382, 0.7406646},
  {0.33805382, 0.74916768},
  {0.36558470, 0.74916768},
  // f 791/723/639 790/721/637 339/722/638
  {0.33805382, 0.72678053},
  {0.33805382, 0.7406646},
  {0.36558470, 0.7406646},
  // f 792/725/641 791/723/639 340/724/640
  {0.33805382, 0.71486926},
  {0.33805382, 0.72678053},
  {0.36558470, 0.72678053},
  // f 793/727/643 792/725/641 341/726/642
  {0.33805382, 0.70472777},
  {0.33805382, 0.71486926},
  {0.36558470, 0.71486926},
  // f 794/729/645 793/727/643 342/728/644
  {0.33805382, 0.69617331},
  {0.33805382, 0.70472777},
  {0.36558470, 0.70472777},
  // f 795/731/647 794/729/645 343/730/646
  {0.33805382, 0.69004178},
  {0.33805382, 0.69617331},
  {0.36558470, 0.69617331},
  // f 796/733/649 795/731/647 344/732/648
  {0.33805382, 0.6841538},
  {0.33805382, 0.69004178},
  {0.36558470, 0.69004178},
  // f 797/735/651 796/733/649 345/734/650
  {0.33805382, 0.67676187},
  {0.33805382, 0.6841538},
  {0.36558470, 0.6841538},
  // f 798/737/653 797/735/651 346/736/652
  {0.33805382, 0.67077994},
  {0.33805382, 0.67676187},
  {0.36558470, 0.67676187},
  // f 799/739/655 798/737/653 347/738/654
  {0.33805382, 0.66525185},
  {0.33805382, 0.67077994},
  {0.36558470, 0.67077994},
  // f 800/742/657 799/741/655 348/740/656
  {0.14242806, 0.79677749},
  {0.13693565, 0.79677749},
  {0.13693565, 0.8244983},
  // f 801/744/659 800/742/657 349/743/658
  {0.14929873, 0.79677749},
  {0.14242806, 0.79677749},
  {0.14242806, 0.8244983},
  // f 802/746/661 801/744/659 350/745/660
  {0.15486726, 0.79677749},
  {0.14929873, 0.79677749},
  {0.14929873, 0.8244983},
  // f 803/748/663 802/746/661 351/747/662
  {0.16181409, 0.79677749},
  {0.15486726, 0.79677749},
  {0.15486726, 0.8244983},
  // f 804/750/665 803/748/663 352/749/664
  {0.16857807, 0.79677749},
  {0.16181409, 0.79677749},
  {0.16181409, 0.8244983},
  // f 805/752/667 804/750/665 353/751/666
  {0.17703494, 0.79677749},
  {0.16857807, 0.79677749},
  {0.16857807, 0.8244983},
  // f 806/754/669 805/752/667 354/753/668
  {0.18748377, 0.79677749},
  {0.17703494, 0.79677749},
  {0.17703494, 0.8244983},
  // f 807/756/671 806/754/669 355/755/670
  {0.20025687, 0.79677749},
  {0.18748377, 0.79677749},
  {0.18748377, 0.8244983},
  // f 808/758/673 807/756/671 356/757/672
  {0.21571974, 0.79677749},
  {0.20025687, 0.79677749},
  {0.20025687, 0.8244983},
  // f 809/760/675 808/758/673 357/759/674
  {0.23921777, 0.79677749},
  {0.21571974, 0.79677749},
  {0.21571974, 0.8244983},
  // f 810/762/677 809/760/675 358/761/676
  {0.25253099, 0.79677749},
  {0.23921777, 0.79677749},
  {0.23921777, 0.8244983},
  // f 811/764/679 810/762/677 359/763/678
  {0.25897923, 0.79677749},
  {0.25253099, 0.79677749},
  {0.25253099, 0.8244983},
  // f 812/766/681 811/764/679 360/765/680
  {0.26458436, 0.79677749},
  {0.25897923, 0.79677749},
  {0.25897923, 0.8244983},
  // f 813/768/683 812/766/681 361/767/682
  {0.27084333, 0.79677749},
  {0.26458436, 0.79677749},
  {0.26458436, 0.8244983},
  // f 814/770/685 813/768/683 362/769/684
  {0.27609938, 0.79677749},
  {0.27084333, 0.79677749},
  {0.27084333, 0.8244983},
  // f 815/772/687 814/770/685 363/771/686
  {0.28099832, 0.79677749},
  {0.27609938, 0.79677749},
  {0.27609938, 0.8244983},
  // f 816/774/689 815/772/687 364/773/688
  {0.28554010, 0.79677749},
  {0.28099832, 0.79677749},
  {0.28099832, 0.8244983},
  // f 817/776/691 816/774/689 365/775/690
  {0.28836957, 0.79677749},
  {0.28554010, 0.79677749},
  {0.28554010, 0.8244983},
  // f 818/779/693 817/778/691 366/777/692
  {0.98208696, 0.96019852},
  {0.98208696, 0.95682156},
  {0.95455611, 0.95682156},
  // f 819/781/695 818/779/693 367/780/694
  {0.98208696, 0.96492302},
  {0.98208696, 0.96019852},
  {0.95455611, 0.96019852},
  // f 820/783/697 819/781/695 368/782/696
  {0.98208696, 0.97068143},
  {0.98208696, 0.96492302},
  {0.95455611, 0.96492302},
  // f 821/785/699 820/783/697 369/784/698
  {0.98208696, 0.97697604},
  {0.98208696, 0.97068143},
  {0.95455611, 0.97068143},
  // f 822/787/701 821/785/699 370/786/700
  {0.98208696, 0.98380661},
  {0.98208696, 0.97697604},
  {0.95455611, 0.97697604},
  // f 823/789/703 822/787/701 371/788/702
  {0.98208696, 0.99193931},
  {0.98208696, 0.98380661},
  {0.95455611, 0.98380661},
  // f 824/792/704 823/791/704 372/790/704
  {0.80262482, 0.98454225},
  {0.86136985, 0.98454225},
  {0.86136985, 0.95682156},
  // f 825/795/707 824/794/706 373/793/705
  {0.00800536, 0.98114777},
  {0.00800536, 0.98460782},
  {0.03553623, 0.98460782},
  // f 826/797/709 825/795/707 374/796/708
  {0.00800536, 0.97795832},
  {0.00800536, 0.98114777},
  {0.03553623, 0.98114777},
  // f 827/799/711 826/797/709 375/798/710
  {0.00800536, 0.975389},
  {0.00800536, 0.97795832},
  {0.03553623, 0.97795832},
  // f 828/801/713 827/799/711 376/800/712
  {0.00800536, 0.97302675},
  {0.00800536, 0.975389},
  {0.03553623, 0.975389},
  // f 829/803/715 828/801/713 377/802/714
  {0.00800536, 0.97087145},
  {0.00800536, 0.97302675},
  {0.03553623, 0.97302675},
  // f 830/806/717 829/805/715 378/804/716
  {0.97533572, 0.42331719},
  {0.97717470, 0.42331719},
  {0.97717470, 0.39559644},
  // f 831/808/719 830/806/717 379/807/718
  {0.97300673, 0.42331719},
  {0.97533572, 0.42331719},
  {0.97533572, 0.39559644},
  // f 832/810/721 831/808/719 380/809/720
  {0.97048539, 0.42331719},
  {0.97300673, 0.42331719},
  {0.97300673, 0.39559644},
  // f 833/812/723 832/810/721 381/811/722
  {0.96777707, 0.42331719},
  {0.97048539, 0.42331719},
  {0.97048539, 0.39559644},
  // f 834/814/725 833/812/723 382/813/724
  {0.96488160, 0.42331719},
  {0.96777707, 0.42331719},
  {0.96777707, 0.39559644},
  // f 835/816/727 834/814/725 383/815/726
  {0.96134341, 0.42331719},
  {0.96488160, 0.42331719},
  {0.96488160, 0.39559644},
  // f 836/818/729 835/816/727 384/817/728
  {0.95756096, 0.42331719},
  {0.96134341, 0.42331719},
  {0.96134341, 0.39559644},
  // f 837/820/731 836/818/729 385/819/730
  {0.95456374, 0.42331719},
  {0.95756096, 0.42331719},
  {0.95756096, 0.39559644},
  // f 838/822/733 837/820/731 386/821/732
  {0.95089334, 0.42331719},
  {0.95456374, 0.42331719},
  {0.95456374, 0.39559644},
  // f 839/824/735 838/822/733 387/823/734
  {0.94620323, 0.42331719},
  {0.95089334, 0.42331719},
  {0.95089334, 0.39559644},
  // f 840/826/737 839/824/735 388/825/736
  {0.94174743, 0.42331719},
  {0.94620323, 0.42331719},
  {0.94620323, 0.39559644},
  // f 841/828/739 840/826/737 389/827/738
  {0.93752587, 0.42331719},
  {0.94174743, 0.42331719},
  {0.94174743, 0.39559644},
  // f 842/830/741 841/828/739 390/829/740
  {0.93418688, 0.42331719},
  {0.93752587, 0.42331719},
  {0.93752587, 0.39559644},
  // f 843/832/743 842/830/741 391/831/742
  {0.93101043, 0.42331719},
  {0.93418688, 0.42331719},
  {0.93418688, 0.39559644},
  // f 844/834/745 843/832/743 392/833/744
  {0.92799670, 0.42331719},
  {0.93101043, 0.42331719},
  {0.93101043, 0.39559644},
  // f 845/836/747 844/834/745 393/835/746
  {0.92514557, 0.42331719},
  {0.92799670, 0.42331719},
  {0.92799670, 0.39559644},
  // f 846/838/749 845/836/747 394/837/748
  {0.92245716, 0.42331719},
  {0.92514557, 0.42331719},
  {0.92514557, 0.39559644},
  // f 847/840/751 846/838/749 395/839/750
  {0.92092216, 0.42331719},
  {0.92245716, 0.42331719},
  {0.92245716, 0.39559644},
  // f 848/843/753 847/842/751 396/841/752
  {0.88556695, 0.01021242},
  {0.88556695, 0.00806057000000004},
  {0.85803604, 0.00806057000000004},
  // f 849/845/755 848/843/753 397/844/754
  {0.88556695, 0.01317734},
  {0.88556695, 0.01021242},
  {0.85803604, 0.01021242},
  // f 850/847/757 849/845/755 398/846/756
  {0.88556695, 0.01646852},
  {0.88556695, 0.01317734},
  {0.85803604, 0.01317734},
  // f 851/849/759 850/847/757 399/848/758
  {0.88556695, 0.02008587},
  {0.88556695, 0.01646852},
  {0.85803604, 0.01646852},
  // f 852/851/761 851/849/759 400/850/760
  {0.88556695, 0.02485734},
  {0.88556695, 0.02008587},
  {0.85803604, 0.02008587},
  // f 853/853/763 852/851/761 401/852/762
  {0.88556695, 0.0310176599999999},
  {0.88556695, 0.02485734},
  {0.85803604, 0.02485734},
  // f 854/855/765 853/853/763 402/854/764
  {0.88556695, 0.03884101},
  {0.88556695, 0.0310176599999999},
  {0.85803604, 0.0310176599999999},
  // f 855/857/767 854/855/765 403/856/766
  {0.88556695, 0.04637146},
  {0.88556695, 0.03884101},
  {0.85803604, 0.03884101},
  // f 856/859/769 855/857/767 404/858/768
  {0.88556695, 0.05454123},
  {0.88556695, 0.04637146},
  {0.85803604, 0.04637146},
  // f 857/861/771 856/859/769 405/860/770
  {0.88556695, 0.06782466},
  {0.88556695, 0.05454123},
  {0.85803604, 0.05454123},
  // f 858/863/773 857/861/771 406/862/772
  {0.88556695, 0.07736754},
  {0.88556695, 0.06782466},
  {0.85803604, 0.06782466},
  // f 859/865/775 858/863/773 407/864/774
  {0.88556695, 0.08396506},
  {0.88556695, 0.07736754},
  {0.85803604, 0.07736754},
  // f 860/867/777 859/865/775 408/866/776
  {0.88556695, 0.08904576},
  {0.88556695, 0.08396506},
  {0.85803604, 0.08396506},
  // f 861/869/779 860/867/777 409/868/778
  {0.88556695, 0.09287471},
  {0.88556695, 0.08904576},
  {0.85803604, 0.08904576},
  // f 862/871/781 861/869/779 410/870/780
  {0.88556695, 0.09633547},
  {0.88556695, 0.09287471},
  {0.85803604, 0.09287471},
  // f 863/873/783 862/871/781 411/872/782
  {0.88556695, 0.09823519},
  {0.88556695, 0.09633547},
  {0.85803604, 0.09633547},
  // f 864/875/785 863/873/783 412/874/784
  {0.88556695, 0.10056198},
  {0.88556695, 0.09823519},
  {0.85803604, 0.09823519},
  // f 865/877/787 864/875/785 413/876/786
  {0.88556695, 0.10318518},
  {0.88556695, 0.10056198},
  {0.85803604, 0.10056198},
  // f 866/880/789 865/879/787 414/878/788
  {0.39397368, 0.88446128},
  {0.39128309, 0.88446128},
  {0.39128309, 0.91218209},
  // f 867/882/791 866/880/789 415/881/790
  {0.39685380, 0.88446128},
  {0.39397368, 0.88446128},
  {0.39397368, 0.91218209},
  // f 868/884/793 867/882/791 416/883/792
  {0.39992347, 0.88446128},
  {0.39685380, 0.88446128},
  {0.39685380, 0.91218209},
  // f 869/886/795 868/884/793 417/885/794
  {0.40318269, 0.88446128},
  {0.39992347, 0.88446128},
  {0.39992347, 0.91218209},
  // f 870/888/797 869/886/795 418/887/796
  {0.40663144, 0.88446128},
  {0.40318269, 0.88446128},
  {0.40318269, 0.91218209},
  // f 871/890/799 870/888/797 419/889/798
  {0.41102013, 0.88446128},
  {0.40663144, 0.88446128},
  {0.40663144, 0.91218209},
  // f 872/892/801 871/890/799 420/891/800
  {0.41568175, 0.88446128},
  {0.41102013, 0.88446128},
  {0.41102013, 0.91218209},
  // f 873/894/803 872/892/801 421/893/802
  {0.41894117, 0.88446128},
  {0.41568175, 0.88446128},
  {0.41568175, 0.91218209},
  // f 874/896/805 873/894/803 422/895/804
  {0.42232189, 0.88446128},
  {0.41894117, 0.88446128},
  {0.41894117, 0.91218209},
  // f 875/898/807 874/896/805 423/897/806
  {0.42836711, 0.88446128},
  {0.42232189, 0.88446128},
  {0.42232189, 0.91218209},
  // f 876/900/809 875/898/807 424/899/808
  {0.43424296, 0.88446128},
  {0.42836711, 0.88446128},
  {0.42836711, 0.91218209},
  // f 877/902/811 876/900/809 425/901/810
  {0.43806610, 0.88446128},
  {0.43424296, 0.88446128},
  {0.43424296, 0.91218209},
  // f 878/904/813 877/902/811 426/903/812
  {0.44181395, 0.88446128},
  {0.43806610, 0.88446128},
  {0.43806610, 0.91218209},
  // f 879/906/815 878/904/813 427/905/814
  {0.44802558, 0.88446128},
  {0.44181395, 0.88446128},
  {0.44181395, 0.91218209},
  // f 880/908/817 879/906/815 428/907/816
  {0.45522878, 0.88446128},
  {0.44802558, 0.88446128},
  {0.44802558, 0.91218209},
  // f 881/910/819 880/908/817 429/909/818
  {0.46305358, 0.88446128},
  {0.45522878, 0.88446128},
  {0.45522878, 0.91218209},
  // f 882/913/72 881/912/72 430/911/72
  {0.91195548, 0.22056758},
  {0.91195548, 0.23924661},
  {0.93948638, 0.23924661},
  // f 760/916/70 882/915/70 431/914/70
  {0.93138647, 0.20133531},
  {0.97211820, 0.20133531},
  {0.97211820, 0.17361462},
  // f 884/919/1 883/918/1 432/917/1
  {0.19396855, 0.92830324},
  {0.13307264, 0.92830324},
  {0.13307264, 0.95602393},
  // f 885/922/71 884/921/71 433/920/71
  {0.90491146, 0.89685845},
  {0.90491146, 0.82742131},
  {0.87738061, 0.82742131},
  // f 886/925/1 885/924/1 434/923/1
  {0.54803050, 0.91687787},
  {0.48148859, 0.91687787},
  {0.48148859, 0.94459867},
  // f 887/928/72 886/927/72 435/926/72
  {0.90157765, 0.00806057000000004},
  {0.90157765, 0.07749778},
  {0.92910850, 0.07749778},
  // f 888/931/1 887/930/1 436/929/1
  {0.93854535, 0.9129796},
  {0.87738061, 0.9129796},
  {0.87738061, 0.94070041},
  // f 889/934/71 888/933/71 437/932/71
  {0.84202534, 0.42104322},
  {0.84202534, 0.2226125},
  {0.81449449, 0.2226125},
  // f 890/937/70 889/936/70 438/935/70
  {0.00800536, 0.9547503},
  {0.06917013, 0.9547503},
  {0.06917013, 0.92702949},
  // f 891/940/72 890/939/72 439/938/72
  {0.87738061, 0.49631464},
  {0.87738061, 0.57658029},
  {0.90491146, 0.57658029},
  // f 892/943/70 891/942/70 440/941/70
  {0.56404120, 0.94459867},
  {0.63058305, 0.94459867},
  {0.63058305, 0.91687787},
  // f 893/946/71 892/945/71 441/944/71
  {0.90491146, 0.3794753},
  {0.90491146, 0.29920965},
  {0.87738061, 0.29920965},
  // f 894/949/70 893/948/70 442/947/70
  {0.36406142, 0.95602393},
  {0.42495733, 0.95602393},
  {0.42495733, 0.92830324},
  // f 883/952/72 894/951/72 443/950/72
  {0.79029745, 0.66186786},
  {0.79029745, 0.86029863},
  {0.81782830, 0.86029863},
  // f 896/955/1 895/954/1 444/953/1
  {0.32204309, 0.66525185},
  {0.13693565, 0.66525185},
  {0.13693565, 0.69297254},
  // f 897/958/71 896/957/820 445/956/71
  {0.94845307, 0.53273833},
  {0.94845307, 0.48373973},
  {0.92092216, 0.48373973},
  // f 898/961/70 897/960/70 446/959/70
  {0.87738061, 0.72453833},
  {0.93948638, 0.72453833},
  {0.93948638, 0.69681752},
  // f 899/964/71 898/963/71 447/962/71
  {0.86136985, 0.64574683},
  {0.86136985, 0.49631464},
  {0.83383900, 0.49631464},
  // f 900/967/70 899/966/70 448/965/70
  {0.72372866, 0.94798255},
  {0.78462458, 0.94798255},
  {0.78462458, 0.92026174},
  // f 901/970/72 900/969/72 449/968/72
  {0.83383900, 0.66186786},
  {0.83383900, 0.81130016},
  {0.86136985, 0.81130016},
  // f 902/973/70 901/972/70 450/971/70
  {0.87738061, 0.28308851},
  {0.93948638, 0.28308851},
  {0.93948638, 0.25536776},
  // f 895/976/72 902/975/821 451/974/72
  {0.92092216, 0.76230145},
  {0.92092216, 0.81130004},
  {0.94845307, 0.81130004},
  // f 939/979/822 940/978/822 941/977/822
  {0.72441387, 0.35417777},
  {0.72441387, 0.42781097},
  {0.78557867, 0.42781097},
  // f 942/980/822 939/979/822 941/977/822
  {0.72441387, 0.31397724},
  {0.72441387, 0.35417777},
  {0.78557867, 0.42781097},
  // f 903/981/822 942/980/822 941/977/822
  {0.78557867, 0.22938025},
  {0.72441387, 0.31397724},
  {0.78557867, 0.42781097},
  // f 943/983/822 938/982/822 939/979/822
  {0.70949239, 0.31397724},
  {0.69107568, 0.35417777},
  {0.72441387, 0.35417777},
  // f 942/980/822 943/983/822 939/979/822
  {0.72441387, 0.31397724},
  {0.70949239, 0.31397724},
  {0.72441387, 0.35417777},
  // f 946/985/822 937/984/822 938/982/822
  {0.69353098, 0.31218141},
  {0.67977262, 0.35374463},
  {0.69107568, 0.35417777},
  // f 945/986/822 946/985/822 938/982/822
  {0.69708192, 0.31299186},
  {0.69353098, 0.31218141},
  {0.69107568, 0.35417777},
  // f 944/987/822 945/986/822 938/982/822
  {0.70225143, 0.31367916},
  {0.69708192, 0.31299186},
  {0.69107568, 0.35417777},
  // f 943/983/822 944/987/822 938/982/822
  {0.70949239, 0.31397724},
  {0.70225143, 0.31367916},
  {0.69107568, 0.35417777},
  // f 948/989/822 936/988/822 937/984/822
  {0.68878132, 0.31042004},
  {0.66943645, 0.35244519},
  {0.67977262, 0.35374463},
  // f 947/990/822 948/989/822 937/984/822
  {0.69081450, 0.31129456},
  {0.68878132, 0.31042004},
  {0.67977262, 0.35374463},
  // f 946/985/822 947/990/822 937/984/822
  {0.69353098, 0.31218141},
  {0.69081450, 0.31129456},
  {0.67977262, 0.35374463},
  // f 950/992/822 935/991/822 936/988/822
  {0.68596464, 0.30876458},
  {0.66118556, 0.35059762},
  {0.66943645, 0.35244519},
  // f 949/993/822 950/992/822 936/988/822
  {0.68695801, 0.30942231},
  {0.68596464, 0.30876458},
  {0.66943645, 0.35244519},
  // f 948/989/822 949/993/822 936/988/822
  {0.68878132, 0.31042004},
  {0.68695801, 0.30942231},
  {0.66943645, 0.35244519},
  // f 952/995/822 934/994/822 935/991/822
  {0.68315023, 0.30629081},
  {0.65470272, 0.34848607},
  {0.66118556, 0.35059762},
  // f 951/996/822 952/995/822 935/991/822
  {0.68475759, 0.30781859},
  {0.68315023, 0.30629081},
  {0.66118556, 0.35059762},
  // f 950/992/822 951/996/822 935/991/822
  {0.68596464, 0.30876458},
  {0.68475759, 0.30781859},
  {0.66118556, 0.35059762},
  // f 953/998/822 933/997/822 934/994/822
  {0.68174809, 0.3046636},
  {0.64876378, 0.34588724},
  {0.65470272, 0.34848607},
  // f 952/995/822 953/998/822 934/994/822
  {0.68315023, 0.30629081},
  {0.68174809, 0.3046636},
  {0.65470272, 0.34848607},
  // f 953/998/822 932/999/822 933/997/822
  {0.68174809, 0.3046636},
  {0.64423025, 0.34334934},
  {0.64876378, 0.34588724},
  // f 954/1001/822 931/1000/822 932/999/822
  {0.68073636, 0.30323172},
  {0.64007437, 0.34047306},
  {0.64423025, 0.34334934},
  // f 953/998/822 954/1001/822 932/999/822
  {0.68174809, 0.3046636},
  {0.68073636, 0.30323172},
  {0.64423025, 0.34334934},
  // f 954/1001/822 930/1002/822 931/1000/822
  {0.68073636, 0.30323172},
  {0.63629627, 0.33725834},
  {0.64007437, 0.34047306},
  // f 955/1004/822 929/1003/822 930/1002/822
  {0.67986721, 0.30173075},
  {0.63288647, 0.33374316},
  {0.63629627, 0.33725834},
  // f 954/1001/822 955/1004/822 930/1002/822
  {0.68073636, 0.30323172},
  {0.67986721, 0.30173075},
  {0.63629627, 0.33725834},
  // f 955/1004/822 928/1005/822 929/1003/822
  {0.67986721, 0.30173075},
  {0.63041711, 0.33074194},
  {0.63288647, 0.33374316},
  // f 956/1007/822 927/1006/822 928/1005/822
  {0.67914051, 0.30016077},
  {0.62765336, 0.32675415},
  {0.63041711, 0.33074194},
  // f 955/1004/822 956/1007/822 928/1005/822
  {0.67986721, 0.30173075},
  {0.67914051, 0.30016077},
  {0.63041711, 0.33074194},
  // f 956/1007/822 926/1008/822 927/1006/822
  {0.67914051, 0.30016077},
  {0.62524861, 0.32250375},
  {0.62765336, 0.32675415},
  // f 957/1010/822 925/1009/822 926/1008/822
  {0.67855626, 0.29852176},
  {0.62320274, 0.31799072},
  {0.62524861, 0.32250375},
  // f 956/1007/822 957/1010/822 926/1008/822
  {0.67914051, 0.30016077},
  {0.67855626, 0.29852176},
  {0.62524861, 0.32250375},
  // f 957/1010/822 924/1011/822 925/1009/822
  {0.67855626, 0.29852176},
  {0.62122148, 0.3122285},
  {0.62320274, 0.31799072},
  // f 958/1013/822 923/1012/822 924/1011/822
  {0.67811453, 0.29681373},
  {0.61975712, 0.30608809},
  {0.62122148, 0.3122285},
  // f 957/1010/822 958/1013/822 924/1011/822
  {0.67855626, 0.29852176},
  {0.67811453, 0.29681373},
  {0.62122148, 0.3122285},
  // f 959/1015/822 922/1014/822 923/1012/822
  {0.67777252, 0.29467291},
  {0.61870188, 0.29844642},
  {0.61975712, 0.30608809},
  // f 958/1013/822 959/1015/822 923/1012/822
  {0.67811453, 0.29681373},
  {0.67777252, 0.29467291},
  {0.61975712, 0.30608809},
  // f 961/1017/822 921/1016/822 922/1014/822
  {0.67763287, 0.29204971},
  {0.61835015, 0.29029006},
  {0.61870188, 0.29844642},
  // f 960/1018/822 961/1017/822 922/1014/822
  {0.67767853, 0.29356521},
  {0.67763287, 0.29204971},
  {0.61870188, 0.29844642},
  // f 959/1015/822 960/1018/822 922/1014/822
  {0.67777252, 0.29467291},
  {0.67767853, 0.29356521},
  {0.61870188, 0.29844642},
  // f 962/1020/822 920/1019/822 921/1016/822
  {0.67775410, 0.28949034},
  {0.61858720, 0.28346622},
  {0.61835015, 0.29029006},
  // f 961/1017/822 962/1020/822 921/1016/822
  {0.67763287, 0.29204971},
  {0.67775410, 0.28949034},
  {0.61835015, 0.29029006},
  // f 963/1022/822 919/1021/822 920/1019/822
  {0.67805094, 0.28738528},
  {0.61929864, 0.27702832},
  {0.61858720, 0.28346622},
  // f 962/1020/822 963/1022/822 920/1019/822
  {0.67775410, 0.28949034},
  {0.67805094, 0.28738528},
  {0.61858720, 0.28346622},
  // f 963/1022/822 918/1023/822 919/1021/822
  {0.67805094, 0.28738528},
  {0.62048429, 0.27097631},
  {0.61929864, 0.27702832},
  // f 964/1025/822 917/1024/822 918/1023/822
  {0.67852581, 0.28536212},
  {0.62214422, 0.26531023},
  {0.62048429, 0.27097631},
  // f 963/1022/822 964/1025/822 918/1023/822
  {0.67805094, 0.28738528},
  {0.67852581, 0.28536212},
  {0.62048429, 0.27097631},
  // f 965/1027/822 916/1026/822 917/1024/822
  {0.67917883, 0.2834208},
  {0.62427843, 0.26003009},
  {0.62214422, 0.26531023},
  // f 964/1025/822 965/1027/822 917/1024/822
  {0.67852581, 0.28536212},
  {0.67917883, 0.2834208},
  {0.62214422, 0.26531023},
  // f 965/1027/822 915/1028/822 916/1026/822
  {0.67917883, 0.2834208},
  {0.62641919, 0.2559247},
  {0.62427843, 0.26003009},
  // f 965/1027/822 914/1029/822 915/1028/822
  {0.67917883, 0.2834208},
  {0.62888926, 0.25208741},
  {0.62641919, 0.2559247},
  // f 966/1031/822 913/1030/822 914/1029/822
  {0.68000990, 0.28156132},
  {0.63168871, 0.24851811},
  {0.62888926, 0.25208741},
  // f 965/1027/822 966/1031/822 914/1029/822
  {0.67917883, 0.2834208},
  {0.68000990, 0.28156132},
  {0.62888926, 0.25208741},
  // f 966/1031/822 912/1032/822 913/1030/822
  {0.68000990, 0.28156132},
  {0.63481754, 0.24521685},
  {0.63168871, 0.24851811},
  // f 967/1034/822 911/1033/822 912/1032/822
  {0.68101907, 0.27978379},
  {0.63827705, 0.24220788},
  {0.63481754, 0.24521685},
  // f 966/1031/822 967/1034/822 912/1032/822
  {0.68000990, 0.28156132},
  {0.68101907, 0.27978379},
  {0.63481754, 0.24521685},
  // f 967/1034/822 910/1035/822 911/1033/822
  {0.68101907, 0.27978379},
  {0.64128369, 0.24002874},
  {0.63827705, 0.24220788},
  // f 967/1034/822 909/1036/822 910/1035/822
  {0.68101907, 0.27978379},
  {0.64534086, 0.23758996},
  {0.64128369, 0.24002874},
  // f 968/1038/822 908/1037/822 909/1036/822
  {0.68220633, 0.27808803},
  {0.65064782, 0.23508143},
  {0.64534086, 0.23758996},
  // f 967/1034/822 968/1038/822 909/1036/822
  {0.68101907, 0.27978379},
  {0.68220633, 0.27808803},
  {0.64534086, 0.23758996},
  // f 969/1040/822 907/1039/822 908/1037/822
  {0.68286681, 0.27727085},
  {0.65643281, 0.23302901},
  {0.65064782, 0.23508143},
  // f 968/1038/822 969/1040/822 908/1037/822
  {0.68220633, 0.27808803},
  {0.68286681, 0.27727085},
  {0.65064782, 0.23508143},
  // f 970/1042/822 906/1041/822 907/1039/822
  {0.68381661, 0.27621317},
  {0.66378635, 0.23121095},
  {0.65643281, 0.23302901},
  // f 969/1040/822 970/1042/822 907/1039/822
  {0.68286681, 0.27727085},
  {0.68381661, 0.27621317},
  {0.65643281, 0.23302901},
  // f 972/1044/822 905/1043/822 906/1041/822
  {0.68607420, 0.2743004},
  {0.67298722, 0.22989333},
  {0.66378635, 0.23121095},
  // f 971/1045/822 972/1044/822 906/1041/822
  {0.68487561, 0.27521521},
  {0.68607420, 0.2743004},
  {0.66378635, 0.23121095},
  // f 970/1042/822 971/1045/822 906/1041/822
  {0.68381661, 0.27621317},
  {0.68487561, 0.27521521},
  {0.66378635, 0.23121095},
  // f 974/1047/822 904/1046/822 905/1043/822
  {0.68928027, 0.27254617},
  {0.68435431, 0.22938025},
  {0.67298722, 0.22989333},
  // f 973/1048/822 974/1047/822 905/1043/822
  {0.68741196, 0.27346879},
  {0.68928027, 0.27254617},
  {0.67298722, 0.22989333},
  // f 972/1044/822 973/1048/822 905/1043/822
  {0.68607420, 0.2743004},
  {0.68741196, 0.27346879},
  {0.67298722, 0.22989333},
  // f 980/1049/822 903/981/822 904/1046/822
  {0.72441387, 0.26971608},
  {0.78557867, 0.22938025},
  {0.68435431, 0.22938025},
  // f 979/1050/822 980/1049/822 904/1046/822
  {0.70707268, 0.26971608},
  {0.72441387, 0.26971608},
  {0.68435431, 0.22938025},
  // f 978/1051/822 979/1050/822 904/1046/822
  {0.70188826, 0.26988238},
  {0.70707268, 0.26971608},
  {0.68435431, 0.22938025},
  // f 977/1052/822 978/1051/822 904/1046/822
  {0.69780916, 0.27030081},
  {0.70188826, 0.26988238},
  {0.68435431, 0.22938025},
  // f 976/1053/822 977/1052/822 904/1046/822
  {0.69415700, 0.27097392},
  {0.69780916, 0.27030081},
  {0.68435431, 0.22938025},
  // f 975/1054/822 976/1053/822 904/1046/822
  {0.69136626, 0.27175355},
  {0.69415700, 0.27097392},
  {0.68435431, 0.22938025},
  // f 974/1047/822 975/1054/822 904/1046/822
  {0.68928027, 0.27254617},
  {0.69136626, 0.27175355},
  {0.68435431, 0.22938025},
  // f 980/1049/822 942/980/822 903/981/822
  {0.72441387, 0.26971608},
  {0.72441387, 0.31397724},
  {0.78557867, 0.22938025},
  // f 982/1057/822 1183/1056/822 1184/1055/822
  {0.32365623, 0.57644486},
  {0.37965760, 0.59281003},
  {0.38073975, 0.58715916},
  // f 981/1058/822 982/1057/822 1184/1055/822
  {0.38159472, 0.58009946},
  {0.32365623, 0.57644486},
  {0.38073975, 0.58715916},
  // f 983/1060/822 1182/1059/822 1183/1056/822
  {0.32252833, 0.58287692},
  {0.37757799, 0.60036635},
  {0.37965760, 0.59281003},
  // f 982/1057/822 983/1060/822 1183/1056/822
  {0.32365623, 0.57644486},
  {0.32252833, 0.58287692},
  {0.37965760, 0.59281003},
  // f 984/1062/822 1181/1061/822 1182/1059/822
  {0.32124540, 0.58764148},
  {0.37483969, 0.6075089},
  {0.37757799, 0.60036635},
  // f 983/1060/822 984/1062/822 1182/1059/822
  {0.32252833, 0.58287692},
  {0.32124540, 0.58764148},
  {0.37757799, 0.60036635},
  // f 984/1062/822 1180/1063/822 1181/1061/822
  {0.32124540, 0.58764148},
  {0.37196827, 0.61330175},
  {0.37483969, 0.6075089},
  // f 985/1065/822 1179/1064/822 1180/1063/822
  {0.31982398, 0.59146667},
  {0.36861300, 0.61879039},
  {0.37196827, 0.61330175},
  // f 984/1062/822 985/1065/822 1180/1063/822
  {0.32124540, 0.58764148},
  {0.31982398, 0.59146667},
  {0.37196827, 0.61330175},
  // f 985/1065/822 1178/1066/822 1179/1064/822
  {0.31982398, 0.59146667},
  {0.36477372, 0.62397504},
  {0.36861300, 0.61879039},
  // f 986/1068/822 1177/1067/822 1178/1066/822
  {0.31834853, 0.59449863},
  {0.35968298, 0.62963951},
  {0.36477372, 0.62397504},
  // f 985/1065/822 986/1068/822 1178/1066/822
  {0.31982398, 0.59146667},
  {0.31834853, 0.59449863},
  {0.36477372, 0.62397504},
  // f 987/1070/822 1176/1069/822 1177/1067/822
  {0.31732503, 0.59621763},
  {0.35724875, 0.6319083},
  {0.35968298, 0.62963951},
  // f 986/1068/822 987/1070/822 1177/1067/822
  {0.31834853, 0.59449863},
  {0.31732503, 0.59621763},
  {0.35968298, 0.62963951},
  // f 987/1070/822 1175/1071/822 1176/1069/822
  {0.31732503, 0.59621763},
  {0.35550228, 0.63334286},
  {0.35724875, 0.6319083},
  // f 988/1073/822 1174/1072/822 1175/1071/822
  {0.31599385, 0.59810174},
  {0.35171244, 0.63602483},
  {0.35550228, 0.63334286},
  // f 987/1070/822 988/1073/822 1175/1071/822
  {0.31732503, 0.59621763},
  {0.31599385, 0.59810174},
  {0.35550228, 0.63334286},
  // f 988/1073/822 1173/1074/822 1174/1072/822
  {0.31599385, 0.59810174},
  {0.34641868, 0.6390264},
  {0.35171244, 0.63602483},
  // f 989/1076/822 1172/1075/822 1173/1074/822
  {0.31403118, 0.60039055},
  {0.34050649, 0.64163828},
  {0.34641868, 0.6390264},
  // f 988/1073/822 989/1076/822 1173/1074/822
  {0.31599385, 0.59810174},
  {0.31403118, 0.60039055},
  {0.34641868, 0.6390264},
  // f 990/1078/822 1171/1077/822 1172/1075/822
  {0.31190723, 0.60243857},
  {0.33259571, 0.6442579},
  {0.34050649, 0.64163828},
  // f 989/1076/822 990/1078/822 1172/1075/822
  {0.31403118, 0.60039055},
  {0.31190723, 0.60243857},
  {0.34050649, 0.64163828},
  // f 991/1080/822 1170/1079/822 1171/1077/822
  {0.30962196, 0.60424554},
  {0.32224095, 0.64660454},
  {0.33259571, 0.6442579},
  // f 990/1078/822 991/1080/822 1171/1077/822
  {0.31190723, 0.60243857},
  {0.30962196, 0.60424554},
  {0.33259571, 0.6442579},
  // f 993/1082/822 1169/1081/822 1170/1079/822
  {0.30456746, 0.60713673},
  {0.30892292, 0.64835095},
  {0.32224095, 0.64660454},
  // f 992/1083/822 993/1082/822 1170/1079/822
  {0.30717537, 0.6058116},
  {0.30456746, 0.60713673},
  {0.32224095, 0.64660454},
  // f 991/1080/822 992/1083/822 1170/1079/822
  {0.30962196, 0.60424554},
  {0.30717537, 0.6058116},
  {0.32224095, 0.64660454},
  // f 996/1085/822 1168/1084/822 1169/1081/822
  {0.29513821, 0.60975802},
  {0.29004923, 0.6491307},
  {0.30892292, 0.64835095},
  // f 995/1086/822 996/1085/822 1169/1081/822
  {0.29886770, 0.6090641},
  {0.29513821, 0.60975802},
  {0.30892292, 0.64835095},
  // f 994/1087/822 995/1086/822 1169/1081/822
  {0.30179822, 0.60822093},
  {0.29886770, 0.6090641},
  {0.30892292, 0.64835095},
  // f 993/1082/822 994/1087/822 1169/1081/822
  {0.30456746, 0.60713673},
  {0.30179822, 0.60822093},
  {0.30892292, 0.64835095},
  // f 999/1089/822 1167/1088/822 1168/1084/822
  {0.28460783, 0.60994875},
  {0.27898848, 0.64879763},
  {0.29004923, 0.6491307},
  // f 998/1090/822 999/1089/822 1168/1084/822
  {0.28910822, 0.61014831},
  {0.28460783, 0.60994875},
  {0.29004923, 0.6491307},
  // f 997/1091/822 998/1090/822 1168/1084/822
  {0.29252270, 0.61002791},
  {0.28910822, 0.61014831},
  {0.29004923, 0.6491307},
  // f 996/1085/822 997/1091/822 1168/1084/822
  {0.29513821, 0.60975802},
  {0.29252270, 0.61002791},
  {0.29004923, 0.6491307},
  // f 1000/1093/822 1166/1092/822 1167/1088/822
  {0.28093287, 0.60943639},
  {0.26871279, 0.64779878},
  {0.27898848, 0.64879763},
  // f 999/1089/822 1000/1093/822 1167/1088/822
  {0.28460783, 0.60994875},
  {0.28093287, 0.60943639},
  {0.27898848, 0.64879763},
  // f 1002/1095/822 1165/1094/822 1166/1092/822
  {0.27561435, 0.60793126},
  {0.25831631, 0.64593077},
  {0.26871279, 0.64779878},
  // f 1001/1096/822 1002/1095/822 1166/1092/822
  {0.27796850, 0.60872936},
  {0.27561435, 0.60793126},
  {0.26871279, 0.64779878},
  // f 1000/1093/822 1001/1096/822 1166/1092/822
  {0.28093287, 0.60943639},
  {0.27796850, 0.60872936},
  {0.26871279, 0.64779878},
  // f 1004/1098/822 1164/1097/822 1165/1094/822
  {0.27142274, 0.60580277},
  {0.24968925, 0.64353335},
  {0.25831631, 0.64593077},
  // f 1003/1099/822 1004/1098/822 1165/1094/822
  {0.27343240, 0.60695565},
  {0.27142274, 0.60580277},
  {0.25831631, 0.64593077},
  // f 1002/1095/822 1003/1099/822 1165/1094/822
  {0.27561435, 0.60793126},
  {0.27343240, 0.60695565},
  {0.25831631, 0.64593077},
  // f 1006/1101/822 1163/1100/822 1164/1097/822
  {0.26901123, 0.60398972},
  {0.24259612, 0.64080632},
  {0.24968925, 0.64353335},
  // f 1005/1102/822 1006/1101/822 1164/1097/822
  {0.27017874, 0.60493565},
  {0.26901123, 0.60398972},
  {0.24968925, 0.64353335},
  // f 1004/1098/822 1005/1102/822 1164/1097/822
  {0.27142274, 0.60580277},
  {0.27017874, 0.60493565},
  {0.24968925, 0.64353335},
  // f 1007/1104/822 1162/1103/822 1163/1100/822
  {0.26717347, 0.60222673},
  {0.23891547, 0.6390636},
  {0.24259612, 0.64080632},
  // f 1006/1101/822 1007/1104/822 1163/1100/822
  {0.26901123, 0.60398972},
  {0.26717347, 0.60222673},
  {0.24259612, 0.64080632},
  // f 1008/1106/822 1161/1105/822 1162/1103/822
  {0.26581845, 0.60065448},
  {0.23540153, 0.63716531},
  {0.23891547, 0.6390636},
  // f 1007/1104/822 1008/1106/822 1162/1103/822
  {0.26717347, 0.60222673},
  {0.26581845, 0.60065448},
  {0.23891547, 0.6390636},
  // f 1008/1106/822 1160/1107/822 1161/1105/822
  {0.26581845, 0.60065448},
  {0.23012595, 0.63380408},
  {0.23540153, 0.63716531},
  // f 1009/1109/822 1159/1108/822 1160/1107/822
  {0.26466668, 0.59902573},
  {0.22527704, 0.63004446},
  {0.23012595, 0.63380408},
  // f 1008/1106/822 1009/1109/822 1160/1107/822
  {0.26581845, 0.60065448},
  {0.26466668, 0.59902573},
  {0.23012595, 0.63380408},
  // f 1010/1111/822 1158/1110/822 1159/1108/822
  {0.26386213, 0.59762514},
  {0.22085488, 0.62588632},
  {0.22527704, 0.63004446},
  // f 1009/1109/822 1010/1111/822 1159/1108/822
  {0.26466668, 0.59902573},
  {0.26386213, 0.59762514},
  {0.22527704, 0.63004446},
  // f 1011/1113/822 1157/1112/822 1158/1110/822
  {0.26319870, 0.59618545},
  {0.21685949, 0.62132978},
  {0.22085488, 0.62588632},
  // f 1010/1111/822 1011/1113/822 1158/1110/822
  {0.26386213, 0.59762514},
  {0.26319870, 0.59618545},
  {0.22085488, 0.62588632},
  // f 1012/1115/822 1156/1114/822 1157/1112/822
  {0.26267648, 0.5947063},
  {0.21329080, 0.61637473},
  {0.21685949, 0.62132978},
  // f 1011/1113/822 1012/1115/822 1157/1112/822
  {0.26319870, 0.59618545},
  {0.26267648, 0.5947063},
  {0.21685949, 0.62132978},
  // f 1013/1117/822 1155/1116/822 1156/1114/822
  {0.26229534, 0.59318805},
  {0.21129858, 0.61312342},
  {0.21329080, 0.61637473},
  // f 1012/1115/822 1013/1117/822 1156/1114/822
  {0.26267648, 0.5947063},
  {0.26229534, 0.59318805},
  {0.21329080, 0.61637473},
  // f 1013/1117/822 1154/1118/822 1155/1116/822
  {0.26229534, 0.59318805},
  {0.20951606, 0.60981262},
  {0.21129858, 0.61312342},
  // f 1014/1120/822 1153/1119/822 1154/1118/822
  {0.26205537, 0.59163058},
  {0.20710024, 0.60439157},
  {0.20951606, 0.60981262},
  // f 1013/1117/822 1014/1120/822 1154/1118/822
  {0.26229534, 0.59318805},
  {0.26205537, 0.59163058},
  {0.20951606, 0.60981262},
  // f 1015/1122/822 1152/1121/822 1153/1119/822
  {0.26197916, 0.59067726},
  {0.20522125, 0.59881794},
  {0.20710024, 0.60439157},
  // f 1014/1120/822 1015/1122/822 1153/1119/822
  {0.26205537, 0.59163058},
  {0.26197916, 0.59067726},
  {0.20710024, 0.60439157},
  // f 1016/1124/822 1151/1123/822 1152/1121/822
  {0.26195377, 0.58970976},
  {0.20387909, 0.59309185},
  {0.20522125, 0.59881794},
  // f 1015/1122/822 1016/1124/822 1152/1121/822
  {0.26197916, 0.59067726},
  {0.26195377, 0.58970976},
  {0.20522125, 0.59881794},
  // f 1017/1126/822 1150/1125/822 1151/1123/822
  {0.26202098, 0.58818293},
  {0.20307383, 0.58721328},
  {0.20387909, 0.59309185},
  // f 1016/1124/822 1017/1126/822 1151/1123/822
  {0.26195377, 0.58970976},
  {0.26202098, 0.58818293},
  {0.20387909, 0.59309185},
  // f 1018/1128/822 1149/1127/822 1150/1125/822
  {0.26222262, 0.58668864},
  {0.20280541, 0.58118236},
  {0.20307383, 0.58721328},
  // f 1017/1126/822 1018/1128/822 1150/1125/822
  {0.26202098, 0.58818293},
  {0.26222262, 0.58668864},
  {0.20307383, 0.58721328},
  // f 1018/1128/822 1148/1129/822 1149/1127/822
  {0.26222262, 0.58668864},
  {0.20306019, 0.57543588},
  {0.20280541, 0.58118236},
  // f 1019/1131/822 1147/1130/822 1148/1129/822
  {0.26255870, 0.58522677},
  {0.20382461, 0.5698911},
  {0.20306019, 0.57543588},
  // f 1018/1128/822 1019/1131/822 1148/1129/822
  {0.26222262, 0.58668864},
  {0.26255870, 0.58522677},
  {0.20306019, 0.57543588},
  // f 1020/1133/822 1146/1132/822 1147/1130/822
  {0.26302919, 0.58379734},
  {0.20493186, 0.56513178},
  {0.20382461, 0.5698911},
  // f 1019/1131/822 1020/1133/822 1147/1130/822
  {0.26255870, 0.58522677},
  {0.26302919, 0.58379734},
  {0.20382461, 0.5698911},
  // f 1020/1133/822 1145/1134/822 1146/1132/822
  {0.26302919, 0.58379734},
  {0.20644176, 0.56053185},
  {0.20493186, 0.56513178},
  // f 1021/1136/822 1144/1135/822 1145/1134/822
  {0.26363412, 0.58240056},
  {0.20835426, 0.55609131},
  {0.20644176, 0.56053185},
  // f 1020/1133/822 1021/1136/822 1145/1134/822
  {0.26302919, 0.58379734},
  {0.26363412, 0.58240056},
  {0.20644176, 0.56053185},
  // f 1021/1136/822 1143/1137/822 1144/1135/822
  {0.26363412, 0.58240056},
  {0.21066946, 0.55181026},
  {0.20835426, 0.55609131},
  // f 1021/1136/822 1142/1138/822 1143/1137/822
  {0.26363412, 0.58240056},
  {0.21198793, 0.54973686},
  {0.21066946, 0.55181026},
  // f 1022/1140/822 1141/1139/822 1142/1138/822
  {0.26453745, 0.58076715},
  {0.21342684, 0.54771793},
  {0.21198793, 0.54973686},
  // f 1021/1136/822 1022/1140/822 1142/1138/822
  {0.26363412, 0.58240056},
  {0.26453745, 0.58076715},
  {0.21198793, 0.54973686},
  // f 1022/1140/822 1140/1141/822 1141/1139/822
  {0.26453745, 0.58076715},
  {0.21623480, 0.54431629},
  {0.21342684, 0.54771793},
  // f 1022/1140/822 1139/1142/822 1140/1141/822
  {0.26453745, 0.58076715},
  {0.21989553, 0.54063332},
  {0.21623480, 0.54431629},
  // f 1023/1144/822 1138/1143/822 1139/1142/822
  {0.26563439, 0.5791806},
  {0.22403808, 0.53716862},
  {0.21989553, 0.54063332},
  // f 1022/1140/822 1023/1144/822 1139/1142/822
  {0.26453745, 0.58076715},
  {0.26563439, 0.5791806},
  {0.21989553, 0.54063332},
  // f 1023/1144/822 1137/1145/822 1138/1143/822
  {0.26563439, 0.5791806},
  {0.22927430, 0.53353179},
  {0.22403808, 0.53716862},
  // f 1024/1147/822 1136/1146/822 1137/1145/822
  {0.26692492, 0.57764077},
  {0.23580751, 0.52981508},
  {0.22927430, 0.53353179},
  // f 1023/1144/822 1024/1147/822 1137/1145/822
  {0.26563439, 0.5791806},
  {0.26692492, 0.57764077},
  {0.22927430, 0.53353179},
  // f 1026/1149/822 1135/1148/822 1136/1146/822
  {0.26986417, 0.57493329},
  {0.24540374, 0.52557111},
  {0.23580751, 0.52981508},
  // f 1025/1150/822 1026/1149/822 1136/1146/822
  {0.26867515, 0.57590353},
  {0.26986417, 0.57493329},
  {0.23580751, 0.52981508},
  // f 1024/1147/822 1025/1150/822 1136/1146/822
  {0.26692492, 0.57764077},
  {0.26867515, 0.57590353},
  {0.23580751, 0.52981508},
  // f 1027/1152/822 1134/1151/822 1135/1148/822
  {0.27166981, 0.57373261},
  {0.25760937, 0.52135062},
  {0.24540374, 0.52557111},
  // f 1026/1149/822 1027/1152/822 1135/1148/822
  {0.26986417, 0.57493329},
  {0.27166981, 0.57373261},
  {0.24540374, 0.52557111},
  // f 1030/1154/822 1133/1153/822 1134/1151/822
  {0.28833970, 0.56725943},
  {0.29300663, 0.51228654},
  {0.25760937, 0.52135062},
  // f 1029/1155/822 1030/1154/822 1134/1151/822
  {0.27983269, 0.56998169},
  {0.28833970, 0.56725943},
  {0.25760937, 0.52135062},
  // f 1028/1156/822 1029/1155/822 1134/1151/822
  {0.27479395, 0.57207453},
  {0.27983269, 0.56998169},
  {0.25760937, 0.52135062},
  // f 1027/1152/822 1028/1156/822 1134/1151/822
  {0.27166981, 0.57373261},
  {0.27479395, 0.57207453},
  {0.25760937, 0.52135062},
  // f 1031/1158/822 1132/1157/822 1133/1153/822
  {0.29986247, 0.56439829},
  {0.30059528, 0.5104903},
  {0.29300663, 0.51228654},
  // f 1030/1154/822 1031/1158/822 1133/1153/822
  {0.28833970, 0.56725943},
  {0.29986247, 0.56439829},
  {0.29300663, 0.51228654},
  // f 1032/1160/822 1131/1159/822 1132/1157/822
  {0.32182935, 0.55842102},
  {0.30658454, 0.50862086},
  {0.30059528, 0.5104903},
  // f 1031/1158/822 1032/1160/822 1132/1157/822
  {0.29986247, 0.56439829},
  {0.32182935, 0.55842102},
  {0.30059528, 0.5104903},
  // f 1032/1160/822 1130/1161/822 1131/1159/822
  {0.32182935, 0.55842102},
  {0.30979401, 0.50728381},
  {0.30658454, 0.50862086},
  // f 1033/1163/822 1129/1162/822 1130/1161/822
  {0.33530867, 0.55335629},
  {0.31150782, 0.50637293},
  {0.30979401, 0.50728381},
  // f 1032/1160/822 1033/1163/822 1130/1161/822
  {0.32182935, 0.55842102},
  {0.33530867, 0.55335629},
  {0.30979401, 0.50728381},
  // f 1033/1163/822 1128/1164/822 1129/1162/822
  {0.33530867, 0.55335629},
  {0.31267574, 0.50560188},
  {0.31150782, 0.50637293},
  // f 1034/1166/822 1127/1165/822 1128/1164/822
  {0.34468210, 0.54878366},
  {0.31343970, 0.50497735},
  {0.31267574, 0.50560188},
  // f 1033/1163/822 1034/1166/822 1128/1164/822
  {0.33530867, 0.55335629},
  {0.34468210, 0.54878366},
  {0.31267574, 0.50560188},
  // f 1034/1166/822 1126/1167/822 1127/1165/822
  {0.34468210, 0.54878366},
  {0.31412113, 0.50436485},
  {0.31343970, 0.50497735},
  // f 1034/1166/822 1125/1168/822 1126/1167/822
  {0.34468210, 0.54878366},
  {0.31474572, 0.50374031},
  {0.31412113, 0.50436485},
  // f 1035/1170/822 1124/1169/822 1125/1168/822
  {0.35094273, 0.54496038},
  {0.31557614, 0.50278068},
  {0.31474572, 0.50374031},
  // f 1034/1166/822 1035/1170/822 1125/1168/822
  {0.34468210, 0.54878366},
  {0.35094273, 0.54496038},
  {0.31474572, 0.50374031},
  // f 1037/1172/822 1123/1171/822 1124/1169/822
  {0.35699442, 0.54030502},
  {0.31627882, 0.50179374},
  {0.31557614, 0.50278068},
  // f 1036/1173/822 1037/1172/822 1124/1169/822
  {0.35411575, 0.54267085},
  {0.35699442, 0.54030502},
  {0.31557614, 0.50278068},
  // f 1035/1170/822 1036/1173/822 1124/1169/822
  {0.35094273, 0.54496038},
  {0.35411575, 0.54267085},
  {0.31557614, 0.50278068},
  // f 1039/1175/822 1122/1174/822 1123/1171/822
  {0.36426255, 0.53261769},
  {0.31676680, 0.50095057},
  {0.31627882, 0.50179374},
  // f 1038/1176/822 1039/1175/822 1123/1171/822
  {0.36087629, 0.53658032},
  {0.36426255, 0.53261769},
  {0.31627882, 0.50179374},
  // f 1037/1172/822 1038/1176/822 1123/1171/822
  {0.35699442, 0.54030502},
  {0.36087629, 0.53658032},
  {0.31627882, 0.50179374},
  // f 1041/1178/822 1121/1177/822 1122/1174/822
  {0.36880517, 0.52548504},
  {0.31716606, 0.50008833},
  {0.31676680, 0.50095057},
  // f 1040/1179/822 1041/1178/822 1122/1174/822
  {0.36670592, 0.52913404},
  {0.36880517, 0.52548504},
  {0.31676680, 0.50095057},
  // f 1039/1175/822 1040/1179/822 1122/1174/822
  {0.36426255, 0.53261769},
  {0.36670592, 0.52913404},
  {0.31676680, 0.50095057},
  // f 1043/1181/822 1120/1180/822 1121/1177/822
  {0.37197116, 0.51769185},
  {0.31747663, 0.49920714},
  {0.31716606, 0.50008833},
  // f 1042/1182/822 1043/1181/822 1121/1177/822
  {0.37056026, 0.52167106},
  {0.37197116, 0.51769185},
  {0.31716606, 0.50008833},
  // f 1041/1178/822 1042/1182/822 1121/1177/822
  {0.36880517, 0.52548504},
  {0.37056026, 0.52167106},
  {0.31716606, 0.50008833},
  // f 1045/1184/822 1119/1183/822 1120/1180/822
  {0.37386388, 0.50835645},
  {0.31769839, 0.49830699},
  {0.31747663, 0.49920714},
  // f 1044/1185/822 1045/1184/822 1120/1180/822
  {0.37303799, 0.51354754},
  {0.37386388, 0.50835645},
  {0.31747663, 0.49920714},
  // f 1043/1181/822 1044/1185/822 1120/1180/822
  {0.37197116, 0.51769185},
  {0.37303799, 0.51354754},
  {0.31747663, 0.49920714},
  // f 1047/1187/822 1118/1186/822 1119/1183/822
  {0.37420118, 0.5019995},
  {0.31783146, 0.49738801},
  {0.31769839, 0.49830699},
  // f 1046/1188/822 1047/1187/822 1119/1183/822
  {0.37409103, 0.50567162},
  {0.37420118, 0.5019995},
  {0.31769839, 0.49830699},
  // f 1045/1184/822 1046/1188/822 1119/1183/822
  {0.37386388, 0.50835645},
  {0.37409103, 0.50567162},
  {0.31769839, 0.49830699},
  // f 1049/1190/822 1117/1189/822 1118/1186/822
  {0.37329954, 0.4923839},
  {0.31787583, 0.49645001},
  {0.31783146, 0.49738801},
  // f 1048/1191/822 1049/1190/822 1118/1186/822
  {0.37397575, 0.49714839},
  {0.37329954, 0.4923839},
  {0.31783146, 0.49738801},
  // f 1047/1187/822 1048/1191/822 1118/1186/822
  {0.37420118, 0.5019995},
  {0.37397575, 0.49714839},
  {0.31783146, 0.49738801},
  // f 1050/1193/822 1116/1192/822 1117/1189/822
  {0.37217253, 0.48770601},
  {0.31784138, 0.49542302},
  {0.31787583, 0.49645001},
  // f 1049/1190/822 1050/1193/822 1117/1189/822
  {0.37329954, 0.4923839},
  {0.37217253, 0.48770601},
  {0.31787583, 0.49645001},
  // f 1051/1195/822 1115/1194/822 1116/1192/822
  {0.37059471, 0.48311478},
  {0.31777042, 0.49466926},
  {0.31784138, 0.49542302},
  // f 1050/1193/822 1051/1195/822 1116/1192/822
  {0.37217253, 0.48770601},
  {0.37059471, 0.48311478},
  {0.31784138, 0.49542302},
  // f 1053/1197/822 1114/1196/822 1115/1194/822
  {0.36706913, 0.47583878},
  {0.31756616, 0.49344444},
  {0.31777042, 0.49466926},
  // f 1052/1198/822 1053/1197/822 1115/1194/822
  {0.36856598, 0.4786101},
  {0.36706913, 0.47583878},
  {0.31777042, 0.49466926},
  // f 1051/1195/822 1052/1198/822 1115/1194/822
  {0.37059471, 0.48311478},
  {0.36856598, 0.4786101},
  {0.31777042, 0.49466926},
  // f 1055/1200/822 1113/1199/822 1114/1196/822
  {0.36275685, 0.46940637},
  {0.31725422, 0.49225885},
  {0.31756616, 0.49344444},
  // f 1054/1201/822 1055/1200/822 1114/1196/822
  {0.36539617, 0.4731012},
  {0.36275685, 0.46940637},
  {0.31756616, 0.49344444},
  // f 1053/1197/822 1054/1201/822 1114/1196/822
  {0.36706913, 0.47583878},
  {0.36539617, 0.4731012},
  {0.31756616, 0.49344444},
  // f 1056/1203/822 1112/1202/822 1113/1199/822
  {0.35976982, 0.4659397},
  {0.31683481, 0.49111253},
  {0.31725422, 0.49225885},
  // f 1055/1200/822 1056/1203/822 1113/1199/822
  {0.36275685, 0.46940637},
  {0.35976982, 0.4659397},
  {0.31725422, 0.49225885},
  // f 1058/1205/822 1111/1204/822 1112/1202/822
  {0.35163608, 0.45887268},
  {0.31630784, 0.49000543},
  {0.31683481, 0.49111253},
  // f 1057/1206/822 1058/1205/822 1112/1202/822
  {0.35593009, 0.46225721},
  {0.35163608, 0.45887268},
  {0.31683481, 0.49111253},
  // f 1056/1203/822 1057/1206/822 1112/1202/822
  {0.35976982, 0.4659397},
  {0.35593009, 0.46225721},
  {0.31683481, 0.49111253},
  // f 1059/1208/822 1110/1207/822 1111/1204/822
  {0.34688789, 0.45578617},
  {0.31567335, 0.48893762},
  {0.31630784, 0.49000543},
  // f 1058/1205/822 1059/1208/822 1111/1204/822
  {0.35163608, 0.45887268},
  {0.34688789, 0.45578617},
  {0.31630784, 0.49000543},
  // f 1060/1210/822 1109/1209/822 1110/1207/822
  {0.34302863, 0.45366681},
  {0.31477004, 0.48770809},
  {0.31567335, 0.48893762},
  // f 1059/1208/822 1060/1210/822 1110/1207/822
  {0.34688789, 0.45578617},
  {0.34302863, 0.45366681},
  {0.31567335, 0.48893762},
  // f 1061/1212/822 1108/1211/822 1109/1209/822
  {0.33891383, 0.45171505},
  {0.31371179, 0.48653501},
  {0.31477004, 0.48770809},
  // f 1060/1210/822 1061/1212/822 1109/1209/822
  {0.34302863, 0.45366681},
  {0.33891383, 0.45171505},
  {0.31477004, 0.48770809},
  // f 1062/1214/822 1107/1213/822 1108/1211/822
  {0.33292156, 0.44942379},
  {0.31249872, 0.48541856},
  {0.31371179, 0.48653501},
  // f 1061/1212/822 1062/1214/822 1108/1211/822
  {0.33891383, 0.45171505},
  {0.33292156, 0.44942379},
  {0.31371179, 0.48653501},
  // f 1062/1214/822 1106/1215/822 1107/1213/822
  {0.33292156, 0.44942379},
  {0.31160486, 0.48472202},
  {0.31249872, 0.48541856},
  // f 1063/1217/822 1105/1216/822 1106/1215/822
  {0.32537815, 0.44732237},
  {0.31064382, 0.48408359},
  {0.31160486, 0.48472202},
  // f 1062/1214/822 1063/1217/822 1106/1215/822
  {0.33292156, 0.44942379},
  {0.32537815, 0.44732237},
  {0.31160486, 0.48472202},
  // f 1063/1217/822 1104/1218/822 1105/1216/822
  {0.32537815, 0.44732237},
  {0.30934817, 0.48336715},
  {0.31064382, 0.48408359},
  // f 1064/1220/822 1103/1219/822 1104/1218/822
  {0.31500494, 0.44543892},
  {0.30765507, 0.48262709},
  {0.30934817, 0.48336715},
  // f 1063/1217/822 1064/1220/822 1104/1218/822
  {0.32537815, 0.44732237},
  {0.31500494, 0.44543892},
  {0.30934817, 0.48336715},
  // f 1064/1220/822 1102/1221/822 1103/1219/822
  {0.31500494, 0.44543892},
  {0.30581093, 0.48201776},
  {0.30765507, 0.48262709},
  // f 1064/1220/822 1101/1222/822 1102/1221/822
  {0.31500494, 0.44543892},
  {0.30346856, 0.48147178},
  {0.30581093, 0.48201776},
  // f 1065/1224/822 1100/1223/822 1101/1222/822
  {0.30336311, 0.44430888},
  {0.30054000, 0.48106551},
  {0.30346856, 0.48147178},
  // f 1064/1220/822 1065/1224/822 1101/1222/822
  {0.31500494, 0.44543892},
  {0.30336311, 0.44430888},
  {0.30346856, 0.48147178},
  // f 1065/1224/822 1099/1225/822 1100/1223/822
  {0.30336311, 0.44430888},
  {0.29650176, 0.48088413},
  {0.30054000, 0.48106551},
  // f 1066/1227/822 1098/1226/822 1099/1225/822
  {0.29045251, 0.44393218},
  {0.29348490, 0.48097181},
  {0.29650176, 0.48088413},
  // f 1065/1224/822 1066/1227/822 1099/1225/822
  {0.30336311, 0.44430888},
  {0.29045251, 0.44393218},
  {0.29650176, 0.48088413},
  // f 1066/1227/822 1097/1228/822 1098/1226/822
  {0.29045251, 0.44393218},
  {0.29109356, 0.48117888},
  {0.29348490, 0.48097181},
  // f 1066/1227/822 1096/1229/822 1097/1228/822
  {0.29045251, 0.44393218},
  {0.28749946, 0.48176366},
  {0.29109356, 0.48117888},
  // f 1067/1231/822 1095/1230/822 1096/1229/822
  {0.27470043, 0.44461673},
  {0.28458896, 0.48253113},
  {0.28749946, 0.48176366},
  // f 1066/1227/822 1067/1231/822 1096/1229/822
  {0.29045251, 0.44393218},
  {0.27470043, 0.44461673},
  {0.28749946, 0.48176366},
  // f 1067/1231/822 1094/1232/822 1095/1230/822
  {0.27470043, 0.44461673},
  {0.28226829, 0.48337901},
  {0.28458896, 0.48253113},
  // f 1067/1231/822 1093/1233/822 1094/1232/822
  {0.27470043, 0.44461673},
  {0.28010836, 0.48440224},
  {0.28226829, 0.48337901},
  // f 1068/1235/822 1092/1234/822 1093/1233/822
  {0.26310447, 0.4461953},
  {0.27810895, 0.48560095},
  {0.28010836, 0.48440224},
  // f 1067/1231/822 1068/1235/822 1093/1233/822
  {0.27470043, 0.44461673},
  {0.26310447, 0.4461953},
  {0.28010836, 0.48440224},
  // f 1068/1235/822 1091/1236/822 1092/1234/822
  {0.26310447, 0.4461953},
  {0.27627036, 0.48697513},
  {0.27810895, 0.48560095},
  // f 1069/1238/822 1090/1237/822 1091/1236/822
  {0.25377321, 0.44836789},
  {0.27513039, 0.4879964},
  {0.27627036, 0.48697513},
  // f 1068/1235/822 1069/1238/822 1091/1236/822
  {0.26310447, 0.4461953},
  {0.25377321, 0.44836789},
  {0.27627036, 0.48697513},
  // f 1069/1238/822 1089/1239/822 1090/1237/822
  {0.25377321, 0.44836789},
  {0.27405500, 0.4891113},
  {0.27513039, 0.4879964},
  // f 1070/1241/822 1088/1240/822 1089/1239/822
  {0.24636212, 0.450863},
  {0.27256286, 0.49095911},
  {0.27405500, 0.4891113},
  // f 1069/1238/822 1070/1241/822 1089/1239/822
  {0.25377321, 0.44836789},
  {0.24636212, 0.450863},
  {0.27405500, 0.4891113},
  // f 1072/1243/822 1087/1242/822 1088/1240/822
  {0.23699513, 0.45538932},
  {0.27121586, 0.49301738},
  {0.27256286, 0.49095911},
  // f 1071/1244/822 1072/1243/822 1088/1240/822
  {0.24056965, 0.45344299},
  {0.23699513, 0.45538932},
  {0.27256286, 0.49095911},
  // f 1070/1241/822 1071/1244/822 1088/1240/822
  {0.24636212, 0.450863},
  {0.24056965, 0.45344299},
  {0.27256286, 0.49095911},
  // f 1074/1246/822 1086/1245/822 1087/1242/822
  {0.22904509, 0.46107286},
  {0.27001408, 0.49528611},
  {0.27121586, 0.49301738},
  // f 1073/1247/822 1074/1246/822 1087/1242/822
  {0.23285009, 0.45807678},
  {0.22904509, 0.46107286},
  {0.27121586, 0.49301738},
  // f 1072/1243/822 1073/1247/822 1087/1242/822
  {0.23699513, 0.45538932},
  {0.23285009, 0.45807678},
  {0.27121586, 0.49301738},
  // f 1076/1249/822 1085/1248/822 1086/1245/822
  {0.22239895, 0.468068},
  {0.26879549, 0.49819905},
  {0.27001408, 0.49528611},
  // f 1075/1250/822 1076/1249/822 1086/1245/822
  {0.22556140, 0.46440327},
  {0.22239895, 0.468068},
  {0.27001408, 0.49528611},
  // f 1074/1246/822 1075/1250/822 1086/1245/822
  {0.22904509, 0.46107286},
  {0.22556140, 0.46440327},
  {0.27001408, 0.49528611},
  // f 1079/1252/822 1084/1251/822 1085/1248/822
  {0.21443824, 0.48204184},
  {0.26764476, 0.50187898},
  {0.26879549, 0.49819905},
  // f 1078/1253/822 1079/1252/822 1085/1248/822
  {0.21703799, 0.47640049},
  {0.21443824, 0.48204184},
  {0.26879549, 0.49819905},
  // f 1077/1254/822 1078/1253/822 1085/1248/822
  {0.21955784, 0.47206712},
  {0.21703799, 0.47640049},
  {0.26879549, 0.49819905},
  // f 1076/1249/822 1077/1254/822 1085/1248/822
  {0.22239895, 0.468068},
  {0.21955784, 0.47206712},
  {0.26879549, 0.49819905},
  // f 1080/1256/822 1083/1255/822 1084/1251/822
  {0.21230118, 0.48816466},
  {0.26665875, 0.50646627},
  {0.26764476, 0.50187898},
  // f 1079/1252/822 1080/1256/822 1084/1251/822
  {0.21443824, 0.48204184},
  {0.21230118, 0.48816466},
  {0.26764476, 0.50187898},
  // f 1081/1258/822 1082/1257/822 1083/1255/822
  {0.21062674, 0.49476892},
  {0.20925796, 0.50308239},
  {0.26665875, 0.50646627},
  // f 1080/1256/822 1081/1258/822 1083/1255/822
  {0.21230118, 0.48816466},
  {0.21062674, 0.49476892},
  {0.26665875, 0.50646627},
  // f 1187/1261/822 1189/1260/822 1190/1259/822
  {0.71339083, 0.59349966},
  {0.61835015, 0.64236283},
  {0.77428669, 0.64236283},
  // f 1185/1262/822 1187/1261/822 1190/1259/822
  {0.77428669, 0.44393218},
  {0.71339083, 0.59349966},
  {0.77428669, 0.64236283},
  // f 1187/1261/822 1188/1263/822 1189/1260/822
  {0.71339083, 0.59349966},
  {0.61835015, 0.59349966},
  {0.61835015, 0.64236283},
  // f 1185/1262/822 1186/1264/822 1187/1261/822
  {0.77428669, 0.44393218},
  {0.71339083, 0.44393218},
  {0.71339083, 0.59349966},
  // f 1191/1267/822 1209/1266/822 1210/1265/822
  {0.43511090, 0.65848398},
  {0.48417714, 0.66571736},
  {0.48041314, 0.65848398},
  // f 1191/1267/822 1208/1268/822 1209/1266/822
  {0.43511090, 0.65848398},
  {0.48847884, 0.67247438},
  {0.48417714, 0.66571736},
  // f 1194/1270/822 1207/1269/822 1208/1268/822
  {0.49049526, 0.72805655},
  {0.49331826, 0.67875481},
  {0.48847884, 0.67247438},
  // f 1191/1267/822 1194/1270/822 1208/1268/822
  {0.43511090, 0.65848398},
  {0.49049526, 0.72805655},
  {0.48847884, 0.67247438},
  // f 1194/1270/822 1206/1271/822 1207/1269/822
  {0.49049526, 0.72805655},
  {0.49869537, 0.68455887},
  {0.49331826, 0.67875481},
  // f 1194/1270/822 1205/1272/822 1206/1271/822
  {0.49049526, 0.72805655},
  {0.50461018, 0.68988645},
  {0.49869537, 0.68455887},
  // f 1194/1270/822 1204/1273/822 1205/1272/822
  {0.49049526, 0.72805655},
  {0.50782704, 0.69239461},
  {0.50461018, 0.68988645},
  // f 1194/1270/822 1203/1274/822 1204/1273/822
  {0.49049526, 0.72805655},
  {0.51129395, 0.69482958},
  {0.50782704, 0.69239461},
  // f 1194/1270/822 1202/1275/822 1203/1274/822
  {0.49049526, 0.72805655},
  {0.51980126, 0.69992948},
  {0.51129395, 0.69482958},
  // f 1194/1270/822 1195/1276/822 1202/1275/822
  {0.49049526, 0.72805655},
  {0.50423950, 0.73769367},
  {0.51980126, 0.69992948},
  // f 1195/1276/822 1201/1277/822 1202/1275/822
  {0.50423950, 0.73769367},
  {0.53141558, 0.70550036},
  {0.51980126, 0.69992948},
  // f 1197/1279/822 1198/1278/822 1201/1277/822
  {0.51650709, 0.74470532},
  {0.53067243, 0.75093174},
  {0.53141558, 0.70550036},
  // f 1196/1280/822 1197/1279/822 1201/1277/822
  {0.51042360, 0.74142563},
  {0.51650709, 0.74470532},
  {0.53141558, 0.70550036},
  // f 1195/1276/822 1196/1280/822 1201/1277/822
  {0.50423950, 0.73769367},
  {0.51042360, 0.74142563},
  {0.53141558, 0.70550036},
  // f 1198/1278/822 1200/1281/822 1201/1277/822
  {0.53067243, 0.75093174},
  {0.54803050, 0.71167862},
  {0.53141558, 0.70550036},
  // f 1198/1278/822 1199/1282/822 1200/1281/822
  {0.53067243, 0.75093174},
  {0.54803050, 0.75688732},
  {0.54803050, 0.71167862},
  // f 1192/1284/822 1193/1283/822 1194/1270/822
  {0.43511090, 0.86029863},
  {0.49049526, 0.86029863},
  {0.49049526, 0.72805655},
  // f 1191/1267/822 1192/1284/822 1194/1270/822
  {0.43511090, 0.65848398},
  {0.43511090, 0.86029863},
  {0.49049526, 0.72805655},
  // f 1214/1287/822 1332/1286/822 1333/1285/822
  {0.00800536, 0.18131524},
  {0.06137330, 0.15668058},
  {0.06137330, 0.13800156},
  // f 1213/1288/822 1214/1287/822 1333/1285/822
  {0.00800536, 0.09671825},
  {0.00800536, 0.18131524},
  {0.06137330, 0.13800156},
  // f 1212/1289/822 1213/1288/822 1333/1285/822
  {0.10210501, 0.09671825},
  {0.00800536, 0.09671825},
  {0.06137330, 0.13800156},
  // f 1211/1290/822 1212/1289/822 1333/1285/822
  {0.10210501, 0.13800156},
  {0.10210501, 0.09671825},
  {0.06137330, 0.13800156},
  // f 1215/1292/822 1216/1291/822 1332/1286/822
  {0.02862125, 0.19425523},
  {0.04277048, 0.201473},
  {0.06137330, 0.15668058},
  // f 1214/1287/822 1215/1292/822 1332/1286/822
  {0.00800536, 0.18131524},
  {0.02862125, 0.19425523},
  {0.06137330, 0.15668058},
  // f 1216/1291/822 1331/1293/822 1332/1286/822
  {0.04277048, 0.201473},
  {0.06919809, 0.16107374},
  {0.06137330, 0.15668058},
  // f 1217/1295/822 1218/1294/822 1331/1293/822
  {0.04898838, 0.20411181},
  {0.05579454, 0.206559},
  {0.06919809, 0.16107374},
  // f 1216/1291/822 1217/1295/822 1331/1293/822
  {0.04277048, 0.201473},
  {0.04898838, 0.20411181},
  {0.06919809, 0.16107374},
  // f 1218/1294/822 1330/1296/822 1331/1293/822
  {0.05579454, 0.206559},
  {0.07640129, 0.16449678},
  {0.06919809, 0.16107374},
  // f 1218/1294/822 1219/1297/822 1330/1296/822
  {0.05579454, 0.206559},
  {0.06620893, 0.20938909},
  {0.07640129, 0.16449678},
  // f 1219/1297/822 1329/1298/822 1330/1296/822
  {0.06620893, 0.20938909},
  {0.08261294, 0.16683221},
  {0.07640129, 0.16449678},
  // f 1219/1297/822 1220/1299/822 1329/1298/822
  {0.06620893, 0.20938909},
  {0.07948450, 0.2117154},
  {0.08261294, 0.16683221},
  // f 1220/1299/822 1328/1300/822 1329/1298/822
  {0.07948450, 0.2117154},
  {0.08636079, 0.16790426},
  {0.08261294, 0.16683221},
  // f 1220/1299/822 1327/1301/822 1328/1300/822
  {0.07948450, 0.2117154},
  {0.09018393, 0.16873801},
  {0.08636079, 0.16790426},
  // f 1221/1303/822 1326/1302/822 1327/1301/822
  {0.09182826, 0.21287316},
  {0.09605978, 0.16954207},
  {0.09018393, 0.16873801},
  // f 1220/1299/822 1221/1303/822 1327/1301/822
  {0.07948450, 0.2117154},
  {0.09182826, 0.21287316},
  {0.09018393, 0.16873801},
  // f 1221/1303/822 1325/1304/822 1326/1302/822
  {0.09182826, 0.21287316},
  {0.10210501, 0.16981006},
  {0.09605978, 0.16954207},
  // f 1222/1306/822 1324/1305/822 1325/1304/822
  {0.10506243, 0.2132591},
  {0.10548571, 0.16972172},
  {0.10210501, 0.16981006},
  // f 1221/1303/822 1222/1306/822 1325/1304/822
  {0.09182826, 0.21287316},
  {0.10506243, 0.2132591},
  {0.10210501, 0.16981006},
  // f 1222/1306/822 1323/1307/822 1324/1305/822
  {0.10506243, 0.2132591},
  {0.10874511, 0.1694566},
  {0.10548571, 0.16972172},
  // f 1223/1309/822 1322/1308/822 1323/1307/822
  {0.12115636, 0.21256518},
  {0.11340676, 0.16872764},
  {0.10874511, 0.1694566},
  // f 1222/1306/822 1223/1309/822 1323/1307/822
  {0.10506243, 0.2132591},
  {0.12115636, 0.21256518},
  {0.10874511, 0.1694566},
  // f 1223/1309/822 1321/1310/822 1322/1308/822
  {0.12115636, 0.21256518},
  {0.11779546, 0.16760105},
  {0.11340676, 0.16872764},
  // f 1224/1312/822 1320/1311/822 1321/1310/822
  {0.13342643, 0.21092689},
  {0.12124421, 0.16635847},
  {0.11779546, 0.16760105},
  // f 1223/1309/822 1224/1312/822 1321/1310/822
  {0.12115636, 0.21256518},
  {0.13342643, 0.21092689},
  {0.11779546, 0.16760105},
  // f 1224/1312/822 1319/1313/822 1320/1311/822
  {0.13342643, 0.21092689},
  {0.12450343, 0.1648398},
  {0.12124421, 0.16635847},
  // f 1225/1315/822 1318/1314/822 1319/1313/822
  {0.14358217, 0.20862836},
  {0.12757307, 0.16304505},
  {0.12450343, 0.1648398},
  // f 1224/1312/822 1225/1315/822 1319/1313/822
  {0.13342643, 0.21092689},
  {0.14358217, 0.20862836},
  {0.12450343, 0.1648398},
  // f 1226/1317/822 1317/1316/822 1318/1314/822
  {0.15190379, 0.20592993},
  {0.13045318, 0.16097409},
  {0.12757307, 0.16304505},
  // f 1225/1315/822 1226/1317/822 1318/1314/822
  {0.14358217, 0.20862836},
  {0.15190379, 0.20592993},
  {0.12757307, 0.16304505},
  // f 1228/1319/822 1316/1318/822 1317/1316/822
  {0.16226156, 0.20121247},
  {0.13314378, 0.15862703},
  {0.13045318, 0.16097409},
  // f 1227/1320/822 1228/1319/822 1317/1316/822
  {0.15676755, 0.2039302},
  {0.16226156, 0.20121247},
  {0.13045318, 0.16097409},
  // f 1226/1317/822 1227/1320/822 1317/1316/822
  {0.15190379, 0.20592993},
  {0.15676755, 0.2039302},
  {0.13045318, 0.16097409},
  // f 1229/1322/822 1315/1321/822 1316/1318/822
  {0.16908321, 0.19705606},
  {0.13564481, 0.15600383},
  {0.13314378, 0.15862703},
  // f 1228/1319/822 1229/1322/822 1316/1318/822
  {0.16226156, 0.20121247},
  {0.16908321, 0.19705606},
  {0.13314378, 0.15862703},
  // f 1230/1324/822 1314/1323/822 1315/1321/822
  {0.17539552, 0.19229674},
  {0.13749284, 0.15367699},
  {0.13564481, 0.15600383},
  // f 1229/1322/822 1230/1324/822 1315/1321/822
  {0.16908321, 0.19705606},
  {0.17539552, 0.19229674},
  {0.13564481, 0.15600383},
  // f 1231/1326/822 1313/1325/822 1314/1323/822
  {0.18119849, 0.18693459},
  {0.13877778, 0.15177727},
  {0.13749284, 0.15367699},
  // f 1230/1324/822 1231/1326/822 1314/1323/822
  {0.17539552, 0.19229674},
  {0.18119849, 0.18693459},
  {0.13749284, 0.15367699},
  // f 1232/1328/822 1312/1327/822 1313/1325/822
  {0.18649217, 0.18096936},
  {0.14072683, 0.1483165},
  {0.13877778, 0.15177727},
  // f 1231/1326/822 1232/1328/822 1313/1325/822
  {0.18119849, 0.18693459},
  {0.18649217, 0.18096936},
  {0.13877778, 0.15177727},
  // f 1233/1330/822 1311/1329/822 1312/1327/822
  {0.19127651, 0.17440128},
  {0.14243528, 0.14448762},
  {0.14072683, 0.1483165},
  // f 1232/1328/822 1233/1330/822 1312/1327/822
  {0.18649217, 0.18096936},
  {0.19127651, 0.17440128},
  {0.14072683, 0.1483165},
  // f 1235/1332/822 1310/1331/822 1311/1329/822
  {0.19654071, 0.16534334},
  {0.14416780, 0.13940692},
  {0.14243528, 0.14448762},
  // f 1234/1333/822 1235/1332/822 1311/1329/822
  {0.19400808, 0.16999006},
  {0.19654071, 0.16534334},
  {0.14243528, 0.14448762},
  // f 1233/1330/822 1234/1333/822 1311/1329/822
  {0.19127651, 0.17440128},
  {0.19400808, 0.16999006},
  {0.14243528, 0.14448762},
  // f 1236/1335/822 1309/1334/822 1310/1331/822
  {0.20052621, 0.1565392},
  {0.14575112, 0.1328094},
  {0.14416780, 0.13940692},
  // f 1235/1332/822 1236/1335/822 1310/1331/822
  {0.19654071, 0.16534334},
  {0.20052621, 0.1565392},
  {0.14416780, 0.13940692},
  // f 1238/1337/822 1308/1336/822 1309/1334/822
  {0.20613219, 0.13756263},
  {0.14709380, 0.12326646},
  {0.14575112, 0.1328094},
  // f 1237/1338/822 1238/1337/822 1309/1334/822
  {0.20372337, 0.1472789},
  {0.20613219, 0.13756263},
  {0.14575112, 0.1328094},
  // f 1236/1335/822 1237/1338/822 1309/1334/822
  {0.20052621, 0.1565392},
  {0.20372337, 0.1472789},
  {0.14575112, 0.1328094},
  // f 1240/1340/822 1307/1339/822 1308/1336/822
  {0.20846801, 0.11916298},
  {0.14767613, 0.10998309},
  {0.14709380, 0.12326646},
  // f 1239/1341/822 1240/1340/822 1308/1336/822
  {0.20788404, 0.12623173},
  {0.20846801, 0.11916298},
  {0.14709380, 0.12326646},
  // f 1238/1337/822 1239/1341/822 1308/1336/822
  {0.20613219, 0.13756263},
  {0.20788404, 0.12623173},
  {0.14709380, 0.12326646},
  // f 1242/1343/822 1306/1342/822 1307/1339/822
  {0.20806234, 0.09677577},
  {0.14744294, 0.10181326},
  {0.14767613, 0.10998309},
  // f 1241/1344/822 1242/1343/822 1307/1339/822
  {0.20870645, 0.11065984},
  {0.20806234, 0.09677577},
  {0.14767613, 0.10998309},
  // f 1240/1340/822 1241/1344/822 1307/1339/822
  {0.20846801, 0.11916298},
  {0.20870645, 0.11065984},
  {0.14767613, 0.10998309},
  // f 1242/1343/822 1305/1345/822 1306/1342/822
  {0.20806234, 0.09677577},
  {0.14674342, 0.0942828100000001},
  {0.14744294, 0.10181326},
  // f 1244/1347/822 1304/1346/822 1305/1345/822
  {0.20391543, 0.07472301},
  {0.14537288, 0.08645952},
  {0.14674342, 0.0942828100000001},
  // f 1243/1348/822 1244/1347/822 1305/1345/822
  {0.20635885, 0.0848645},
  {0.20391543, 0.07472301},
  {0.14674342, 0.0942828100000001},
  // f 1242/1343/822 1243/1348/822 1305/1345/822
  {0.20806234, 0.09677577},
  {0.20635885, 0.0848645},
  {0.14674342, 0.0942828100000001},
  // f 1245/1350/822 1303/1349/822 1304/1346/822
  {0.20101954, 0.06616855},
  {0.14367402, 0.08029914},
  {0.14537288, 0.08645952},
  // f 1244/1347/822 1245/1350/822 1304/1346/822
  {0.20391543, 0.07472301},
  {0.20101954, 0.06616855},
  {0.14537288, 0.08645952},
  // f 1247/1352/822 1302/1351/822 1303/1349/822
  {0.19539808, 0.05414903},
  {0.14184664, 0.07552773},
  {0.14367402, 0.08029914},
  // f 1246/1353/822 1247/1352/822 1303/1349/822
  {0.19840044, 0.06003696},
  {0.19539808, 0.05414903},
  {0.14367402, 0.08029914},
  // f 1245/1350/822 1246/1353/822 1303/1349/822
  {0.20101954, 0.06616855},
  {0.19840044, 0.06003696},
  {0.14367402, 0.08029914},
  // f 1248/1355/822 1301/1354/822 1302/1351/822
  {0.19082280, 0.0467571},
  {0.14006212, 0.07191032},
  {0.14184664, 0.07552773},
  // f 1247/1352/822 1248/1355/822 1302/1351/822
  {0.19539808, 0.05414903},
  {0.19082280, 0.0467571},
  {0.14184664, 0.07552773},
  // f 1250/1357/822 1300/1356/822 1301/1354/822
  {0.18129237, 0.03524703},
  {0.13803966, 0.0686191900000001},
  {0.14006212, 0.07191032},
  // f 1249/1358/822 1250/1357/822 1301/1354/822
  {0.18629998, 0.04077524},
  {0.18129237, 0.03524703},
  {0.14006212, 0.07191032},
  // f 1248/1355/822 1249/1358/822 1301/1354/822
  {0.19082280, 0.0467571},
  {0.18629998, 0.04077524},
  {0.14006212, 0.07191032},
  // f 1251/1360/822 1299/1359/822 1300/1356/822
  {0.17579997, 0.03017253},
  {0.13577923, 0.06565422},
  {0.13803966, 0.0686191900000001},
  // f 1250/1357/822 1251/1360/822 1300/1356/822
  {0.18129237, 0.03524703},
  {0.17579997, 0.03017253},
  {0.13803966, 0.0686191900000001},
  // f 1252/1362/822 1298/1361/822 1299/1359/822
  {0.16892931, 0.02492851},
  {0.13382369, 0.06350243},
  {0.13577923, 0.06565422},
  // f 1251/1360/822 1252/1362/822 1299/1359/822
  {0.17579997, 0.03017253},
  {0.16892931, 0.02492851},
  {0.13577923, 0.06565422},
  // f 1253/1364/822 1297/1363/822 1298/1361/822
  {0.16336076, 0.02138436},
  {0.13228871, 0.06200624},
  {0.13382369, 0.06350243},
  // f 1252/1362/822 1253/1364/822 1298/1361/822
  {0.16892931, 0.02492851},
  {0.16336076, 0.02138436},
  {0.13382369, 0.06350243},
  // f 1254/1366/822 1296/1365/822 1297/1363/822
  {0.15641394, 0.01767081},
  {0.12960027, 0.05973673},
  {0.13228871, 0.06200624},
  // f 1253/1364/822 1254/1366/822 1297/1363/822
  {0.16336076, 0.02138436},
  {0.15641394, 0.01767081},
  {0.13228871, 0.06200624},
  // f 1255/1368/822 1295/1367/822 1296/1365/822
  {0.14964996, 0.01484156},
  {0.12674919, 0.05774736},
  {0.12960027, 0.05973673},
  // f 1254/1366/822 1255/1368/822 1296/1365/822
  {0.15641394, 0.01767081},
  {0.14964996, 0.01484156},
  {0.12960027, 0.05973673},
  // f 1256/1370/822 1294/1369/822 1295/1367/822
  {0.14119309, 0.01224679},
  {0.12373545, 0.05603826},
  {0.12674919, 0.05774736},
  // f 1255/1368/822 1256/1370/822 1295/1367/822
  {0.14964996, 0.01484156},
  {0.14119309, 0.01224679},
  {0.12674919, 0.05774736},
  // f 1256/1370/822 1293/1371/822 1294/1369/822
  {0.14119309, 0.01224679},
  {0.12055904, 0.0546093},
  {0.12373545, 0.05603826},
  // f 1257/1373/822 1292/1372/822 1293/1371/822
  {0.13074426, 0.01009411},
  {0.11721999, 0.0534606},
  {0.12055904, 0.0546093},
  // f 1256/1370/822 1257/1373/822 1293/1371/822
  {0.14119309, 0.01224679},
  {0.13074426, 0.01009411},
  {0.12055904, 0.0546093},
  // f 1258/1375/822 1291/1374/822 1292/1372/822
  {0.11797117, 0.00861411999999995},
  {0.11299841, 0.05245191},
  {0.11721999, 0.0534606},
  // f 1257/1373/822 1258/1375/822 1292/1372/822
  {0.13074426, 0.01009411},
  {0.11797117, 0.00861411999999995},
  {0.11721999, 0.0534606},
  // f 1258/1375/822 1290/1376/822 1291/1374/822
  {0.11797117, 0.00861411999999995},
  {0.10854261, 0.05184668},
  {0.11299841, 0.05245191},
  // f 1259/1378/822 1289/1377/822 1290/1376/822
  {0.10250829, 0.00806057000000004},
  {0.10385258, 0.05164498},
  {0.10854261, 0.05184668},
  // f 1258/1375/822 1259/1378/822 1290/1376/822
  {0.11797117, 0.00861411999999995},
  {0.10250829, 0.00806057000000004},
  {0.10854261, 0.05184668},
  // f 1259/1378/822 1288/1379/822 1289/1377/822
  {0.10250829, 0.00806057000000004},
  {0.10018212, 0.05176169},
  {0.10385258, 0.05164498},
  // f 1259/1378/822 1287/1380/822 1288/1379/822
  {0.10250829, 0.00806057000000004},
  {0.09718492, 0.0520475500000001},
  {0.10018212, 0.05176169},
  // f 1260/1382/822 1286/1381/822 1287/1380/822
  {0.07901026, 0.00901346999999997},
  {0.09340245, 0.0526955099999999},
  {0.09718492, 0.0520475500000001},
  // f 1259/1378/822 1260/1382/822 1287/1380/822
  {0.10250829, 0.00806057000000004},
  {0.07901026, 0.00901346999999997},
  {0.09718492, 0.0520475500000001},
  // f 1260/1382/822 1285/1383/822 1286/1381/822
  {0.07901026, 0.00901346999999997},
  {0.08986428, 0.05364841},
  {0.09340245, 0.0526955099999999},
  // f 1260/1382/822 1284/1384/822 1285/1383/822
  {0.07901026, 0.00901346999999997},
  {0.08696882, 0.05473238},
  {0.08986428, 0.05364841},
  // f 1261/1386/822 1283/1385/822 1284/1384/822
  {0.06569706, 0.01081449},
  {0.08426046, 0.0560497599999999},
  {0.08696882, 0.05473238},
  // f 1260/1382/822 1261/1386/822 1284/1384/822
  {0.07901026, 0.00901346999999997},
  {0.06569706, 0.01081449},
  {0.08696882, 0.05473238},
  // f 1261/1386/822 1282/1387/822 1283/1385/822
  {0.06569706, 0.01081449},
  {0.08173916, 0.05760062},
  {0.08426046, 0.0560497599999999},
  // f 1262/1389/822 1281/1388/822 1282/1387/822
  {0.05924879, 0.01226288},
  {0.07941019, 0.05937165},
  {0.08173916, 0.05760062},
  // f 1261/1386/822 1262/1389/822 1282/1387/822
  {0.06569706, 0.01081449},
  {0.05924879, 0.01226288},
  {0.08173916, 0.05760062},
  // f 1264/1391/822 1280/1390/822 1281/1388/822
  {0.04738472, 0.01666105},
  {0.07757119, 0.06105435},
  {0.07941019, 0.05937165},
  // f 1263/1392/822 1264/1391/822 1281/1388/822
  {0.05364369, 0.01401621},
  {0.04738472, 0.01666105},
  {0.07941019, 0.05937165},
  // f 1262/1389/822 1263/1392/822 1281/1388/822
  {0.05924879, 0.01226288},
  {0.05364369, 0.01401621},
  {0.07941019, 0.05937165},
  // f 1265/1394/822 1279/1393/822 1280/1390/822
  {0.04212864, 0.01949042},
  {0.07560924, 0.06320971},
  {0.07757119, 0.06105435},
  // f 1264/1391/822 1265/1394/822 1280/1390/822
  {0.04738472, 0.01666105},
  {0.04212864, 0.01949042},
  {0.07757119, 0.06105435},
  // f 1267/1396/822 1278/1395/822 1279/1393/822
  {0.03268793, 0.02642542},
  {0.07384486, 0.06557196},
  {0.07560924, 0.06320971},
  // f 1266/1397/822 1267/1396/822 1279/1393/822
  {0.03722971, 0.02274519},
  {0.03268793, 0.02642542},
  {0.07560924, 0.06320971},
  // f 1265/1394/822 1266/1397/822 1279/1393/822
  {0.04212864, 0.01949042},
  {0.03722971, 0.02274519},
  {0.07560924, 0.06320971},
  // f 1269/1399/822 1277/1398/822 1278/1395/822
  {0.02675808, 0.03249216},
  {0.07227813, 0.06814122},
  {0.07384486, 0.06557196},
  // f 1268/1400/822 1269/1399/822 1278/1395/822
  {0.02985847, 0.02911514},
  {0.02675808, 0.03249216},
  {0.07384486, 0.06557196},
  // f 1267/1396/822 1268/1400/822 1278/1395/822
  {0.03268793, 0.02642542},
  {0.02985847, 0.02911514},
  {0.07384486, 0.06557196},
  // f 1271/1402/822 1276/1401/822 1277/1398/822
  {0.01940268, 0.0429751900000001},
  {0.07072949, 0.07133079},
  {0.07227813, 0.06814122},
  // f 1270/1403/822 1271/1402/822 1277/1398/822
  {0.02308279, 0.0372166},
  {0.01940268, 0.0429751900000001},
  {0.07227813, 0.06814122},
  // f 1269/1399/822 1270/1403/822 1277/1398/822
  {0.02675808, 0.03249216},
  {0.02308279, 0.0372166},
  {0.07227813, 0.06814122},
  // f 1272/1405/822 1275/1404/822 1276/1401/822
  {0.01614736, 0.04926974},
  {0.06943899, 0.07479072},
  {0.07072949, 0.07133079},
  // f 1271/1402/822 1272/1405/822 1276/1401/822
  {0.01940268, 0.0429751900000001},
  {0.01614736, 0.04926974},
  {0.07072949, 0.07133079},
  // f 1273/1407/822 1274/1406/822 1275/1404/822
  {0.01331687, 0.05610031},
  {0.01069393, 0.06423301},
  {0.06943899, 0.07479072},
  // f 1272/1405/822 1273/1407/822 1275/1404/822
  {0.01614736, 0.04926974},
  {0.01331687, 0.05610031},
  {0.06943899, 0.07479072},
  // f 1343/1410/822 1344/1409/822 1345/1408/822
  {0.34032533, 0.34754533},
  {0.34032533, 0.42781097},
  {0.40122122, 0.42781097},
  // f 1334/1411/822 1343/1410/822 1345/1408/822
  {0.40122122, 0.22938025},
  {0.34032533, 0.34754533},
  {0.40122122, 0.42781097},
  // f 1337/1413/822 1342/1412/822 1343/1410/822
  {0.27378342, 0.29881746},
  {0.27378342, 0.34754533},
  {0.34032533, 0.34754533},
  // f 1336/1414/822 1337/1413/822 1343/1410/822
  {0.34032533, 0.29881746},
  {0.27378342, 0.29881746},
  {0.34032533, 0.34754533},
  // f 1334/1411/822 1336/1414/822 1343/1410/822
  {0.40122122, 0.22938025},
  {0.34032533, 0.29881746},
  {0.34032533, 0.34754533},
  // f 1340/1416/822 1341/1415/822 1342/1412/822
  {0.21261866, 0.42781097},
  {0.27378342, 0.42781097},
  {0.27378342, 0.34754533},
  // f 1339/1417/822 1340/1416/822 1342/1412/822
  {0.21261866, 0.22938025},
  {0.21261866, 0.42781097},
  {0.27378342, 0.34754533},
  // f 1337/1413/822 1339/1417/822 1342/1412/822
  {0.27378342, 0.29881746},
  {0.21261866, 0.22938025},
  {0.27378342, 0.34754533},
  // f 1337/1413/822 1338/1418/822 1339/1417/822
  {0.27378342, 0.29881746},
  {0.27378342, 0.22938025},
  {0.21261866, 0.22938025},
  // f 1334/1411/822 1335/1419/822 1336/1414/822
  {0.40122122, 0.22938025},
  {0.34032533, 0.22938025},
  {0.34032533, 0.29881746},
  // f 1346/1422/822 1352/1421/822 1353/1420/822
  {0.60233939, 0.22938025},
  {0.54023361, 0.27837884},
  {0.60233939, 0.27837884},
  // f 1349/1424/822 1351/1423/822 1352/1421/822
  {0.47933772, 0.27837884},
  {0.54023361, 0.42781097},
  {0.54023361, 0.27837884},
  // f 1346/1422/822 1349/1424/822 1352/1421/822
  {0.60233939, 0.22938025},
  {0.47933772, 0.27837884},
  {0.54023361, 0.27837884},
  // f 1349/1424/822 1350/1425/822 1351/1423/822
  {0.47933772, 0.27837884},
  {0.47933772, 0.42781097},
  {0.54023361, 0.42781097},
  // f 1347/1427/822 1348/1426/822 1349/1424/822
  {0.41723198, 0.22938025},
  {0.41723198, 0.27837884},
  {0.47933772, 0.27837884},
  // f 1346/1422/822 1347/1427/822 1349/1424/822
  {0.60233939, 0.22938025},
  {0.41723198, 0.22938025},
  {0.47933772, 0.27837884},
  // f 1431/1430/823 1355/1429/823 1354/1428/823
  {0.49627566, 0.48426801},
  {0.53633523, 0.44393218},
  {0.43511090, 0.44393218},
  // f 1393/1431/823 1431/1430/823 1354/1428/823
  {0.49627566, 0.52852917},
  {0.49627566, 0.48426801},
  {0.43511090, 0.44393218},
  // f 1392/1432/823 1393/1431/823 1354/1428/823
  {0.43511090, 0.64236283},
  {0.49627566, 0.52852917},
  {0.43511090, 0.44393218},
  // f 1425/1434/823 1356/1433/823 1355/1429/823
  {0.53140926, 0.4870981},
  {0.54770231, 0.44444525},
  {0.53633523, 0.44393218},
  // f 1426/1435/823 1425/1434/823 1355/1429/823
  {0.52932328, 0.48630542},
  {0.53140926, 0.4870981},
  {0.53633523, 0.44393218},
  // f 1427/1436/823 1426/1435/823 1355/1429/823
  {0.52653253, 0.48552579},
  {0.52932328, 0.48630542},
  {0.53633523, 0.44393218},
  // f 1428/1437/823 1427/1436/823 1355/1429/823
  {0.52288038, 0.48485273},
  {0.52653253, 0.48552579},
  {0.53633523, 0.44393218},
  // f 1429/1438/823 1428/1437/823 1355/1429/823
  {0.51880127, 0.48443431},
  {0.52288038, 0.48485273},
  {0.53633523, 0.44393218},
  // f 1430/1439/823 1429/1438/823 1355/1429/823
  {0.51361686, 0.48426801},
  {0.51880127, 0.48443431},
  {0.53633523, 0.44393218},
  // f 1431/1430/823 1430/1439/823 1355/1429/823
  {0.49627566, 0.48426801},
  {0.51361686, 0.48426801},
  {0.53633523, 0.44393218},
  // f 1423/1441/823 1357/1440/823 1356/1433/823
  {0.53461534, 0.48885232},
  {0.55690318, 0.44576281},
  {0.54770231, 0.44444525},
  // f 1424/1442/823 1423/1441/823 1356/1433/823
  {0.53327751, 0.48802072},
  {0.53461534, 0.48885232},
  {0.54770231, 0.44444525},
  // f 1425/1434/823 1424/1442/823 1356/1433/823
  {0.53140926, 0.4870981},
  {0.53327751, 0.48802072},
  {0.54770231, 0.44444525},
  // f 1421/1444/823 1358/1443/823 1357/1440/823
  {0.53687292, 0.49076504},
  {0.56425667, 0.44758093},
  {0.55690318, 0.44576281},
  // f 1422/1445/823 1421/1444/823 1357/1440/823
  {0.53581387, 0.48976713},
  {0.53687292, 0.49076504},
  {0.55690318, 0.44576281},
  // f 1423/1441/823 1422/1445/823 1357/1440/823
  {0.53461534, 0.48885232},
  {0.53581387, 0.48976713},
  {0.55690318, 0.44576281},
  // f 1420/1447/823 1359/1446/823 1358/1443/823
  {0.53782272, 0.49182278},
  {0.57004172, 0.4496333},
  {0.56425667, 0.44758093},
  // f 1421/1444/823 1420/1447/823 1358/1443/823
  {0.53687292, 0.49076504},
  {0.53782272, 0.49182278},
  {0.56425667, 0.44758093},
  // f 1419/1449/823 1360/1448/823 1359/1446/823
  {0.53848320, 0.49263996},
  {0.57534862, 0.45214182},
  {0.57004172, 0.4496333},
  // f 1420/1447/823 1419/1449/823 1359/1446/823
  {0.53782272, 0.49182278},
  {0.53848320, 0.49263996},
  {0.57004172, 0.4496333},
  // f 1418/1451/823 1361/1450/823 1360/1448/823
  {0.53967041, 0.49433571},
  {0.57940578, 0.45458061},
  {0.57534862, 0.45214182},
  // f 1419/1449/823 1418/1451/823 1360/1448/823
  {0.53848320, 0.49263996},
  {0.53967041, 0.49433571},
  {0.57534862, 0.45214182},
  // f 1418/1451/823 1362/1452/823 1361/1450/823
  {0.53967041, 0.49433571},
  {0.58241248, 0.45675975},
  {0.57940578, 0.45458061},
  // f 1418/1451/823 1363/1453/823 1362/1452/823
  {0.53967041, 0.49433571},
  {0.58587193, 0.45976871},
  {0.58241248, 0.45675975},
  // f 1417/1455/823 1364/1454/823 1363/1453/823
  {0.54067963, 0.49611324},
  {0.58900076, 0.46307003},
  {0.58587193, 0.45976871},
  // f 1418/1451/823 1417/1455/823 1363/1453/823
  {0.53967041, 0.49433571},
  {0.54067963, 0.49611324},
  {0.58587193, 0.45976871},
  // f 1417/1455/823 1365/1456/823 1364/1454/823
  {0.54067963, 0.49611324},
  {0.59180021, 0.46663928},
  {0.58900076, 0.46307003},
  // f 1416/1458/823 1366/1457/823 1365/1456/823
  {0.54151070, 0.49797273},
  {0.59427029, 0.47047663},
  {0.59180021, 0.46663928},
  // f 1417/1455/823 1416/1458/823 1365/1456/823
  {0.54067963, 0.49611324},
  {0.54151070, 0.49797273},
  {0.59180021, 0.46663928},
  // f 1416/1458/823 1367/1459/823 1366/1457/823
  {0.54151070, 0.49797273},
  {0.59641111, 0.47458202},
  {0.59427029, 0.47047663},
  // f 1416/1458/823 1368/1460/823 1367/1459/823
  {0.54151070, 0.49797273},
  {0.59854531, 0.47986209},
  {0.59641111, 0.47458202},
  // f 1415/1462/823 1369/1461/823 1368/1460/823
  {0.54216373, 0.49991399},
  {0.60020524, 0.48552823},
  {0.59854531, 0.47986209},
  // f 1416/1458/823 1415/1462/823 1368/1460/823
  {0.54151070, 0.49797273},
  {0.54216373, 0.49991399},
  {0.59854531, 0.47986209},
  // f 1414/1464/823 1370/1463/823 1369/1461/823
  {0.54263860, 0.50193715},
  {0.60139090, 0.49158019},
  {0.60020524, 0.48552823},
  // f 1415/1462/823 1414/1464/823 1369/1461/823
  {0.54216373, 0.49991399},
  {0.54263860, 0.50193715},
  {0.60020524, 0.48552823},
  // f 1414/1464/823 1371/1465/823 1370/1463/823
  {0.54263860, 0.50193715},
  {0.60210228, 0.49801809},
  {0.60139090, 0.49158019},
  // f 1413/1467/823 1372/1466/823 1371/1465/823
  {0.54293543, 0.50404227},
  {0.60233939, 0.50484204},
  {0.60210228, 0.49801809},
  // f 1414/1464/823 1413/1467/823 1371/1465/823
  {0.54263860, 0.50193715},
  {0.54293543, 0.50404227},
  {0.60210228, 0.49801809},
  // f 1412/1469/823 1373/1468/823 1372/1466/823
  {0.54305661, 0.50660157},
  {0.60198766, 0.51299834},
  {0.60233939, 0.50484204},
  // f 1413/1467/823 1412/1469/823 1372/1466/823
  {0.54293543, 0.50404227},
  {0.54305661, 0.50660157},
  {0.60233939, 0.50484204},
  // f 1410/1471/823 1374/1470/823 1373/1468/823
  {0.54291701, 0.50922477},
  {0.60093242, 0.52064002},
  {0.60198766, 0.51299834},
  // f 1411/1472/823 1410/1471/823 1373/1468/823
  {0.54301101, 0.50811708},
  {0.54291701, 0.50922477},
  {0.60198766, 0.51299834},
  // f 1412/1469/823 1411/1472/823 1373/1468/823
  {0.54305661, 0.50660157},
  {0.54301101, 0.50811708},
  {0.60198766, 0.51299834},
  // f 1409/1474/823 1375/1473/823 1374/1470/823
  {0.54257500, 0.51136565},
  {0.59946799, 0.52678037},
  {0.60093242, 0.52064002},
  // f 1410/1471/823 1409/1474/823 1374/1470/823
  {0.54291701, 0.50922477},
  {0.54257500, 0.51136565},
  {0.60093242, 0.52064002},
  // f 1408/1476/823 1376/1475/823 1375/1473/823
  {0.54213321, 0.51307356},
  {0.59748673, 0.53254259},
  {0.59946799, 0.52678037},
  // f 1409/1474/823 1408/1476/823 1375/1473/823
  {0.54257500, 0.51136565},
  {0.54213321, 0.51307356},
  {0.59946799, 0.52678037},
  // f 1408/1476/823 1377/1477/823 1376/1475/823
  {0.54213321, 0.51307356},
  {0.59544086, 0.53705561},
  {0.59748673, 0.53254259},
  // f 1407/1479/823 1378/1478/823 1377/1477/823
  {0.54154903, 0.51471269},
  {0.59303612, 0.54130602},
  {0.59544086, 0.53705561},
  // f 1408/1476/823 1407/1479/823 1377/1477/823
  {0.54213321, 0.51307356},
  {0.54154903, 0.51471269},
  {0.59544086, 0.53705561},
  // f 1407/1479/823 1379/1480/823 1378/1478/823
  {0.54154903, 0.51471269},
  {0.59027237, 0.54529393},
  {0.59303612, 0.54130602},
  // f 1406/1482/823 1380/1481/823 1379/1480/823
  {0.54082233, 0.51628268},
  {0.58780301, 0.54829502},
  {0.59027237, 0.54529393},
  // f 1407/1479/823 1406/1482/823 1379/1480/823
  {0.54154903, 0.51471269},
  {0.54082233, 0.51628268},
  {0.59027237, 0.54529393},
  // f 1406/1482/823 1381/1483/823 1380/1481/823
  {0.54082233, 0.51628268},
  {0.58439326, 0.55181026},
  {0.58780301, 0.54829502},
  // f 1405/1485/823 1382/1484/823 1381/1483/823
  {0.53995317, 0.51778364},
  {0.58061516, 0.55502498},
  {0.58439326, 0.55181026},
  // f 1406/1482/823 1405/1485/823 1381/1483/823
  {0.54082233, 0.51628268},
  {0.53995317, 0.51778364},
  {0.58439326, 0.55181026},
  // f 1405/1485/823 1383/1486/823 1382/1484/823
  {0.53995317, 0.51778364},
  {0.57645929, 0.55790126},
  {0.58061516, 0.55502498},
  // f 1404/1488/823 1384/1487/823 1383/1486/823
  {0.53894138, 0.51921546},
  {0.57192570, 0.56043911},
  {0.57645929, 0.55790126},
  // f 1405/1485/823 1404/1488/823 1383/1486/823
  {0.53995317, 0.51778364},
  {0.53894138, 0.51921546},
  {0.57645929, 0.55790126},
  // f 1404/1488/823 1385/1489/823 1384/1487/823
  {0.53894138, 0.51921546},
  {0.56598675, 0.56303799},
  {0.57192570, 0.56043911},
  // f 1403/1491/823 1386/1490/823 1385/1489/823
  {0.53753930, 0.52084267},
  {0.55950397, 0.56514955},
  {0.56598675, 0.56303799},
  // f 1404/1488/823 1403/1491/823 1385/1489/823
  {0.53894138, 0.51921546},
  {0.53753930, 0.52084267},
  {0.56598675, 0.56303799},
  // f 1401/1493/823 1387/1492/823 1386/1490/823
  {0.53472489, 0.5233165},
  {0.55125308, 0.56699705},
  {0.55950397, 0.56514955},
  // f 1402/1494/823 1401/1493/823 1386/1490/823
  {0.53593194, 0.52237046},
  {0.53472489, 0.5233165},
  {0.55950397, 0.56514955},
  // f 1403/1491/823 1402/1494/823 1386/1490/823
  {0.53753930, 0.52084267},
  {0.53593194, 0.52237046},
  {0.55950397, 0.56514955},
  // f 1399/1496/823 1388/1495/823 1387/1492/823
  {0.53190815, 0.52497184},
  {0.54091692, 0.56829655},
  {0.55125308, 0.56699705},
  // f 1400/1497/823 1399/1496/823 1387/1492/823
  {0.53373152, 0.52397418},
  {0.53190815, 0.52497184},
  {0.55125308, 0.56699705},
  // f 1401/1493/823 1400/1497/823 1387/1492/823
  {0.53472489, 0.5233165},
  {0.53373152, 0.52397418},
  {0.55125308, 0.56699705},
  // f 1397/1499/823 1389/1498/823 1388/1495/823
  {0.52715856, 0.52673328},
  {0.52961379, 0.56872964},
  {0.54091692, 0.56829655},
  // f 1398/1500/823 1397/1499/823 1388/1495/823
  {0.52987504, 0.52584648},
  {0.52715856, 0.52673328},
  {0.54091692, 0.56829655},
  // f 1399/1496/823 1398/1500/823 1388/1495/823
  {0.53190815, 0.52497184},
  {0.52987504, 0.52584648},
  {0.54091692, 0.56829655},
  // f 1394/1502/823 1390/1501/823 1389/1498/823
  {0.51119715, 0.52852917},
  {0.49627566, 0.56872964},
  {0.52961379, 0.56872964},
  // f 1395/1503/823 1394/1502/823 1389/1498/823
  {0.51843810, 0.52823102},
  {0.51119715, 0.52852917},
  {0.52961379, 0.56872964},
  // f 1396/1504/823 1395/1503/823 1389/1498/823
  {0.52360761, 0.52754378},
  {0.51843810, 0.52823102},
  {0.52961379, 0.56872964},
  // f 1397/1499/823 1396/1504/823 1389/1498/823
  {0.52715856, 0.52673328},
  {0.52360761, 0.52754378},
  {0.52961379, 0.56872964},
  // f 1394/1502/823 1393/1431/823 1390/1501/823
  {0.51119715, 0.52852917},
  {0.49627566, 0.52852917},
  {0.49627566, 0.56872964},
  // f 1392/1432/823 1391/1505/823 1390/1501/823
  {0.43511090, 0.64236283},
  {0.49627566, 0.64236283},
  {0.49627566, 0.56872964},
  // f 1393/1431/823 1392/1432/823 1390/1501/823
  {0.49627566, 0.52852917},
  {0.43511090, 0.64236283},
  {0.49627566, 0.56872964},
  // f 1635/1508/823 1433/1507/823 1432/1506/823
  {0.00886033, 0.58715916},
  {0.06594386, 0.57644486},
  {0.00800536, 0.58009946},
  // f 1634/1510/823 1434/1509/823 1433/1507/823
  {0.00994247, 0.59281003},
  {0.06707175, 0.58287692},
  {0.06594386, 0.57644486},
  // f 1635/1508/823 1634/1510/823 1433/1507/823
  {0.00886033, 0.58715916},
  {0.00994247, 0.59281003},
  {0.06594386, 0.57644486},
  // f 1633/1512/823 1435/1511/823 1434/1509/823
  {0.01202207, 0.60036635},
  {0.06835470, 0.58764148},
  {0.06707175, 0.58287692},
  // f 1634/1510/823 1633/1512/823 1434/1509/823
  {0.00994247, 0.59281003},
  {0.01202207, 0.60036635},
  {0.06707175, 0.58287692},
  // f 1631/1514/823 1436/1513/823 1435/1511/823
  {0.01763180, 0.61330175},
  {0.06977611, 0.59146667},
  {0.06835470, 0.58764148},
  // f 1632/1515/823 1631/1514/823 1435/1511/823
  {0.01476037, 0.6075089},
  {0.01763180, 0.61330175},
  {0.06835470, 0.58764148},
  // f 1633/1512/823 1632/1515/823 1435/1511/823
  {0.01202207, 0.60036635},
  {0.01476037, 0.6075089},
  {0.06835470, 0.58764148},
  // f 1629/1517/823 1437/1516/823 1436/1513/823
  {0.02482639, 0.62397504},
  {0.07125158, 0.59449863},
  {0.06977611, 0.59146667},
  // f 1630/1518/823 1629/1517/823 1436/1513/823
  {0.02098712, 0.61879039},
  {0.02482639, 0.62397504},
  {0.06977611, 0.59146667},
  // f 1631/1514/823 1630/1518/823 1436/1513/823
  {0.01763180, 0.61330175},
  {0.02098712, 0.61879039},
  {0.06977611, 0.59146667},
  // f 1628/1520/823 1438/1519/823 1437/1516/823
  {0.02991714, 0.62963951},
  {0.07227508, 0.59621763},
  {0.07125158, 0.59449863},
  // f 1629/1517/823 1628/1520/823 1437/1516/823
  {0.02482639, 0.62397504},
  {0.02991714, 0.62963951},
  {0.07125158, 0.59449863},
  // f 1626/1522/823 1439/1521/823 1438/1519/823
  {0.03409785, 0.63334286},
  {0.07360625, 0.59810174},
  {0.07227508, 0.59621763},
  // f 1627/1523/823 1626/1522/823 1438/1519/823
  {0.03235138, 0.6319083},
  {0.03409785, 0.63334286},
  {0.07227508, 0.59621763},
  // f 1628/1520/823 1627/1523/823 1438/1519/823
  {0.02991714, 0.62963951},
  {0.03235138, 0.6319083},
  {0.07227508, 0.59621763},
  // f 1624/1525/823 1440/1524/823 1439/1521/823
  {0.04318140, 0.6390264},
  {0.07556891, 0.60039055},
  {0.07360625, 0.59810174},
  // f 1625/1526/823 1624/1525/823 1439/1521/823
  {0.03788767, 0.63602483},
  {0.04318140, 0.6390264},
  {0.07360625, 0.59810174},
  // f 1626/1522/823 1625/1526/823 1439/1521/823
  {0.03409785, 0.63334286},
  {0.03788767, 0.63602483},
  {0.07360625, 0.59810174},
  // f 1623/1528/823 1441/1527/823 1440/1524/823
  {0.04909359, 0.64163828},
  {0.07769287, 0.60243857},
  {0.07556891, 0.60039055},
  // f 1624/1525/823 1623/1528/823 1440/1524/823
  {0.04318140, 0.6390264},
  {0.04909359, 0.64163828},
  {0.07556891, 0.60039055},
  // f 1622/1530/823 1442/1529/823 1441/1527/823
  {0.05700439, 0.6442579},
  {0.07997815, 0.60424554},
  {0.07769287, 0.60243857},
  // f 1623/1528/823 1622/1530/823 1441/1527/823
  {0.04909359, 0.64163828},
  {0.05700439, 0.6442579},
  {0.07769287, 0.60243857},
  // f 1621/1532/823 1443/1531/823 1442/1529/823
  {0.06735913, 0.64660454},
  {0.08242474, 0.6058116},
  {0.07997815, 0.60424554},
  // f 1622/1530/823 1621/1532/823 1442/1529/823
  {0.05700439, 0.6442579},
  {0.06735913, 0.64660454},
  {0.07997815, 0.60424554},
  // f 1621/1532/823 1444/1533/823 1443/1531/823
  {0.06735913, 0.64660454},
  {0.08503264, 0.60713673},
  {0.08242474, 0.6058116},
  // f 1620/1535/823 1445/1534/823 1444/1533/823
  {0.08067717, 0.64835095},
  {0.08780186, 0.60822093},
  {0.08503264, 0.60713673},
  // f 1621/1532/823 1620/1535/823 1444/1533/823
  {0.06735913, 0.64660454},
  {0.08067717, 0.64835095},
  {0.08503264, 0.60713673},
  // f 1620/1535/823 1446/1536/823 1445/1534/823
  {0.08067717, 0.64835095},
  {0.09073240, 0.6090641},
  {0.08780186, 0.60822093},
  // f 1620/1535/823 1447/1537/823 1446/1536/823
  {0.08067717, 0.64835095},
  {0.09446190, 0.60975802},
  {0.09073240, 0.6090641},
  // f 1619/1539/823 1448/1538/823 1447/1537/823
  {0.09955087, 0.6491307},
  {0.09707740, 0.61002791},
  {0.09446190, 0.60975802},
  // f 1620/1535/823 1619/1539/823 1447/1537/823
  {0.08067717, 0.64835095},
  {0.09955087, 0.6491307},
  {0.09446190, 0.60975802},
  // f 1619/1539/823 1449/1540/823 1448/1538/823
  {0.09955087, 0.6491307},
  {0.10049187, 0.61014831},
  {0.09707740, 0.61002791},
  // f 1619/1539/823 1450/1541/823 1449/1540/823
  {0.09955087, 0.6491307},
  {0.10499227, 0.60994875},
  {0.10049187, 0.61014831},
  // f 1618/1543/823 1451/1542/823 1450/1541/823
  {0.11061162, 0.64879763},
  {0.10866724, 0.60943639},
  {0.10499227, 0.60994875},
  // f 1619/1539/823 1618/1543/823 1450/1541/823
  {0.09955087, 0.6491307},
  {0.11061162, 0.64879763},
  {0.10499227, 0.60994875},
  // f 1617/1545/823 1452/1544/823 1451/1542/823
  {0.12088729, 0.64779878},
  {0.11163159, 0.60872936},
  {0.10866724, 0.60943639},
  // f 1618/1543/823 1617/1545/823 1451/1542/823
  {0.11061162, 0.64879763},
  {0.12088729, 0.64779878},
  {0.10866724, 0.60943639},
  // f 1617/1545/823 1453/1546/823 1452/1544/823
  {0.12088729, 0.64779878},
  {0.11398576, 0.60793126},
  {0.11163159, 0.60872936},
  // f 1616/1548/823 1454/1547/823 1453/1546/823
  {0.13128379, 0.64593077},
  {0.11616769, 0.60695565},
  {0.11398576, 0.60793126},
  // f 1617/1545/823 1616/1548/823 1453/1546/823
  {0.12088729, 0.64779878},
  {0.13128379, 0.64593077},
  {0.11398576, 0.60793126},
  // f 1616/1548/823 1455/1549/823 1454/1547/823
  {0.13128379, 0.64593077},
  {0.11817735, 0.60580277},
  {0.11616769, 0.60695565},
  // f 1615/1551/823 1456/1550/823 1455/1549/823
  {0.13991085, 0.64353335},
  {0.11942136, 0.60493565},
  {0.11817735, 0.60580277},
  // f 1616/1548/823 1615/1551/823 1455/1549/823
  {0.13128379, 0.64593077},
  {0.13991085, 0.64353335},
  {0.11817735, 0.60580277},
  // f 1615/1551/823 1457/1552/823 1456/1550/823
  {0.13991085, 0.64353335},
  {0.12058887, 0.60398972},
  {0.11942136, 0.60493565},
  // f 1614/1554/823 1458/1553/823 1457/1552/823
  {0.14700398, 0.64080632},
  {0.12242662, 0.60222673},
  {0.12058887, 0.60398972},
  // f 1615/1551/823 1614/1554/823 1457/1552/823
  {0.13991085, 0.64353335},
  {0.14700398, 0.64080632},
  {0.12058887, 0.60398972},
  // f 1613/1556/823 1459/1555/823 1458/1553/823
  {0.15068462, 0.6390636},
  {0.12378166, 0.60065448},
  {0.12242662, 0.60222673},
  // f 1614/1554/823 1613/1556/823 1458/1553/823
  {0.14700398, 0.64080632},
  {0.15068462, 0.6390636},
  {0.12242662, 0.60222673},
  // f 1611/1558/823 1460/1557/823 1459/1555/823
  {0.15947416, 0.63380408},
  {0.12493341, 0.59902573},
  {0.12378166, 0.60065448},
  // f 1612/1559/823 1611/1558/823 1459/1555/823
  {0.15419857, 0.63716531},
  {0.15947416, 0.63380408},
  {0.12378166, 0.60065448},
  // f 1613/1556/823 1612/1559/823 1459/1555/823
  {0.15068462, 0.6390636},
  {0.15419857, 0.63716531},
  {0.12378166, 0.60065448},
  // f 1610/1561/823 1461/1560/823 1460/1557/823
  {0.16432306, 0.63004446},
  {0.12573797, 0.59762514},
  {0.12493341, 0.59902573},
  // f 1611/1558/823 1610/1561/823 1460/1557/823
  {0.15947416, 0.63380408},
  {0.16432306, 0.63004446},
  {0.12493341, 0.59902573},
  // f 1609/1563/823 1462/1562/823 1461/1560/823
  {0.16874522, 0.62588632},
  {0.12640141, 0.59618545},
  {0.12573797, 0.59762514},
  // f 1610/1561/823 1609/1563/823 1461/1560/823
  {0.16432306, 0.63004446},
  {0.16874522, 0.62588632},
  {0.12573797, 0.59762514},
  // f 1608/1565/823 1463/1564/823 1462/1562/823
  {0.17274061, 0.62132978},
  {0.12692362, 0.5947063},
  {0.12640141, 0.59618545},
  // f 1609/1563/823 1608/1565/823 1462/1562/823
  {0.16874522, 0.62588632},
  {0.17274061, 0.62132978},
  {0.12640141, 0.59618545},
  // f 1607/1567/823 1464/1566/823 1463/1564/823
  {0.17630929, 0.61637473},
  {0.12730476, 0.59318805},
  {0.12692362, 0.5947063},
  // f 1608/1565/823 1607/1567/823 1463/1564/823
  {0.17274061, 0.62132978},
  {0.17630929, 0.61637473},
  {0.12692362, 0.5947063},
  // f 1605/1569/823 1465/1568/823 1464/1566/823
  {0.18008402, 0.60981262},
  {0.12754472, 0.59163058},
  {0.12730476, 0.59318805},
  // f 1606/1570/823 1605/1569/823 1464/1566/823
  {0.17830151, 0.61312342},
  {0.18008402, 0.60981262},
  {0.12730476, 0.59318805},
  // f 1607/1567/823 1606/1570/823 1464/1566/823
  {0.17630929, 0.61637473},
  {0.17830151, 0.61312342},
  {0.12730476, 0.59318805},
  // f 1604/1572/823 1466/1571/823 1465/1568/823
  {0.18249984, 0.60439157},
  {0.12762092, 0.59067726},
  {0.12754472, 0.59163058},
  // f 1605/1569/823 1604/1572/823 1465/1568/823
  {0.18008402, 0.60981262},
  {0.18249984, 0.60439157},
  {0.12754472, 0.59163058},
  // f 1603/1574/823 1467/1573/823 1466/1571/823
  {0.18437883, 0.59881794},
  {0.12764634, 0.58970976},
  {0.12762092, 0.59067726},
  // f 1604/1572/823 1603/1574/823 1466/1571/823
  {0.18249984, 0.60439157},
  {0.18437883, 0.59881794},
  {0.12762092, 0.59067726},
  // f 1603/1574/823 1602/1575/823 1467/1573/823
  {0.18437883, 0.59881794},
  {0.18572100, 0.59309185},
  {0.12764634, 0.58970976},
  // f 1602/1575/823 1468/1576/823 1467/1573/823
  {0.18572100, 0.59309185},
  {0.12757912, 0.58818293},
  {0.12764634, 0.58970976},
  // f 1602/1575/823 1601/1577/823 1468/1576/823
  {0.18572100, 0.59309185},
  {0.18652625, 0.58721328},
  {0.12757912, 0.58818293},
  // f 1601/1577/823 1469/1578/823 1468/1576/823
  {0.18652625, 0.58721328},
  {0.12737748, 0.58668864},
  {0.12757912, 0.58818293},
  // f 1600/1580/823 1599/1579/823 1469/1578/823
  {0.18679467, 0.58118236},
  {0.18653989, 0.57543588},
  {0.12737748, 0.58668864},
  // f 1601/1577/823 1600/1580/823 1469/1578/823
  {0.18652625, 0.58721328},
  {0.18679467, 0.58118236},
  {0.12737748, 0.58668864},
  // f 1599/1579/823 1470/1581/823 1469/1578/823
  {0.18653989, 0.57543588},
  {0.12704141, 0.58522677},
  {0.12737748, 0.58668864},
  // f 1599/1579/823 1598/1582/823 1470/1581/823
  {0.18653989, 0.57543588},
  {0.18577547, 0.5698911},
  {0.12704141, 0.58522677},
  // f 1598/1582/823 1471/1583/823 1470/1581/823
  {0.18577547, 0.5698911},
  {0.12657091, 0.58379734},
  {0.12704141, 0.58522677},
  // f 1597/1585/823 1596/1584/823 1471/1583/823
  {0.18466823, 0.56513178},
  {0.18315834, 0.56053185},
  {0.12657091, 0.58379734},
  // f 1598/1582/823 1597/1585/823 1471/1583/823
  {0.18577547, 0.5698911},
  {0.18466823, 0.56513178},
  {0.12657091, 0.58379734},
  // f 1596/1584/823 1472/1586/823 1471/1583/823
  {0.18315834, 0.56053185},
  {0.12596598, 0.58240056},
  {0.12657091, 0.58379734},
  // f 1594/1588/823 1593/1587/823 1472/1586/823
  {0.17893064, 0.55181026},
  {0.17761217, 0.54973686},
  {0.12596598, 0.58240056},
  // f 1595/1589/823 1594/1588/823 1472/1586/823
  {0.18124582, 0.55609131},
  {0.17893064, 0.55181026},
  {0.12596598, 0.58240056},
  // f 1596/1584/823 1595/1589/823 1472/1586/823
  {0.18315834, 0.56053185},
  {0.18124582, 0.55609131},
  {0.12596598, 0.58240056},
  // f 1593/1587/823 1473/1590/823 1472/1586/823
  {0.17761217, 0.54973686},
  {0.12506263, 0.58076715},
  {0.12596598, 0.58240056},
  // f 1591/1592/823 1590/1591/823 1473/1590/823
  {0.17336529, 0.54431629},
  {0.16970457, 0.54063332},
  {0.12506263, 0.58076715},
  // f 1592/1593/823 1591/1592/823 1473/1590/823
  {0.17617325, 0.54771793},
  {0.17336529, 0.54431629},
  {0.12506263, 0.58076715},
  // f 1593/1587/823 1592/1593/823 1473/1590/823
  {0.17761217, 0.54973686},
  {0.17617325, 0.54771793},
  {0.12506263, 0.58076715},
  // f 1590/1591/823 1474/1594/823 1473/1590/823
  {0.16970457, 0.54063332},
  {0.12396570, 0.5791806},
  {0.12506263, 0.58076715},
  // f 1589/1596/823 1588/1595/823 1474/1594/823
  {0.16556203, 0.53716862},
  {0.16032580, 0.5335319},
  {0.12396570, 0.5791806},
  // f 1590/1591/823 1589/1596/823 1474/1594/823
  {0.16970457, 0.54063332},
  {0.16556203, 0.53716862},
  {0.12396570, 0.5791806},
  // f 1588/1595/823 1475/1597/823 1474/1594/823
  {0.16032580, 0.5335319},
  {0.12267519, 0.57764077},
  {0.12396570, 0.5791806},
  // f 1588/1595/823 1587/1598/823 1475/1597/823
  {0.16032580, 0.5335319},
  {0.15379260, 0.52981508},
  {0.12267519, 0.57764077},
  // f 1587/1598/823 1476/1599/823 1475/1597/823
  {0.15379260, 0.52981508},
  {0.12092493, 0.57590353},
  {0.12267519, 0.57764077},
  // f 1587/1598/823 1477/1600/823 1476/1599/823
  {0.15379260, 0.52981508},
  {0.11973592, 0.57493329},
  {0.12092493, 0.57590353},
  // f 1587/1598/823 1586/1601/823 1477/1600/823
  {0.15379260, 0.52981508},
  {0.14419636, 0.52557111},
  {0.11973592, 0.57493329},
  // f 1586/1601/823 1478/1602/823 1477/1600/823
  {0.14419636, 0.52557111},
  {0.11793030, 0.57373261},
  {0.11973592, 0.57493329},
  // f 1586/1601/823 1585/1603/823 1478/1602/823
  {0.14419636, 0.52557111},
  {0.13199072, 0.52135062},
  {0.11793030, 0.57373261},
  // f 1585/1603/823 1479/1604/823 1478/1602/823
  {0.13199072, 0.52135062},
  {0.11480615, 0.57207453},
  {0.11793030, 0.57373261},
  // f 1585/1603/823 1480/1605/823 1479/1604/823
  {0.13199072, 0.52135062},
  {0.10976741, 0.56998169},
  {0.11480615, 0.57207453},
  // f 1585/1603/823 1481/1606/823 1480/1605/823
  {0.13199072, 0.52135062},
  {0.10126038, 0.56725943},
  {0.10976741, 0.56998169},
  // f 1584/1608/823 1482/1607/823 1481/1606/823
  {0.09659346, 0.51228654},
  {0.08973763, 0.56439829},
  {0.10126038, 0.56725943},
  // f 1585/1603/823 1584/1608/823 1481/1606/823
  {0.13199072, 0.52135062},
  {0.09659346, 0.51228654},
  {0.10126038, 0.56725943},
  // f 1583/1610/823 1483/1609/823 1482/1607/823
  {0.08900483, 0.5104903},
  {0.06777073, 0.55842102},
  {0.08973763, 0.56439829},
  // f 1584/1608/823 1583/1610/823 1482/1607/823
  {0.09659346, 0.51228654},
  {0.08900483, 0.5104903},
  {0.08973763, 0.56439829},
  // f 1582/1612/823 1581/1611/823 1483/1609/823
  {0.08301555, 0.50862086},
  {0.07980608, 0.50728381},
  {0.06777073, 0.55842102},
  // f 1583/1610/823 1582/1612/823 1483/1609/823
  {0.08900483, 0.5104903},
  {0.08301555, 0.50862086},
  {0.06777073, 0.55842102},
  // f 1581/1611/823 1484/1613/823 1483/1609/823
  {0.07980608, 0.50728381},
  {0.05429142, 0.55335629},
  {0.06777073, 0.55842102},
  // f 1580/1615/823 1579/1614/823 1484/1613/823
  {0.07809229, 0.50637293},
  {0.07692436, 0.50560188},
  {0.05429142, 0.55335629},
  // f 1581/1611/823 1580/1615/823 1484/1613/823
  {0.07980608, 0.50728381},
  {0.07809229, 0.50637293},
  {0.05429142, 0.55335629},
  // f 1579/1614/823 1485/1616/823 1484/1613/823
  {0.07692436, 0.50560188},
  {0.04491801, 0.54878366},
  {0.05429142, 0.55335629},
  // f 1577/1618/823 1576/1617/823 1485/1616/823
  {0.07547896, 0.50436485},
  {0.07485438, 0.50374031},
  {0.04491801, 0.54878366},
  // f 1578/1619/823 1577/1618/823 1485/1616/823
  {0.07616039, 0.50497735},
  {0.07547896, 0.50436485},
  {0.04491801, 0.54878366},
  // f 1579/1614/823 1578/1619/823 1485/1616/823
  {0.07692436, 0.50560188},
  {0.07616039, 0.50497735},
  {0.04491801, 0.54878366},
  // f 1576/1617/823 1486/1620/823 1485/1616/823
  {0.07485438, 0.50374031},
  {0.03865736, 0.54496038},
  {0.04491801, 0.54878366},
  // f 1576/1617/823 1575/1621/823 1486/1620/823
  {0.07485438, 0.50374031},
  {0.07402395, 0.50278068},
  {0.03865736, 0.54496038},
  // f 1575/1621/823 1487/1622/823 1486/1620/823
  {0.07402395, 0.50278068},
  {0.03548435, 0.54267085},
  {0.03865736, 0.54496038},
  // f 1575/1621/823 1488/1623/823 1487/1622/823
  {0.07402395, 0.50278068},
  {0.03260570, 0.54030502},
  {0.03548435, 0.54267085},
  // f 1575/1621/823 1574/1624/823 1488/1623/823
  {0.07402395, 0.50278068},
  {0.07332127, 0.50179374},
  {0.03260570, 0.54030502},
  // f 1574/1624/823 1489/1625/823 1488/1623/823
  {0.07332127, 0.50179374},
  {0.02872384, 0.53658032},
  {0.03260570, 0.54030502},
  // f 1574/1624/823 1490/1626/823 1489/1625/823
  {0.07332127, 0.50179374},
  {0.02533755, 0.53261769},
  {0.02872384, 0.53658032},
  // f 1574/1624/823 1573/1627/823 1490/1626/823
  {0.07332127, 0.50179374},
  {0.07283329, 0.50095057},
  {0.02533755, 0.53261769},
  // f 1573/1627/823 1491/1628/823 1490/1626/823
  {0.07283329, 0.50095057},
  {0.02289420, 0.52913404},
  {0.02533755, 0.53261769},
  // f 1573/1627/823 1492/1629/823 1491/1628/823
  {0.07283329, 0.50095057},
  {0.02079493, 0.52548504},
  {0.02289420, 0.52913404},
  // f 1573/1627/823 1572/1630/823 1492/1629/823
  {0.07283329, 0.50095057},
  {0.07243405, 0.50008833},
  {0.02079493, 0.52548504},
  // f 1572/1630/823 1493/1631/823 1492/1629/823
  {0.07243405, 0.50008833},
  {0.01903985, 0.52167106},
  {0.02079493, 0.52548504},
  // f 1572/1630/823 1494/1632/823 1493/1631/823
  {0.07243405, 0.50008833},
  {0.01762890, 0.51769185},
  {0.01903985, 0.52167106},
  // f 1572/1630/823 1571/1633/823 1494/1632/823
  {0.07243405, 0.50008833},
  {0.07212348, 0.49920714},
  {0.01762890, 0.51769185},
  // f 1571/1633/823 1495/1634/823 1494/1632/823
  {0.07212348, 0.49920714},
  {0.01656209, 0.51354754},
  {0.01762890, 0.51769185},
  // f 1571/1633/823 1496/1635/823 1495/1634/823
  {0.07212348, 0.49920714},
  {0.01573620, 0.50835645},
  {0.01656209, 0.51354754},
  // f 1571/1633/823 1570/1636/823 1496/1635/823
  {0.07212348, 0.49920714},
  {0.07190171, 0.49830699},
  {0.01573620, 0.50835645},
  // f 1570/1636/823 1497/1637/823 1496/1635/823
  {0.07190171, 0.49830699},
  {0.01550905, 0.50567162},
  {0.01573620, 0.50835645},
  // f 1570/1636/823 1498/1638/823 1497/1637/823
  {0.07190171, 0.49830699},
  {0.01539891, 0.5019995},
  {0.01550905, 0.50567162},
  // f 1569/1640/823 1499/1639/823 1498/1638/823
  {0.07176863, 0.49738801},
  {0.01562433, 0.49714839},
  {0.01539891, 0.5019995},
  // f 1570/1636/823 1569/1640/823 1498/1638/823
  {0.07190171, 0.49830699},
  {0.07176863, 0.49738801},
  {0.01539891, 0.5019995},
  // f 1569/1640/823 1500/1641/823 1499/1639/823
  {0.07176863, 0.49738801},
  {0.01630054, 0.4923839},
  {0.01562433, 0.49714839},
  // f 1568/1643/823 1501/1642/823 1500/1641/823
  {0.07172427, 0.49645001},
  {0.01742755, 0.48770601},
  {0.01630054, 0.4923839},
  // f 1569/1640/823 1568/1643/823 1500/1641/823
  {0.07176863, 0.49738801},
  {0.07172427, 0.49645001},
  {0.01630054, 0.4923839},
  // f 1567/1645/823 1502/1644/823 1501/1642/823
  {0.07175871, 0.49542302},
  {0.01900540, 0.48311478},
  {0.01742755, 0.48770601},
  // f 1568/1643/823 1567/1645/823 1501/1642/823
  {0.07172427, 0.49645001},
  {0.07175871, 0.49542302},
  {0.01742755, 0.48770601},
  // f 1566/1647/823 1503/1646/823 1502/1644/823
  {0.07182966, 0.49466926},
  {0.02103413, 0.4786101},
  {0.01900540, 0.48311478},
  // f 1567/1645/823 1566/1647/823 1502/1644/823
  {0.07175871, 0.49542302},
  {0.07182966, 0.49466926},
  {0.01900540, 0.48311478},
  // f 1566/1647/823 1504/1648/823 1503/1646/823
  {0.07182966, 0.49466926},
  {0.02253099, 0.47583878},
  {0.02103413, 0.4786101},
  // f 1565/1650/823 1505/1649/823 1504/1648/823
  {0.07203395, 0.49344444},
  {0.02420394, 0.4731012},
  {0.02253099, 0.47583878},
  // f 1566/1647/823 1565/1650/823 1504/1648/823
  {0.07182966, 0.49466926},
  {0.07203395, 0.49344444},
  {0.02253099, 0.47583878},
  // f 1565/1650/823 1506/1651/823 1505/1649/823
  {0.07203395, 0.49344444},
  {0.02684327, 0.46940637},
  {0.02420394, 0.4731012},
  // f 1564/1653/823 1507/1652/823 1506/1651/823
  {0.07234587, 0.49225885},
  {0.02983030, 0.4659397},
  {0.02684327, 0.46940637},
  // f 1565/1650/823 1564/1653/823 1506/1651/823
  {0.07203395, 0.49344444},
  {0.07234587, 0.49225885},
  {0.02684327, 0.46940637},
  // f 1563/1655/823 1508/1654/823 1507/1652/823
  {0.07276528, 0.49111253},
  {0.03367003, 0.46225721},
  {0.02983030, 0.4659397},
  // f 1564/1653/823 1563/1655/823 1507/1652/823
  {0.07234587, 0.49225885},
  {0.07276528, 0.49111253},
  {0.02983030, 0.4659397},
  // f 1563/1655/823 1509/1656/823 1508/1654/823
  {0.07276528, 0.49111253},
  {0.03796400, 0.45887268},
  {0.03367003, 0.46225721},
  // f 1562/1658/823 1510/1657/823 1509/1656/823
  {0.07329224, 0.49000543},
  {0.04271221, 0.45578617},
  {0.03796400, 0.45887268},
  // f 1563/1655/823 1562/1658/823 1509/1656/823
  {0.07276528, 0.49111253},
  {0.07329224, 0.49000543},
  {0.03796400, 0.45887268},
  // f 1561/1660/823 1511/1659/823 1510/1657/823
  {0.07392674, 0.48893762},
  {0.04657147, 0.45366681},
  {0.04271221, 0.45578617},
  // f 1562/1658/823 1561/1660/823 1510/1657/823
  {0.07329224, 0.49000543},
  {0.07392674, 0.48893762},
  {0.04271221, 0.45578617},
  // f 1560/1662/823 1512/1661/823 1511/1659/823
  {0.07483006, 0.48770809},
  {0.05068627, 0.45171505},
  {0.04657147, 0.45366681},
  // f 1561/1660/823 1560/1662/823 1511/1659/823
  {0.07392674, 0.48893762},
  {0.07483006, 0.48770809},
  {0.04657147, 0.45366681},
  // f 1559/1664/823 1513/1663/823 1512/1661/823
  {0.07588830, 0.48653501},
  {0.05667853, 0.44942379},
  {0.05068627, 0.45171505},
  // f 1560/1662/823 1559/1664/823 1512/1661/823
  {0.07483006, 0.48770809},
  {0.07588830, 0.48653501},
  {0.05068627, 0.45171505},
  // f 1557/1666/823 1514/1665/823 1513/1663/823
  {0.07799525, 0.48472202},
  {0.06422196, 0.44732237},
  {0.05667853, 0.44942379},
  // f 1558/1667/823 1557/1666/823 1513/1663/823
  {0.07710139, 0.48541856},
  {0.07799525, 0.48472202},
  {0.05667853, 0.44942379},
  // f 1559/1664/823 1558/1667/823 1513/1663/823
  {0.07588830, 0.48653501},
  {0.07710139, 0.48541856},
  {0.05667853, 0.44942379},
  // f 1555/1669/823 1515/1668/823 1514/1665/823
  {0.08025192, 0.48336715},
  {0.07459515, 0.44543892},
  {0.06422196, 0.44732237},
  // f 1556/1670/823 1555/1669/823 1514/1665/823
  {0.07895628, 0.48408359},
  {0.08025192, 0.48336715},
  {0.06422196, 0.44732237},
  // f 1557/1666/823 1556/1670/823 1514/1665/823
  {0.07799525, 0.48472202},
  {0.07895628, 0.48408359},
  {0.06422196, 0.44732237},
  // f 1552/1672/823 1516/1671/823 1515/1668/823
  {0.08613155, 0.48147178},
  {0.08623699, 0.44430888},
  {0.07459515, 0.44543892},
  // f 1553/1673/823 1552/1672/823 1515/1668/823
  {0.08378918, 0.48201776},
  {0.08613155, 0.48147178},
  {0.07459515, 0.44543892},
  // f 1554/1674/823 1553/1673/823 1515/1668/823
  {0.08194504, 0.48262709},
  {0.08378918, 0.48201776},
  {0.07459515, 0.44543892},
  // f 1555/1669/823 1554/1674/823 1515/1668/823
  {0.08025192, 0.48336715},
  {0.08194504, 0.48262709},
  {0.07459515, 0.44543892},
  // f 1550/1676/823 1517/1675/823 1516/1671/823
  {0.09309833, 0.48088413},
  {0.09914760, 0.44393218},
  {0.08623699, 0.44430888},
  // f 1551/1677/823 1550/1676/823 1516/1671/823
  {0.08906011, 0.48106551},
  {0.09309833, 0.48088413},
  {0.08623699, 0.44430888},
  // f 1552/1672/823 1551/1677/823 1516/1671/823
  {0.08613155, 0.48147178},
  {0.08906011, 0.48106551},
  {0.08623699, 0.44430888},
  // f 1547/1679/823 1518/1678/823 1517/1675/823
  {0.10210064, 0.48176366},
  {0.11489967, 0.44461673},
  {0.09914760, 0.44393218},
  // f 1548/1680/823 1547/1679/823 1517/1675/823
  {0.09850654, 0.48117888},
  {0.10210064, 0.48176366},
  {0.09914760, 0.44393218},
  // f 1549/1681/823 1548/1680/823 1517/1675/823
  {0.09611519, 0.48097181},
  {0.09850654, 0.48117888},
  {0.09914760, 0.44393218},
  // f 1550/1676/823 1549/1681/823 1517/1675/823
  {0.09309833, 0.48088413},
  {0.09611519, 0.48097181},
  {0.09914760, 0.44393218},
  // f 1544/1683/823 1519/1682/823 1518/1678/823
  {0.10949175, 0.48440224},
  {0.12649563, 0.4461953},
  {0.11489967, 0.44461673},
  // f 1545/1684/823 1544/1683/823 1518/1678/823
  {0.10733183, 0.48337901},
  {0.10949175, 0.48440224},
  {0.11489967, 0.44461673},
  // f 1546/1685/823 1545/1684/823 1518/1678/823
  {0.10501114, 0.48253113},
  {0.10733183, 0.48337901},
  {0.11489967, 0.44461673},
  // f 1547/1679/823 1546/1685/823 1518/1678/823
  {0.10210064, 0.48176366},
  {0.10501114, 0.48253113},
  {0.11489967, 0.44461673},
  // f 1542/1687/823 1520/1686/823 1519/1682/823
  {0.11332975, 0.48697513},
  {0.13582687, 0.44836789},
  {0.12649563, 0.4461953},
  // f 1543/1688/823 1542/1687/823 1519/1682/823
  {0.11149115, 0.48560095},
  {0.11332975, 0.48697513},
  {0.12649563, 0.4461953},
  // f 1544/1683/823 1543/1688/823 1519/1682/823
  {0.10949175, 0.48440224},
  {0.11149115, 0.48560095},
  {0.12649563, 0.4461953},
  // f 1540/1690/823 1521/1689/823 1520/1686/823
  {0.11554509, 0.4891113},
  {0.14323798, 0.450863},
  {0.13582687, 0.44836789},
  // f 1541/1691/823 1540/1690/823 1520/1686/823
  {0.11446970, 0.4879964},
  {0.11554509, 0.4891113},
  {0.13582687, 0.44836789},
  // f 1542/1687/823 1541/1691/823 1520/1686/823
  {0.11332975, 0.48697513},
  {0.11446970, 0.4879964},
  {0.13582687, 0.44836789},
  // f 1539/1693/823 1522/1692/823 1521/1689/823
  {0.11703724, 0.49095911},
  {0.14903045, 0.45344299},
  {0.14323798, 0.450863},
  // f 1540/1690/823 1539/1693/823 1521/1689/823
  {0.11554509, 0.4891113},
  {0.11703724, 0.49095911},
  {0.14323798, 0.450863},
  // f 1539/1693/823 1523/1694/823 1522/1692/823
  {0.11703724, 0.49095911},
  {0.15260497, 0.45538932},
  {0.14903045, 0.45344299},
  // f 1538/1696/823 1524/1695/823 1523/1694/823
  {0.11838425, 0.49301738},
  {0.15675001, 0.45807678},
  {0.15260497, 0.45538932},
  // f 1539/1693/823 1538/1696/823 1523/1694/823
  {0.11703724, 0.49095911},
  {0.11838425, 0.49301738},
  {0.15260497, 0.45538932},
  // f 1538/1696/823 1525/1697/823 1524/1695/823
  {0.11838425, 0.49301738},
  {0.16055501, 0.46107286},
  {0.15675001, 0.45807678},
  // f 1537/1699/823 1526/1698/823 1525/1697/823
  {0.11958604, 0.49528611},
  {0.16403870, 0.46440327},
  {0.16055501, 0.46107286},
  // f 1538/1696/823 1537/1699/823 1525/1697/823
  {0.11838425, 0.49301738},
  {0.11958604, 0.49528611},
  {0.16055501, 0.46107286},
  // f 1537/1699/823 1527/1700/823 1526/1698/823
  {0.11958604, 0.49528611},
  {0.16720115, 0.468068},
  {0.16403870, 0.46440327},
  // f 1536/1702/823 1528/1701/823 1527/1700/823
  {0.12080462, 0.49819905},
  {0.17004226, 0.47206712},
  {0.16720115, 0.468068},
  // f 1537/1699/823 1536/1702/823 1527/1700/823
  {0.11958604, 0.49528611},
  {0.12080462, 0.49819905},
  {0.16720115, 0.468068},
  // f 1536/1702/823 1529/1703/823 1528/1701/823
  {0.12080462, 0.49819905},
  {0.17256212, 0.47640049},
  {0.17004226, 0.47206712},
  // f 1536/1702/823 1530/1704/823 1529/1703/823
  {0.12080462, 0.49819905},
  {0.17516185, 0.48204184},
  {0.17256212, 0.47640049},
  // f 1535/1706/823 1531/1705/823 1530/1704/823
  {0.12195533, 0.50187898},
  {0.17729890, 0.48816466},
  {0.17516185, 0.48204184},
  // f 1536/1702/823 1535/1706/823 1530/1704/823
  {0.12080462, 0.49819905},
  {0.12195533, 0.50187898},
  {0.17516185, 0.48204184},
  // f 1534/1708/823 1532/1707/823 1531/1705/823
  {0.12294136, 0.50646627},
  {0.17897336, 0.49476892},
  {0.17729890, 0.48816466},
  // f 1535/1706/823 1534/1708/823 1531/1705/823
  {0.12195533, 0.50187898},
  {0.12294136, 0.50646627},
  {0.17729890, 0.48816466},
  // f 1534/1708/823 1533/1709/823 1532/1707/823
  {0.12294136, 0.50646627},
  {0.18034212, 0.50308239},
  {0.17897336, 0.49476892},
  // f 1638/1712/823 1637/1711/823 1636/1710/823
  {0.70344305, 0.15762806},
  {0.70344305, 0.00806057000000004},
  {0.64254719, 0.00806057000000004},
  // f 1641/1713/823 1638/1712/823 1636/1710/823
  {0.64254719, 0.20649135},
  {0.70344305, 0.15762806},
  {0.64254719, 0.00806057000000004},
  // f 1640/1715/823 1639/1714/823 1638/1712/823
  {0.79848373, 0.20649135},
  {0.79848373, 0.15762806},
  {0.70344305, 0.15762806},
  // f 1641/1713/823 1640/1715/823 1638/1712/823
  {0.64254719, 0.20649135},
  {0.79848373, 0.20649135},
  {0.70344305, 0.15762806},
  // f 1645/1718/823 1643/1717/823 1642/1716/823
  {0.06554057, 0.7348243},
  {0.12092493, 0.86706638},
  {0.12092493, 0.66525185},
  // f 1659/1719/823 1645/1718/823 1642/1716/823
  {0.06755699, 0.67924213},
  {0.06554057, 0.7348243},
  {0.12092493, 0.66525185},
  // f 1660/1720/823 1659/1719/823 1642/1716/823
  {0.07185869, 0.67248523},
  {0.06755699, 0.67924213},
  {0.12092493, 0.66525185},
  // f 1661/1721/823 1660/1720/823 1642/1716/823
  {0.07562268, 0.66525185},
  {0.07185869, 0.67248523},
  {0.12092493, 0.66525185},
  // f 1645/1718/823 1644/1722/823 1643/1717/823
  {0.06554057, 0.7348243},
  {0.06554057, 0.86706638},
  {0.12092493, 0.86706638},
  // f 1653/1724/823 1646/1723/823 1645/1718/823
  {0.03623455, 0.70669723},
  {0.05179632, 0.74446142},
  {0.06554057, 0.7348243},
  // f 1654/1725/823 1653/1724/823 1645/1718/823
  {0.04474187, 0.70159733},
  {0.03623455, 0.70669723},
  {0.06554057, 0.7348243},
  // f 1655/1726/823 1654/1725/823 1645/1718/823
  {0.04820875, 0.69916236},
  {0.04474187, 0.70159733},
  {0.06554057, 0.7348243},
  // f 1656/1727/823 1655/1726/823 1645/1718/823
  {0.05142563, 0.6966542},
  {0.04820875, 0.69916236},
  {0.06554057, 0.7348243},
  // f 1657/1728/823 1656/1727/823 1645/1718/823
  {0.05734046, 0.69132662},
  {0.05142563, 0.6966542},
  {0.06554057, 0.7348243},
  // f 1658/1729/823 1657/1728/823 1645/1718/823
  {0.06271759, 0.68552256},
  {0.05734046, 0.69132662},
  {0.06554057, 0.7348243},
  // f 1659/1719/823 1658/1729/823 1645/1718/823
  {0.06755699, 0.67924213},
  {0.06271759, 0.68552256},
  {0.06554057, 0.7348243},
  // f 1652/1731/823 1647/1730/823 1646/1723/823
  {0.02462024, 0.71226811},
  {0.04561220, 0.74819338},
  {0.05179632, 0.74446142},
  // f 1653/1724/823 1652/1731/823 1646/1723/823
  {0.03623455, 0.70669723},
  {0.02462024, 0.71226811},
  {0.05179632, 0.74446142},
  // f 1652/1731/823 1648/1732/823 1647/1730/823
  {0.02462024, 0.71226811},
  {0.03952874, 0.75147307},
  {0.04561220, 0.74819338},
  // f 1652/1731/823 1649/1733/823 1648/1732/823
  {0.02462024, 0.71226811},
  {0.02536339, 0.75769949},
  {0.03952874, 0.75147307},
  // f 1651/1735/823 1650/1734/823 1649/1733/823
  {0.00800536, 0.71844637},
  {0.00800536, 0.76365507},
  {0.02536339, 0.75769949},
  // f 1652/1731/823 1651/1735/823 1649/1733/823
  {0.02462024, 0.71226811},
  {0.00800536, 0.71844637},
  {0.02536339, 0.75769949},
  // f 1784/1738/823 1663/1737/823 1662/1736/823
  {0.37205037, 0.13800156},
  {0.33131865, 0.09671825},
  {0.33131865, 0.13800156},
  // f 1784/1738/823 1664/1739/823 1663/1737/823
  {0.37205037, 0.13800156},
  {0.42541826, 0.09671825},
  {0.33131865, 0.09671825},
  // f 1784/1738/823 1665/1740/823 1664/1739/823
  {0.37205037, 0.13800156},
  {0.42541826, 0.18131524},
  {0.42541826, 0.09671825},
  // f 1783/1742/823 1666/1741/823 1665/1740/823
  {0.37205037, 0.15668058},
  {0.40480238, 0.19425523},
  {0.42541826, 0.18131524},
  // f 1784/1738/823 1783/1742/823 1665/1740/823
  {0.37205037, 0.13800156},
  {0.37205037, 0.15668058},
  {0.42541826, 0.18131524},
  // f 1783/1742/823 1667/1743/823 1666/1741/823
  {0.37205037, 0.15668058},
  {0.39065319, 0.201473},
  {0.40480238, 0.19425523},
  // f 1782/1745/823 1668/1744/823 1667/1743/823
  {0.36422557, 0.16107374},
  {0.38443527, 0.20411181},
  {0.39065319, 0.201473},
  // f 1783/1742/823 1782/1745/823 1667/1743/823
  {0.37205037, 0.15668058},
  {0.36422557, 0.16107374},
  {0.39065319, 0.201473},
  // f 1782/1745/823 1669/1746/823 1668/1744/823
  {0.36422557, 0.16107374},
  {0.37762913, 0.206559},
  {0.38443527, 0.20411181},
  // f 1781/1748/823 1670/1747/823 1669/1746/823
  {0.35702235, 0.16449678},
  {0.36721474, 0.20938909},
  {0.37762913, 0.206559},
  // f 1782/1745/823 1781/1748/823 1669/1746/823
  {0.36422557, 0.16107374},
  {0.35702235, 0.16449678},
  {0.37762913, 0.206559},
  // f 1780/1750/823 1671/1749/823 1670/1747/823
  {0.35081071, 0.16683221},
  {0.35393918, 0.21171534},
  {0.36721474, 0.20938909},
  // f 1781/1748/823 1780/1750/823 1670/1747/823
  {0.35702235, 0.16449678},
  {0.35081071, 0.16683221},
  {0.36721474, 0.20938909},
  // f 1778/1752/823 1672/1751/823 1671/1749/823
  {0.34323972, 0.16873801},
  {0.34159538, 0.21287316},
  {0.35393918, 0.21171534},
  // f 1779/1753/823 1778/1752/823 1671/1749/823
  {0.34706286, 0.16790426},
  {0.34323972, 0.16873801},
  {0.35393918, 0.21171534},
  // f 1780/1750/823 1779/1753/823 1671/1749/823
  {0.35081071, 0.16683221},
  {0.34706286, 0.16790426},
  {0.35393918, 0.21171534},
  // f 1776/1755/823 1673/1754/823 1672/1751/823
  {0.33131865, 0.16981006},
  {0.32836121, 0.2132591},
  {0.34159538, 0.21287316},
  // f 1777/1756/823 1776/1755/823 1672/1751/823
  {0.33736387, 0.16954207},
  {0.33131865, 0.16981006},
  {0.34159538, 0.21287316},
  // f 1778/1752/823 1777/1756/823 1672/1751/823
  {0.34323972, 0.16873801},
  {0.33736387, 0.16954207},
  {0.34159538, 0.21287316},
  // f 1774/1758/823 1674/1757/823 1673/1754/823
  {0.32467854, 0.1694566},
  {0.31226727, 0.21256518},
  {0.32836121, 0.2132591},
  // f 1775/1759/823 1774/1758/823 1673/1754/823
  {0.32793793, 0.16972172},
  {0.32467854, 0.1694566},
  {0.32836121, 0.2132591},
  // f 1776/1755/823 1775/1759/823 1673/1754/823
  {0.33131865, 0.16981006},
  {0.32793793, 0.16972172},
  {0.32836121, 0.2132591},
  // f 1773/1761/823 1772/1760/823 1674/1757/823
  {0.32001689, 0.16872764},
  {0.31562820, 0.16760105},
  {0.31226727, 0.21256518},
  // f 1774/1758/823 1773/1761/823 1674/1757/823
  {0.32467854, 0.1694566},
  {0.32001689, 0.16872764},
  {0.31226727, 0.21256518},
  // f 1772/1760/823 1675/1762/823 1674/1757/823
  {0.31562820, 0.16760105},
  {0.29999721, 0.21092689},
  {0.31226727, 0.21256518},
  // f 1771/1764/823 1770/1763/823 1675/1762/823
  {0.31217945, 0.16635847},
  {0.30892023, 0.1648398},
  {0.29999721, 0.21092689},
  // f 1772/1760/823 1771/1764/823 1675/1762/823
  {0.31562820, 0.16760105},
  {0.31217945, 0.16635847},
  {0.29999721, 0.21092689},
  // f 1770/1763/823 1676/1765/823 1675/1762/823
  {0.30892023, 0.1648398},
  {0.28984147, 0.20862836},
  {0.29999721, 0.21092689},
  // f 1770/1763/823 1769/1766/823 1676/1765/823
  {0.30892023, 0.1648398},
  {0.30585057, 0.16304499},
  {0.28984147, 0.20862836},
  // f 1769/1766/823 1677/1767/823 1676/1765/823
  {0.30585057, 0.16304499},
  {0.28151986, 0.20592993},
  {0.28984147, 0.20862836},
  // f 1769/1766/823 1768/1768/823 1677/1767/823
  {0.30585057, 0.16304499},
  {0.30297044, 0.16097403},
  {0.28151986, 0.20592993},
  // f 1768/1768/823 1678/1769/823 1677/1767/823
  {0.30297044, 0.16097403},
  {0.27665609, 0.2039302},
  {0.28151986, 0.20592993},
  // f 1768/1768/823 1679/1770/823 1678/1769/823
  {0.30297044, 0.16097403},
  {0.27116209, 0.20121247},
  {0.27665609, 0.2039302},
  // f 1768/1768/823 1767/1771/823 1679/1770/823
  {0.30297044, 0.16097403},
  {0.30027989, 0.15862703},
  {0.27116209, 0.20121247},
  // f 1767/1771/823 1680/1772/823 1679/1770/823
  {0.30027989, 0.15862703},
  {0.26434046, 0.19705606},
  {0.27116209, 0.20121247},
  // f 1767/1771/823 1766/1773/823 1680/1772/823
  {0.30027989, 0.15862703},
  {0.29777884, 0.15600383},
  {0.26434046, 0.19705606},
  // f 1766/1773/823 1681/1774/823 1680/1772/823
  {0.29777884, 0.15600383},
  {0.25802812, 0.19229674},
  {0.26434046, 0.19705606},
  // f 1766/1773/823 1765/1775/823 1681/1774/823
  {0.29777884, 0.15600383},
  {0.29593080, 0.15367699},
  {0.25802812, 0.19229674},
  // f 1765/1775/823 1682/1776/823 1681/1774/823
  {0.29593080, 0.15367699},
  {0.25222513, 0.18693453},
  {0.25802812, 0.19229674},
  // f 1765/1775/823 1764/1777/823 1682/1776/823
  {0.29593080, 0.15367699},
  {0.29464588, 0.15177727},
  {0.25222513, 0.18693453},
  // f 1764/1777/823 1683/1778/823 1682/1776/823
  {0.29464588, 0.15177727},
  {0.24693146, 0.18096936},
  {0.25222513, 0.18693453},
  // f 1764/1777/823 1763/1779/823 1683/1778/823
  {0.29464588, 0.15177727},
  {0.29269680, 0.1483165},
  {0.24693146, 0.18096936},
  // f 1763/1779/823 1684/1780/823 1683/1778/823
  {0.29269680, 0.1483165},
  {0.24214713, 0.17440128},
  {0.24693146, 0.18096936},
  // f 1763/1779/823 1762/1781/823 1684/1780/823
  {0.29269680, 0.1483165},
  {0.29098836, 0.14448762},
  {0.24214713, 0.17440128},
  // f 1762/1781/823 1685/1782/823 1684/1780/823
  {0.29098836, 0.14448762},
  {0.23941556, 0.16999006},
  {0.24214713, 0.17440128},
  // f 1762/1781/823 1686/1783/823 1685/1782/823
  {0.29098836, 0.14448762},
  {0.23688293, 0.16534334},
  {0.23941556, 0.16999006},
  // f 1762/1781/823 1761/1784/823 1686/1783/823
  {0.29098836, 0.14448762},
  {0.28925586, 0.13940692},
  {0.23688293, 0.16534334},
  // f 1761/1784/823 1687/1785/823 1686/1783/823
  {0.28925586, 0.13940692},
  {0.23289743, 0.1565392},
  {0.23688293, 0.16534334},
  // f 1761/1784/823 1760/1786/823 1687/1785/823
  {0.28925586, 0.13940692},
  {0.28767255, 0.1328094},
  {0.23289743, 0.1565392},
  // f 1760/1786/823 1688/1787/823 1687/1785/823
  {0.28767255, 0.1328094},
  {0.22970027, 0.1472789},
  {0.23289743, 0.1565392},
  // f 1760/1786/823 1689/1788/823 1688/1787/823
  {0.28767255, 0.1328094},
  {0.22729145, 0.13756263},
  {0.22970027, 0.1472789},
  // f 1760/1786/823 1759/1789/823 1689/1788/823
  {0.28767255, 0.1328094},
  {0.28632984, 0.12326646},
  {0.22729145, 0.13756263},
  // f 1759/1789/823 1690/1790/823 1689/1788/823
  {0.28632984, 0.12326646},
  {0.22553959, 0.12623173},
  {0.22729145, 0.13756263},
  // f 1759/1789/823 1691/1791/823 1690/1790/823
  {0.28632984, 0.12326646},
  {0.22495563, 0.11916298},
  {0.22553959, 0.12623173},
  // f 1759/1789/823 1758/1792/823 1691/1791/823
  {0.28632984, 0.12326646},
  {0.28574753, 0.10998303},
  {0.22495563, 0.11916298},
  // f 1758/1792/823 1692/1793/823 1691/1791/823
  {0.28574753, 0.10998303},
  {0.22471718, 0.11065984},
  {0.22495563, 0.11916298},
  // f 1758/1792/823 1693/1794/823 1692/1793/823
  {0.28574753, 0.10998303},
  {0.22536130, 0.09677577},
  {0.22471718, 0.11065984},
  // f 1756/1796/823 1694/1795/823 1693/1794/823
  {0.28668022, 0.0942828100000001},
  {0.22706479, 0.0848645},
  {0.22536130, 0.09677577},
  // f 1757/1797/823 1756/1796/823 1693/1794/823
  {0.28598070, 0.10181326},
  {0.28668022, 0.0942828100000001},
  {0.22536130, 0.09677577},
  // f 1758/1792/823 1757/1797/823 1693/1794/823
  {0.28574753, 0.10998303},
  {0.28598070, 0.10181326},
  {0.22536130, 0.09677577},
  // f 1756/1796/823 1695/1798/823 1694/1795/823
  {0.28668022, 0.0942828100000001},
  {0.22950821, 0.07472301},
  {0.22706479, 0.0848645},
  // f 1755/1800/823 1696/1799/823 1695/1798/823
  {0.28805077, 0.08645952},
  {0.23240410, 0.06616855},
  {0.22950821, 0.07472301},
  // f 1756/1796/823 1755/1800/823 1695/1798/823
  {0.28668022, 0.0942828100000001},
  {0.28805077, 0.08645952},
  {0.22950821, 0.07472301},
  // f 1754/1802/823 1697/1801/823 1696/1799/823
  {0.28974965, 0.08029914},
  {0.23502320, 0.06003696},
  {0.23240410, 0.06616855},
  // f 1755/1800/823 1754/1802/823 1696/1799/823
  {0.28805077, 0.08645952},
  {0.28974965, 0.08029914},
  {0.23240410, 0.06616855},
  // f 1754/1802/823 1698/1803/823 1697/1801/823
  {0.28974965, 0.08029914},
  {0.23802556, 0.05414903},
  {0.23502320, 0.06003696},
  // f 1753/1805/823 1699/1804/823 1698/1803/823
  {0.29157701, 0.07552767},
  {0.24260084, 0.0467571},
  {0.23802556, 0.05414903},
  // f 1754/1802/823 1753/1805/823 1698/1803/823
  {0.28974965, 0.08029914},
  {0.29157701, 0.07552767},
  {0.23802556, 0.05414903},
  // f 1752/1807/823 1700/1806/823 1699/1804/823
  {0.29336154, 0.07191032},
  {0.24712366, 0.04077524},
  {0.24260084, 0.0467571},
  // f 1753/1805/823 1752/1807/823 1699/1804/823
  {0.29157701, 0.07552767},
  {0.29336154, 0.07191032},
  {0.24260084, 0.0467571},
  // f 1752/1807/823 1701/1808/823 1700/1806/823
  {0.29336154, 0.07191032},
  {0.25213128, 0.03524703},
  {0.24712366, 0.04077524},
  // f 1751/1810/823 1702/1809/823 1701/1808/823
  {0.29538399, 0.0686191900000001},
  {0.25762367, 0.03017253},
  {0.25213128, 0.03524703},
  // f 1752/1807/823 1751/1810/823 1701/1808/823
  {0.29336154, 0.07191032},
  {0.29538399, 0.0686191900000001},
  {0.25213128, 0.03524703},
  // f 1750/1812/823 1703/1811/823 1702/1809/823
  {0.29764441, 0.06565422},
  {0.26449436, 0.02492851},
  {0.25762367, 0.03017253},
  // f 1751/1810/823 1750/1812/823 1702/1809/823
  {0.29538399, 0.0686191900000001},
  {0.29764441, 0.06565422},
  {0.25762367, 0.03017253},
  // f 1749/1814/823 1704/1813/823 1703/1811/823
  {0.29959995, 0.06350243},
  {0.27006289, 0.02138436},
  {0.26449436, 0.02492851},
  // f 1750/1812/823 1749/1814/823 1703/1811/823
  {0.29764441, 0.06565422},
  {0.29959995, 0.06350243},
  {0.26449436, 0.02492851},
  // f 1748/1816/823 1705/1815/823 1704/1813/823
  {0.30113494, 0.06200624},
  {0.27700970, 0.01767081},
  {0.27006289, 0.02138436},
  // f 1749/1814/823 1748/1816/823 1704/1813/823
  {0.29959995, 0.06350243},
  {0.30113494, 0.06200624},
  {0.27006289, 0.02138436},
  // f 1747/1818/823 1706/1817/823 1705/1815/823
  {0.30382338, 0.05973673},
  {0.28377369, 0.01484156},
  {0.27700970, 0.01767081},
  // f 1748/1816/823 1747/1818/823 1705/1815/823
  {0.30113494, 0.06200624},
  {0.30382338, 0.05973673},
  {0.27700970, 0.01767081},
  // f 1746/1820/823 1707/1819/823 1706/1817/823
  {0.30667448, 0.05774736},
  {0.29223055, 0.01224679},
  {0.28377369, 0.01484156},
  // f 1747/1818/823 1746/1820/823 1706/1817/823
  {0.30382338, 0.05973673},
  {0.30667448, 0.05774736},
  {0.28377369, 0.01484156},
  // f 1744/1822/823 1708/1821/823 1707/1819/823
  {0.31286460, 0.0546093},
  {0.30267936, 0.01009411},
  {0.29223055, 0.01224679},
  // f 1745/1823/823 1744/1822/823 1707/1819/823
  {0.30968821, 0.05603826},
  {0.31286460, 0.0546093},
  {0.29223055, 0.01224679},
  // f 1746/1820/823 1745/1823/823 1707/1819/823
  {0.30667448, 0.05774736},
  {0.30968821, 0.05603826},
  {0.29223055, 0.01224679},
  // f 1743/1825/823 1709/1824/823 1708/1821/823
  {0.31620365, 0.05346054},
  {0.31545249, 0.00861411999999995},
  {0.30267936, 0.01009411},
  // f 1744/1822/823 1743/1825/823 1708/1821/823
  {0.31286460, 0.0546093},
  {0.31620365, 0.05346054},
  {0.30267936, 0.01009411},
  // f 1741/1827/823 1710/1826/823 1709/1824/823
  {0.32488105, 0.05184668},
  {0.33091536, 0.00806057000000004},
  {0.31545249, 0.00861411999999995},
  // f 1742/1828/823 1741/1827/823 1709/1824/823
  {0.32042524, 0.05245185},
  {0.32488105, 0.05184668},
  {0.31545249, 0.00861411999999995},
  // f 1743/1825/823 1742/1828/823 1709/1824/823
  {0.31620365, 0.05346054},
  {0.32042524, 0.05245185},
  {0.31545249, 0.00861411999999995},
  // f 1738/1830/823 1711/1829/823 1710/1826/823
  {0.33623871, 0.0520475500000001},
  {0.35441339, 0.00901346999999997},
  {0.33091536, 0.00806057000000004},
  // f 1739/1831/823 1738/1830/823 1710/1826/823
  {0.33324152, 0.05176169},
  {0.33623871, 0.0520475500000001},
  {0.33091536, 0.00806057000000004},
  // f 1740/1832/823 1739/1831/823 1710/1826/823
  {0.32957107, 0.05164492},
  {0.33324152, 0.05176169},
  {0.33091536, 0.00806057000000004},
  // f 1741/1827/823 1740/1832/823 1710/1826/823
  {0.32488105, 0.05184668},
  {0.32957107, 0.05164492},
  {0.33091536, 0.00806057000000004},
  // f 1735/1834/823 1712/1833/823 1711/1829/823
  {0.34645483, 0.0547323199999999},
  {0.36772659, 0.01081449},
  {0.35441339, 0.00901346999999997},
  // f 1736/1835/823 1735/1834/823 1711/1829/823
  {0.34355935, 0.05364841},
  {0.34645483, 0.0547323199999999},
  {0.35441339, 0.00901346999999997},
  // f 1737/1836/823 1736/1835/823 1711/1829/823
  {0.34002122, 0.0526955099999999},
  {0.34355935, 0.05364841},
  {0.35441339, 0.00901346999999997},
  // f 1738/1830/823 1737/1836/823 1711/1829/823
  {0.33623871, 0.0520475500000001},
  {0.34002122, 0.0526955099999999},
  {0.35441339, 0.00901346999999997},
  // f 1733/1838/823 1713/1837/823 1712/1833/823
  {0.35168448, 0.05760056},
  {0.37417486, 0.01226288},
  {0.36772659, 0.01081449},
  // f 1734/1839/823 1733/1838/823 1712/1833/823
  {0.34916317, 0.0560497},
  {0.35168448, 0.05760056},
  {0.36772659, 0.01081449},
  // f 1735/1834/823 1734/1839/823 1712/1833/823
  {0.34645483, 0.0547323199999999},
  {0.34916317, 0.0560497},
  {0.36772659, 0.01081449},
  // f 1732/1841/823 1714/1840/823 1713/1837/823
  {0.35401347, 0.05937159},
  {0.37977996, 0.01401621},
  {0.37417486, 0.01226288},
  // f 1733/1838/823 1732/1841/823 1713/1837/823
  {0.35168448, 0.05760056},
  {0.35401347, 0.05937159},
  {0.37417486, 0.01226288},
  // f 1732/1841/823 1715/1842/823 1714/1840/823
  {0.35401347, 0.05937159},
  {0.38603893, 0.01666105},
  {0.37977996, 0.01401621},
  // f 1731/1844/823 1716/1843/823 1715/1842/823
  {0.35585245, 0.06105435},
  {0.39129499, 0.01949042},
  {0.38603893, 0.01666105},
  // f 1732/1841/823 1731/1844/823 1715/1842/823
  {0.35401347, 0.05937159},
  {0.35585245, 0.06105435},
  {0.38603893, 0.01666105},
  // f 1730/1846/823 1717/1845/823 1716/1843/823
  {0.35781443, 0.06320971},
  {0.39619392, 0.02274519},
  {0.39129499, 0.01949042},
  // f 1731/1844/823 1730/1846/823 1716/1843/823
  {0.35585245, 0.06105435},
  {0.35781443, 0.06320971},
  {0.39129499, 0.01949042},
  // f 1730/1846/823 1718/1847/823 1717/1845/823
  {0.35781443, 0.06320971},
  {0.40073571, 0.02642542},
  {0.39619392, 0.02274519},
  // f 1729/1849/823 1719/1848/823 1718/1847/823
  {0.35957879, 0.06557196},
  {0.40356517, 0.02911514},
  {0.40073571, 0.02642542},
  // f 1730/1846/823 1729/1849/823 1718/1847/823
  {0.35781443, 0.06320971},
  {0.35957879, 0.06557196},
  {0.40073571, 0.02642542},
  // f 1729/1849/823 1720/1850/823 1719/1848/823
  {0.35957879, 0.06557196},
  {0.40666553, 0.03249216},
  {0.40356517, 0.02911514},
  // f 1728/1852/823 1721/1851/823 1720/1850/823
  {0.36114553, 0.06814122},
  {0.41034085, 0.0372166},
  {0.40666553, 0.03249216},
  // f 1729/1849/823 1728/1852/823 1720/1850/823
  {0.35957879, 0.06557196},
  {0.36114553, 0.06814122},
  {0.40666553, 0.03249216},
  // f 1728/1852/823 1722/1853/823 1721/1851/823
  {0.36114553, 0.06814122},
  {0.41402096, 0.04297513},
  {0.41034085, 0.0372166},
  // f 1727/1855/823 1723/1854/823 1722/1853/823
  {0.36269417, 0.07133079},
  {0.41727629, 0.04926974},
  {0.41402096, 0.04297513},
  // f 1728/1852/823 1727/1855/823 1722/1853/823
  {0.36114553, 0.06814122},
  {0.36269417, 0.07133079},
  {0.41402096, 0.04297513},
  // f 1726/1857/823 1724/1856/823 1723/1854/823
  {0.36398467, 0.07479072},
  {0.42010674, 0.05610031},
  {0.41727629, 0.04926974},
  // f 1727/1855/823 1726/1857/823 1723/1854/823
  {0.36269417, 0.07133079},
  {0.36398467, 0.07479072},
  {0.41727629, 0.04926974},
  // f 1726/1857/823 1725/1858/823 1724/1856/823
  {0.36398467, 0.07479072},
  {0.42272970, 0.06423301},
  {0.42010674, 0.05610031},
  // f 1787/1861/823 1786/1860/823 1785/1859/823
  {0.06890127, 0.29881746},
  {0.06890127, 0.22938025},
  {0.00800536, 0.22938025},
  // f 1794/1862/823 1787/1861/823 1785/1859/823
  {0.06890127, 0.34754533},
  {0.06890127, 0.29881746},
  {0.00800536, 0.22938025},
  // f 1796/1863/823 1794/1862/823 1785/1859/823
  {0.00800536, 0.42781097},
  {0.06890127, 0.34754533},
  {0.00800536, 0.22938025},
  // f 1794/1862/823 1788/1864/823 1787/1861/823
  {0.06890127, 0.34754533},
  {0.13544317, 0.29881746},
  {0.06890127, 0.29881746},
  // f 1790/1866/823 1789/1865/823 1788/1864/823
  {0.19660792, 0.22938025},
  {0.13544317, 0.22938025},
  {0.13544317, 0.29881746},
  // f 1793/1867/823 1790/1866/823 1788/1864/823
  {0.13544317, 0.34754533},
  {0.19660792, 0.22938025},
  {0.13544317, 0.29881746},
  // f 1794/1862/823 1793/1867/823 1788/1864/823
  {0.06890127, 0.34754533},
  {0.13544317, 0.34754533},
  {0.13544317, 0.29881746},
  // f 1793/1867/823 1791/1868/823 1790/1866/823
  {0.13544317, 0.34754533},
  {0.19660792, 0.42781097},
  {0.19660792, 0.22938025},
  // f 1793/1867/823 1792/1869/823 1791/1868/823
  {0.13544317, 0.34754533},
  {0.13544317, 0.42781097},
  {0.19660792, 0.42781097},
  // f 1796/1863/823 1795/1870/823 1794/1862/823
  {0.00800536, 0.42781097},
  {0.06890127, 0.42781097},
  {0.06890127, 0.34754533},
  // f 1803/1873/823 1798/1872/823 1797/1871/823
  {0.50353479, 0.05705917},
  {0.62653643, 0.00806057000000004},
  {0.44142902, 0.00806057000000004},
  // f 1804/1874/823 1803/1873/823 1797/1871/823
  {0.44142902, 0.05705917},
  {0.50353479, 0.05705917},
  {0.44142902, 0.00806057000000004},
  // f 1800/1876/823 1799/1875/823 1798/1872/823
  {0.56443071, 0.05705917},
  {0.62653643, 0.05705917},
  {0.62653643, 0.00806057000000004},
  // f 1803/1873/823 1800/1876/823 1798/1872/823
  {0.50353479, 0.05705917},
  {0.56443071, 0.05705917},
  {0.62653643, 0.00806057000000004},
  // f 1802/1878/823 1801/1877/823 1800/1876/823
  {0.50353479, 0.20649135},
  {0.56443071, 0.20649135},
  {0.56443071, 0.05705917},
  // f 1803/1873/823 1802/1878/823 1800/1876/823
  {0.50353479, 0.05705917},
  {0.50353479, 0.20649135},
  {0.56443071, 0.05705917},
  // f 2/4/1 453/3/1 1/1/1
  {0.23816000, 0.86834013},
  {0.23816000, 0.84061944},
  {0.13693565, 0.86834013},
  // f 3/6/4 454/5/3 2/4/2
  {0.24952707, 0.86834013},
  {0.24952707, 0.84061944},
  {0.23816000, 0.86834013},
  // f 4/8/6 455/7/5 3/6/4
  {0.25872794, 0.86834013},
  {0.25872794, 0.84061944},
  {0.24952707, 0.86834013},
  // f 5/10/8 456/9/7 4/8/6
  {0.26608145, 0.86834013},
  {0.26608145, 0.84061944},
  {0.25872794, 0.86834013},
  // f 6/12/10 457/11/9 5/10/824
  {0.27186647, 0.86834013},
  {0.27186647, 0.84061944},
  {0.26608145, 0.86834013},
  // f 7/14/12 458/13/11 6/12/825
  {0.27717340, 0.86834013},
  {0.27717340, 0.84061944},
  {0.27186647, 0.86834013},
  // f 8/16/14 459/15/13 7/14/826
  {0.28123057, 0.86834013},
  {0.28123057, 0.84061944},
  {0.27717340, 0.86834013},
  // f 9/18/16 460/17/15 8/16/14
  {0.28423724, 0.86834013},
  {0.28423724, 0.84061944},
  {0.28123057, 0.86834013},
  // f 10/1879/18 461/19/17 9/18/16
  {0.28769672, 0.86834013},
  {0.28769672, 0.84061944},
  {0.28423724, 0.86834013},
  // f 11/23/20 462/22/19 10/20/827
  {0.85803604, 0.12260771},
  {0.88556695, 0.12260771},
  {0.85803604, 0.11930633},
  // f 12/25/22 463/24/21 11/23/828
  {0.85803604, 0.12617695},
  {0.88556695, 0.12617695},
  {0.85803604, 0.12260771},
  // f 13/27/24 464/26/23 12/25/829
  {0.85803604, 0.1300143},
  {0.88556695, 0.1300143},
  {0.85803604, 0.12617695},
  // f 14/29/26 465/28/25 13/27/830
  {0.85803604, 0.13411963},
  {0.88556695, 0.13411963},
  {0.85803604, 0.1300143},
  // f 15/31/28 466/30/27 14/29/831
  {0.85803604, 0.13939977},
  {0.88556695, 0.13939977},
  {0.85803604, 0.13411963},
  // f 16/33/30 467/32/29 15/31/28
  {0.85803604, 0.14506578},
  {0.88556695, 0.14506578},
  {0.85803604, 0.13939977},
  // f 17/35/32 468/34/31 16/33/832
  {0.85803604, 0.1511178},
  {0.88556695, 0.1511178},
  {0.85803604, 0.14506578},
  // f 18/37/34 469/36/33 17/35/833
  {0.85803604, 0.15755576},
  {0.88556695, 0.15755576},
  {0.85803604, 0.1511178},
  // f 19/39/36 470/38/35 18/37/34
  {0.85803604, 0.1643796},
  {0.88556695, 0.1643796},
  {0.85803604, 0.15755576},
  // f 20/41/38 471/40/37 19/39/834
  {0.85803604, 0.17253596},
  {0.88556695, 0.17253596},
  {0.85803604, 0.1643796},
  // f 21/43/40 472/42/39 20/41/835
  {0.85803604, 0.18017763},
  {0.88556695, 0.18017763},
  {0.85803604, 0.17253596},
  // f 22/45/42 473/44/41 21/43/836
  {0.85803604, 0.18631804},
  {0.88556695, 0.18631804},
  {0.85803604, 0.18017763},
  // f 23/47/44 474/46/43 22/45/42
  {0.85803604, 0.19208026},
  {0.88556695, 0.19208026},
  {0.85803604, 0.18631804},
  // f 24/49/46 475/48/45 23/47/837
  {0.85803604, 0.19659328},
  {0.88556695, 0.19659328},
  {0.85803604, 0.19208026},
  // f 25/51/48 476/50/47 24/49/838
  {0.85803604, 0.20084369},
  {0.88556695, 0.20084369},
  {0.85803604, 0.19659328},
  // f 26/53/50 477/52/49 25/51/839
  {0.85803604, 0.20483148},
  {0.88556695, 0.20483148},
  {0.85803604, 0.20084369},
  // f 27/55/52 478/54/51 26/53/50
  {0.85803604, 0.20783269},
  {0.88556695, 0.20783269},
  {0.85803604, 0.20483148},
  // f 28/1880/54 479/56/53 27/55/52
  {0.85803604, 0.21134788},
  {0.88556695, 0.21134788},
  {0.85803604, 0.20783269},
  // f 29/60/56 480/59/55 28/57/840
  {0.37149429, 0.88446128},
  {0.37149429, 0.91218209},
  {0.37527239, 0.88446128},
  // f 30/62/58 481/61/57 29/60/56
  {0.36733842, 0.88446128},
  {0.36733842, 0.91218209},
  {0.37149429, 0.88446128},
  // f 31/64/60 482/63/59 30/62/58
  {0.36280486, 0.88446128},
  {0.36280486, 0.91218209},
  {0.36733842, 0.88446128},
  // f 32/66/62 483/65/61 31/64/60
  {0.35686591, 0.88446128},
  {0.35686591, 0.91218209},
  {0.36280486, 0.88446128},
  // f 33/68/64 484/67/63 32/66/62
  {0.35038304, 0.88446128},
  {0.35038304, 0.91218209},
  {0.35686591, 0.88446128},
  // f 34/70/66 485/69/65 33/68/64
  {0.34213221, 0.88446128},
  {0.34213221, 0.91218209},
  {0.35038304, 0.88446128},
  // f 35/72/68 486/71/67 34/70/66
  {0.33179602, 0.88446128},
  {0.33179602, 0.91218209},
  {0.34213221, 0.88446128},
  // f 36/74/69 487/73/69 35/72/68
  {0.32049292, 0.88446128},
  {0.32049292, 0.91218209},
  {0.33179602, 0.88446128},
  // f 37/1881/70 488/75/70 36/74/70
  {0.28715476, 0.88446128},
  {0.28715476, 0.91218209},
  {0.32049292, 0.88446128},
  // f 38/1882/71 489/78/71 37/76/71
  {0.87738061, 0.66633463},
  {0.90491146, 0.66633463},
  {0.87738061, 0.59270144},
  // f 39/1883/70 490/81/70 38/79/70
  {0.20997930, 0.92830324},
  {0.20997930, 0.95602393},
  {0.27114406, 0.92830324},
  // f 1/1884/72 452/84/72 39/82/72
  {0.59157205, 0.65848398},
  {0.56404120, 0.65848398},
  {0.59157205, 0.85691476},
  // f 41/88/1 492/87/1 40/85/1
  {0.94630802, 0.15749347},
  {0.94630802, 0.12977266},
  {0.93138647, 0.15749347},
  // f 42/90/75 493/89/74 41/88/73
  {0.95354897, 0.15749347},
  {0.95354897, 0.12977266},
  {0.94630802, 0.15749347},
  // f 43/92/77 494/91/76 42/90/75
  {0.95871848, 0.15749347},
  {0.95871848, 0.12977266},
  {0.95354897, 0.15749347},
  // f 44/94/79 495/93/78 43/92/77
  {0.96226937, 0.15749347},
  {0.96226937, 0.12977266},
  {0.95871848, 0.15749347},
  // f 45/96/81 496/95/80 44/94/79
  {0.96498585, 0.15749347},
  {0.96498585, 0.12977266},
  {0.96226937, 0.15749347},
  // f 46/98/83 497/97/82 45/96/81
  {0.96701896, 0.15749347},
  {0.96701896, 0.12977266},
  {0.96498585, 0.15749347},
  // f 47/100/85 498/99/84 46/98/83
  {0.96884233, 0.15749347},
  {0.96884233, 0.12977266},
  {0.96701896, 0.15749347},
  // f 48/102/87 499/101/86 47/100/85
  {0.96983576, 0.15749347},
  {0.96983576, 0.12977266},
  {0.96884233, 0.15749347},
  // f 49/104/89 500/103/88 48/102/841
  {0.97104275, 0.15749347},
  {0.97104275, 0.12977266},
  {0.96983576, 0.15749347},
  // f 50/1885/91 501/105/90 49/104/842
  {0.97265011, 0.15749347},
  {0.97265011, 0.12977266},
  {0.97104275, 0.15749347},
  // f 51/109/93 502/108/92 50/106/843
  {0.99199462, 0.6211518},
  {0.96446377, 0.6211518},
  {0.99199462, 0.62277901},
  // f 52/111/95 503/110/94 51/109/93
  {0.99199462, 0.61971998},
  {0.96446377, 0.61971998},
  {0.99199462, 0.6211518},
  // f 53/113/97 504/112/96 52/111/95
  {0.99199462, 0.61821902},
  {0.96446377, 0.61821902},
  {0.99199462, 0.61971998},
  // f 54/115/99 505/114/98 53/113/844
  {0.99199462, 0.61664903},
  {0.96446377, 0.61664903},
  {0.99199462, 0.61821902},
  // f 55/117/101 506/116/100 54/115/845
  {0.99199462, 0.61501002},
  {0.96446377, 0.61501002},
  {0.99199462, 0.61664903},
  // f 56/119/103 507/118/102 55/117/101
  {0.99199462, 0.61330199},
  {0.96446377, 0.61330199},
  {0.99199462, 0.61501002},
  // f 57/121/105 508/120/104 56/119/846
  {0.99199462, 0.61116123},
  {0.96446377, 0.61116123},
  {0.99199462, 0.61330199},
  // f 58/123/107 509/122/106 57/121/105
  {0.99199462, 0.61005354},
  {0.96446377, 0.61005354},
  {0.99199462, 0.61116123},
  // f 59/125/109 510/124/108 58/123/107
  {0.99199462, 0.60853791},
  {0.96446377, 0.60853791},
  {0.99199462, 0.61005354},
  // f 60/127/111 511/126/110 59/125/847
  {0.99199462, 0.60597861},
  {0.96446377, 0.60597861},
  {0.99199462, 0.60853791},
  // f 61/129/113 512/128/112 60/127/111
  {0.99199462, 0.60387349},
  {0.96446377, 0.60387349},
  {0.99199462, 0.60597861},
  // f 62/131/115 513/130/114 61/129/113
  {0.99199462, 0.60185039},
  {0.96446377, 0.60185039},
  {0.99199462, 0.60387349},
  // f 63/133/117 514/132/116 62/131/848
  {0.99199462, 0.59990907},
  {0.96446377, 0.59990907},
  {0.99199462, 0.60185039},
  // f 64/135/119 515/134/118 63/133/117
  {0.99199462, 0.59804964},
  {0.96446377, 0.59804964},
  {0.99199462, 0.59990907},
  // f 65/137/121 516/136/120 64/135/849
  {0.99199462, 0.59627199},
  {0.96446377, 0.59627199},
  {0.99199462, 0.59804964},
  // f 66/139/123 517/138/122 65/137/850
  {0.99199462, 0.59457624},
  {0.96446377, 0.59457624},
  {0.99199462, 0.59627199},
  // f 67/141/125 518/140/124 66/139/851
  {0.99199462, 0.59375906},
  {0.96446377, 0.59375906},
  {0.99199462, 0.59457624},
  // f 68/1886/127 519/142/126 67/141/852
  {0.99199462, 0.59270144},
  {0.96446377, 0.59270144},
  {0.99199462, 0.59375906},
  // f 69/146/129 520/145/128 68/143/853
  {0.70665878, 0.9641037},
  {0.70665878, 0.99182439},
  {0.70771784, 0.9641037},
  // f 70/148/131 521/147/130 69/146/854
  {0.70546031, 0.9641037},
  {0.70546031, 0.99182439},
  {0.70665878, 0.9641037},
  // f 71/150/133 522/149/132 70/148/855
  {0.70412248, 0.9641037},
  {0.70412248, 0.99182439},
  {0.70546031, 0.9641037},
  // f 72/152/135 523/151/134 71/150/133
  {0.70225424, 0.9641037},
  {0.70225424, 0.99182439},
  {0.70412248, 0.9641037},
  // f 73/154/137 524/153/136 72/152/135
  {0.70016825, 0.9641037},
  {0.70016825, 0.99182439},
  {0.70225424, 0.9641037},
  // f 74/156/139 525/155/138 73/154/137
  {0.69737750, 0.9641037},
  {0.69737750, 0.99182439},
  {0.70016825, 0.9641037},
  // f 75/158/141 526/157/140 74/156/139
  {0.69372529, 0.9641037},
  {0.69372529, 0.99182439},
  {0.69737750, 0.9641037},
  // f 76/160/143 527/159/142 75/158/141
  {0.68964624, 0.9641037},
  {0.68964624, 0.99182439},
  {0.69372529, 0.9641037},
  // f 77/162/144 528/161/144 76/160/143
  {0.68446183, 0.9641037},
  {0.68446183, 0.99182439},
  {0.68964624, 0.9641037},
  // f 78/1887/70 529/163/70 77/162/70
  {0.66712064, 0.9641037},
  {0.66712064, 0.99182439},
  {0.68446183, 0.9641037},
  // f 40/1888/71 491/166/71 78/164/71
  {0.94511920, 0.11365151},
  {0.97265011, 0.11365151},
  {0.94511920, 0.06939036},
  // f 80/1889/145 531/169/145 79/167/145
  {0.93531907, 0.98454225},
  {0.93531907, 0.95682156},
  {0.87738061, 0.98454225},
  // f 81/173/149 532/172/148 80/170/856
  {0.72372866, 0.97431076},
  {0.75125951, 0.97431076},
  {0.72372866, 0.9678787},
  // f 82/175/151 533/174/150 81/173/149
  {0.72372866, 0.97907531},
  {0.75125951, 0.97907531},
  {0.72372866, 0.97431076},
  // f 83/177/153 534/176/152 82/175/857
  {0.72372866, 0.98290038},
  {0.75125951, 0.98290038},
  {0.72372866, 0.97907531},
  // f 84/179/155 535/178/154 83/177/858
  {0.72372866, 0.98593235},
  {0.75125951, 0.98593235},
  {0.72372866, 0.98290038},
  // f 85/181/157 536/180/156 84/179/859
  {0.72372866, 0.98765135},
  {0.75125951, 0.98765135},
  {0.72372866, 0.98593235},
  // f 86/183/159 537/182/158 85/181/157
  {0.72372866, 0.98953557},
  {0.75125951, 0.98953557},
  {0.72372866, 0.98765135},
  // f 87/1890/161 538/184/160 86/183/860
  {0.72372866, 0.99182439},
  {0.75125951, 0.99182439},
  {0.72372866, 0.98953557},
  // f 88/188/163 539/187/162 87/185/161
  {0.50329673, 0.98844051},
  {0.50329673, 0.9607197},
  {0.50117278, 0.98844051},
  // f 89/190/165 540/189/164 88/188/861
  {0.50558197, 0.98844051},
  {0.50558197, 0.9607197},
  {0.50329673, 0.98844051},
  // f 90/192/167 541/191/166 89/190/165
  {0.50802857, 0.98844051},
  {0.50802857, 0.9607197},
  {0.50558197, 0.98844051},
  // f 91/194/169 542/193/168 90/192/862
  {0.51063651, 0.98844051},
  {0.51063651, 0.9607197},
  {0.50802857, 0.98844051},
  // f 92/196/171 543/195/170 91/194/169
  {0.51340568, 0.98844051},
  {0.51340568, 0.9607197},
  {0.51063651, 0.98844051},
  // f 93/198/173 544/197/172 92/196/863
  {0.51633626, 0.98844051},
  {0.51633626, 0.9607197},
  {0.51340568, 0.98844051},
  // f 94/200/175 545/199/174 93/198/173
  {0.52006572, 0.98844051},
  {0.52006572, 0.9607197},
  {0.51633626, 0.98844051},
  // f 95/202/177 546/201/176 94/200/175
  {0.52268124, 0.98844051},
  {0.52268124, 0.9607197},
  {0.52006572, 0.98844051},
  // f 96/204/179 547/203/178 95/202/177
  {0.52609569, 0.98844051},
  {0.52609569, 0.9607197},
  {0.52268124, 0.98844051},
  // f 97/206/181 548/205/180 96/204/179
  {0.53059614, 0.98844051},
  {0.53059614, 0.9607197},
  {0.52609569, 0.98844051},
  // f 98/208/183 549/207/182 97/206/181
  {0.53427106, 0.98844051},
  {0.53427106, 0.9607197},
  {0.53059614, 0.98844051},
  // f 99/210/185 550/209/184 98/208/183
  {0.53723544, 0.98844051},
  {0.53723544, 0.9607197},
  {0.53427106, 0.98844051},
  // f 100/212/187 551/211/186 99/210/864
  {0.53958958, 0.98844051},
  {0.53958958, 0.9607197},
  {0.53723544, 0.98844051},
  // f 101/214/189 552/213/188 100/212/187
  {0.54177153, 0.98844051},
  {0.54177153, 0.9607197},
  {0.53958958, 0.98844051},
  // f 102/216/191 553/215/190 101/214/189
  {0.54378122, 0.98844051},
  {0.54378122, 0.9607197},
  {0.54177153, 0.98844051},
  // f 103/218/193 554/217/192 102/216/191
  {0.54502523, 0.98844051},
  {0.54502523, 0.9607197},
  {0.54378122, 0.98844051},
  // f 104/220/195 555/219/194 103/218/865
  {0.54619271, 0.98844051},
  {0.54619271, 0.9607197},
  {0.54502523, 0.98844051},
  // f 105/1891/197 556/221/196 104/220/866
  {0.54803050, 0.98844051},
  {0.54803050, 0.9607197},
  {0.54619271, 0.98844051},
  // f 106/225/199 557/224/198 105/222/197
  {0.48516199, 0.98373342},
  {0.45763114, 0.98373342},
  {0.48516199, 0.98530579},
  // f 107/227/201 558/226/200 106/225/867
  {0.48516199, 0.98210466},
  {0.45763114, 0.98210466},
  {0.48516199, 0.98373342},
  // f 108/229/203 559/228/202 107/227/868
  {0.48516199, 0.98070407},
  {0.45763114, 0.98070407},
  {0.48516199, 0.98210466},
  // f 109/231/205 560/230/204 108/229/869
  {0.48516199, 0.97926438},
  {0.45763114, 0.97926438},
  {0.48516199, 0.98070407},
  // f 110/233/207 561/232/206 109/231/205
  {0.48516199, 0.97778535},
  {0.45763114, 0.97778535},
  {0.48516199, 0.97926438},
  // f 111/235/209 562/234/208 110/233/870
  {0.48516199, 0.9762671},
  {0.45763114, 0.9762671},
  {0.48516199, 0.97778535},
  // f 112/237/211 563/236/210 111/235/871
  {0.48516199, 0.97470951},
  {0.45763114, 0.97470951},
  {0.48516199, 0.9762671},
  // f 113/239/213 564/238/212 112/237/872
  {0.48516199, 0.97375619},
  {0.45763114, 0.97375619},
  {0.48516199, 0.97470951},
  // f 114/241/215 565/240/214 113/239/873
  {0.48516199, 0.97278869},
  {0.45763114, 0.97278869},
  {0.48516199, 0.97375619},
  // f 115/243/217 566/242/216 114/241/215
  {0.48516199, 0.97126186},
  {0.45763114, 0.97126186},
  {0.48516199, 0.97278869},
  // f 116/245/219 567/244/218 115/243/874
  {0.48516199, 0.96976757},
  {0.45763114, 0.96976757},
  {0.48516199, 0.97126186},
  // f 117/247/221 568/246/220 116/245/875
  {0.48516199, 0.96830571},
  {0.45763114, 0.96830571},
  {0.48516199, 0.96976757},
  // f 118/249/223 569/248/222 117/247/876
  {0.48516199, 0.96687639},
  {0.45763114, 0.96687639},
  {0.48516199, 0.96830571},
  // f 119/251/225 570/250/224 118/249/877
  {0.48516199, 0.96547949},
  {0.45763114, 0.96547949},
  {0.48516199, 0.96687639},
  // f 120/253/227 571/252/226 119/251/225
  {0.48516199, 0.96384609},
  {0.45763114, 0.96384609},
  {0.48516199, 0.96547949},
  // f 121/255/229 572/254/228 120/253/878
  {0.48516199, 0.96225965},
  {0.45763114, 0.96225965},
  {0.48516199, 0.96384609},
  // f 122/1892/231 573/256/230 121/255/879
  {0.48516199, 0.9607197},
  {0.45763114, 0.9607197},
  {0.48516199, 0.96225965},
  // f 123/260/233 574/259/232 122/257/880
  {0.10020646, 0.88318765},
  {0.10020646, 0.91090834},
  {0.10195672, 0.88318765},
  // f 124/262/235 575/261/234 123/260/881
  {0.09901745, 0.88318765},
  {0.09901745, 0.91090834},
  {0.10020646, 0.88318765},
  // f 125/264/237 576/263/236 124/262/882
  {0.09721183, 0.88318765},
  {0.09721183, 0.91090834},
  {0.09901745, 0.88318765},
  // f 126/266/239 577/265/238 125/264/237
  {0.09408768, 0.88318765},
  {0.09408768, 0.91090834},
  {0.09721183, 0.88318765},
  // f 127/268/241 578/267/240 126/266/239
  {0.08904894, 0.88318765},
  {0.08904894, 0.91090834},
  {0.09408768, 0.88318765},
  // f 128/270/243 579/269/242 127/268/241
  {0.08054190, 0.88318765},
  {0.08054190, 0.91090834},
  {0.08904894, 0.88318765},
  // f 129/272/245 580/271/244 128/270/243
  {0.06901915, 0.88318765},
  {0.06901915, 0.91090834},
  {0.08054190, 0.88318765},
  // f 130/274/247 581/273/246 129/272/883
  {0.04705226, 0.88318765},
  {0.04705226, 0.91090834},
  {0.06901915, 0.88318765},
  // f 131/276/249 582/275/248 130/274/884
  {0.03357295, 0.88318765},
  {0.03357295, 0.91090834},
  {0.04705226, 0.88318765},
  // f 132/278/251 583/277/250 131/276/249
  {0.02419953, 0.88318765},
  {0.02419953, 0.91090834},
  {0.03357295, 0.88318765},
  // f 133/280/253 584/279/252 132/278/251
  {0.01793888, 0.88318765},
  {0.01793888, 0.91090834},
  {0.02419953, 0.88318765},
  // f 134/282/255 585/281/254 133/280/253
  {0.01476588, 0.88318765},
  {0.01476588, 0.91090834},
  {0.01793888, 0.88318765},
  // f 135/284/257 586/283/256 134/282/885
  {0.01188723, 0.88318765},
  {0.01188723, 0.91090834},
  {0.01476588, 0.88318765},
  // f 136/1893/259 587/285/258 135/284/257
  {0.00800536, 0.88318765},
  {0.00800536, 0.91090834},
  {0.01188723, 0.88318765},
  // f 137/289/261 588/288/260 136/286/259
  {0.90491146, 0.80733752},
  {0.87738061, 0.80733752},
  {0.90491146, 0.81130004},
  // f 138/291/263 589/290/262 137/289/886
  {0.90491146, 0.80385375},
  {0.87738061, 0.80385375},
  {0.90491146, 0.80733752},
  // f 139/293/265 590/292/264 138/291/887
  {0.90491146, 0.80020487},
  {0.87738061, 0.80020487},
  {0.90491146, 0.80385375},
  // f 140/295/267 591/294/266 139/293/888
  {0.90491146, 0.79639077},
  {0.87738061, 0.79639077},
  {0.90491146, 0.80020487},
  // f 141/297/269 592/296/268 140/295/889
  {0.90491146, 0.79241168},
  {0.87738061, 0.79241168},
  {0.90491146, 0.79639077},
  // f 142/299/271 593/298/270 141/297/890
  {0.90491146, 0.78826737},
  {0.87738061, 0.78826737},
  {0.90491146, 0.79241168},
  // f 143/301/273 594/300/272 142/299/891
  {0.90491146, 0.78307617},
  {0.87738061, 0.78307617},
  {0.90491146, 0.78826737},
  // f 144/303/275 595/302/274 143/301/892
  {0.90491146, 0.78039145},
  {0.87738061, 0.78039145},
  {0.90491146, 0.78307617},
  // f 145/305/277 596/304/276 144/303/275
  {0.90491146, 0.77671933},
  {0.87738061, 0.77671933},
  {0.90491146, 0.78039145},
  // f 146/307/279 597/306/278 145/305/893
  {0.90491146, 0.77186823},
  {0.87738061, 0.77186823},
  {0.90491146, 0.77671933},
  // f 147/309/281 598/308/280 146/307/279
  {0.90491146, 0.76710367},
  {0.87738061, 0.76710367},
  {0.90491146, 0.77186823},
  // f 148/311/283 599/310/282 147/309/281
  {0.90491146, 0.76242578},
  {0.87738061, 0.76242578},
  {0.90491146, 0.76710367},
  // f 149/313/285 600/312/284 148/311/894
  {0.90491146, 0.75783455},
  {0.87738061, 0.75783455},
  {0.90491146, 0.76242578},
  // f 150/315/287 601/314/286 149/313/895
  {0.90491146, 0.75332999},
  {0.87738061, 0.75332999},
  {0.90491146, 0.75783455},
  // f 151/317/289 602/316/288 150/315/896
  {0.90491146, 0.7505585},
  {0.87738061, 0.7505585},
  {0.90491146, 0.75332999},
  // f 152/319/291 603/318/290 151/317/897
  {0.90491146, 0.74782097},
  {0.87738061, 0.74782097},
  {0.90491146, 0.7505585},
  // f 153/321/293 604/320/292 152/319/898
  {0.90491146, 0.74412608},
  {0.87738061, 0.74412608},
  {0.90491146, 0.74782097},
  // f 154/1894/295 605/322/294 153/321/899
  {0.90491146, 0.74065948},
  {0.87738061, 0.74065948},
  {0.90491146, 0.74412608},
  // f 155/326/297 606/325/296 154/323/900
  {0.14077538, 0.91218209},
  {0.14077538, 0.88446128},
  {0.13693565, 0.91218209},
  // f 156/328/299 607/327/298 155/326/297
  {0.14506935, 0.91218209},
  {0.14506935, 0.88446128},
  {0.14077538, 0.91218209},
  // f 157/330/301 608/329/300 156/328/901
  {0.14981756, 0.91218209},
  {0.14981756, 0.88446128},
  {0.14506935, 0.91218209},
  // f 158/332/303 609/331/302 157/330/902
  {0.15367682, 0.91218209},
  {0.15367682, 0.88446128},
  {0.14981756, 0.91218209},
  // f 159/334/305 610/333/304 158/332/903
  {0.15779163, 0.91218209},
  {0.15779163, 0.88446128},
  {0.15367682, 0.91218209},
  // f 160/336/307 611/335/306 159/334/305
  {0.16378388, 0.91218209},
  {0.16378388, 0.88446128},
  {0.15779163, 0.91218209},
  // f 161/338/309 612/337/308 160/336/307
  {0.17132731, 0.91218209},
  {0.17132731, 0.88446128},
  {0.16378388, 0.91218209},
  // f 162/340/311 613/339/310 161/338/309
  {0.18170050, 0.91218209},
  {0.18170050, 0.88446128},
  {0.17132731, 0.91218209},
  // f 163/342/313 614/341/312 162/340/311
  {0.19334236, 0.91218209},
  {0.19334236, 0.88446128},
  {0.18170050, 0.91218209},
  // f 164/344/315 615/343/314 163/342/313
  {0.20625295, 0.91218209},
  {0.20625295, 0.88446128},
  {0.19334236, 0.91218209},
  // f 165/346/317 616/345/316 164/344/315
  {0.22200502, 0.91218209},
  {0.22200502, 0.88446128},
  {0.20625295, 0.91218209},
  // f 166/348/319 617/347/318 165/346/317
  {0.23360099, 0.91218209},
  {0.23360099, 0.88446128},
  {0.22200502, 0.91218209},
  // f 167/350/321 618/349/320 166/348/319
  {0.24293223, 0.91218209},
  {0.24293223, 0.88446128},
  {0.23360099, 0.91218209},
  // f 168/352/323 619/351/322 167/350/321
  {0.25034332, 0.91218209},
  {0.25034332, 0.88446128},
  {0.24293223, 0.91218209},
  // f 169/354/325 620/353/324 168/352/323
  {0.25613579, 0.91218209},
  {0.25613579, 0.88446128},
  {0.25034332, 0.91218209},
  // f 170/356/327 621/355/326 169/354/325
  {0.25971031, 0.91218209},
  {0.25971031, 0.88446128},
  {0.25613579, 0.91218209},
  // f 171/358/329 622/357/328 170/356/904
  {0.26385537, 0.91218209},
  {0.26385537, 0.88446128},
  {0.25971031, 0.91218209},
  // f 172/360/331 623/359/330 171/358/329
  {0.26766035, 0.91218209},
  {0.26766035, 0.88446128},
  {0.26385537, 0.91218209},
  // f 173/1895/333 624/361/332 172/360/331
  {0.27114406, 0.91218209},
  {0.27114406, 0.88446128},
  {0.26766035, 0.91218209},
  // f 174/365/335 625/364/334 173/362/333
  {0.95455611, 0.90568602},
  {0.98208696, 0.90568602},
  {0.95455611, 0.90202129},
  // f 175/367/337 626/366/336 174/365/905
  {0.95455611, 0.90968513},
  {0.98208696, 0.90968513},
  {0.95455611, 0.90568602},
  // f 176/369/339 627/368/338 175/367/906
  {0.95455611, 0.91401851},
  {0.98208696, 0.91401851},
  {0.95455611, 0.90968513},
  // f 177/371/341 628/370/340 176/369/907
  {0.95455611, 0.91965985},
  {0.98208696, 0.91965985},
  {0.95455611, 0.91401851},
  // f 178/373/343 629/372/342 177/371/341
  {0.95455611, 0.92578268},
  {0.98208696, 0.92578268},
  {0.95455611, 0.91965985},
  // f 179/375/345 630/374/344 178/373/343
  {0.95455611, 0.93238699},
  {0.98208696, 0.93238699},
  {0.95455611, 0.92578268},
  // f 180/1896/346 631/376/346 179/375/345
  {0.95455611, 0.94070041},
  {0.98208696, 0.94070041},
  {0.95455611, 0.93238699},
  // f 181/1897/347 632/379/347 180/377/347
  {0.92092216, 0.30791265},
  {0.92092216, 0.3356334},
  {0.97832298, 0.30791265},
  // f 182/383/351 633/382/350 181/380/908
  {0.94845307, 0.67224514},
  {0.92092216, 0.67224514},
  {0.94845307, 0.67683244},
  // f 183/385/353 634/384/352 182/383/351
  {0.94845307, 0.66856515},
  {0.92092216, 0.66856515},
  {0.94845307, 0.67224514},
  // f 184/387/355 635/386/354 183/385/909
  {0.94845307, 0.66565228},
  {0.92092216, 0.66565228},
  {0.94845307, 0.66856515},
  // f 185/389/357 636/388/356 184/387/355
  {0.94845307, 0.66338348},
  {0.92092216, 0.66338348},
  {0.94845307, 0.66565228},
  // f 186/391/359 637/390/358 185/389/910
  {0.94845307, 0.66132522},
  {0.92092216, 0.66132522},
  {0.94845307, 0.66338348},
  // f 187/393/361 638/392/360 186/391/359
  {0.94845307, 0.65947747},
  {0.92092216, 0.65947747},
  {0.94845307, 0.66132522},
  // f 188/1898/363 639/394/362 187/393/911
  {0.94845307, 0.65836251},
  {0.92092216, 0.65836251},
  {0.94845307, 0.65947747},
  // f 189/398/365 640/397/364 188/395/912
  {0.68403524, 0.91687787},
  {0.68403524, 0.94459867},
  {0.68517518, 0.91687787},
  // f 190/400/367 641/399/366 189/398/913
  {0.68219662, 0.91687787},
  {0.68219662, 0.94459867},
  {0.68403524, 0.91687787},
  // f 191/402/369 642/401/368 190/400/914
  {0.68019724, 0.91687787},
  {0.68019724, 0.94459867},
  {0.68219662, 0.91687787},
  // f 192/404/371 643/403/370 191/402/369
  {0.67803729, 0.91687787},
  {0.67803729, 0.94459867},
  {0.68019724, 0.91687787},
  // f 193/406/373 644/405/372 192/404/371
  {0.67571664, 0.91687787},
  {0.67571664, 0.94459867},
  {0.67803729, 0.91687787},
  // f 194/408/375 645/407/374 193/406/915
  {0.67280614, 0.91687787},
  {0.67280614, 0.94459867},
  {0.67571664, 0.91687787},
  // f 195/410/377 646/409/376 194/408/375
  {0.66921204, 0.91687787},
  {0.66921204, 0.94459867},
  {0.67280614, 0.91687787},
  // f 196/412/379 647/411/378 195/410/377
  {0.66682070, 0.91687787},
  {0.66682070, 0.94459867},
  {0.66921204, 0.91687787},
  // f 197/414/381 648/413/380 196/412/379
  {0.66380382, 0.91687787},
  {0.66380382, 0.94459867},
  {0.66682070, 0.91687787},
  // f 198/416/383 649/415/382 197/414/381
  {0.65976560, 0.91687787},
  {0.65976560, 0.94459867},
  {0.66380382, 0.91687787},
  // f 199/418/385 650/417/384 198/416/383
  {0.65683705, 0.91687787},
  {0.65683705, 0.94459867},
  {0.65976560, 0.91687787},
  // f 200/420/387 651/419/386 199/418/385
  {0.65449470, 0.91687787},
  {0.65449470, 0.94459867},
  {0.65683705, 0.91687787},
  // f 201/422/389 652/421/388 200/420/387
  {0.65265054, 0.91687787},
  {0.65265054, 0.94459867},
  {0.65449470, 0.91687787},
  // f 202/424/391 653/423/390 201/422/389
  {0.65095741, 0.91687787},
  {0.65095741, 0.94459867},
  {0.65265054, 0.91687787},
  // f 203/426/393 654/425/392 202/424/391
  {0.64966178, 0.91687787},
  {0.64966178, 0.94459867},
  {0.65095741, 0.91687787},
  // f 204/428/395 655/427/394 203/426/916
  {0.64870071, 0.91687787},
  {0.64870071, 0.94459867},
  {0.64966178, 0.91687787},
  // f 205/430/397 656/429/396 204/428/917
  {0.64780688, 0.91687787},
  {0.64780688, 0.94459867},
  {0.64870071, 0.91687787},
  // f 206/1899/399 657/431/398 205/430/397
  {0.64659381, 0.91687787},
  {0.64659381, 0.94459867},
  {0.64780688, 0.91687787},
  // f 207/435/401 658/434/400 206/432/918
  {0.08518086, 0.97039855},
  {0.11271173, 0.97039855},
  {0.08518086, 0.96922553},
  // f 208/437/403 659/436/402 207/435/401
  {0.08518086, 0.97162807},
  {0.11271173, 0.97162807},
  {0.08518086, 0.97039855},
  // f 209/439/405 660/438/404 208/437/919
  {0.08518086, 0.97269595},
  {0.11271173, 0.97269595},
  {0.08518086, 0.97162807},
  // f 210/441/407 661/440/406 209/439/920
  {0.08518086, 0.97380292},
  {0.11271173, 0.97380292},
  {0.08518086, 0.97269595},
  // f 211/443/409 662/442/408 210/441/921
  {0.08518086, 0.97494936},
  {0.11271173, 0.97494936},
  {0.08518086, 0.97380292},
  // f 212/445/411 663/444/410 211/443/409
  {0.08518086, 0.9761349},
  {0.11271173, 0.9761349},
  {0.08518086, 0.97494936},
  // f 213/447/413 664/446/412 212/445/411
  {0.08518086, 0.97735977},
  {0.11271173, 0.97735977},
  {0.08518086, 0.9761349},
  // f 214/449/415 665/448/414 213/447/922
  {0.08518086, 0.97811341},
  {0.11271173, 0.97811341},
  {0.08518086, 0.97735977},
  // f 215/451/417 666/450/416 214/449/923
  {0.08518086, 0.9791404},
  {0.11271173, 0.9791404},
  {0.08518086, 0.97811341},
  // f 216/453/419 667/452/418 215/451/417
  {0.08518086, 0.98007846},
  {0.11271173, 0.98007846},
  {0.08518086, 0.9791404},
  // f 217/455/421 668/454/420 216/453/924
  {0.08518086, 0.98099744},
  {0.11271173, 0.98099744},
  {0.08518086, 0.98007846},
  // f 218/457/423 669/456/422 217/455/925
  {0.08518086, 0.98189759},
  {0.11271173, 0.98189759},
  {0.08518086, 0.98099744},
  // f 219/459/425 670/458/424 218/457/423
  {0.08518086, 0.98277879},
  {0.11271173, 0.98277879},
  {0.08518086, 0.98189759},
  // f 220/461/427 671/460/426 219/459/926
  {0.08518086, 0.98364103},
  {0.11271173, 0.98364103},
  {0.08518086, 0.98277879},
  // f 221/463/336 672/462/428 220/461/927
  {0.08518086, 0.9844842},
  {0.11271173, 0.9844842},
  {0.08518086, 0.98364103},
  // f 222/465/430 673/464/429 221/463/336
  {0.08518086, 0.98547113},
  {0.11271173, 0.98547113},
  {0.08518086, 0.9844842},
  // f 223/1900/432 674/466/431 222/465/928
  {0.08518086, 0.98643076},
  {0.11271173, 0.98643076},
  {0.08518086, 0.98547113},
  // f 224/470/434 675/469/433 223/467/929
  {0.64852452, 0.70855165},
  {0.64852452, 0.68083084},
  {0.64789993, 0.70855165},
  // f 225/472/436 676/471/435 224/470/930
  {0.64920592, 0.70855165},
  {0.64920592, 0.68083084},
  {0.64852452, 0.70855165},
  // f 226/474/438 677/473/437 225/472/436
  {0.64996994, 0.70855165},
  {0.64996994, 0.68083084},
  {0.64920592, 0.70855165},
  // f 227/476/440 678/475/439 226/474/438
  {0.65113783, 0.70855165},
  {0.65113783, 0.68083084},
  {0.64996994, 0.70855165},
  // f 228/478/442 679/477/441 227/476/931
  {0.65285164, 0.70855165},
  {0.65285164, 0.68083084},
  {0.65113783, 0.70855165},
  // f 229/480/444 680/479/443 228/478/442
  {0.65606111, 0.70855165},
  {0.65606111, 0.68083084},
  {0.65285164, 0.70855165},
  // f 230/482/446 681/481/445 229/480/444
  {0.66205037, 0.70855165},
  {0.66205037, 0.68083084},
  {0.65606111, 0.70855165},
  // f 231/484/448 682/483/447 230/482/932
  {0.66963899, 0.70855165},
  {0.66963899, 0.68083084},
  {0.66205037, 0.70855165},
  // f 232/486/450 683/485/449 231/484/448
  {0.70503628, 0.70855165},
  {0.70503628, 0.68083084},
  {0.66963899, 0.70855165},
  // f 233/488/452 684/487/451 232/486/933
  {0.71724188, 0.70855165},
  {0.71724188, 0.68083084},
  {0.70503628, 0.70855165},
  // f 234/490/454 685/489/453 233/488/452
  {0.72683811, 0.70855165},
  {0.72683811, 0.68083084},
  {0.71724188, 0.70855165},
  // f 235/492/456 686/491/455 234/490/454
  {0.73337132, 0.70855165},
  {0.73337132, 0.68083084},
  {0.72683811, 0.70855165},
  // f 236/494/458 687/493/457 235/492/934
  {0.73860759, 0.70855165},
  {0.73860759, 0.68083084},
  {0.73337132, 0.70855165},
  // f 237/496/460 688/495/459 236/494/458
  {0.74275011, 0.70855165},
  {0.74275011, 0.68083084},
  {0.73860759, 0.70855165},
  // f 238/1901/462 689/497/461 237/496/935
  {0.74641085, 0.70855165},
  {0.74641085, 0.68083084},
  {0.74275011, 0.70855165},
  // f 239/501/464 690/500/463 238/498/936
  {0.83383900, 0.83082294},
  {0.86136985, 0.83082294},
  {0.83383900, 0.82742131},
  // f 240/503/466 691/502/465 239/501/464
  {0.83383900, 0.83284175},
  {0.86136985, 0.83284175},
  {0.83383900, 0.83082294},
  // f 241/505/468 692/504/467 240/503/937
  {0.83383900, 0.83491516},
  {0.86136985, 0.83491516},
  {0.83383900, 0.83284175},
  // f 242/507/470 693/506/469 241/505/938
  {0.83383900, 0.83919632},
  {0.86136985, 0.83919632},
  {0.83383900, 0.83491516},
  // f 243/509/472 694/508/471 242/507/470
  {0.83383900, 0.84363687},
  {0.86136985, 0.84363687},
  {0.83383900, 0.83919632},
  // f 244/511/474 695/510/473 243/509/939
  {0.83383900, 0.84823668},
  {0.86136985, 0.84823668},
  {0.83383900, 0.84363687},
  // f 245/513/476 696/512/475 244/511/940
  {0.83383900, 0.85299599},
  {0.86136985, 0.85299599},
  {0.83383900, 0.84823668},
  // f 246/515/478 697/514/477 245/513/941
  {0.83383900, 0.85854077},
  {0.86136985, 0.85854077},
  {0.83383900, 0.85299599},
  // f 247/517/480 698/516/479 246/515/942
  {0.83383900, 0.86428726},
  {0.86136985, 0.86428726},
  {0.83383900, 0.85854077},
  // f 248/519/482 699/518/481 247/517/480
  {0.83383900, 0.87031829},
  {0.86136985, 0.87031829},
  {0.83383900, 0.86428726},
  // f 249/521/484 700/520/483 248/519/482
  {0.83383900, 0.87619674},
  {0.86136985, 0.87619674},
  {0.83383900, 0.87031829},
  // f 250/523/486 701/522/485 249/521/943
  {0.83383900, 0.88192284},
  {0.86136985, 0.88192284},
  {0.83383900, 0.87619674},
  // f 251/525/488 702/524/487 250/523/944
  {0.83383900, 0.88749647},
  {0.86136985, 0.88749647},
  {0.83383900, 0.88192284},
  // f 252/527/490 703/526/489 251/525/945
  {0.83383900, 0.89291763},
  {0.86136985, 0.89291763},
  {0.83383900, 0.88749647},
  // f 253/529/492 704/528/491 252/527/490
  {0.83383900, 0.89622843},
  {0.86136985, 0.89622843},
  {0.83383900, 0.89291763},
  // f 254/531/494 705/530/493 253/529/492
  {0.83383900, 0.89947963},
  {0.86136985, 0.89947963},
  {0.83383900, 0.89622843},
  // f 255/533/496 706/532/495 254/531/946
  {0.83383900, 0.90443468},
  {0.86136985, 0.90443468},
  {0.83383900, 0.89947963},
  // f 256/1902/498 707/534/497 255/533/947
  {0.83383900, 0.90899134},
  {0.86136985, 0.90899134},
  {0.83383900, 0.90443468},
  // f 257/538/500 708/537/499 256/535/948
  {0.69844711, 0.87303603},
  {0.69844711, 0.90075672},
  {0.70286930, 0.87303603},
  // f 258/540/502 709/539/501 257/538/500
  {0.69359821, 0.87303603},
  {0.69359821, 0.90075672},
  {0.69844711, 0.87303603},
  // f 259/542/504 710/541/503 258/540/502
  {0.68832260, 0.87303603},
  {0.68832260, 0.90075672},
  {0.69359821, 0.87303603},
  // f 260/544/506 711/543/505 259/542/949
  {0.68480867, 0.87303603},
  {0.68480867, 0.90075672},
  {0.68832260, 0.87303603},
  // f 261/546/508 712/545/507 260/544/506
  {0.68112803, 0.87303603},
  {0.68112803, 0.90075672},
  {0.68480867, 0.87303603},
  // f 262/548/510 713/547/509 261/546/508
  {0.67403489, 0.87303603},
  {0.67403489, 0.90075672},
  {0.68112803, 0.87303603},
  // f 263/550/512 714/549/511 262/548/510
  {0.66540784, 0.87303603},
  {0.66540784, 0.90075672},
  {0.67403489, 0.87303603},
  // f 264/552/514 715/551/513 263/550/512
  {0.65501136, 0.87303603},
  {0.65501136, 0.90075672},
  {0.66540784, 0.87303603},
  // f 265/554/516 716/553/515 264/552/514
  {0.64473563, 0.87303603},
  {0.64473563, 0.90075672},
  {0.65501136, 0.87303603},
  // f 266/556/518 717/555/517 265/554/516
  {0.63367492, 0.87303603},
  {0.63367492, 0.90075672},
  {0.64473563, 0.87303603},
  // f 267/558/520 718/557/519 266/556/518
  {0.61480123, 0.87303603},
  {0.61480123, 0.90075672},
  {0.63367492, 0.87303603},
  // f 268/560/522 719/559/521 267/558/520
  {0.60148317, 0.87303603},
  {0.60148317, 0.90075672},
  {0.61480123, 0.87303603},
  // f 269/562/524 720/561/523 268/560/522
  {0.59112841, 0.87303603},
  {0.59112841, 0.90075672},
  {0.60148317, 0.87303603},
  // f 270/564/526 721/563/525 269/562/524
  {0.58321762, 0.87303603},
  {0.58321762, 0.90075672},
  {0.59112841, 0.87303603},
  // f 271/566/528 722/565/527 270/564/526
  {0.57730544, 0.87303603},
  {0.57730544, 0.90075672},
  {0.58321762, 0.87303603},
  // f 272/568/530 723/567/529 271/566/528
  {0.57201171, 0.87303603},
  {0.57201171, 0.90075672},
  {0.57730544, 0.87303603},
  // f 273/570/532 724/569/531 272/568/950
  {0.56822187, 0.87303603},
  {0.56822187, 0.90075672},
  {0.57201171, 0.87303603},
  // f 274/572/534 725/571/533 273/570/532
  {0.56647545, 0.87303603},
  {0.56647545, 0.90075672},
  {0.56822187, 0.87303603},
  // f 275/1903/536 726/573/535 274/572/534
  {0.56404120, 0.87303603},
  {0.56404120, 0.90075672},
  {0.56647545, 0.87303603},
  // f 276/577/538 727/576/537 275/574/951
  {0.94845307, 0.63657701},
  {0.92092216, 0.63657701},
  {0.94845307, 0.64224136},
  // f 277/579/540 728/578/539 276/577/952
  {0.94845307, 0.63139236},
  {0.92092216, 0.63139236},
  {0.94845307, 0.63657701},
  // f 278/581/542 729/580/541 277/579/953
  {0.94845307, 0.62590361},
  {0.92092216, 0.62590361},
  {0.94845307, 0.63139236},
  // f 279/583/544 730/582/543 278/581/954
  {0.94845307, 0.62011087},
  {0.92092216, 0.62011087},
  {0.94845307, 0.62590361},
  // f 280/585/546 731/584/545 279/583/955
  {0.94845307, 0.61296833},
  {0.92092216, 0.61296833},
  {0.94845307, 0.62011087},
  // f 281/587/548 732/586/547 280/585/546
  {0.94845307, 0.60541189},
  {0.92092216, 0.60541189},
  {0.94845307, 0.61296833},
  // f 282/589/550 733/588/549 281/587/956
  {0.94845307, 0.59976113},
  {0.92092216, 0.59976113},
  {0.94845307, 0.60541189},
  // f 79/1904/551 530/590/551 282/589/550
  {0.94845307, 0.59270144},
  {0.92092216, 0.59270144},
  {0.94845307, 0.59976113},
  // f 284/1905/1 735/593/1 283/591/1
  {0.34805068, 0.95602393},
  {0.34805068, 0.92830324},
  {0.28715476, 0.95602393},
  // f 285/1906/71 736/596/71 284/594/71
  {0.38159543, 0.81481934},
  {0.40912628, 0.81481934},
  {0.38159543, 0.66525185},
  // f 286/1907/1 737/599/1 285/597/1
  {0.39874810, 0.86834013},
  {0.39874810, 0.84061944},
  {0.30370745, 0.86834013},
  // f 287/1908/71 738/602/71 286/600/71
  {0.92092216, 0.87628448},
  {0.94845307, 0.87628448},
  {0.92092216, 0.82742131},
  // f 288/1909/70 739/605/70 287/603/70
  {0.13693565, 0.75293565},
  {0.13693565, 0.78065634},
  {0.29287222, 0.75293565},
  // f 283/1910/72 734/608/72 288/606/72
  {0.84202534, 0.00806057000000004},
  {0.81449449, 0.00806057000000004},
  {0.84202534, 0.20649135},
  // f 290/1911/71 741/611/71 289/609/71
  {0.79029745, 0.64574683},
  {0.81782830, 0.64574683},
  {0.79029745, 0.44393218},
  // f 291/1912/70 742/614/70 290/612/70
  {0.92092216, 0.43943834},
  {0.92092216, 0.46715909},
  {0.97630656, 0.43943834},
  // f 292/1913/72 743/617/72 291/615/72
  {0.74641085, 0.72467268},
  {0.71888000, 0.72467268},
  {0.74641085, 0.85691488},
  // f 293/621/555 744/620/554 292/618/957
  {0.96471310, 0.35175455},
  {0.96471310, 0.3794753},
  {0.97845739, 0.35175455},
  // f 294/623/557 745/622/556 293/621/958
  {0.95852900, 0.35175455},
  {0.95852900, 0.3794753},
  {0.96471310, 0.35175455},
  // f 295/625/559 746/624/558 294/623/557
  {0.95244557, 0.35175455},
  {0.95244557, 0.3794753},
  {0.95852900, 0.35175455},
  // f 296/627/561 747/626/560 295/625/959
  {0.93828022, 0.35175455},
  {0.93828022, 0.3794753},
  {0.95244557, 0.35175455},
  // f 297/1914/562 748/628/562 296/627/561
  {0.92092216, 0.35175455},
  {0.92092216, 0.3794753},
  {0.93828022, 0.35175455},
  // f 298/1915/72 749/631/72 297/629/72
  {0.97265011, 0.00806057000000004},
  {0.94511920, 0.00806057000000004},
  {0.97265011, 0.05326921},
  // f 299/635/565 750/634/564 298/632/563
  {0.93753707, 0.57658029},
  {0.93753707, 0.54885948},
  {0.92092216, 0.57658029},
  // f 300/637/567 751/636/566 299/635/960
  {0.94915140, 0.57658029},
  {0.94915140, 0.54885948},
  {0.93753707, 0.57658029},
  // f 301/639/569 752/638/568 300/637/961
  {0.95765865, 0.57658029},
  {0.95765865, 0.54885948},
  {0.94915140, 0.57658029},
  // f 302/641/571 753/640/570 301/639/962
  {0.96112555, 0.57658029},
  {0.96112555, 0.54885948},
  {0.95765865, 0.57658029},
  // f 303/643/573 754/642/572 302/641/963
  {0.96434242, 0.57658029},
  {0.96434242, 0.54885948},
  {0.96112555, 0.57658029},
  // f 304/1916/575 755/644/574 303/643/964
  {0.97025728, 0.57658029},
  {0.97025728, 0.54885948},
  {0.96434242, 0.57658029},
  // f 305/648/577 756/647/576 304/645/965
  {0.11271173, 0.94730031},
  {0.08518086, 0.94730031},
  {0.11271173, 0.95310438},
  // f 306/650/579 757/649/578 305/648/577
  {0.11271173, 0.94101977},
  {0.08518086, 0.94101977},
  {0.11271173, 0.94730031},
  // f 307/652/581 758/651/580 306/650/966
  {0.11271173, 0.93426287},
  {0.08518086, 0.93426287},
  {0.11271173, 0.94101977},
  // f 308/1917/582 759/653/582 307/652/967
  {0.11271173, 0.92702949},
  {0.08518086, 0.92702949},
  {0.11271173, 0.93426287},
  // f 289/1918/1 740/656/1 308/654/1
  {0.60934347, 0.98844051},
  {0.60934347, 0.9607197},
  {0.56404120, 0.98844051},
  // f 310/1919/72 761/659/72 309/657/72
  {0.86136985, 0.43891007},
  {0.83383900, 0.43891007},
  {0.86136985, 0.48019344},
  // f 311/1920/1 762/662/1 310/660/1
  {0.81782830, 0.90414059},
  {0.81782830, 0.8764199},
  {0.72372866, 0.90414059},
  // f 312/1921/71 763/665/71 311/663/71
  {0.87738061, 0.4801935},
  {0.90491146, 0.4801935},
  {0.87738061, 0.39559644},
  // f 313/669/586 764/668/585 312/666/968
  {0.28951290, 0.70909369},
  {0.28951290, 0.7368145},
  {0.31012878, 0.70909369},
  // f 314/671/588 765/670/587 313/669/586
  {0.27536368, 0.70909369},
  {0.27536368, 0.7368145},
  {0.28951290, 0.70909369},
  // f 315/673/590 766/672/589 314/671/588
  {0.26914576, 0.70909369},
  {0.26914576, 0.7368145},
  {0.27536368, 0.70909369},
  // f 316/675/592 767/674/591 315/673/590
  {0.26233962, 0.70909369},
  {0.26233962, 0.7368145},
  {0.26914576, 0.70909369},
  // f 317/677/594 768/676/593 316/675/969
  {0.25192523, 0.70909369},
  {0.25192523, 0.7368145},
  {0.26233962, 0.70909369},
  // f 318/679/596 769/678/595 317/677/594
  {0.23864965, 0.70909369},
  {0.23864965, 0.7368145},
  {0.25192523, 0.70909369},
  // f 319/681/598 770/680/597 318/679/596
  {0.22630590, 0.70909369},
  {0.22630590, 0.7368145},
  {0.23864965, 0.70909369},
  // f 320/683/600 771/682/599 319/681/598
  {0.21307173, 0.70909369},
  {0.21307173, 0.7368145},
  {0.22630590, 0.70909369},
  // f 321/685/602 772/684/601 320/683/600
  {0.19697779, 0.70909369},
  {0.19697779, 0.7368145},
  {0.21307173, 0.70909369},
  // f 322/687/604 773/686/603 321/685/602
  {0.18470772, 0.70909369},
  {0.18470772, 0.7368145},
  {0.19697779, 0.70909369},
  // f 323/689/606 774/688/605 322/687/604
  {0.17455199, 0.70909369},
  {0.17455199, 0.7368145},
  {0.18470772, 0.70909369},
  // f 324/691/608 775/690/607 323/689/970
  {0.16623035, 0.70909369},
  {0.16623035, 0.7368145},
  {0.17455199, 0.70909369},
  // f 325/693/610 776/692/609 324/691/971
  {0.16136661, 0.70909369},
  {0.16136661, 0.7368145},
  {0.16623035, 0.70909369},
  // f 326/695/612 777/694/611 325/693/610
  {0.15587258, 0.70909369},
  {0.15587258, 0.7368145},
  {0.16136661, 0.70909369},
  // f 327/697/614 778/696/613 326/695/972
  {0.14905095, 0.70909369},
  {0.14905095, 0.7368145},
  {0.15587258, 0.70909369},
  // f 328/699/616 779/698/615 327/697/973
  {0.14273864, 0.70909369},
  {0.14273864, 0.7368145},
  {0.14905095, 0.70909369},
  // f 329/1922/618 780/700/617 328/699/616
  {0.13693565, 0.70909369},
  {0.13693565, 0.7368145},
  {0.14273864, 0.70909369},
  // f 330/704/620 781/703/619 329/701/974
  {0.36558470, 0.81097412},
  {0.33805382, 0.81097412},
  {0.36558470, 0.81693923},
  // f 331/706/622 782/705/621 330/704/975
  {0.36558470, 0.80440605},
  {0.33805382, 0.80440605},
  {0.36558470, 0.81097412},
  // f 332/708/624 783/707/623 331/706/976
  {0.36558470, 0.79999483},
  {0.33805382, 0.79999483},
  {0.36558470, 0.80440605},
  // f 333/710/626 784/709/625 332/708/977
  {0.36558470, 0.79534805},
  {0.33805382, 0.79534805},
  {0.36558470, 0.79999483},
  // f 334/712/628 785/711/627 333/710/978
  {0.36558470, 0.78654385},
  {0.33805382, 0.78654385},
  {0.36558470, 0.79534805},
  // f 335/714/630 786/713/629 334/712/628
  {0.36558470, 0.77728367},
  {0.33805382, 0.77728367},
  {0.36558470, 0.78654385},
  // f 336/716/632 787/715/631 335/714/979
  {0.36558470, 0.7675674},
  {0.33805382, 0.7675674},
  {0.36558470, 0.77728367},
  // f 337/718/634 788/717/633 336/716/980
  {0.36558470, 0.75623655},
  {0.33805382, 0.75623655},
  {0.36558470, 0.7675674},
  // f 338/720/636 789/719/635 337/718/981
  {0.36558470, 0.74916768},
  {0.33805382, 0.74916768},
  {0.36558470, 0.75623655},
  // f 339/722/638 790/721/637 338/720/636
  {0.36558470, 0.7406646},
  {0.33805382, 0.7406646},
  {0.36558470, 0.74916768},
  // f 340/724/640 791/723/639 339/722/982
  {0.36558470, 0.72678053},
  {0.33805382, 0.72678053},
  {0.36558470, 0.7406646},
  // f 341/726/642 792/725/641 340/724/983
  {0.36558470, 0.71486926},
  {0.33805382, 0.71486926},
  {0.36558470, 0.72678053},
  // f 342/728/644 793/727/643 341/726/984
  {0.36558470, 0.70472777},
  {0.33805382, 0.70472777},
  {0.36558470, 0.71486926},
  // f 343/730/646 794/729/645 342/728/985
  {0.36558470, 0.69617331},
  {0.33805382, 0.69617331},
  {0.36558470, 0.70472777},
  // f 344/732/648 795/731/647 343/730/986
  {0.36558470, 0.69004178},
  {0.33805382, 0.69004178},
  {0.36558470, 0.69617331},
  // f 345/734/650 796/733/649 344/732/987
  {0.36558470, 0.6841538},
  {0.33805382, 0.6841538},
  {0.36558470, 0.69004178},
  // f 346/736/652 797/735/651 345/734/650
  {0.36558470, 0.67676187},
  {0.33805382, 0.67676187},
  {0.36558470, 0.6841538},
  // f 347/738/654 798/737/653 346/736/988
  {0.36558470, 0.67077994},
  {0.33805382, 0.67077994},
  {0.36558470, 0.67676187},
  // f 348/1923/656 799/739/655 347/738/989
  {0.36558470, 0.66525185},
  {0.33805382, 0.66525185},
  {0.36558470, 0.67077994},
  // f 349/743/658 800/742/657 348/740/656
  {0.14242806, 0.8244983},
  {0.14242806, 0.79677749},
  {0.13693565, 0.8244983},
  // f 350/745/660 801/744/659 349/743/658
  {0.14929873, 0.8244983},
  {0.14929873, 0.79677749},
  {0.14242806, 0.8244983},
  // f 351/747/662 802/746/661 350/745/990
  {0.15486726, 0.8244983},
  {0.15486726, 0.79677749},
  {0.14929873, 0.8244983},
  // f 352/749/664 803/748/663 351/747/662
  {0.16181409, 0.8244983},
  {0.16181409, 0.79677749},
  {0.15486726, 0.8244983},
  // f 353/751/666 804/750/665 352/749/991
  {0.16857807, 0.8244983},
  {0.16857807, 0.79677749},
  {0.16181409, 0.8244983},
  // f 354/753/668 805/752/667 353/751/666
  {0.17703494, 0.8244983},
  {0.17703494, 0.79677749},
  {0.16857807, 0.8244983},
  // f 355/755/670 806/754/669 354/753/668
  {0.18748377, 0.8244983},
  {0.18748377, 0.79677749},
  {0.17703494, 0.8244983},
  // f 356/757/672 807/756/671 355/755/670
  {0.20025687, 0.8244983},
  {0.20025687, 0.79677749},
  {0.18748377, 0.8244983},
  // f 357/759/674 808/758/673 356/757/672
  {0.21571974, 0.8244983},
  {0.21571974, 0.79677749},
  {0.20025687, 0.8244983},
  // f 358/761/676 809/760/675 357/759/674
  {0.23921777, 0.8244983},
  {0.23921777, 0.79677749},
  {0.21571974, 0.8244983},
  // f 359/763/678 810/762/677 358/761/676
  {0.25253099, 0.8244983},
  {0.25253099, 0.79677749},
  {0.23921777, 0.8244983},
  // f 360/765/680 811/764/679 359/763/678
  {0.25897923, 0.8244983},
  {0.25897923, 0.79677749},
  {0.25253099, 0.8244983},
  // f 361/767/682 812/766/681 360/765/680
  {0.26458436, 0.8244983},
  {0.26458436, 0.79677749},
  {0.25897923, 0.8244983},
  // f 362/769/684 813/768/683 361/767/682
  {0.27084333, 0.8244983},
  {0.27084333, 0.79677749},
  {0.26458436, 0.8244983},
  // f 363/771/686 814/770/685 362/769/992
  {0.27609938, 0.8244983},
  {0.27609938, 0.79677749},
  {0.27084333, 0.8244983},
  // f 364/773/688 815/772/687 363/771/993
  {0.28099832, 0.8244983},
  {0.28099832, 0.79677749},
  {0.27609938, 0.8244983},
  // f 365/775/690 816/774/689 364/773/994
  {0.28554010, 0.8244983},
  {0.28554010, 0.79677749},
  {0.28099832, 0.8244983},
  // f 366/1924/692 817/776/691 365/775/690
  {0.28836957, 0.8244983},
  {0.28836957, 0.79677749},
  {0.28554010, 0.8244983},
  // f 367/780/694 818/779/693 366/777/995
  {0.95455611, 0.96019852},
  {0.98208696, 0.96019852},
  {0.95455611, 0.95682156},
  // f 368/782/696 819/781/695 367/780/996
  {0.95455611, 0.96492302},
  {0.98208696, 0.96492302},
  {0.95455611, 0.96019852},
  // f 369/784/698 820/783/697 368/782/997
  {0.95455611, 0.97068143},
  {0.98208696, 0.97068143},
  {0.95455611, 0.96492302},
  // f 370/786/700 821/785/699 369/784/698
  {0.95455611, 0.97697604},
  {0.98208696, 0.97697604},
  {0.95455611, 0.97068143},
  // f 371/788/702 822/787/701 370/786/998
  {0.95455611, 0.98380661},
  {0.98208696, 0.98380661},
  {0.95455611, 0.97697604},
  // f 372/1925/703 823/789/703 371/788/702
  {0.95455611, 0.99193931},
  {0.98208696, 0.99193931},
  {0.95455611, 0.98380661},
  // f 373/1926/704 824/792/704 372/790/704
  {0.80262482, 0.95682156},
  {0.80262482, 0.98454225},
  {0.86136985, 0.95682156},
  // f 374/796/708 825/795/707 373/793/999
  {0.03553623, 0.98114777},
  {0.00800536, 0.98114777},
  {0.03553623, 0.98460782},
  // f 375/798/710 826/797/709 374/796/708
  {0.03553623, 0.97795832},
  {0.00800536, 0.97795832},
  {0.03553623, 0.98114777},
  // f 376/800/712 827/799/711 375/798/1000
  {0.03553623, 0.975389},
  {0.00800536, 0.975389},
  {0.03553623, 0.97795832},
  // f 377/802/714 828/801/713 376/800/1001
  {0.03553623, 0.97302675},
  {0.00800536, 0.97302675},
  {0.03553623, 0.975389},
  // f 378/1927/716 829/803/715 377/802/1002
  {0.03553623, 0.97087145},
  {0.00800536, 0.97087145},
  {0.03553623, 0.97302675},
  // f 379/807/718 830/806/717 378/804/1003
  {0.97533572, 0.39559644},
  {0.97533572, 0.42331719},
  {0.97717470, 0.39559644},
  // f 380/809/720 831/808/719 379/807/718
  {0.97300673, 0.39559644},
  {0.97300673, 0.42331719},
  {0.97533572, 0.39559644},
  // f 381/811/722 832/810/721 380/809/1004
  {0.97048539, 0.39559644},
  {0.97048539, 0.42331719},
  {0.97300673, 0.39559644},
  // f 382/813/724 833/812/723 381/811/722
  {0.96777707, 0.39559644},
  {0.96777707, 0.42331719},
  {0.97048539, 0.39559644},
  // f 383/815/726 834/814/725 382/813/724
  {0.96488160, 0.39559644},
  {0.96488160, 0.42331719},
  {0.96777707, 0.39559644},
  // f 384/817/728 835/816/727 383/815/726
  {0.96134341, 0.39559644},
  {0.96134341, 0.42331719},
  {0.96488160, 0.39559644},
  // f 385/819/730 836/818/729 384/817/728
  {0.95756096, 0.39559644},
  {0.95756096, 0.42331719},
  {0.96134341, 0.39559644},
  // f 386/821/732 837/820/731 385/819/730
  {0.95456374, 0.39559644},
  {0.95456374, 0.42331719},
  {0.95756096, 0.39559644},
  // f 387/823/734 838/822/733 386/821/732
  {0.95089334, 0.39559644},
  {0.95089334, 0.42331719},
  {0.95456374, 0.39559644},
  // f 388/825/736 839/824/735 387/823/734
  {0.94620323, 0.39559644},
  {0.94620323, 0.42331719},
  {0.95089334, 0.39559644},
  // f 389/827/738 840/826/737 388/825/736
  {0.94174743, 0.39559644},
  {0.94174743, 0.42331719},
  {0.94620323, 0.39559644},
  // f 390/829/740 841/828/739 389/827/738
  {0.93752587, 0.39559644},
  {0.93752587, 0.42331719},
  {0.94174743, 0.39559644},
  // f 391/831/742 842/830/741 390/829/740
  {0.93418688, 0.39559644},
  {0.93418688, 0.42331719},
  {0.93752587, 0.39559644},
  // f 392/833/744 843/832/743 391/831/742
  {0.93101043, 0.39559644},
  {0.93101043, 0.42331719},
  {0.93418688, 0.39559644},
  // f 393/835/746 844/834/745 392/833/1005
  {0.92799670, 0.39559644},
  {0.92799670, 0.42331719},
  {0.93101043, 0.39559644},
  // f 394/837/748 845/836/747 393/835/1006
  {0.92514557, 0.39559644},
  {0.92514557, 0.42331719},
  {0.92799670, 0.39559644},
  // f 395/839/750 846/838/749 394/837/748
  {0.92245716, 0.39559644},
  {0.92245716, 0.42331719},
  {0.92514557, 0.39559644},
  // f 396/1928/752 847/840/751 395/839/1007
  {0.92092216, 0.39559644},
  {0.92092216, 0.42331719},
  {0.92245716, 0.39559644},
  // f 397/844/754 848/843/753 396/841/752
  {0.85803604, 0.01021242},
  {0.88556695, 0.01021242},
  {0.85803604, 0.00806057000000004},
  // f 398/846/756 849/845/755 397/844/754
  {0.85803604, 0.01317734},
  {0.88556695, 0.01317734},
  {0.85803604, 0.01021242},
  // f 399/848/758 850/847/757 398/846/756
  {0.85803604, 0.01646852},
  {0.88556695, 0.01646852},
  {0.85803604, 0.01317734},
  // f 400/850/760 851/849/759 399/848/1008
  {0.85803604, 0.02008587},
  {0.88556695, 0.02008587},
  {0.85803604, 0.01646852},
  // f 401/852/762 852/851/761 400/850/1009
  {0.85803604, 0.02485734},
  {0.88556695, 0.02485734},
  {0.85803604, 0.02008587},
  // f 402/854/764 853/853/763 401/852/1010
  {0.85803604, 0.0310176599999999},
  {0.88556695, 0.0310176599999999},
  {0.85803604, 0.02485734},
  // f 403/856/766 854/855/765 402/854/764
  {0.85803604, 0.03884101},
  {0.88556695, 0.03884101},
  {0.85803604, 0.0310176599999999},
  // f 404/858/768 855/857/767 403/856/1011
  {0.85803604, 0.04637146},
  {0.88556695, 0.04637146},
  {0.85803604, 0.03884101},
  // f 405/860/770 856/859/769 404/858/1012
  {0.85803604, 0.05454123},
  {0.88556695, 0.05454123},
  {0.85803604, 0.04637146},
  // f 406/862/772 857/861/771 405/860/770
  {0.85803604, 0.06782466},
  {0.88556695, 0.06782466},
  {0.85803604, 0.05454123},
  // f 407/864/774 858/863/773 406/862/1013
  {0.85803604, 0.07736754},
  {0.88556695, 0.07736754},
  {0.85803604, 0.06782466},
  // f 408/866/776 859/865/775 407/864/774
  {0.85803604, 0.08396506},
  {0.88556695, 0.08396506},
  {0.85803604, 0.07736754},
  // f 409/868/778 860/867/777 408/866/1014
  {0.85803604, 0.08904576},
  {0.88556695, 0.08904576},
  {0.85803604, 0.08396506},
  // f 410/870/780 861/869/779 409/868/1015
  {0.85803604, 0.09287471},
  {0.88556695, 0.09287471},
  {0.85803604, 0.08904576},
  // f 411/872/782 862/871/781 410/870/1016
  {0.85803604, 0.09633547},
  {0.88556695, 0.09633547},
  {0.85803604, 0.09287471},
  // f 412/874/784 863/873/783 411/872/782
  {0.85803604, 0.09823519},
  {0.88556695, 0.09823519},
  {0.85803604, 0.09633547},
  // f 413/876/786 864/875/785 412/874/784
  {0.85803604, 0.10056198},
  {0.88556695, 0.10056198},
  {0.85803604, 0.09823519},
  // f 414/1929/788 865/877/787 413/876/1017
  {0.85803604, 0.10318518},
  {0.88556695, 0.10318518},
  {0.85803604, 0.10056198},
  // f 415/881/790 866/880/789 414/878/1018
  {0.39397368, 0.91218209},
  {0.39397368, 0.88446128},
  {0.39128309, 0.91218209},
  // f 416/883/792 867/882/791 415/881/1019
  {0.39685380, 0.91218209},
  {0.39685380, 0.88446128},
  {0.39397368, 0.91218209},
  // f 417/885/794 868/884/793 416/883/1020
  {0.39992347, 0.91218209},
  {0.39992347, 0.88446128},
  {0.39685380, 0.91218209},
  // f 418/887/796 869/886/795 417/885/1021
  {0.40318269, 0.91218209},
  {0.40318269, 0.88446128},
  {0.39992347, 0.91218209},
  // f 419/889/798 870/888/797 418/887/796
  {0.40663144, 0.91218209},
  {0.40663144, 0.88446128},
  {0.40318269, 0.91218209},
  // f 420/891/800 871/890/799 419/889/798
  {0.41102013, 0.91218209},
  {0.41102013, 0.88446128},
  {0.40663144, 0.91218209},
  // f 421/893/802 872/892/801 420/891/800
  {0.41568175, 0.91218209},
  {0.41568175, 0.88446128},
  {0.41102013, 0.91218209},
  // f 422/895/804 873/894/803 421/893/802
  {0.41894117, 0.91218209},
  {0.41894117, 0.88446128},
  {0.41568175, 0.91218209},
  // f 423/897/806 874/896/805 422/895/804
  {0.42232189, 0.91218209},
  {0.42232189, 0.88446128},
  {0.41894117, 0.91218209},
  // f 424/899/808 875/898/807 423/897/806
  {0.42836711, 0.91218209},
  {0.42836711, 0.88446128},
  {0.42232189, 0.91218209},
  // f 425/901/810 876/900/809 424/899/808
  {0.43424296, 0.91218209},
  {0.43424296, 0.88446128},
  {0.42836711, 0.91218209},
  // f 426/903/812 877/902/811 425/901/810
  {0.43806610, 0.91218209},
  {0.43806610, 0.88446128},
  {0.43424296, 0.91218209},
  // f 427/905/814 878/904/813 426/903/1022
  {0.44181395, 0.91218209},
  {0.44181395, 0.88446128},
  {0.43806610, 0.91218209},
  // f 428/907/816 879/906/815 427/905/814
  {0.44802558, 0.91218209},
  {0.44802558, 0.88446128},
  {0.44181395, 0.91218209},
  // f 429/909/818 880/908/817 428/907/816
  {0.45522878, 0.91218209},
  {0.45522878, 0.88446128},
  {0.44802558, 0.91218209},
  // f 430/1930/819 881/910/819 429/909/1023
  {0.46305358, 0.91218209},
  {0.46305358, 0.88446128},
  {0.45522878, 0.91218209},
  // f 431/1931/72 882/913/72 430/911/72
  {0.93948638, 0.22056758},
  {0.91195548, 0.22056758},
  {0.93948638, 0.23924661},
  // f 309/1932/70 760/916/70 431/914/70
  {0.93138647, 0.17361462},
  {0.93138647, 0.20133531},
  {0.97211820, 0.17361462},
  // f 433/1933/1 884/919/1 432/917/1
  {0.19396855, 0.95602393},
  {0.19396855, 0.92830324},
  {0.13307264, 0.95602393},
  // f 434/1934/71 885/922/71 433/920/71
  {0.87738061, 0.89685845},
  {0.90491146, 0.89685845},
  {0.87738061, 0.82742131},
  // f 435/1935/1 886/925/1 434/923/1
  {0.54803050, 0.94459867},
  {0.54803050, 0.91687787},
  {0.48148859, 0.94459867},
  // f 436/1936/72 887/928/72 435/926/72
  {0.92910850, 0.00806057000000004},
  {0.90157765, 0.00806057000000004},
  {0.92910850, 0.07749778},
  // f 437/1937/1 888/931/1 436/929/1
  {0.93854535, 0.94070041},
  {0.93854535, 0.9129796},
  {0.87738061, 0.94070041},
  // f 438/1938/71 889/934/71 437/932/71
  {0.81449449, 0.42104322},
  {0.84202534, 0.42104322},
  {0.81449449, 0.2226125},
  // f 439/1939/70 890/937/70 438/935/70
  {0.00800536, 0.92702949},
  {0.00800536, 0.9547503},
  {0.06917013, 0.92702949},
  // f 440/1940/72 891/940/72 439/938/72
  {0.90491146, 0.49631464},
  {0.87738061, 0.49631464},
  {0.90491146, 0.57658029},
  // f 441/1941/70 892/943/70 440/941/70
  {0.56404120, 0.91687787},
  {0.56404120, 0.94459867},
  {0.63058305, 0.91687787},
  // f 442/1942/71 893/946/71 441/944/71
  {0.87738061, 0.3794753},
  {0.90491146, 0.3794753},
  {0.87738061, 0.29920965},
  // f 443/1943/70 894/949/70 442/947/70
  {0.36406142, 0.92830324},
  {0.36406142, 0.95602393},
  {0.42495733, 0.92830324},
  // f 432/1944/72 883/952/72 443/950/72
  {0.81782830, 0.66186786},
  {0.79029745, 0.66186786},
  {0.81782830, 0.86029863},
  // f 445/1945/1 896/955/1 444/953/1
  {0.32204309, 0.69297254},
  {0.32204309, 0.66525185},
  {0.13693565, 0.69297254},
  // f 446/1946/820 897/958/71 445/956/71
  {0.92092216, 0.53273833},
  {0.94845307, 0.53273833},
  {0.92092216, 0.48373973},
  // f 447/1947/70 898/961/70 446/959/70
  {0.87738061, 0.69681752},
  {0.87738061, 0.72453833},
  {0.93948638, 0.69681752},
  // f 448/1948/71 899/964/71 447/962/71
  {0.83383900, 0.64574683},
  {0.86136985, 0.64574683},
  {0.83383900, 0.49631464},
  // f 449/1949/70 900/967/70 448/965/70
  {0.72372866, 0.92026174},
  {0.72372866, 0.94798255},
  {0.78462458, 0.92026174},
  // f 450/1950/72 901/970/72 449/968/72
  {0.86136985, 0.66186786},
  {0.83383900, 0.66186786},
  {0.86136985, 0.81130016},
  // f 451/1951/70 902/973/70 450/971/70
  {0.87738061, 0.25536776},
  {0.87738061, 0.28308851},
  {0.93948638, 0.25536776},
  // f 444/1952/821 895/976/72 451/974/72
  {0.94845307, 0.76230145},
  {0.92092216, 0.76230145},
  {0.94845307, 0.81130004}
 
//	{0.0,0.0},
//	{1.0,0.0},
//	{1.0,1.0},
//	{0.0,1.0}
};

struct Vector *VertexBufferRSX;	//Vertex buffer in RSX memory
struct TexST *TexCoordBufferRSX;	//Vertex buffer in RSX memory

float ang=0;

void drawFrame(int buffer, long frame) 
{

	realityFrontFace(context, REALITY_FRONT_FACE_CCW);
	realityCullFace(context, REALITY_CULL_FACE_BACK); 
	realityCullEnable(context, 1);

	realityDepthTestFunc(context, REALITY_ZFUNC_LESSOREQUAL);
	realityDepthWriteEnable(context, 1);
	realityDepthTestEnable(context, 1);

	//Find the constant where the matrix is stored in the shader (shouldn't actually do this every frame :) )
	int matrixParam=realityVertexProgramGetConstant((realityVertexProgram*)vshader_bin,"TransMatrix");
	int positionAttr,textureAttr;
	realityMatrix matrix,tmp;
	ang+=0.008;
	unsigned int offset;


	//Build a transform matrix to rotate the quad by its center
	//translate to center the quad at 0,0
	matrix=realityMatrixScale(700.0,700.0,700.0);

	//rotate through x axis
	tmp=realityMatrixRotationX(ang);
	matrix=realityMatrixMul(&matrix,&tmp);

	//rotate through y axis
	tmp=realityMatrixRotationY(ang*2);
	matrix=realityMatrixMul(&matrix,&tmp);

	//rotate through z axis
	tmp=realityMatrixRotationZ(ang*3);
	matrix=realityMatrixMul(&matrix,&tmp);

	//move to original position
	tmp=realityMatrixTranslation(600+400,300+300,0);
	matrix=realityMatrixMul(&matrix,&tmp);

	realityViewportTranslate(context, 0.0, 0.0, 0.0, 0.0);
	realityViewportScale(context, 1.0, 1.0, 1.0, 0.0); 

	realityZControl(context, 0, 1, 1); // disable viewport culling

	// Enable alpha blending.
	realityBlendFunc(context,
		NV30_3D_BLEND_FUNC_SRC_RGB_SRC_ALPHA |
		NV30_3D_BLEND_FUNC_SRC_ALPHA_SRC_ALPHA,
		NV30_3D_BLEND_FUNC_DST_RGB_ONE_MINUS_SRC_ALPHA |
		NV30_3D_BLEND_FUNC_DST_ALPHA_ZERO);
	realityBlendEquation(context, NV40_3D_BLEND_EQUATION_RGB_FUNC_ADD |
		NV40_3D_BLEND_EQUATION_ALPHA_FUNC_ADD);
	realityBlendEnable(context, 1);

	realityViewport(context, res.width, res.height);

	setupRenderTarget(buffer);

	// set the clear color
	realitySetClearColor(context, 0x00000000); // Black, because it looks cool
	// and the depth clear value
	realitySetClearDepthValue(context, 0xffff);
	// Clear the buffers
	realityClearBuffers(context, REALITY_CLEAR_BUFFERS_COLOR_R |
				     REALITY_CLEAR_BUFFERS_COLOR_G |
				     REALITY_CLEAR_BUFFERS_COLOR_B |
				     NV30_3D_CLEAR_BUFFERS_COLOR_A |
				     NV30_3D_CLEAR_BUFFERS_STENCIL |
				     REALITY_CLEAR_BUFFERS_DEPTH);

	// Load shaders, because the rsx won't do anything without them.
	realityLoadVertexProgram(context, (realityVertexProgram*)vshader_bin);
	realityLoadFragmentProgram(context, &nv30_fp); 

	//Pass the matrix to the shader
	realitySetVertexProgramConstant4fBlock(context,matrixParam,4,(float*)(matrix.data));

	// Load texture
	load_tex(0, tx_offset, dice.width, dice.height, dice.width*4,  NV40_3D_TEX_FORMAT_FORMAT_A8R8G8B8, 1);

	//Get the input attributes by name
	//shouldn't do attribute scan every frame but it's just a sample :)
	positionAttr = realityVertexProgramGetInputAttribute((realityVertexProgram*)vshader_bin,"inputvertex.vertex");
	textureAttr = realityVertexProgramGetInputAttribute((realityVertexProgram*)vshader_bin,"inputvertex.texcoord");

	//Bind the memory array to the input attributes
	//rsx requires the offset in his memory area
	realityAddressToOffset(&VertexBufferRSX[0].x,&offset);

	//stride is the distance (in bytes) from the attribute in a vertex to the same attribute in the next vertex (that is, the size of a single vertex struct)
	//size is the number of components of this attribute that will be passed to this input parameter in the vertex program (max 4)
	realityBindVertexBufferAttribute(context,positionAttr,offset,sizeof(struct Vector),4,REALITY_BUFFER_DATATYPE_FLOAT,REALITY_RSX_MEMORY);

	//now the texture coords
	realityAddressToOffset(&TexCoordBufferRSX[0].s,&offset);
	realityBindVertexBufferAttribute(context,textureAttr,offset,sizeof(struct TexST),2,REALITY_BUFFER_DATATYPE_FLOAT,REALITY_RSX_MEMORY);

	//Now request the draw of the bound buffer (or a part of it)
	realityDrawVertexBuffer(context,REALITY_TRIANGLES,0,5340);
}

s32 main(s32 argc, const char* argv[])
{
	PadInfo padinfo;
	PadData paddata;
	int i;
	
	init_screen();
	ioPadInit(7);

	// Load texture
	dice = loadPng(dice_bin);
	assert(realityAddressToOffset(dice.data, &tx_offset) == 0);

	// install fragment shader in rsx memory
	u32 *frag_mem = rsxMemAlign(256, 256);
	printf("frag_mem = 0x%08lx\n", (u64) frag_mem);
	realityInstallFragmentProgram(context, &nv30_fp, frag_mem);

	//Transfer the vertex buffer to RSX memory (main memory mapping is not supported yet)
	VertexBufferRSX = rsxMemAlign(64,5340*sizeof(struct Vector));
	memcpy(VertexBufferRSX,VertexBufferMain,5340*sizeof(struct Vector));

	TexCoordBufferRSX = rsxMemAlign(64,5340*sizeof(struct TexST));
	memcpy(TexCoordBufferRSX,TexCoordBufferMain,5340*sizeof(struct TexST));

	long frame = 0; // To keep track of how many frames we have rendered.
	
	// Ok, everything is setup. Now for the main loop.
	while(1){
		// Check the pads.
		ioPadGetInfo(&padinfo);
		for(i=0; i<MAX_PADS; i++){
			if(padinfo.status[i]){
				ioPadGetData(i, &paddata);
				
				if(paddata.BTN_CROSS){
					return 0;
				}
			}
			
		}

		waitFlip(); // Wait for the last flip to finish, so we can draw to the old buffer
		drawFrame(currentBuffer, frame++); // Draw into the unused buffer
		flip(currentBuffer); // Flip buffer onto screen
		currentBuffer = !currentBuffer;

	}
	
	return 0;
}

